/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PetalsComponent.java 601 2006-06-13 12:53:50Z wjoseph $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;

import org.objectweb.petals.component.common.listener.AbstractInternalMEProcessor;
import org.objectweb.petals.component.common.listener.MessageExchangeListener;
import org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager;
import org.objectweb.petals.component.common.util.ComponentLogger;

import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

/**
 * This class provides a basic component, allowing developper to create simply a
 * new JBI component
 * 
 * @author wjoseph - eBMWebSourcing
 * 
 */
public abstract class PetalsComponent implements Component, ComponentLifeCycle {

    /**
     * The component name
     */
    protected String componentName;

    /**
     * The component's delivery channel
     */
    protected DeliveryChannel channel;

    /**
     * the componentContext
     */
    private ComponentContext context;

    /**
     * The message exhange listener
     */
    private MessageExchangeListener listener;

    /**
     * The component's logger
     */
    protected Logger logger;

    /**
     * The component's service unit manager
     */
    protected PetalsServiceUnitManager serviceUnitManager;

    /**
     * Default contructor
     */
    public PetalsComponent() {
        this.serviceUnitManager = createServiceUnitManager();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#init(javax.jbi.component.ComponentContext)
     */
    public void init(ComponentContext ctx) throws JBIException {
        this.context = ctx;
        Logger log = context.getLogger("", null);
        logger = new ComponentLogger(log, log.getName(), log
            .getResourceBundleName(), componentName);
        channel = ctx.getDeliveryChannel();
        logger.log(Level.INFO, componentName + "init");
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#shutDown()
     */
    public void shutDown() throws JBIException {
        logger.log(Level.INFO, componentName + " shutdown");
        listener = null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#start()
     */
    public void start() throws JBIException {
        logger.log(Level.INFO, componentName + " start");
        AbstractInternalMEProcessor processor = createMessageExchangeProcessor();
        listener = createMessageExchangeListener(processor);
        listener.listen();
    }

    protected abstract AbstractInternalMEProcessor createMessageExchangeProcessor();

    /**
     * Creates a new MessageExchangeListener
     * 
     * @param processor
     *            the messageExchange processor
     * @return
     */
    protected MessageExchangeListener createMessageExchangeListener(
        AbstractInternalMEProcessor processor) {
        return new MessageExchangeListener(channel, processor,
            MessageExchangeListener.IgnoredStatus.DONE_AND_ERROR_IGNORED, 0,
            context.getComponentName());
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ComponentLifeCycle#stop()
     */
    public void stop() throws JBIException {
        logger.log(Level.INFO, componentName + " stop");
        this.listener.terminate();
    }

    /**
     * Create a new service unit manager
     * 
     * @return new instance of service unit manager
     */
    protected PetalsServiceUnitManager createServiceUnitManager() {
        return new PetalsServiceUnitManager();
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#getLifeCycle()
     */
    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#getServiceDescription(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public Document getServiceDescription(ServiceEndpoint endpoint) {
        return serviceUnitManager.getServiceDescription(endpoint);
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#getServiceUnitManager()
     */
    public ServiceUnitManager getServiceUnitManager() {
        return serviceUnitManager;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#isExchangeWithConsumerOkay(javax.jbi.servicedesc.ServiceEndpoint,
     *      javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithConsumerOkay(ServiceEndpoint endpoint,
        MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#isExchangeWithProviderOkay(javax.jbi.servicedesc.ServiceEndpoint,
     *      javax.jbi.messaging.MessageExchange)
     */
    public boolean isExchangeWithProviderOkay(ServiceEndpoint endpoint,
        MessageExchange exchange) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.Component#resolveEndpointReference(org.w3c.dom.DocumentFragment)
     */
    public ServiceEndpoint resolveEndpointReference(DocumentFragment epr) {
        return null;
    }

    /**
     * Returns the component's name
     * 
     * @return the componentName
     */
    public String getComponentName() {
        return componentName;
    }

    /**
     * Set componentName
     * 
     * @param componentName
     */
    public void setComponentName(String componentName) {
        this.componentName = componentName;
    }

}
