/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: AbstractServiceUnitManager.java 601 2006-06-13 12:53:50Z wjoseph $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.basic;

import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.common.util.ManagementMessageUtil;
import org.objectweb.petals.component.common.util.WSDLHelper;
import org.objectweb.petals.component.common.util.XMLHelper;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Generic service unit manager. Only handles one type of service unit.
 * Endpoints are activated by this manager. An abstract deploy method allows you
 * to add extra information in the service unit descriptor.
 * 
 * @version $Rev: 250 $ $Date: 2006-04-21 14:20:57 +0200 (ven, 21 avr 2006) $
 * @since Petals 1.0
 * @author wjoseph - eBMWebsourcing
 */
public abstract class AbstractServiceUnitManager implements ServiceUnitManager {

    /**
     * Namespace of the extension elements in the jbi descriptor file
     */
    public static final String EXTENSIONS_XMLNS = "http://petals.objectweb.org/extensions";

    /**
     * Error messages
     */
    protected static final String ERROR_DUPLICATE_SERVICE_UNIT = "A service unit with the given name is already registered ";

    protected static final String FAILED_ACTIVATE_ENDPOINT = "Activate new endpoint failed ";

    protected static final String FAILED_DEACTIVATE_ENDPOINT = "Deactivate endpoint failed ";

    protected static final String INCOMPLETE_SERVICE_UNIT_PACKAGE = "The deployed service package is incomplete ";

    /**
     * The component's context
     */
    protected ComponentContext context;

    /**
     * The component's Logger
     */
    protected Logger logger;

    protected Map<ServiceEndpoint, Document> endpointServiceDescriptionMap;

    protected Map<QName, Set<ServiceEndpoint>> servicesByInterfaceMap;

    protected Map<String, String> serviceUnitInstallationRootPath;

    protected Map<String, ServiceEndpoint> serviceUnitMap;

    /**
     * Default constructor
     */
    public AbstractServiceUnitManager() {
        endpointServiceDescriptionMap = new HashMap<ServiceEndpoint, Document>();
        serviceUnitMap = new HashMap<String, ServiceEndpoint>();
        servicesByInterfaceMap = new HashMap<QName, Set<ServiceEndpoint>>();
        serviceUnitInstallationRootPath = new HashMap<String, String>();
    }

    /**
     * Creates a new service unit manager
     * 
     * @param context
     *            Component context of the callind component
     * @param logger
     *            Logger of the calling component
     */
    public AbstractServiceUnitManager(ComponentContext context, Logger logger)
        throws DeploymentException {
        this.context = context;
        this.logger = logger;
        endpointServiceDescriptionMap = new HashMap<ServiceEndpoint, Document>();
        serviceUnitMap = new HashMap<String, ServiceEndpoint>();
        servicesByInterfaceMap = new HashMap<QName, Set<ServiceEndpoint>>();
        serviceUnitInstallationRootPath = new HashMap<String, String>();
    }

    /**
     * Add a service endpoint for an interface QName
     * 
     * @param interfaceName
     *            the interface QName
     * @param serviceEndpoint
     *            the endpoint to add
     */
    public void addServiceForInterface(QName interfaceName,
            ServiceEndpoint serviceEndpoint) {
        Set<ServiceEndpoint> endpoints = null;
        endpoints = servicesByInterfaceMap.get(interfaceName);
        if (endpoints == null) {
            endpoints = new HashSet<ServiceEndpoint>();
        }
        endpoints.add(serviceEndpoint);
        servicesByInterfaceMap.put(interfaceName, endpoints);
    }

    /**
     * Deploy a service unit. Before deploying, the service unit type if
     * verified
     * 
     * @param serviceUnitName
     *            name of the service unit to deploy
     * @param serviceUnitRootPath
     *            root path of the service unit to deploy
     * @return a valid management message
     * @throws DeploymentException
     *             If an error occurs during deployment
     */
    public String deploy(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {
        String result = null;
        if (serviceUnitMap.get(serviceUnitName) == null) {
            File jbiXmlFile = new File(serviceUnitRootPath + File.separator
                    + "META-INF" + File.separator + "jbi.xml");
            Node servicesNode = getNode(jbiXmlFile, "services");
            Map<String, String> extensions = getExtensions(servicesNode);
            logger.log(Level.FINE, "deploy  serviceUnitName  "
                    + serviceUnitName + "serviceUnitRootPath"
                    + serviceUnitRootPath);
            result = ManagementMessageUtil.getComponentTaskResult(context
                    .getComponentName(), "deploy",
                    ManagementMessageUtil.TASK_RESULT_SUCCESS);
            deploy(extensions);
        } else {
            throw new DeploymentException(ERROR_DUPLICATE_SERVICE_UNIT);
        }
        return result;
    }

    /**
     * Get the service description for one of the component endpoint
     * 
     * @param se
     *            Service endpoint to get description for
     * @return The service description or null
     */
    public Document getServiceDescription(ServiceEndpoint se) {
        Document result = null;
        for (ServiceEndpoint ep : endpointServiceDescriptionMap.keySet()) {
            if (se.getServiceName().equals(ep.getServiceName())) {
                result = endpointServiceDescriptionMap.get(ep);
                break;
            }
        }
        return result;
    }

    /**
     * Return the endpoint for a given service unit name
     * 
     * @param serviceUnitName
     *            The service unit name
     * @return The matching endpoint or null if no matching endpoint found
     */
    public ServiceEndpoint getServiceUnitEndpoint(String serviceUnitName) {
        return serviceUnitMap.get(serviceUnitName);
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#init(java.lang.String,
     *      java.lang.String)
     * 
     * The service unit root path is stored in a Map to support multithreading
     */
    public void init(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {
        logger.log(Level.FINE, "init  serviceUnitName  " + serviceUnitName
                + "  serviceUnitRootPath " + serviceUnitRootPath);
        if (serviceUnitInstallationRootPath.put(serviceUnitName,
                serviceUnitRootPath) != null) {
            logger.log(Level.SEVERE, ERROR_DUPLICATE_SERVICE_UNIT);
            throw new DeploymentException(ERROR_DUPLICATE_SERVICE_UNIT);
        }
    }

    /**
     * Remove the service description for an endpoint
     * 
     * @param ep
     *            the service endpoint for which the description will be removed
     */
    public void removeServiceDescription(ServiceEndpoint ep) {
        endpointServiceDescriptionMap.remove(ep);
    }

    /**
     * Remove the service description for an endpoint
     * 
     * @param ep
     *            the service endpoint for which the description will be removed
     */
    public void removeServiceUnitEndpoint(String suName) {
        serviceUnitMap.remove(suName);
    }

    /**
     * Set the service description for a given endpoint
     * 
     * @param se
     *            The service endpoint
     * @param serviceDesc
     *            The dom document representation of the service description
     */
    public void setServiceDescription(ServiceEndpoint se, Document serviceDesc) {
        endpointServiceDescriptionMap.put(se, serviceDesc);
    }

    /**
     * Set the endpoint for a given service unit name
     * 
     * @param serviceUnitName
     *            The name of the service unit
     * @param serviceEP
     *            The service endpoint
     */
    public void setServiceUnitEndpoint(String serviceUnitName,
            ServiceEndpoint serviceEP) {
        serviceUnitMap.put(serviceUnitName, serviceEP);
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#shutDown(java.lang.String)
     */
    public void shutDown(String serviceUnitName) throws DeploymentException {
        logger.log(Level.FINE, "shutdown  serviceUnitName  " + serviceUnitName);
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#start(java.lang.String)
     * 
     * Endpoints are activated here
     */
    public void start(String serviceUnitName) throws DeploymentException {
        logger.log(Level.FINE, "start  serviceUnitName  " + serviceUnitName);
        String suRootPath = serviceUnitInstallationRootPath
                .get(serviceUnitName);
        Document serviceDesc = getServiceUnitWSDL(suRootPath);
        if (serviceDesc != null) {
            try {
                activateEndpointsFromJBIDescription(serviceDesc,
                        serviceUnitName, suRootPath);
            } catch (Exception ex) {
                logger.log(Level.SEVERE, FAILED_ACTIVATE_ENDPOINT + ex);
                throw new DeploymentException(FAILED_ACTIVATE_ENDPOINT, ex);
            }
        } else {
            throw new DeploymentException(INCOMPLETE_SERVICE_UNIT_PACKAGE);
        }
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#stop(java.lang.String)
     * 
     * Endpoints are deactivated here
     */
    public void stop(String serviceUnitName) throws DeploymentException {
        try {
            serviceUnitInstallationRootPath.remove(serviceUnitName);
            deactivateEndpointsFromJBIDescription(serviceUnitName);
        } catch (Exception e) {
            throw new DeploymentException(FAILED_DEACTIVATE_ENDPOINT, e);
        }
        logger.log(Level.FINE, "stop  serviceUnitName  " + serviceUnitName);
    }

    /**
     * (non-Javadoc)
     * 
     * @see javax.jbi.component.ServiceUnitManager#undeploy(java.lang.String,
     *      java.lang.String)
     */
    public String undeploy(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {
        return ManagementMessageUtil.getComponentTaskResult(context
                .getComponentName(), "undeploy",
                ManagementMessageUtil.TASK_RESULT_SUCCESS);
    }

    /**
     * Activate an endpoint in Petals platform, put the activate endpoint add
     * the service description of this endpoint in HashMap, add the service unit
     * name and the endpoint in a HashMap and add the interface QName and the
     * endpoint in a HashMap. Only the first provides node is activated
     * 
     * @param serviceDesc
     *            the service unit service description
     * @param serviceUnitName
     *            the service unit name
     * @param suRootPath
     *            the service unit root path
     * @throws Exception
     *             If errors occur during processing
     * @return the activated endpoint
     */
    protected ServiceEndpoint activateEndpointsFromJBIDescription(
            Document serviceDesc, String serviceUnitName, String suRootPath)
        throws Exception {
        File jbiXmlFile = new File(suRootPath + File.separator + "META-INF"
                + File.separator + "jbi.xml");
        Node providesNode = getNode(jbiXmlFile, "provides");
        ServiceEndpoint ep = null;
        if (providesNode != null) {
            QName interfaceName = getInterfaceNameFromJbiXml(providesNode);
            String endpointName = getEndpointNameFromJbiXml(providesNode);
            QName serviceName = getServiceNameFromJbiXml(providesNode);
            ep = context.activateEndpoint(serviceName, endpointName);
            setServiceDescription(ep, serviceDesc);
            setServiceUnitEndpoint(serviceUnitName, ep);
            addServiceForInterface(interfaceName, ep);
        }
        return ep;
    }

    /**
     * Deactivate an endpoint in Petals platform, remove the activate endpoint
     * from various HashMap in which it was registered
     * 
     * @param ep
     *            the service endpoint to be deactivated
     * @throws Exception
     *             If errors occur during processing
     */
    protected void deactivateEndpointsFromJBIDescription(String serviceUnitName)
        throws JBIException {
        ServiceEndpoint ep = getServiceUnitEndpoint(serviceUnitName);
        if (ep != null) {
            context.deactivateEndpoint(ep);
            removeServiceDescription(ep);
            removeServiceUnitEndpoint(serviceUnitName);
        }
    }

    /**
     * Do extra computation with the extension fragment of the jbi descriptor.
     * 
     * @param jbiExtensionFragment
     *            the extension frangment of the jbi descriptor
     */
    protected abstract void deploy(Map<String, String> extensions);

    /**
     * Return the endpoint name parse from the jbi descriptor file
     * 
     * @param jbiXml
     *            jbi descriptor file
     * @return the endpoint name if found, else null
     */
    protected String getEndpointNameFromJbiXml(Node providesNode) {
        String result = new String();
        result = XMLHelper.getAttributeValue(providesNode, "endpoint-name");
        return result;
    }

    /**
     * Returns a Map with all extension elements. These elements are xml node
     * placed in the jbi.xml service unit descriptor file. They use the
     * EXTENSIONS_XMLNS namespace.
     * 
     * @param servicesNode
     * @return Map containing all extensions element with their value
     */
    protected Map<String, String> getExtensions(Node servicesNode) {
        Map<String, String> extensions = new HashMap<String, String>();
        NodeList nl = servicesNode.getChildNodes();
        String prefix = XMLHelper.getPrefixForNamespaceURI(servicesNode,
                EXTENSIONS_XMLNS, false);
        for (int i = 0; i < nl.getLength(); i++) {
            Node node = nl.item(i);
            if (node.getNodeType() == Node.ELEMENT_NODE
                    && node.getNodeName().startsWith(prefix)) {
                extensions.put(node.getNodeName().replaceFirst(prefix, ""),
                        node.getTextContent());
            }
        }
        return extensions;
    }

    /**
     * Return the interface QName parse from the jbi descriptor file
     * 
     * @param jbiXml
     *            jbi descriptor file
     * @return the interface QName if found, else null
     */
    protected QName getInterfaceNameFromJbiXml(Node providesNode) {
        QName result = null;
        result = QName.valueOf(XMLHelper.getAttributeValue(providesNode,
                "interface-name"));
        return result;
    }

    /**
     * Parse the JBI descriptor into a Dom Document and search for the first
     * node whom name matches the NodeName
     * 
     * @param jbiXmlFile
     *            the service unit description file
     * @param nodeName
     *            Name of the node to search
     * @return node with the right name if found, else null
     */
    protected Node getNode(File jbiXmlFile, String nodeName) {
        Document jbiDoc = WSDLHelper.createDocumentFromWSDL(jbiXmlFile);
        Node providesNode = XMLHelper.findChild(jbiDoc, nodeName, true);
        return providesNode;
    }

    /**
     * Return the service QName parse from the jbi descriptor file
     * 
     * @param jbiXml
     *            jbi descriptor file
     * @return the service QName if found, else null
     */
    protected QName getServiceNameFromJbiXml(Node providesNode) {
        QName result = null;
        result = QName.valueOf(XMLHelper.getAttributeValue(providesNode,
                "service-name"));
        return result;
    }

    /**
     * Returns A Dom Document represenation of the service description file
     * 
     * @param suRootPath
     * @return
     */
    protected Document getServiceUnitWSDL(String suRootPath) {
        Document result = null;
        File[] files = new File(suRootPath).listFiles();
        for (File file : files) {
            if (file.getName().endsWith(".wsdl")) {
                result = WSDLHelper.createDocumentFromWSDL(file);
                break;
            }
        }
        return result;
    }
}
