/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SimpleBindingComponentServiceUnitManager.java 601 2006-06-13 12:53:50Z wjoseph $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.basic.binding;

import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.management.DeploymentException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.common.basic.AbstractServiceUnitManager;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * Generic service unit manager. Only handles one type of service unit.
 * Endpoints are activated by this manager. An abstract deploy method allows you
 * to add extra information in the service unit descriptor.
 * 
 * @version $Rev: 250 $ $Date: 2006-04-21 14:20:57 +0200 (ven, 21 avr 2006) $
 * @since Petals 1.0
 * @author wjoseph - eBMWebsourcing
 */
public class SimpleBindingComponentServiceUnitManager extends
    AbstractServiceUnitManager {

    protected static final String NO_ENDPOINT_FOR_CONSUMES_NODE = "No endpoint was found matching information from the consumes node of the JBI descriptor";

    protected Map<ServiceEndpoint, String> endpointAddress;

    protected Map<String, Set<String>> extAddressesForSU;

    protected ExternalListenerManager externalListenerManager;

    /**
     * Default contructor
     */
    public SimpleBindingComponentServiceUnitManager() {
        super();
        endpointAddress = new HashMap<ServiceEndpoint, String>();
        extAddressesForSU = new HashMap<String, Set<String>>();
    }

    public void setExternalListenerManager(
        ExternalListenerManager externalListenerManager) {
        this.externalListenerManager = externalListenerManager;
    }

    /**
     * Create a new SimpleBindingComponentServiceUnitManager
     * 
     * @param context
     *            Component's context
     * @param log
     *            Component's logger
     * 
     * @throws DeploymentException
     */
    public SimpleBindingComponentServiceUnitManager(ComponentContext context,
        Logger log)
        throws DeploymentException {
        super(context, log);
        endpointAddress = new HashMap<ServiceEndpoint, String>();
        extAddressesForSU = new HashMap<String, Set<String>>();
    }

    /**
     * Returns the distant address for an endpoint. May be null if this endpoint
     * has no address registered for it
     * 
     * @param ep
     *            an Endpoint
     * @return the distant address for this endpoint, or null
     */
    public String getAddressForEndpoint(ServiceEndpoint ep) {
        return endpointAddress.get(ep);
    }

    /**
     * Return the endpoint associated with a distant address. May be null if
     * this address does not match any endpoint
     * 
     * @param address
     *            The distant address
     * @return the endpoint associated with the distant address, else null
     */
    public ServiceEndpoint getEndpointForAddress(String address) {
        for (Map.Entry<ServiceEndpoint, String> entry : endpointAddress.entrySet()) {
            if (entry.getValue() != null && entry.getValue().equals(address)){
                return entry.getKey();
            }
        }
        return null;
    }

    /**
     * Same process as for AbstractServiceUnitManager except the registered
     * endpoint is stored with the distant address in a Map (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.basic.AbstractServiceUnitManager#start(java.lang.String)
     */
    @Override
    public void start(String serviceUnitName) throws DeploymentException {
        logger.log(Level.FINE, "start  serviceUnitName  " + serviceUnitName);
        String suRootPath = serviceUnitInstallationRootPath
            .get(serviceUnitName);
        Document serviceDesc = getServiceUnitWSDL(suRootPath);
        try {
            ServiceEndpoint ep = activateEndpointsFromJBIDescription(
                serviceDesc, serviceUnitName, suRootPath);
            registerAddressForEndpoint(ep, suRootPath);
            String address = registerEndpointForAddress(suRootPath,
                serviceUnitName);
            if (address != null) {
                externalListenerManager.startListening(address);
            }
        } catch (Exception ex) {
            logger.log(Level.SEVERE, FAILED_ACTIVATE_ENDPOINT + ex, ex);
            throw new DeploymentException(FAILED_ACTIVATE_ENDPOINT, ex);
        }
    }

    /**
     * The distant address registered for this endpoint is removed (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.basic.AbstractServiceUnitManager#stop(java.lang.String)
     */
    public void stop(String serviceUnitName) throws DeploymentException {
        try {
            serviceUnitInstallationRootPath.remove(serviceUnitName);
            deactivateEndpointsFromJBIDescription(serviceUnitName);
            ServiceEndpoint ep = serviceUnitMap.get(serviceUnitName);
            endpointAddress.remove(ep);
            Set<String> addresses = extAddressesForSU.get(serviceUnitName);
            if (addresses != null) {
                for (String address : addresses) {
                    ServiceEndpoint seToRemove = getEndpointForAddress(address);
                    if(seToRemove!= null){
                        endpointAddress.remove(seToRemove);
                    }
                    externalListenerManager.stopListening(address);
                }

            }
        } catch (Exception e) {
            throw new DeploymentException(FAILED_DEACTIVATE_ENDPOINT, e);
        }
        logger.log(Level.FINE, "stop  serviceUnitName  " + serviceUnitName);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.basic.AbstractServiceUnitManager#deploy(java.util.Map)
     */
    @Override
    protected void deploy(Map<String, String> extensions) {
    }

    /**
     * Register a distant address for the new endpoint. The address is found in
     * the jbiXml file as an extension element.
     * 
     * @param ep
     *            the new endpoint
     * @param suRootPath
     *            the service unit's root path
     */
    protected void registerAddressForEndpoint(ServiceEndpoint ep,
        String suRootPath) {
        File jbiXmlFile = new File(suRootPath + File.separator + "META-INF"
            + File.separator + "jbi.xml");
        Node providesNode = getNode(jbiXmlFile, "provides");
        if (providesNode != null) {
            Map<String, String> extensions = getExtensions(providesNode);
            endpointAddress.put(ep, extensions.get("address"));
        }
    }

    /**
     * Register an endpoint for the distant address. The address is found in the
     * jbiXml file as an extension element. We search in the container if there
     * is an endpoint for the serviceName and endpointName found in the consumes
     * node. If no endpoint is found, we search for endpoints for the Interface
     * Name found in the consumes node. If still no endpoint is foud an
     * Exception is raised
     * 
     * @param suRootPath
     *            the service unit's root path
     * @param serviceUnitName
     *            the service unit name
     * @return the registered address (listened address). Can be null if no
     *         "consumes" node can be found in the JBI descriptor or if no
     *         external address is provided
     * @throws Exception
     *             If bo endpoint is found for this serviceName or Interface
     *             Name
     */
    protected String registerEndpointForAddress(String suRootPath,
        String serviceUnitName) throws Exception {
        String address = null;

        File jbiXmlFile = new File(suRootPath + File.separator + "META-INF"
            + File.separator + "jbi.xml");
        Node consumesNode = getNode(jbiXmlFile, "consumes");
        if (consumesNode != null) {
            Map<String, String> extensions = getExtensions(consumesNode);
            address = extensions.get("address");
            if (address != null) {
                QName interfaceName = getInterfaceNameFromJbiXml(consumesNode);
                String endpointName = getEndpointNameFromJbiXml(consumesNode);
                QName serviceName = getServiceNameFromJbiXml(consumesNode);
                ServiceEndpoint endpoint = context.getEndpoint(serviceName,
                    endpointName);
                if (endpoint == null) {
                    ServiceEndpoint[] endpoints = context
                        .getEndpoints(interfaceName);
                    if (endpoints.length == 0) {
                        throw new Exception(NO_ENDPOINT_FOR_CONSUMES_NODE);
                    } else {
                        endpointAddress.put(endpoints[0],address);
                    }
                } else {
                    endpointAddress.put(endpoint,address);
                }
                registerAddressForSU(address, serviceUnitName);
            }   
        }
        return address;
    }

    /**
     * Add this address to the su (identified by its su name) address set
     * 
     * @param address
     *            the address to register
     * @param serviceUnitName
     *            the service unit name
     */
    protected void registerAddressForSU(String address, String serviceUnitName) {
        // Retrieve the address set for this suName
        Set<String> addresses = extAddressesForSU.get(serviceUnitName);

        if (addresses == null) {
            addresses = new HashSet<String>();
            addresses.add(address);
            extAddressesForSU.put(serviceUnitName, addresses);
        } else {
            // Add this address to the su set
            addresses.add(address);
        }
    }
}
