/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PetalsServiceUnitManager.java 154 2006-03-27 15:30:10Z wjoseph $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.serviceunitmanager.manager;

import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.management.DeploymentException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.common.serviceunitmanager.handler.ExternalServiceManager;
import org.objectweb.petals.component.common.serviceunitmanager.handler.PetalsServiceUnitHandler;
import org.objectweb.petals.component.common.util.ManagementMessageUtil;
import org.objectweb.petals.component.common.util.WSDLHelper;
import org.objectweb.petals.component.common.util.XMLHelper;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * General service unit manager in charge of delegating the acutal deployment,
 * installation, etc to dedicated ServiceUnitHandler
 * 
 * @version $Rev: 250 $ $Date: 2006-04-21 14:20:57 +0200 (ven, 21 avr 2006) $
 * @since Petals 1.0
 * @author wjoseph - eBMWebsourcing
 */
public class PetalsServiceUnitManager implements ServiceUnitManager {

    private static final String NO_HANDLER_FOR_THIS_TYPE = "No registered handlers for this type of service unit ";

    private static final String DUPLICATE_SERVICE_UNIT = "Another service unit with the same name is already deployed ";

    private static final String HANDLER_FOR_THIS_SERVICE_NOT_FOUND = "Handler for this service name unit not found ";
    
    public static final String EXTENSIONS_XMLNS = "http://petals.objectweb.org/extensions";

    private Map<ServiceEndpoint, Document> serviceDescriptionMap;

    private Map<String, ServiceEndpoint> serviceUnitMap;

    private Map<String, PetalsServiceUnitHandler> handlersMap;

    private Map<QName, Set<ServiceEndpoint>> servicesByInterfaceMap;

    private Set<PetalsServiceUnitHandler> handlersSet;
    
    private Map<ServiceEndpoint, PetalsServiceUnitHandler> endpointSUHandler;

    protected ComponentContext context;

    protected Logger logger;

    /**
     * Creates a new service unit manager
     * 
     * @param context
     *            Component context of the callind component
     * @param logger
     *            Logger of the calling component
     */
    public PetalsServiceUnitManager(ComponentContext context, Logger logger)
        throws DeploymentException {
        this.context = context;
        this.logger = logger;
        serviceDescriptionMap = new HashMap<ServiceEndpoint, Document>();
        serviceUnitMap = new HashMap<String, ServiceEndpoint>();
        handlersMap = new HashMap<String, PetalsServiceUnitHandler>();
        handlersSet = new HashSet<PetalsServiceUnitHandler>();
        handlersSet.add(new ExternalServiceManager(context, logger));
        servicesByInterfaceMap = new HashMap<QName, Set<ServiceEndpoint>>();
        endpointSUHandler = new HashMap<ServiceEndpoint, PetalsServiceUnitHandler>();
    }
    
    /**
     * Default constructor
     */
    public PetalsServiceUnitManager() {
    serviceDescriptionMap = new HashMap<ServiceEndpoint, Document>();
    serviceUnitMap = new HashMap<String, ServiceEndpoint>();
    handlersMap = new HashMap<String, PetalsServiceUnitHandler>();
    handlersSet = new HashSet<PetalsServiceUnitHandler>();
    handlersSet.add(new ExternalServiceManager(context, logger));
    servicesByInterfaceMap = new HashMap<QName, Set<ServiceEndpoint>>();
}


    /**
     * Add a new handler in the handlersSet
     * 
     * @param suh
     *            The new handler to be added
     */
    public void addNewHandler(PetalsServiceUnitHandler suh) {
        if (!isHandlerPresent(suh.getClass().getName())) {
            handlersSet.add(suh);
        }
    }

    /**
     * Add a service endpoint for an interface QName
     * @param interfaceName the interface QName
     * @param serviceEndpoint the endpoint to add
     */
    public void addServiceForInterface(QName interfaceName,
            ServiceEndpoint serviceEndpoint) {
        Set<ServiceEndpoint> endpoints = null;
        endpoints = servicesByInterfaceMap.get(interfaceName);
        if (endpoints == null) {
            endpoints = new HashSet<ServiceEndpoint>();
        }
        endpoints.add(serviceEndpoint);
        servicesByInterfaceMap.put(interfaceName, endpoints);
    }
    
    
    /**
     * Add a PetalsServiceUnitHandler for a ServiceEnpoint
     * @param ep the endpoint
     * @param handler the PetalsServiceUnitHandler
     */
    public void addSUHandlerForEndpoint(ServiceEndpoint ep, PetalsServiceUnitHandler handler) {
        endpointSUHandler.put(ep, handler);
    }
    
    public String deploy(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {
        String result = null;
        if (serviceUnitMap.get(serviceUnitName) == null) {
            String serviceUnitType = getServiceUnitType(serviceUnitRootPath);
            PetalsServiceUnitHandler serviceUnitManager = null;
            Iterator<PetalsServiceUnitHandler> it = handlersSet.iterator();
            while (result == null && it.hasNext()) {
                serviceUnitManager = it.next();
                result = serviceUnitManager.deploy(serviceUnitName,
                         serviceUnitType, serviceUnitRootPath);
            }
            if (result != null) {
                handlersMap.put(serviceUnitName, serviceUnitManager);
            } else {
                logger.log(Level.SEVERE, NO_HANDLER_FOR_THIS_TYPE);
                throw new DeploymentException(NO_HANDLER_FOR_THIS_TYPE);
            }
        } else {
            throw new DeploymentException(DUPLICATE_SERVICE_UNIT);
        }
        return result;
    }

    public ServiceEndpoint findEndpointForService(QName serviceName) {
        ServiceEndpoint result = null;
        for(ServiceEndpoint endpoint : serviceUnitMap.values()) {
            if (endpoint.getServiceName().equals(serviceName)){
                result = endpoint;
                break;
            }
        }
        return result;
    }
    
    /**
     * Get the service description for one of the component endpoint
     * 
     * @param se
     *            Service endpoint to get description for
     * @return The service description or null if this endpoint is not found
     */
    public Document getServiceDescription(ServiceEndpoint se) {
        Document result = null;
        for (ServiceEndpoint ep : serviceDescriptionMap.keySet()) {
            if (se.getServiceName().equals(ep.getServiceName())) {
                result = serviceDescriptionMap.get(ep);
                break;
            }
        }
        return result;
    }

    public Set<ServiceEndpoint> getServicesForInterface(String interfaceName) {
        return servicesByInterfaceMap.get(interfaceName);
    }

    /**
     * Return the endpoint for a given service unit name
     * 
     * @param serviceUnitName
     *            The service unit name
     * @return The matching endpoint or null if no matching endpoint found
     */
    public ServiceEndpoint getServiceUnitEndpoint(String serviceUnitName) {
        return serviceUnitMap.get(serviceUnitName);
    }
    
    
    /**
     * Returns the PetalsServiceUnitHandler used to register this endpoint
     * @param ep the endpoint
     * @return a PetalsServiceUnitHandler, or null if the endpoint was not registered by this component
     */
    public PetalsServiceUnitHandler getServiceUnitHandlerForEndpoint(ServiceEndpoint ep) {
        return endpointSUHandler.get(ep);
    }

    public void init(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {

        PetalsServiceUnitHandler serviceUnitManager = handlersMap
                .get(serviceUnitName);
        if (serviceUnitManager != null) {
            serviceUnitManager.init(serviceUnitName, serviceUnitRootPath);
        } else {
            logger.log(Level.SEVERE, HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
            throw new DeploymentException(HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
        }
    }

    /**
     * Remove the service description for an endpoint 
     * @param ep the service endpoint for which the description will be removed
     */
    public void removeServiceDescription (ServiceEndpoint ep) {
        serviceDescriptionMap.remove(ep);
    }

    public void removeServiceForInterface(ServiceEndpoint serviceEndpoint) {
        for (Set<ServiceEndpoint> endpoints : servicesByInterfaceMap.values()){
            endpoints.remove(serviceEndpoint);
        }
    }
    
    /**
     * Remove the service description for an endpoint 
     * @param ep the service endpoint for which the description will be removed
     */
    public void removeServiceUnitEndpoint (String suName) {
        serviceUnitMap.remove(suName);
    }
    
    /**
     * Remove the PetalsServiceUnitHandler for a ServiceEnpoint
     * @param ep the endpoint
     */
    public void removeSUHandlerForEndpoint(ServiceEndpoint ep) {
        endpointSUHandler.remove(ep);
    }

    /**
     * Set the service description for a given endpoint
     * 
     * @param se
     *            The service endpoint
     * @param serviceDesc
     *            The dom document representation of the service description
     */
    public void setServiceDescription(ServiceEndpoint se, Document serviceDesc) {
        serviceDescriptionMap.put(se, serviceDesc);
    }
    
    /**
     * Set the endpoint for a given service unit name
     * 
     * @param serviceUnitName
     *            The name of the service unit
     * @param serviceEP
     *            The service endpoint
     */
    public void setServiceUnitEndpoint(String serviceUnitName,
            ServiceEndpoint serviceEP) {
        serviceUnitMap.put(serviceUnitName, serviceEP);
    }

    public void shutDown(String serviceUnitName) throws DeploymentException {
        PetalsServiceUnitHandler serviceUnitManager = handlersMap
                .get(serviceUnitName);
        if (serviceUnitManager != null) {
            serviceUnitManager.shutDown(serviceUnitName, this);
        } else {
            logger.log(Level.SEVERE, HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
            throw new DeploymentException(HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
        }
    }

    public void start(String serviceUnitName) throws DeploymentException {
        PetalsServiceUnitHandler serviceUnitManager = handlersMap
                .get(serviceUnitName);
        if (serviceUnitManager != null) {
            serviceUnitManager.start(serviceUnitName, this);
        } else {
            logger.log(Level.SEVERE, HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
            throw new DeploymentException(HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
        }
    }

    public void stop(String serviceUnitName) throws DeploymentException {
        PetalsServiceUnitHandler serviceUnitManager = handlersMap
                .get(serviceUnitName);
        if (serviceUnitManager != null) {
            serviceUnitManager.stop(serviceUnitName, this);
        } else {
            logger.log(Level.SEVERE, HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
            throw new DeploymentException(HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
        }
    }
    
    public String undeploy(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {
        PetalsServiceUnitHandler serviceUnitManager = handlersMap
                .get(serviceUnitName);
        if (serviceUnitManager != null) {
            String result = serviceUnitManager.undeploy(serviceUnitName,
                    serviceUnitRootPath);
            if (ManagementMessageUtil.getComponentTaskResult(
                    context.getComponentName(), "undeploy",
                    ManagementMessageUtil.TASK_RESULT_SUCCESS).equals(result)) {
                handlersMap.remove(serviceUnitName);
            }
            return result;
        } else {
            logger.log(Level.SEVERE, HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
            throw new DeploymentException(HANDLER_FOR_THIS_SERVICE_NOT_FOUND);
        }
    }

    /**
     * Checks if a handler with the given class name is already present in the
     * handlerSet
     * 
     * @param className
     *            Fully qualified class name
     * @return true if class name found; false otherwise
     */
    protected boolean isHandlerPresent(String className) {
        for (PetalsServiceUnitHandler suh : handlersSet) {
            if (suh.getClass().getName().equals(className)) {
                return true;
            }
        }
        return false;
    }


    /**
     * Parse the jbi descriptor file of the service unit looking for a type node containing the service unit type.
     * If no type node is found, null is returned
     * 
     * @param suRootPath the service unit root path
     * @return the value of the type node if found, else null
     */
    protected String getServiceUnitType(String suRootPath) {
        String result = null;
        File jbiXmlFile = new File(suRootPath + File.separator + "META-INF"
                + File.separator + "jbi.xml");
        if (jbiXmlFile.exists()) {
            Document configDoc = WSDLHelper.createDocumentFromWSDL(jbiXmlFile);
            Node node = XMLHelper.findChild(configDoc, "services", true);
            if (node != null) {
                if (node != null) {
                    Node typeNode = XMLHelper.findChild(node, EXTENSIONS_XMLNS,
                            "type", true);
                    if (typeNode != null
                            && typeNode.getNodeType() == Node.ELEMENT_NODE) {
                        result = typeNode.getTextContent();
                    }
                }
            }
        }
        return result;
    }
    
}
