/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XMLHelper.java 17:06:54 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.component.common.util;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.StringWriter;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.sun.org.apache.xpath.internal.NodeSet;

/**
 * Util class to manipulate Dom document
 * 
 * @author ddesjardins - eBMWebsourcing
 * @author wjoseph - eBMWebsourcing
 */
public class XMLHelper {

    /**
     * Namespace for the soap 1.1
     */

    protected XMLHelper() {
    }

    /**
     * Create a dom document from a string
     * 
     * @param source
     *            The input string representing xml flow
     * @return dom document
     */
    public static Document createDocumentFromString(String source) {
        Document doc = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory
                    .newInstance();
            factory.setNamespaceAware(true);
            DocumentBuilder builder = factory.newDocumentBuilder();
            InputStream is = new ByteArrayInputStream(source.getBytes());
            doc = builder.parse(is);
            doc.normalize();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return doc;
    }

    /**
     * Create a String result from a DOM document
     * 
     * @param document
     *            the DOM Document
     * @return a String representation of the DOM Document
     * @throws TransformerException
     */
    public static String createStringFromDOMDocument(Document document)
        throws TransformerException {
        document.normalize();
        Source source = new DOMSource(document);
        StringWriter out = new StringWriter();
        Result resultStream = new StreamResult(out);
        TransformerFactory tFactory = TransformerFactory.newInstance();
        Transformer transformer;
        transformer = tFactory.newTransformer();
        transformer.transform(source, resultStream);
        String result = out.toString();

        return result;
    }

    /**
     * Create a String result from a DOM Node
     * 
     * @param node
     *            the DOM Node
     * @return a String representation of the DOM Document
     * @throws TransformerException
     */
    public static String createStringFromDOMNode(Node node)
        throws TransformerException {
        // node.normalize();
        Source source = new DOMSource(node);
        StringWriter out = new StringWriter();
        Result resultStream = new StreamResult(out);
        TransformerFactory tFactory = TransformerFactory.newInstance();
        Transformer transformer;
        transformer = tFactory.newTransformer();
        transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
        transformer.transform(source, resultStream);
        String result = out.toString();
        return result;
    }

    /**
     * Create a DOM Node result from an XML Source
     * 
     * @param node
     *            the XML Source
     * @return a DOM representation of the XML Document
     * @throws TransformerException
     */
    public static Node createDOMNodeFromSource(Source source)
        throws TransformerException {
        DOMResult resultDom = new DOMResult();
        TransformerFactory tFactory = TransformerFactory.newInstance();
        Transformer transformer;
        transformer = tFactory.newTransformer();
        transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
        transformer.transform(source, resultDom);
        Node resultNode = resultDom.getNode();
        return resultNode;
    }

    /**
     * Search for a child with the given nodeName. If recursive, search in all
     * the child of firdt level, then if not found, search in the 2nd level of
     * the first child, ...
     * 
     * @param parent
     *            parent node
     * @param nodeName
     *            node name
     * @param recursive
     *            boolean to know if we got through the xml tree
     * @return a node
     */
    public static Node findChild(Node parent, String nodeName, boolean recursive) {
        parent.normalize();
        Node result = null;
        if (parent != null && nodeName != null) {
            NodeList nl = parent.getChildNodes();
            for (int i = 0; i < nl.getLength() && result == null; i++) {
                if (nodeName.equals(nl.item(i).getNodeName())) {
                    result = nl.item(i);
                }
            }
            // now, search recursively if required
            if (result == null && recursive) {
                for (int i = 0; i < nl.getLength() && result == null; i++) {
                    result = findChild(nl.item(i), nodeName, true);
                }
            }
        }
        return result;
    }

    /**
     * Search for a child with the given nodeName. If recursive, search in all
     * the child of firdt level, then if not found, search in the 2nd level of
     * the first child, ...
     * 
     * @param parent
     *            parent node
     * @param namespaceURI
     *            The namespaceURI of the node
     * @param nodeName
     *            node name
     * @param recursive
     *            boolean to know if we got through the xml tree
     * @return a node
     */
    public static Node findChild(Node parent, String namespaceURI,
            String nodeName, boolean recursive) {
        parent.normalize();
        String prefix = getPrefixForNamespaceURI(parent, namespaceURI, false);
        return (findChild(parent, prefix + nodeName, recursive));
    }

    /**
     * Search for children with the given nodeName. If recursive, search in all
     * the children of first level, then search in the 2nd level of the first
     * children, ...
     * 
     * @param parent
     *            parent node
     * @param nodeName
     *            node name
     * @param recursive
     *            boolean to know if we got through the xml tree
     * @return a node list of nodes with this name
     */
    public static NodeList findChildren(Node parent, String nodeName,
            boolean recursive) {
        parent.normalize();
        NodeSet nodeList = new NodeSet();
        if (parent != null && nodeName != null) {
            NodeList nl = parent.getChildNodes();
            for (int i = 0; i < nl.getLength(); i++) {
                if (nodeName.equals(nl.item(i).getNodeName())) {
                    nodeList.addElement(nl.item(i));
                }
            }
            // now, search recursively if required
            if (recursive) {
                for (int i = 0; i < nl.getLength(); i++) {
                    nodeList = findChildren(nl.item(i), nodeName, nodeList);
                }
            }
        }
        return nodeList;
    }

    /**
     * Return the value of the attribute in the node
     * 
     * @param n
     *            the node
     * @param attName
     *            the name of the attribute
     * @return the value of the attribute, null if not found
     */
    public static String getAttributeValue(Node n, String attName) {
        NamedNodeMap atts = n.getAttributes();
        Node att = atts.getNamedItem(attName);
        if (att != null) {
            return att.getNodeValue();
        }
        return null;
    }

    /**
     * Return the first child of a node, regardless <i>text</i> node
     * 
     * @param node
     * @return
     */
    public static Node getFirstChild(Node node) {
        node.normalize();
        Node result = node.getFirstChild();
        while (result != null && result.getNodeType() == Node.TEXT_NODE) {
            result = result.getNextSibling();
        }
        return result;
    }

    /**
     * Return the next sibling of a node, regardless <i>text</i> node
     * 
     * @param node
     * @return
     */
    public static Node getNextSibling(Node node) {
        node.normalize();
        Node result = node.getNextSibling();
        while (result != null && result.getNodeType() == Node.TEXT_NODE) {
            result = result.getNextSibling();
        }
        return result;
    }

    /**
     * Search a document to see if a namespace is declared in it and if it is
     * returns this namespace URI
     * 
     * @param node
     *            The node for which we will lookup the prefix
     * @param namespaceURI
     *            The URI for which we lookup the prefix
     * @param deep
     *            false if we want to search all the document
     * @return the prefix if the namespace was found, else null
     */
    public static String getPrefixForNamespaceURI(Node node,
            String namespaceURI, boolean deep) {
        String result = "";
        if (!deep) {
            if (namespaceURI != null && !"".equals(namespaceURI)) {
                while (node.getParentNode() != null) {
                    node = node.getParentNode();
                }
                // Search in root node attributes
                NamedNodeMap attributes = node.getAttributes();
                int i = 0;
                if (attributes != null) {
                    while (i < attributes.getLength()) {
                        Node attr = attributes.item(i++);
                        if (namespaceURI.equals(attr.getNodeValue())) {
                            String nodeName = attr.getNodeName();
                            if (nodeName.startsWith("xmlns:")) {
                                result = nodeName.replaceFirst("xmlns:", "")
                                        + ":";
                                return result;
                            } else if (nodeName.startsWith("xmlns")) {
                                return result;
                            }
                        }
                    }
                }
                // Search in child nodes attributes
                i = 0;
                NodeList nl = node.getChildNodes();
                while (i < nl.getLength()) {
                    Node tmpNode = nl.item(i++);
                    String prefix = getPrefixForNamespaceURI(tmpNode,
                            namespaceURI, true);
                    if (prefix != null) {
                        return prefix;
                    }
                }
            }
        } else {
            // Search in root node attributes
            NamedNodeMap attributes = node.getAttributes();
            int i = 0;
            if (attributes != null) {
                while (i < attributes.getLength()) {
                    Node attr = attributes.item(i++);
                    if (namespaceURI.equals(attr.getNodeValue())) {
                        String nodeName = attr.getNodeName();
                        if (nodeName.startsWith("xmlns:")) {
                            result = nodeName.replaceFirst("xmlns:", "") + ":";
                            return result;
                        } else if (nodeName.startsWith("xmlns")) {
                            return result;
                        }
                    }
                }
            }
            // Search in child nodes attributes
            i = 0;
            NodeList nl = node.getChildNodes();
            while (i < nl.getLength()) {
                Node tmpNode = nl.item(i++);
                String prefix = getPrefixForNamespaceURI(tmpNode, namespaceURI,
                        true);
                if (prefix != null) {
                    return prefix;
                }
            }
            return null;
        }
        return result;
    }

    /**
     * Search a document to see if a prefix is declared in it and if it is
     * returns its namespace URI
     * 
     * @param node
     *            The node for which we will lookup the Uri
     * @param pref
     *            The prefix for which we lookup the Uri
     * @param deep
     *            false if we want to search all the document
     * @return the Uri if the prefix was found, else null
     */
    public static String getNamespaceURIForPrefix(Node node, String pref,
            boolean deep) {
        String result = "";
        if (!deep) {
            if (pref != null && !"".equals(pref)) {
                while (node.getParentNode() != null) {
                    node = node.getParentNode();
                }
                // Search in root node attributes
                NamedNodeMap attributes = node.getAttributes();
                int i = 0;
                if (attributes != null) {
                    while (i < attributes.getLength()) {
                        Node attr = attributes.item(i++);
                        if (attr.getNodeName().startsWith("xmlns:" + pref)) {
                            result = attr.getNodeValue();
                            return result;
                        }
                    }
                }
                // Search in child nodes attributes
                i = 0;
                NodeList nl = node.getChildNodes();
                while (i < nl.getLength()) {
                    Node tmpNode = nl.item(i++);
                    String uri = getNamespaceURIForPrefix(tmpNode, pref, true);
                    if (uri != null) {
                        return uri;
                    }
                }
            }
        } else {
            // Search in root node attributes
            NamedNodeMap attributes = node.getAttributes();
            int i = 0;
            if (attributes != null) {
                while (i < attributes.getLength()) {
                    Node attr = attributes.item(i++);
                    if (attr.getNodeName().startsWith("xmlns:" + pref)) {
                        result = attr.getNodeValue();
                        return result;
                    }
                }
            }
            // Search in child nodes attributes
            i = 0;
            NodeList nl = node.getChildNodes();
            while (i < nl.getLength()) {
                Node tmpNode = nl.item(i++);
                String uri = getNamespaceURIForPrefix(tmpNode, pref, true);
                if (uri != null) {
                    return uri;
                }
            }
            return null;
        }
        return result;
    }

    /**
     * Return the <i>text</i> element of a node
     * 
     * @param node
     * @return
     */
    public static String getTextContent(Node node) {
        node.normalize();
        NodeList list = node.getChildNodes();
        for (int i = 0; i < list.getLength(); i++) {
            if (list.item(i).getNodeType() == Node.TEXT_NODE) {
                return list.item(i).getNodeValue();
            }
        }
        return null;
    }

    /**
     * Transform an XML {@link Node} into a String
     * 
     * @param node
     *            the XML {@link Node} to parse
     * @return the resulting String, null if node is null
     * @throws TransformerException
     *             if errors occured during transformation
     */
    public static String parseToString(Node node) throws TransformerException {
        String result = null;
        if (node != null) {
            node.normalize();
            Transformer transformer = TransformerFactory.newInstance()
                    .newTransformer();
            StringWriter stringWriter = new StringWriter(128);
            transformer.transform(new DOMSource(node), new StreamResult(
                    stringWriter));
            StringBuffer buffer = stringWriter.getBuffer();
            result = buffer.toString();
        }
        return result;
    }

    /**
     * Search recursively for children with the given nodeName.
     * 
     * @param parent
     *            parent node
     * @param nodeName
     *            node name
     * @param nodeList
     *            Set of nodes to be completed
     * @return a node list of nodes with this name
     */
    private static NodeSet findChildren(Node parent, String nodeName,
            NodeSet nodeList) {
        if (parent != null && nodeName != null) {
            NodeList nl = parent.getChildNodes();
            for (int i = 0; i < nl.getLength(); i++) {
                if (nodeName.equals(nl.item(i).getNodeName())) {
                    nodeList.addElement(nl.item(i));
                }
            }
            for (int i = 0; i < nl.getLength(); i++) {
                nodeList = findChildren(nl.item(i), nodeName, nodeList);
            }
        }
        return nodeList;
    }

}
