/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JBIListener.java 154 27 sept. 06 alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.axis2.listener.jbi;

import java.io.StringReader;
import java.net.URI;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axiom.soap.SOAP11Constants;
import org.apache.axiom.soap.SOAP12Constants;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.client.Options;
import org.apache.axis2.client.ServiceClient;
import org.apache.axis2.description.AxisOperation;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.OutInAxisOperation;
import org.apache.axis2.description.OutOnlyAxisOperation;
import org.apache.axis2.description.RobustOutOnlyAxisOperation;
import org.apache.axis2.transport.http.HTTPConstants;
import org.apache.axis2.transport.http.HttpTransportProperties;
import org.objectweb.petals.component.common.HandlingException;
import org.objectweb.petals.component.common.MEPConstants;
import org.objectweb.petals.component.common.su.SimpleServiceUnitManager;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.tools.jbicommon.descriptor.Extensions;

/**
 * listener for incoming JBI messages
 * 
 * @author alouis - EBM Websourcing
 * 
 */
public class JBIListener implements
		org.objectweb.petals.component.common.bc.JBIListener {

	protected Logger log;

	protected SimpleServiceUnitManager bindingSUM;

	public static final String SOAP_VERSION_11 = "11";

	public static final String SOAP_VERSION_12 = "12";

	public static final String ANONYMOUS_PROXY = "anonymous:anonymous:anonymous";

	/**
	 * Creates a new instance of JBIListener
	 * 
	 * @param bindingSUM
	 * @param log
	 */
	public JBIListener(SimpleServiceUnitManager bindingSUM, Logger log) {
		super();
		this.log = log;
		this.bindingSUM = bindingSUM;
	}

	/**
	 * Called by the JBI channel listener when a jbi message is accepted. A
	 * remote webservice call is performed.
	 * 
	 * @param address
	 * @param in
	 * @param out
	 * @param operation
	 * @param messageExchangePattern
	 * @return true if a response has been provided by the external webservice
	 * @throws MessagingException
	 */
	public boolean onJBIMessage(String address,
			MessageExchangeWrapper exchange, Extensions extensions)
			throws HandlingException {

		log.log(Level.FINE, "a JBI message is received");
		boolean responseSet = false;

		// Get the incoming Normalized message
		NormalizedMessage in = exchange.getInMessage();
		if (in == null) {
			throw new HandlingException(
					"Message exchange must handle an In normalized message");
		}

		// Get the outgoing Normalized message
		NormalizedMessage out = exchange.getOutMessage();

		// Get operation
		QName operation = exchange.getOperation();

		// Get MEP
		URI mep = exchange.getExchangePattern();

		try {
			// create Options for the stub
			Options options = createOptions(address, operation, extensions);

			ServiceClient serviceClient = createServiceClient(mep, operation);
			serviceClient.setOptions(options);

			// transform xml content to a string
			String content = SourceHelper.createString(in.getContent());

			// create the in body
			OMElement inBodyElement = createInMessage(content);

			// set Addressing mode
			serviceClient.engageModule(new QName(Constants.MODULE_ADDRESSING));
			// send as InOnly message
			if (MEPConstants.IN_ONLY_PATTERN.value().equals(mep)) {
				serviceClient.fireAndForget(operation, inBodyElement);
				responseSet = false;

			}
			// send as RobustInOnly message
			else if (MEPConstants.ROBUST_IN_ONLY_PATTERN.value().equals(mep)) {
				// If the robust in only exchange throws a Fault
				// It will be catch by the component-common framework
				serviceClient.sendRobust(operation, inBodyElement);
				responseSet = false;

			}
			// send as In(optional)Out message
			else if (MEPConstants.IN_OPTIONAL_OUT_PATTERN.value().equals(mep)
					|| MEPConstants.IN_OUT_PATTERN.value().equals(mep)) {
				OMElement outBodyElement = serviceClient.sendReceive(operation,
						inBodyElement);
				// FIXME : The following method seems to work better; the
				// previous one rise a operation not found exception
				// OMElement outBodyElement = serviceClient
				// .sendReceive(inBodyElement);

				// if msg exchange required a response, set it
				if (outBodyElement != null
						&& outBodyElement.toString().length() > 0) {
					Source responseSource = SourceHelper
							.createSource(outBodyElement.toString());
					if (out == null) {
						out = exchange.createMessage();
						exchange.setOutMessage(out);
					}
					out.setContent(responseSource);
					responseSet = true;
				}
			} else
				throw new MessagingException(
						"JBIListener - MEP not recognized.");
		} catch (Throwable e) {
			log.log(Level.WARNING, e.getMessage(), e);
			throw new HandlingException(e);
		}
		return responseSet;
	}

	/**
	 * Create a ServiceClient with an AxisService set with the good operation
	 * 
	 * @param mep
	 *            the message exchange pattern used. Non null
	 * @param operation
	 *            the target operation QName. Non null
	 * @return a ServiceClient. Not null.
	 * @throws HandlingException
	 */
	private ServiceClient createServiceClient(URI mep, QName operation)
			throws HandlingException {

		AxisService service = new AxisService("anonService" + this.hashCode()
				+ System.currentTimeMillis());
		AxisOperation axisOperation = null;
		if (MEPConstants.IN_ONLY_PATTERN.value().equals(mep)) {
			axisOperation = new OutOnlyAxisOperation(operation);
		} else if (MEPConstants.ROBUST_IN_ONLY_PATTERN.value().equals(mep)) {
			axisOperation = new RobustOutOnlyAxisOperation(operation);
		} else if (MEPConstants.IN_OPTIONAL_OUT_PATTERN.value().equals(mep)
				|| MEPConstants.IN_OUT_PATTERN.value().equals(mep)) {
			axisOperation = new OutInAxisOperation(operation);
		}
		service.addOperation(axisOperation);

		ServiceClient client = null;
		try {
			client = new ServiceClient(null, service);
		} catch (AxisFault e) {
			throw new HandlingException("Can't create ServiceClient", e);
		}
		return client;
	}

	/**
	 * Create an Axis OMElement representing the given body TODO create from a
	 * Source element, not a string
	 * 
	 * @param body
	 * @return
	 * @throws FactoryConfigurationError
	 * @throws XMLStreamException
	 */
	protected static OMElement createInMessage(String body)
			throws XMLStreamException, FactoryConfigurationError {
		XMLStreamReader parser = XMLInputFactory.newInstance()
				.createXMLStreamReader(new StringReader(body));
		StAXOMBuilder builder = new StAXOMBuilder(parser);
		return builder.getDocumentElement();

	}

	protected Options createOptions(String address, QName operation,
			Extensions extensions) {

		// Get the soapEnvelopeNamespaceURI version to use - optional, default
		// is 1.1 version
		String soapEnvelopeNamespaceURI = retrieveSoapEnvelopeNamespaceURI(extensions);

		// get proxy settings if they are defined in the extensions
		HttpTransportProperties.ProxyProperties proxyProperties = retrieveProxySettings(extensions);

		Options options = new Options();

		// set destination address
		options.setTo(new EndpointReference(address));

		// set operation
		options.setAction(operation.toString());

		options.setSoapVersionURI(soapEnvelopeNamespaceURI);
		options.setTransportInProtocol(Constants.TRANSPORT_HTTP);

		// set proxy settings
		if (proxyProperties != null) {
			options.setProperty(HTTPConstants.PROXY, proxyProperties);
		}

		// TODO set From address
		// options.setFrom(new EndpointReference(fromAddress));

		// FIXME
		// options.setProperty(
		// org.apache.axis2.transport.http.HTTPConstants.CHUNKED,
		// org.apache.axis2.Constants.VALUE_FALSE);
		return options;
	}

	/**
	 * Retrieve, in the jbi.xml descriptor, the SOAP version to use for
	 * SoapEnvelopeNamespaceURI Search a <em>soap-version</em> tag. <br>
	 * Legal values are <b>11</b> and <b>12</b>. 11 is used if no tag is
	 * found.
	 * 
	 * @param extensions
	 *            must be non null. Contains a <em>soap-version</em> optional
	 *            tag.
	 * @return the Axis2 soapEnvelopeNamespaceURI constant.
	 * @see org.apache.axiom.soap.SOAP11Constants.SOAP_ENVELOPE_NAMESPACE_URI
	 * @see org.apache.axiom.soap.SOAP12Constants.SOAP_ENVELOPE_NAMESPACE_URI
	 */
	protected String retrieveSoapEnvelopeNamespaceURI(Extensions extensions) {
		String soapEnvelopeNamespaceURI = null;
		String soapVersion = PetalsExtensionsUtil
				.extractValueFromKeyValueExtension(extensions, "soap-version");

		if (SOAP_VERSION_12.equals(soapVersion)) {
			soapEnvelopeNamespaceURI = SOAP12Constants.SOAP_ENVELOPE_NAMESPACE_URI;
		} else {
			soapEnvelopeNamespaceURI = SOAP11Constants.SOAP_ENVELOPE_NAMESPACE_URI;
		}
		return soapEnvelopeNamespaceURI;
	}

	/**
	 * Retrieve the proxy settings from the extensions. The proxy-host value is
	 * required. The other ones will be set to default values by Axis if they
	 * have not been setted by the SU.
	 * 
	 * 
	 * @param extensions
	 *            must be not null. Contains <em>proxy-*</em> optional tags.
	 * 
	 * @return the proxy settings if they are present in the extensions, null
	 *         otherwise.
	 * @see org.apache.axis2.transport.http.HttpTransportProperties.ProxyProperties
	 * 
	 */
	protected HttpTransportProperties.ProxyProperties retrieveProxySettings(
			Extensions extensions) {

		HttpTransportProperties.ProxyProperties proxyProperties = null;

		if (PetalsExtensionsUtil.hasKeyValueExtension(extensions, "proxy-host")) {
			// proxy host is required, if it is not set the proxy mode is not
			// activated

			proxyProperties = new HttpTransportProperties.ProxyProperties();

			String domain = PetalsExtensionsUtil
					.extractValueFromKeyValueExtension(extensions,
							"proxy-domain");
			String password = PetalsExtensionsUtil
					.extractValueFromKeyValueExtension(extensions,
							"proxy-password");
			String proxyHost = PetalsExtensionsUtil
					.extractValueFromKeyValueExtension(extensions, "proxy-host");

			int proxyPort = -1;
			String tmp = PetalsExtensionsUtil
					.extractValueFromKeyValueExtension(extensions, "proxy-port");
			if (tmp != null) {
				try {
					Integer.parseInt(tmp);
				} catch (NumberFormatException e) {
				}
			}

			String proxyUserName = PetalsExtensionsUtil
					.extractValueFromKeyValueExtension(extensions, "proxy-user");

			proxyProperties.setDomain(domain);
			proxyProperties.setPassWord(password);
			proxyProperties.setProxyName(proxyHost);
			proxyProperties.setProxyPort(proxyPort);
			proxyProperties.setUserName(proxyUserName);
		}

		return proxyProperties;
	}

}
