package org.objectweb.petals.component.common;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.component.ComponentContext;
import javax.jbi.component.ComponentLifeCycle;
import javax.jbi.component.ServiceUnitManager;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;
import javax.xml.transform.Source;

import org.objectweb.petals.component.common.listener.MessageExchangeListener;
import org.objectweb.petals.component.common.listener.MessageExchangeProcessor;
import org.objectweb.petals.component.common.listener.ProcessingResult;
import org.objectweb.petals.component.common.su.ServiceUnitListener;
import org.objectweb.petals.component.common.su.SimpleServiceUnitManager;
import org.objectweb.petals.component.common.util.ComponentLogger;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.component.common.util.MessageExchangeWrapperImpl;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.tools.jbicommon.descriptor.Consumes;
import org.objectweb.petals.tools.jbicommon.descriptor.Extensions;
import org.objectweb.petals.tools.jbicommon.descriptor.Provides;
import org.objectweb.petals.tools.jbicommon.util.StringHelper;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

/**
 * An abstract class that can be used to easily create a new JBI Component. It
 * implements Component, ComponentLifeCycle and MessageExchangeProcessor.
 * 
 * @author ofabre
 * 
 */
public abstract class AbstractComponent implements Component,
    ComponentLifeCycle, MessageExchangeProcessor {

    private static final String MESSAGE_EXCHANGE_SETTING_ERROR = "Error during message exchange setting";

    /**
     * The component's delivery channel
     */
    private DeliveryChannel channel;

    /**
     * The Component's Context
     */
    private ComponentContext context;

    /**
     * The message exchange listener
     */
    private MessageExchangeListener listener;

    /**
     * The component's logger
     */
    private Logger log; // NOPMD by ofabre

    /**
     * The component's service unit manager
     */
    private SimpleServiceUnitManager suManager;

    /**
     * The ServiceUnit lifecycle listener
     */
    private ServiceUnitListener serviceUnitListener;

    public ComponentLifeCycle getLifeCycle() {
        return this;
    }

    public Document getServiceDescription(ServiceEndpoint endpoint) {
        return suManager.getServiceDescription(endpoint);
    }

    public ServiceUnitManager getServiceUnitManager() {
        return suManager;
    }

    public ServiceUnitListener getServiceUnitListener() {
        return serviceUnitListener;
    }

    public void setServiceUnitListener(ServiceUnitListener serviceUnitListener) {
        this.serviceUnitListener = serviceUnitListener;
    }

    public boolean isExchangeWithConsumerOkay(ServiceEndpoint endpoint,
        MessageExchange exchange) {
        return true;
    }

    public boolean isExchangeWithProviderOkay(ServiceEndpoint endpoint,
        MessageExchange exchange) {
        return true;
    }

    public ServiceEndpoint resolveEndpointReference(DocumentFragment epr) {
        return null;
    }

    /**
     * Returns null by default as this class does not provides JMX extension
     * MBean
     * 
     * @see javax.jbi.component.ComponentLifeCycle#getExtensionMBeanName()
     */
    public ObjectName getExtensionMBeanName() {
        return null;
    }

    public void init(ComponentContext compContext) throws JBIException {
        this.context = compContext;
        Logger logger = context.getLogger("", getResourceBundleName());
        this.log = new ComponentLogger(logger, logger.getName(), logger
            .getResourceBundleName(), context.getComponentName());
        channel = context.getDeliveryChannel();
        suManager = createServiceUnitManager();
        log.log(Level.INFO, "Component initialized");
        init();
    }

    /**
     * This method is called at the end of the component classical
     * initialization. It could be overidden to make additional initialization.
     * When this method is called, the component's delivery channel, context,
     * logger and service unit manager are available.
     * 
     */
    protected void init() {

    }

    /**
     * Can be overidden to return a custom resource bundle name
     * 
     * @return the resource bundle name or null if no resource bundle name has
     *         to be used (default value)
     */
    protected String getResourceBundleName() {
        return null;
    }

    public void shutDown() throws JBIException {
        log.log(Level.INFO, "shutdown");
        listener = null; // NOPMD frees the memory while component is
        // shutdown
        context = null; // NOPMD idem
        log = null; // NOPMD idem
        channel = null; // NOPMD idem
        suManager = null; // NOPMD idem

    }

    public void start() throws JBIException {
        log.log(Level.INFO, "start");
        listener = createMessageExchangeListener(this);
        listener.listen();

    }

    public void stop() throws JBIException {
        log.log(Level.INFO, "stop");
        // FIXME : call terminate or close?
        this.listener.terminate();

    }

    public ProcessingResult process(final MessageExchange exchange)
        throws JBIException {
        if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
            if (MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {
                if (MEPConstants.IN_ONLY_PATTERN.value().equals(
                    exchange.getPattern())) {
                    processInOnly(exchange);
                } else if (MEPConstants.ROBUST_IN_ONLY_PATTERN.value().equals(
                    exchange.getPattern())) {
                    processRobustInOnly(exchange);
                } else if (MEPConstants.IN_OUT_PATTERN.value().equals(
                    exchange.getPattern())
                    || MEPConstants.IN_OPTIONAL_OUT_PATTERN.value().equals(
                        exchange.getPattern())) {
                    processInOut(exchange);
                } else {
                    MessagingException e = new MessagingException(
                        "MessageExchangePattern not recognized :"
                            + exchange.getPattern());
                    exchange.setError(e);
                    channel.send(exchange);
                    throw e;
                }
                channel.send(exchange);
            }
        } else {
            log.log(Level.INFO, "No processing on Unactive MEP");
        }
        return ProcessingResult.SUCESS;
    }

    /**
     * Process "in-only" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws PEtALSComponentSDKException
     *             When errors occur during processing
     */
    private void processInOnly(final MessageExchange exchange)
        throws PEtALSComponentSDKException {
        try {
            Extensions extensions = getExchangeEpJBIExtensions(exchange);
            onExchange(new MessageExchangeWrapperImpl(exchange), extensions);
        } catch (HandlingException e) {
            throw new PEtALSComponentSDKException(
                "Processing error but InOnly message can not contains Fault.",
                e);
        } finally {
            try {
                if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
                    exchange.setStatus(ExchangeStatus.DONE);
                }
            } catch (MessagingException e) {
                log.severe(MESSAGE_EXCHANGE_SETTING_ERROR);
                throw new PEtALSComponentSDKException(
                    MESSAGE_EXCHANGE_SETTING_ERROR, e);
            }
        }
    }

    /**
     * Process "robust-in-only" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws PEtALSComponentSDKException
     *             When errors occur during processing
     * 
     */
    private void processRobustInOnly(final MessageExchange exchange)
        throws PEtALSComponentSDKException {
        try {
            Extensions extensions = getExchangeEpJBIExtensions(exchange);
            onExchange(new MessageExchangeWrapperImpl(exchange), extensions);
        } catch (HandlingException e) {
            try {
                log.log(Level.SEVERE,
                    "Error during message exchange processing", e);
                exchange.setFault(createFault(e, exchange));
            } catch (MessagingException e1) {
                log.severe(MESSAGE_EXCHANGE_SETTING_ERROR);
                throw new PEtALSComponentSDKException(
                    MESSAGE_EXCHANGE_SETTING_ERROR, e1);
            }
        }
    }

    /**
     * Process "in-out" MessageExchange
     * 
     * @param exchange
     *            the MessageExchange
     * @throws PEtALSComponentSDKException
     *             When errors occur during processing
     * 
     */
    private void processInOut(final MessageExchange exchange)
        throws PEtALSComponentSDKException {
        try {
            if (!ackFaultReception(exchange)) {

                Extensions extensions = getExchangeEpJBIExtensions(exchange);

                boolean outTreated = onExchange(new MessageExchangeWrapperImpl(
                    exchange), extensions);

                if (!outTreated) {
                    exchange.setStatus(ExchangeStatus.DONE);
                }
            }
        } catch (HandlingException e) {
            try {
                log.log(Level.SEVERE,
                    "Error during message exchange processing", e);
                exchange.setFault(createFault(e, exchange));
            } catch (MessagingException e1) {
                log.severe(MESSAGE_EXCHANGE_SETTING_ERROR);
                throw new PEtALSComponentSDKException(
                    MESSAGE_EXCHANGE_SETTING_ERROR, e1);
            }
        } catch (MessagingException e) {
            log.severe(MESSAGE_EXCHANGE_SETTING_ERROR);
            throw new PEtALSComponentSDKException(
                MESSAGE_EXCHANGE_SETTING_ERROR, e);
        }
    }

    private Extensions getExchangeEpJBIExtensions(final MessageExchange exchange) {
        ServiceEndpoint ep = exchange.getEndpoint();
        Extensions extensions = null;
        if (ep != null) {
            Provides provides = suManager.getJBIDescription(ep);
            if (provides != null) {
                extensions = provides.getExtensions();
            }
        }
        return extensions;
    }

    /**
     * Creates a Fault from an exception
     * 
     * @param e
     *            The {@link Exception} used to create a new {@link Fault}
     * @param exchange
     *            The {@link MessageExchange} that allows to create a new
     *            {@link Fault}
     * @return the created Fault
     * @throws PEtALSComponentSDKException
     *             if an error occured during Fault creation or filling
     */
    private Fault createFault(final JBIException e,
        final MessageExchange exchange) throws PEtALSComponentSDKException {
        String msg = "";
        Fault f = null;
        try {
            f = exchange.createFault();
        } catch (MessagingException e1) {
            msg = "Error during Fault creation";
            log.severe(msg);
            throw new PEtALSComponentSDKException(msg, e1);
        }
        // String faultString = SourceHelper.createSoapFault(e, null);
        String faultString = SourceHelper.createSoapFault(e, "0");
        Source content = SourceHelper.createSource(faultString);
        try {
            f.setContent(content);
        } catch (MessagingException e1) {
            msg = "Error during Fault content setting";
            log.severe(msg);
            throw new PEtALSComponentSDKException(msg, e1);
        }
        return f;
    }

    /**
     * If the received exchange is a fault response from the consumer,
     * acknowledge it : status DONE
     * 
     * @param ex
     *            The processed {@link MessageExchange}
     * @return true if the ack. of the Fault is done; false if no fault is
     *         present in the exchange
     * @throws MessagingException
     */
    protected boolean ackFaultReception(final MessageExchange ex)
        throws MessagingException {
        boolean result = false;

        if (ex.getFault() != null) {
            ex.setStatus(ExchangeStatus.DONE);

            result = true;
        }
        return result;
    }

    /**
     * Called to process a message exchange. MessageExchange handles all
     * information of the exchange between provider and consumer (operation,
     * MessageExchangePattern, payload...). Optionnal extensions can be used to
     * handle complementary informations
     * 
     * @param exchangeWrapper
     *            the message exchange to process, wrapped to add some util
     *            methods. Not null
     * @param extensions
     *            JBI descriptor extensions. The message exchange targets a
     *            specific endpoint, this endpoint has been activated during the
     *            processing of a provides node of a service unit jbi
     *            descriptor. Theses extensions are the extensions of this
     *            provides node. Can be null
     * @return true if the out response has been set, false otherwise (Useful
     *         for a in-out mep)
     * 
     * @see MEPConstants
     * @throws HandlingException
     *             a Fault will be generated with the exception, the "out" will
     *             not be treated
     */
    protected abstract boolean onExchange(MessageExchangeWrapper exchangeWrapper,
        Extensions extensions) throws HandlingException;

    /**
     * Can be overidden to provide a specific service unit manager
     * 
     * @return the service unit manager used by this component.
     */
    protected SimpleServiceUnitManager createServiceUnitManager() {
        return new SimpleServiceUnitManager(this, context, log);
    }

    /**
     * Creates a new MessageExchangeListener
     * 
     * @param processor
     *            the messageExchange processor
     * @return the created {@link MessageExchangeListener}
     */
    protected MessageExchangeListener createMessageExchangeListener(
        final MessageExchangeProcessor processor) {
        return new MessageExchangeListener(channel, processor,
            MessageExchangeListener.IgnoredStatus.DONE_AND_ERROR_IGNORED, 0,
            context.getComponentName(), log);
    }

    /**
     * Returns the component logger
     * 
     * @return component logger. Not null.
     */
    public Logger getLogger() {
        return log;
    }

    /**
     * Returns the component delivery channel, used to send messages to petals
     * container
     * 
     * @return the delivery channel. Not null.
     */
    public DeliveryChannel getChannel() {
        return channel;
    }

    /**
     * The component context
     * 
     * @return the component context. Not null.
     */
    public ComponentContext getContext() {
        return context;
    }

    /**
     * Send a message to the endpoint associated to the mapped address. For
     * exemple, this method can be used by a listener, that listen to messages
     * from external services (linked to the given address), in order to send
     * these messages to internal JBI Components.
     * 
     * @param address
     *            the distant address of the external consumer service. Must be
     *            non null and non empty.
     * @param messageExchange
     *            the message exchange to be sent to the internal JBI Component.
     *            Must be non null.
     * @throws MessagingException
     *             if the address has no endpoint associated to it, or if an
     *             error occurs while sending the message on the container
     */
    public void sendMessage(final String address,
        final MessageExchange messageExchange) throws MessagingException {

        if (StringHelper.isNullOrEmpty(address)) {
            throw new MessagingException(
                "You must provide a non null and non empty address");
        }
        if (messageExchange == null) {
            throw new MessagingException(
                "You must provide a non null MessageExchange");
        }

        Consumes consumes = suManager.getConsumesFromAddress(address);

        ServiceEndpoint ep = null;
        if (consumes.getEndpointName() != null
            && consumes.getServiceName() != null) {
            ep = context.getEndpoint(consumes.getServiceName(), consumes
                .getEndpointName());
        }

        messageExchange.setInterfaceName(consumes.getInterfaceName());
        messageExchange.setService(consumes.getServiceName());
        messageExchange.setEndpoint(ep);

        try {
            channel.send(messageExchange);
        } catch (JBIException e) {
            throw new MessagingException(
                "Failed to send the mapped message to the delivery channel.");
        }

    }
}
