/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: CsvSUHandler.java 1168 2006-10-12 15:50:18Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.engine.csv;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.management.DeploymentException;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.common.serviceunitmanager.handler.PetalsServiceUnitHandler;
import org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager;
import org.objectweb.petals.component.common.util.ManagementMessageUtil;
import org.objectweb.petals.component.common.util.WSDLHelper;

import org.w3c.dom.Document;

/**
 * This is the SU handler for the CSV service engine.
 * 
 * At startup, lookups for a single *.wsdl file and a single
 * *.properties file. The first provides the service name, and the
 * second provides properties that are given to the CSV
 * component and from there to the CsvProcessor to use (look at
 * the processor's documentation to learn which properties
 * may be set).
 * 
 * @version $Rev: 1068 $Date: {date}
 * @since   Petals 1.0
 * @author Marc Dutoo - Open Wide
 *
 */
public class CsvSUHandler extends PetalsServiceUnitHandler {

    /** csv SE map */
    protected Map<String, Properties> mapServiceCsv;
    /** csv SU map. Needed to be able to remove csv SE */
    protected Map<String, String> mapServiceUnitCsv;


    private CsvSUHandler() {
        super();
        SERVICE_UNIT_TYPE="csvServiceUnit";
    }

    public CsvSUHandler(ComponentContext context,
            HashMap<String, Properties> mapServiceCsv, Logger logger) {
        super(context, logger);
        this.mapServiceCsv= mapServiceCsv;
        this.mapServiceUnitCsv = new HashMap<String, String>();
        SERVICE_UNIT_TYPE="csvServiceUnit";
    }
    

    /*
     *  (non-Javadoc)
     * @see org.objectweb.petals.component.common.serviceunitmanager.handler.PetalsServiceUnitHandler#deploy(java.lang.String, java.lang.String, java.lang.String)
     */
    public String deploy(String serviceUnitName, String serviceUnitType,
            String serviceUnitRootPath) {
        if (SERVICE_UNIT_TYPE.equals(serviceUnitType)) {
            logger.info(" DEPLOY  serviceUnitName  " + serviceUnitName
                    + "  serviceUnitRootPath " + serviceUnitRootPath);
            return ManagementMessageUtil.getComponentTaskResult(context
                    .getComponentName(), "deploy",
                    ManagementMessageUtil.TASK_RESULT_SUCCESS);
        } else {
            return null;
        }
    }

    /* (non-Javadoc)
     * @see org.objectweb.petals.component.common.serviceunitmanager.handler.PetalsServiceUnitHandler#init(java.lang.String, java.lang.String)
     */
    public void init(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {
        this.serviceUnitInstallationRootPath.put(serviceUnitName, serviceUnitRootPath);
    }

    /* (non-Javadoc)
     * @see org.objectweb.petals.component.common.serviceunitmanager.handler.PetalsServiceUnitHandler#undeploy(java.lang.String, java.lang.String)
     */
    public String undeploy(String serviceUnitName, String serviceUnitRootPath)
        throws DeploymentException {
        return ManagementMessageUtil.getComponentTaskResult(context
                .getComponentName(), "undeploy",
                ManagementMessageUtil.TASK_RESULT_SUCCESS);
    }
    
    /* (non-Javadoc)
     * @see org.objectweb.petals.component.common.serviceunitmanager.handler.PetalsServiceUnitHandler#shutDown(java.lang.String, org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager)
     */
    public void shutDown(String serviceUnitName, PetalsServiceUnitManager epHandler) throws DeploymentException {
        String csv = this.mapServiceUnitCsv.remove(serviceUnitName);
        this.mapServiceCsv.remove(csv);
    }

    
    /* (non-Javadoc)
     * @see org.objectweb.petals.component.common.serviceunitmanager.handler.PetalsServiceUnitHandler#start(java.lang.String, org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager)
     */
    public void start(String serviceUnitName, PetalsServiceUnitManager epHandler) throws DeploymentException {
        logger.log(Level.FINE, "start  serviceUnitName  " + serviceUnitName);
        
        // looking for configuration files : one *.wsdl and one *.properties
        // TODO use special tags in jbi.xml to set component properties instead
        String serviceUnitRootPath = serviceUnitInstallationRootPath.get(serviceUnitName);
        File[] files = new File(serviceUnitRootPath).listFiles();
        Document serviceDesc = null;
        String propFileName = null;
        QName service=  null;
        
        for (File file : files) {
            if (file.getName().endsWith(".wsdl")) {
                // found the service definition
                serviceDesc = WSDLHelper.createDocumentFromWSDL(file);
            } else if (file.getName().endsWith(".properties")){
                // found the service configuration
                propFileName = file.getAbsolutePath();
            }
        }
        
        // activating found endpoints
        if (serviceDesc != null) {
            try {
                activateEndpointsFromJBIDescription(epHandler, serviceDesc,
                        serviceUnitName, serviceUnitRootPath);
                service = WSDLHelper.getServiceNameFromWSDLDocument(serviceDesc)[0];
            } catch (Exception ex) {
                logger.log(Level.SEVERE, FAILED_ACTIVATE_ENDPOINT + ex);
                throw new DeploymentException(FAILED_ACTIVATE_ENDPOINT, ex);
            }
        } else {
            throw new DeploymentException(INCOMPLETE_SERVICE_UNIT_PACKAGE);
        }
        
        // configuring the component for the service & properties found
        if (service!=null && propFileName != null) {
            Properties csvReaderProps = new Properties();
            try {
                csvReaderProps.load(new FileInputStream(propFileName));
            } catch (IOException e) {
                logger.log(Level.SEVERE, FAILED_ACTIVATE_ENDPOINT, e);
                throw new DeploymentException(FAILED_ACTIVATE_ENDPOINT, e);
            }
            mapServiceUnitCsv.put(serviceUnitName, service.getLocalPart());
            mapServiceCsv.put(service.getLocalPart(), csvReaderProps);
        }
    }

    
    /* (non-Javadoc)
     * @see org.objectweb.petals.component.common.serviceunitmanager.handler.PetalsServiceUnitHandler#stop(java.lang.String, org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager)
     */
    public void stop(String serviceUnitName, PetalsServiceUnitManager epHandler) throws DeploymentException {
        // When service unit is stopped, so properties of directory are deleted
        String csv = this.mapServiceUnitCsv.remove(serviceUnitName);
        this.mapServiceCsv.remove(csv);
    }
}
