/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JmsConnexion.java 676 2006-06-27 15:44:03Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.jms;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.JMSSecurityException;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.objectweb.petals.component.common.PEtALSComponentSDKException;
import org.objectweb.petals.tools.jbicommon.util.StringHelper;

/**
 * This class is used to create a JMS consumer or producer.
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.1
 * @author alouis
 */
public class JMSConnection {

    public enum Role {
        JMS_PRODUCER, JMS_CONSUMER
    };

    private Context context;

    private String connectionFactoryName;

    private String destinationName;

    private String user;

    private String password;

    private Connection connection;

    private Session session;

    private Destination destination;

    private MessageConsumer messageConsumer;

    private MessageProducer messageProducer;

    private boolean transacted;

    private boolean started = false;

    private Role role;

    /**
     * Constructor. All required information to create the producer or consumer
     * are set here.
     * 
     * @param jndiContext
     *            the context where the connectionFactory and the destination
     *            can be retrieved.
     * @param connectionFactoryName
     *            name of the jms connectionfactory
     * @param destinationName
     *            name of the destination (queue or topic)
     * @param transacted
     *            created a transacted session or not
     * @param user
     *            user to use to connect to the connection factory if required
     * @param password
     *            password to use to connect to the connection factory if
     *            required
     * @param role
     *            JSM role : consumer or producer
     */
    public JMSConnection(Context jndiContext, String connectionFactoryName,
        String destinationName, boolean transacted, String user,
        String password, Role role) {
        this.context = jndiContext;
        this.connectionFactoryName = connectionFactoryName;
        this.destinationName = destinationName;
        this.user = user;
        this.password = password;
        this.transacted = transacted;
        this.role = role;
    }

    /**
     * Retrieve the connection factory and the destination from the JNDI
     * context. Create a connection, using user/pwd if required, but do not
     * start it.
     * 
     * @throws PEtALSComponentSDKException
     */
    protected void init() throws PEtALSComponentSDKException {

        checkJNDIAccess(context);

        ConnectionFactory connFact = lookupConnectionFactory(context,
            connectionFactoryName);

        destination = lookupDestination(context, destinationName);

        connection = createConnection(connFact, user, password);

        try {
            context.close();
        } catch (NamingException e) {
            throw new PEtALSComponentSDKException("JNDI context exception.", e);
        }
    }

    /**
     * Create a consumer/producer, using the specified transacted mode. Start
     * the jms connection.
     * 
     * @throws PEtALSComponentSDKException
     *             jmsException on starting the consumer/producer, or connection
     *             already started
     */
    public void start() throws PEtALSComponentSDKException {

        if (!started) {

            init();

            try {
                session = connection.createSession(transacted,
                    Session.AUTO_ACKNOWLEDGE);

                if (Role.JMS_CONSUMER.equals(role)) {
                    try{
                        messageConsumer = session.createConsumer(destination);
                    } catch (JMSSecurityException e) {
                        throw new PEtALSComponentSDKException("Can not start JMS consumer, READ right not granted.", e);
                    }

                } else if (Role.JMS_PRODUCER.equals(role)) {
                    messageProducer = session.createProducer(destination);

                } else {
                    throw new PEtALSComponentSDKException("Unknown JMS Role :"
                        + role);

                }
                connection.start();

                started = true;

            } catch (JMSException e) {
                throw new PEtALSComponentSDKException(
                    "Can not start the JMS consumer/producer.", e);
            }

        } else {
            throw new PEtALSComponentSDKException(
                "JMS connection already started for the '" + destinationName
                    + "' destination.");

        }
    }

    /**
     * Close messageConsumer/messageProducer, session and connection.
     * 
     * @throws PEtALSComponentSDKException
     *             jmsException on close
     */
    public void stop() throws PEtALSComponentSDKException {
        try {
            if (messageConsumer != null) {
                messageConsumer.close();
                messageConsumer = null;
            }
            if (messageProducer != null) {
                messageProducer.close();
                messageProducer = null;
            }
            if (session != null) {
                session.close();
                session = null;
            }
            if (connection != null) {
                connection.close();
                connection = null;
            }
            started = false;
        } catch (JMSException e) {
            throw new PEtALSComponentSDKException(
                "Can no stop the JMS connection.", e);
        }
    }

    public MessageConsumer getMessageConsumer() {
        return messageConsumer;
    }

    public MessageProducer getMessageProducer() {
        return messageProducer;
    }

    public Session getSession() {
        return session;
    }

    public Role getRole() {
        return role;
    }

    public String getDestinationName() {
        return destinationName;
    }

    public boolean isTransacted() {
        return transacted;
    }

    /**
     * Test the JNDI access. A <code>listBindings("")</code> call is performed
     * to check if no exception is thrown.
     * 
     * @param ctx
     * @throws PEtALSComponentSDKException
     *             the JNDI can not be accessed
     */
    private static void checkJNDIAccess(Context ctx)
        throws PEtALSComponentSDKException {
        try {
            // test the JNDI access
            ctx.listBindings("");
        } catch (NamingException e) {
            String providerUrl = "[JNDI provider URL not found]";
            try {
                providerUrl = ctx.getEnvironment().get(
                    InitialContext.PROVIDER_URL).toString();
            } catch (NamingException e1) {
                throw new PEtALSComponentSDKException(
                    "Exception during JNDI Context access.", e);
            }
            throw new PEtALSComponentSDKException(
                "The JNDI can not be accessed at " + providerUrl);
        }
    }

    /**
     * Retrieve the connectionFactory in the JNDI context
     * 
     * @param ctx
     *            must be connected, must be non null
     * @param connectionFactoryName
     *            must be non null
     * @return the connectionFactory
     * @throws PEtALSComponentSDKException
     *             the factory is not found
     */
    private static ConnectionFactory lookupConnectionFactory(Context ctx,
        String connectionFactoryName) throws PEtALSComponentSDKException {
        ConnectionFactory connFact = null;
        try {
            connFact = (ConnectionFactory) ctx.lookup(connectionFactoryName);
        } catch (NamingException e) {
            throw new PEtALSComponentSDKException("ConnectionFactory '"
                + connectionFactoryName + "' not found in the JNDI context.", e);
        }
        return connFact;
    }

    /**
     * Retrieve the destination in the JNDI context
     * 
     * @param ctx
     *            must be connected, must be non null
     * @param destination
     *            must be non null
     * @return the destination
     * @throws PEtALSComponentSDKException
     *             the destination is not found
     */
    private static Destination lookupDestination(Context ctx,
        String destinationName) throws PEtALSComponentSDKException {
        Destination dest = null;
        try {
            dest = (Destination) ctx.lookup(destinationName);
        } catch (NamingException e) {
            throw new PEtALSComponentSDKException("Destination '"
                + destinationName + "' not found in the JNDI context.", e);
        }
        return dest;
    }

    /**
     * Create a JMS Connection.<br>
     * If user is null, call a <code>connectionFactory.createConnection()</code>.<br>
     * If user is not null, call a
     * <code>connectionFactory.createConnection(user,password)</code>.
     * 
     * @param connectionFactory
     *            must be non null
     * @param user
     *            can be null
     * @param password
     *            can be null
     * @return a JMS connection
     * @throws PEtALSComponentSDKException
     *             JMS security exception or connectionfactory exception
     */
    private static Connection createConnection(
        ConnectionFactory connectionFactory, String user, String password)
        throws PEtALSComponentSDKException {
        Connection conn = null;
        try {
            if (StringHelper.isNullOrEmpty(user)) {
                conn = connectionFactory.createConnection();
            } else {
                conn = connectionFactory.createConnection(user, password);
            }

        } catch (JMSSecurityException e) {
            throw new PEtALSComponentSDKException(
                "Invalid user/password to connect to the JMS connectionFactory.",
                e);
        } catch (JMSException e) {
            throw new PEtALSComponentSDKException(
                "JMS connectionFactory error.", e);
        }
        return conn;
    }
}