/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JMSMessageListener.java 12:18:02 alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.jms.incoming;

import java.util.Enumeration;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.TextMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.objectweb.petals.binding.jms.JmsBC;
import org.objectweb.petals.component.common.PEtALSComponentSDKException;
import org.objectweb.petals.component.common.util.SourceHelper;

/**
 * This class is the JMS message listener for the JmsBC component
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.1
 * @author alouis
 */
public class JMSMessageListener implements MessageListener {

    public static final String PROPERTY_OPERATION = "javax.jbi.messaging.operation";

    protected JmsBC jmsBC;
    
    protected String address;
    
    protected Logger l;
    
    protected MessageExchangeFactory exchangeFactory;

    public JMSMessageListener(JmsBC jmsBC, String address) {
        this.jmsBC=jmsBC;
        this.address = address;
        l=jmsBC.getLogger();
    }

    public void onMessage(Message jmsExchange) {
        l.log(Level.FINE, "onMessage");
        l.log(Level.INFO, "A JMS message is received from '"
            + address+ "'.");

        try {
            // 1. Only TextMessage are read.
            TextMessage textMessage;
            textMessage = (TextMessage) jmsExchange;

            // 2. Create an exchange
            InOnly exchange = createInOnly();

            // 3. transform textMessage->normalizedMessage
            transform(textMessage, exchange);

            // 4. send the exchange, use the framework's send() method
            jmsBC.sendMessage(address, exchange);
            
            l.log(Level.INFO, "the message has been sent.");

        } catch (ClassCastException e) {
            l
            .log(
                Level.SEVERE,
                "Only TextMessage are processed by petals-binding-jms component.",
                e);
        } catch (Exception e) {
            l.log(Level.SEVERE, e.getClass() + " " + e.getMessage(), e);
        }
    }

    /**
     * Create an InOnly exchange,
     * @return
     * @throws MessagingException
     */
    protected InOnly createInOnly() throws MessagingException {
        MessageExchangeFactory factory = getMessageExchangeFactory();
        InOnly exchange = factory.createInOnlyExchange();
        NormalizedMessage nm = exchange.createMessage();
        exchange.setInMessage(nm);
        return exchange;
    }



    /**
     * Return an unique MessageExchangeFactory created by the
     *  DeliveryChannel. Lazyloading pattern.
     * @return
     */
    protected MessageExchangeFactory getMessageExchangeFactory() {
        if (exchangeFactory == null) {
            exchangeFactory = jmsBC.getChannel().createExchangeFactory();
        }
        return exchangeFactory;
    }

    /**
     * Transform the TextMessage
     * into a NormalizedMessage of the JBI MsgExch .
     * The content and properties are set. 
     * @param jms
     * @param jbi
     * @throws PEtALSComponentSDKException can not create a Source object from the jms text content
     * @throws JMSException getting jms content or properties
     * @throws MessagingException setting NormalizedMessage properties or content
     */
    protected void transform(TextMessage jms, MessageExchange jbi)
    throws JMSException, MessagingException,PEtALSComponentSDKException {
        l.log(Level.FINE, "transform");

        //set content
        String content = jms.getText();
        Source source = SourceHelper.createSource(content);

        jbi.getMessage("in").setContent(source);

        // set operation
        String operation = jms.getStringProperty(PROPERTY_OPERATION);
        if (operation != null){
            jbi.setOperation(new QName(operation));
        }

        // set properties
        for (Enumeration it = jms.getPropertyNames(); it.hasMoreElements();) {
            String pName = (String) it.nextElement();
            jbi.setProperty(pName, jms.getObjectProperty(pName));
        }
    }
}
