/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JBIExchangeProcessor.java 12:18:02 alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.jms.outgoing;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jms.JMSException;
import javax.jms.JMSSecurityException;
import javax.jms.TextMessage;
import javax.xml.namespace.QName;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;

import org.objectweb.petals.binding.jms.JMSConnection;
import org.objectweb.petals.component.common.HandlingException;
import org.objectweb.petals.component.common.MEPConstants;
import org.objectweb.petals.component.common.util.MessageExchangeWrapper;
import org.objectweb.petals.tools.jbicommon.descriptor.Extensions;

/**
 * This class is the JBI message processor for the JmsBC component. It receives
 * messages from JBI and send them to a JMS destination.
 * 
 * @version $Revision: $ $Date: $
 * @since Petals 1.0
 * @author alouis
 */
public class JBIListener implements
    org.objectweb.petals.component.common.bc.JBIListener {

    private static final String PROPERTY_OPERATION = "javax.jbi.messaging.operation";

    private Logger logger;

    private Map<String, JMSConnection> jmsConnections;

    public JBIListener(Map<String, JMSConnection> jmsConnections, Logger log) {
        this.jmsConnections = jmsConnections;
        logger = log;
    }

    /**
     * Process ACTIVE messages as a PROVIDER. Only InOnly and RobustInOnly
     * messages are processed. Otherwise, set the Exchange to ERROR status and
     * send it, and throw an Exception
     * 
     * If the specified exchange endpoint is not registered, set the Exchange to
     * ERROR status and send it, and throw an Exception * Else, send a
     * <em>JMS message</em> to the corresponding <em>JMS Queue<em>,
     *   
     *   If the JMS msg has not been correctly sent
     *     set the Exchange to ERROR status and send it, and throw an Exception     *
     *      
     * @param exchange
     * @throws Exception exhange not correct, destination not found, problem while sending the JMS message,
     */
    public boolean onJBIMessage(String address, MessageExchangeWrapper exchange,
        Extensions extensions) throws HandlingException {
        logger.log(Level.FINE, "onJBIMessage");

        if (ExchangeStatus.ACTIVE.equals(exchange.getStatus())) {
            if (MessageExchange.Role.PROVIDER.equals(exchange.getRole())) {
                if (MEPConstants.IN_ONLY_PATTERN.value().equals(exchange.getPattern())
                    || MEPConstants.ROBUST_IN_ONLY_PATTERN.value().equals(exchange
                        .getPattern())) {
                    sendJMSMessage(address, exchange);
                } else {
                    Exception e = new Exception(
                        "MessageExchangePattern not allowed :"
                            + exchange.getPattern());
                    try {
                        setErrorStatus(exchange, e, true);
                    } catch (MessagingException e1) {
                        throw new HandlingException(e1);
                    }
                }
            }
        }
        return true;
    }

    /**
     * Set the exchange to "ERROR" status. An exception is created with the
     * given reason as message. This exception is set in the MessageExchange
     * object.
     * 
     * @param msgExchange
     *            current MessageExchange used to send and receive messages
     * @param ex
     *            The exception
     * @param severe
     *            Log the error at warning or severe level
     * @throws MessagingException
     *             If an error occured during calls on the JBI container
     */
    protected void setErrorStatus(MessageExchange msgExchange, Exception ex,
        boolean severe) throws MessagingException {
        msgExchange.setError(ex);
        msgExchange.setStatus(ExchangeStatus.ERROR);
        if (!severe) {
            logger.log(Level.WARNING, ex.getMessage(), ex);
        } else {
            logger.log(Level.SEVERE, ex.getMessage(), ex);
        }
    }

    /**
     * Send a JMS message to the JMS queue corresponding to the specified
     * exchange's endpoint.
     * 
     * @param address
     *            JMS destination
     * @param exchange
     * @throws Exception
     *             no JMS destination found, or problem while sending the
     *             message to JMS
     */
    protected void sendJMSMessage(String address, MessageExchange exchange)
        throws HandlingException {
        logger.log(Level.FINE, "sendJMSMessage");
        logger.log(Level.INFO, "A JBI message is received for '" + address
            + "' JMS provider.");

        // find the connection to use for this exchange
        JMSConnection conn = jmsConnections.get(address);

        if (conn == null) {
            // No such endpoint registered
            throw new HandlingException("'" + address
                + "' JMS provider not found.");
        }

        // transform the jbi msg into a jms msg
        TextMessage msg = null;
        try {
            msg = conn.getSession().createTextMessage();
            transform(exchange, msg);
        } catch (JMSException e) {
            logger.log(Level.SEVERE, "Can not create a JMS text message.", e);
            throw new HandlingException(e);
        } catch (Exception e) {
            logger.log(Level.SEVERE,
                "Can not transform the JBI message to a JMS message.", e);
            throw new HandlingException(e);
        }

        // send
        try {
            conn.getMessageProducer().send(msg);
            if(conn.isTransacted()){
                conn.getSession().commit();
            }
        } catch (JMSSecurityException e) {
            logger.log(Level.SEVERE, "Can not send the JMS message, WRITE right not granted.", e);
            throw new HandlingException(e);
        } catch (JMSException e) {
            logger.log(Level.SEVERE, "Can not send the JMS message.", e);
            throw new HandlingException(e);
        }
        logger.log(Level.INFO, "Message sent to '" + conn.getDestinationName()
            + "'.");
    }

    /**
     * Transform the NormalizedMessage of a JBI MsgExch into a TextMessage. The
     * content , properties and attachments of the IN NormalizedMessage are set.
     * 
     * @param exchange
     * @param jms
     * @throws JMSException
     *             writing properties to the TextMessage
     * @throws TransformerException
     *             NormalizedMessage content access error
     * @throws IOException
     *             NormalizedMessage content access error
     */
    protected void transform(MessageExchange exchange, TextMessage jms) throws JMSException, TransformerException, IOException {
        logger.log(Level.FINE, "transform");

        NormalizedMessage in = exchange.getMessage("in");
        QName operation = exchange.getOperation();

        // set normalized properties
        for (Iterator it = in.getPropertyNames().iterator(); it.hasNext();) {
            String pName = (String) it.next();
            jms.setObjectProperty(pName, in.getProperty(pName));
        }
        // set operation
        if (operation != null) {
            jms.setStringProperty(PROPERTY_OPERATION, operation.toString());
        }

        // set content
        StreamResult stream = new StreamResult(new ByteArrayOutputStream());
        Transformer transformer = TransformerFactory.newInstance()
            .newTransformer();
        transformer.transform(in.getContent(), stream);
        stream.getOutputStream().close();
        jms.setText(stream.getOutputStream().toString());
    }

}
