/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: MailSessionManager.java 154 28 sept. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.mail.listeners;

import java.util.Properties;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.mail.Folder;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Store;
import javax.mail.Transport;
import javax.mail.internet.MimeMessage;

import org.objectweb.petals.binding.mail.MailBCImpl;
import org.objectweb.petals.tools.jbicommon.util.StringHelper;

/**
 * This manager can be used to manage session with mail servers : create and
 * open sessions, send and receive mails
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class MailSessionManager {

    protected Logger log;

    /**
     * 
     * @param log
     *            the component logger
     */
    public MailSessionManager(Logger log) {
        super();
        this.log = log;
    }

    /**
     * Create a {@link Properties} object with given properties. Then create a
     * new {@link Session} initialized with the given properties
     * 
     * @param sessionDescriptor
     *            a {@link SessionDescriptor} build from an address URI
     * @return the Mail Server {@link Session}
     */
    public Session createSessionPropertiesFromDescriptor(
            SessionDescriptor sessionDescriptor) {
        Properties props = new Properties();

        if (MailBCImpl.MAIL_SCHEME_SMTP.equals(sessionDescriptor.getScheme())) {
            props.put(MailBCImpl.MAIL_TRANSPORT_PROTOCOL_KEY, sessionDescriptor
                    .getScheme());
        } else {
            props.put(MailBCImpl.MAIL_STORE_PROTOCOL_KEY, sessionDescriptor
                    .getScheme());
        }
        props.put(MailBCImpl.MAIL_HOST_KEY, sessionDescriptor.getHostname());
        if (sessionDescriptor.getUsername() != null) {
            props
                    .put(MailBCImpl.MAIL_USER_KEY, sessionDescriptor
                            .getUsername());
        }
        if (sessionDescriptor.getFromAddress() != null) {
            props.put(MailBCImpl.MAIL_FROM_KEY, sessionDescriptor
                    .getFromAddress());
        }

        return Session.getInstance(props);
    }

    /**
     * Send an email based on the given MimeMessage
     * 
     * @param mimeMessage
     *            the {@link MimeMessage} to send. The resulting email will
     *            conform to the RFC822 and MIME standards
     * @param sessionDescriptor
     *            a {@link SessionDescriptor} build from an address URI
     * @param session
     *            the mail {@link Session}
     * @throws JBIException
     *             if error occurs during message sending
     * 
     * @see Transport#send(javax.mail.Message);
     */
    public void sendMail(MimeMessage mimeMessage,
            SessionDescriptor sessionDescriptor, Session session)
        throws JBIException {

        try {
            Transport transport = session
                    .getTransport(MailBCImpl.MAIL_SCHEME_SMTP);
            if (!StringHelper.isNullOrEmpty(sessionDescriptor.getUsername())
                    && !StringHelper.isNullOrEmpty(sessionDescriptor.getPassword())) {
                transport.connect(sessionDescriptor.getHostname(), Integer
                        .parseInt(sessionDescriptor.getPort()),
                        sessionDescriptor.getUsername(), sessionDescriptor
                                .getPassword());
            } else if (!StringHelper.isNullOrEmpty(sessionDescriptor.getUsername())) {

                transport.connect(sessionDescriptor.getHostname(), Integer
                        .parseInt(sessionDescriptor.getPort()),
                        sessionDescriptor.getUsername(), null);
            } else {
                transport.connect(sessionDescriptor.getHostname(), Integer
                        .parseInt(sessionDescriptor.getPort()), null, null);
            }
            mimeMessage.saveChanges();
            transport.sendMessage(mimeMessage, mimeMessage.getAllRecipients());
            transport.close();
        } catch (javax.mail.MessagingException e) {
            String msg = "Error sending the email";
            log.severe(msg);
            throw new JBIException(msg, e);
        }
    }

    /**
     * Retrieve a {@link Store} from the given Session. Then connect to it.
     * 
     * @param session
     *            the mail {@link Session}
     * @param sessionDescriptor
     *            the {@link SessionDescriptor} associated to this session
     * @return the mail {@link Store}
     * @throws JBIException
     *             if error occurs during mail {@link Store} retrieval
     */
    public Store getStoreAndConnect(Session session,
            SessionDescriptor sessionDescriptor) throws JBIException {
        Store store = null;
        try {
            // Get a Store object
            store = session.getStore(sessionDescriptor.getScheme());

            // Connect to this Store
            if (!StringHelper.isNullOrEmpty(sessionDescriptor.getUsername())
                    && !StringHelper.isNullOrEmpty(sessionDescriptor.getPassword())) {
                store.connect(sessionDescriptor.getHostname(), Integer
                        .parseInt(sessionDescriptor.getPort()),
                        sessionDescriptor.getUsername(), sessionDescriptor
                                .getPassword());
            } else if (!StringHelper.isNullOrEmpty(sessionDescriptor.getUsername())) {

                store.connect(sessionDescriptor.getHostname(), Integer
                        .parseInt(sessionDescriptor.getPort()),
                        sessionDescriptor.getUsername(), null);
            } else {
                store.connect(sessionDescriptor.getHostname(), Integer
                        .parseInt(sessionDescriptor.getPort()), null, null);
            }
        } catch (MessagingException e) {
            String msg = "Error connecting to mail store";
            log.severe(msg);
            throw new JBIException(msg);
        }
        return store;
    }

    /**
     * Retrieve a specified {@link Folder} from the given {@link Store} and open
     * it.
     * 
     * @param store
     *            the mail {@link Store}
     * @param sessionDescriptor
     *            the mail {@link SessionDescriptor}
     * @return the opened {@link Folder}
     * @throws JBIException
     *             if {@link Folder} retrieval failed
     */
    public Folder getFolderAndOpen(Store store,
            SessionDescriptor sessionDescriptor) throws JBIException {
        String msg = "";
        // Search in the specified folder of this Store
        Folder folder = null;
        try {
            folder = store.getDefaultFolder();
            if (folder == null) {
                msg = "No default folder for this store : " + store.toString();
                log.severe(msg);
                throw new JBIException(msg);
            }
            folder = folder.getFolder(sessionDescriptor.getFolder());
            if (folder == null) {
                msg = "Invalid folder : " + sessionDescriptor.getFolder();
                log.severe(msg);
                throw new JBIException(msg);
            }
            folder.open(Folder.READ_WRITE);
        } catch (MessagingException e) {
            msg = "Error retrieving folder on store : " + store.toString();
            log.severe(msg);
            throw new JBIException(msg);
        }
        return folder;
    }

    /**
     * Retrieve new mails into the given folder. All mails, read or unread are
     * considered as new mail.
     * 
     * @param folder
     *            the folder used to search for new mail
     * @return the new mail Array
     * @throws JBIException
     *             if new mail retrieval fails
     */
    public Message[] getNewMails(Folder folder) throws JBIException {

        // Check for new mail to process
        Message[] messages = null;
        try {
            messages = folder.getMessages();
        } catch (MessagingException e) {
            String msg = "Error retrieving new mails from folder : "
                    + folder.toString();
            log.severe(msg);
            throw new JBIException(msg);
        }
        return messages;
    }
}
