/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XlinkerBCListener.java 154 19 avr. 2006 wjoseph $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.xlinker;

import java.io.File;
import java.net.URL;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.messaging.RobustInOnly;
import javax.wsdl.WSDLException;
import javax.xml.namespace.QName;
import javax.xml.soap.AttachmentPart;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.SOAPConnection;
import javax.xml.soap.SOAPConnectionFactory;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPMessage;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;

import org.act.soap.server.StandaloneWSServer;
import org.act.soap.util.Constants;
import org.objectweb.petals.binding.xlinker.utils.SOAPHelper;
import org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.component.common.util.WSDLHelper;
import org.w3c.dom.Document;

/**
 * Xlinker binding component listener this a JBI listener, which listens for
 * messaging for registered external service. Those messages are transmitted to
 * the remote webservice, and the response is forwarded to the JBI consumer
 * 
 * @version $Rev: 250 $ $Date: 2006-04-21 14:20:57 +0200 (ven, 21 avr 2006) $
 * @since Petals 1.0
 * @author wjoseph - EBM Websourcing
 * 
 */
public class XlinkerBCListener implements Runnable {
    private static final String ERROR_PARSE_WSDL = "Error parsing the service description ";

    private static final String ERROR_REMOTE_CALL = "Problem during remote call";

    private static final String ERROR_UNREGISTERED_ENDPOINT = "No such endpoint registered";

    private static final String ERROR_UNKNOWN_OPERATION = "Operation not recognized";

    private static final String SERVICE_FACTORY_IMPL_ACT = "org.act.soap.client.ServiceFactory";

    private static final String MESSAGE_FACTORY_PROPERTY = "javax.xml.soap.MessageFactory";

    private static final String MESSAGE_FACTORY_IMPL_ACT = "org.act.soap.message.MessageFactoryImpl";

    private static final String SOAP_CON_FACTORY_PROPERTY = "javax.xml.soap.SOAPConnectionFactory";

    private static final String SOAP_CON_FACTORY_IMPL_ACT = "org.act.soap.message.transport.ConnectionFactoryImpl";

    private static final String SERVICE_FACTORY_PROPERTY = "javax.xml.rpc.ServiceFactory";

    protected StandaloneWSServer receiver;

    protected DeliveryChannel channel;

    protected ComponentContext context;

    protected MessageExchange messageExchange;

    protected Logger logger;

    protected boolean running;

    protected PetalsServiceUnitManager suManager;

    /**
     * Default constructor
     * 
     */
    public XlinkerBCListener() {
    }

    /**
     * XlinkerBCListener Class constructor
     * 
     * @param channel
     *            the component delivery channel
     * @param context
     *            the component context
     * @param sum
     *            Service unit manager
     * @param se
     *            Service endpoint
     * @param log
     *            Component logger
     */
    public XlinkerBCListener(DeliveryChannel channel, ComponentContext context,
            PetalsServiceUnitManager sum, Logger log) {
        super();
        this.channel = channel;
        this.context = context;
        this.suManager = sum;
        this.logger = log;

        // setProperty of Factory
        System.setProperty(SERVICE_FACTORY_PROPERTY, SERVICE_FACTORY_IMPL_ACT);
        System.setProperty(MESSAGE_FACTORY_PROPERTY, MESSAGE_FACTORY_IMPL_ACT);
        System
                .setProperty(SOAP_CON_FACTORY_PROPERTY,
                        SOAP_CON_FACTORY_IMPL_ACT);
    }

    /**
     * Returns the running StandaloneWSServer
     * 
     * @return The receiver
     */
    public StandaloneWSServer getReceiver() {
        return receiver;
    }

    /**
     * The main loop of the listener
     * 
     *  (non-Javadoc)
     * @see java.lang.Runnable#run()
     */
    public void run() {
        running = true;

        // Start the soap receiver
        startSOAPReceiver();
        logger.log(Level.INFO, "SOAPServer Started");

        // Listen incoming messages
        while (running) {
            try {
                messageExchange = channel.accept();
                // If message is to be treated
                if (messageExchange != null
                        && !messageExchange.getStatus().equals(
                                ExchangeStatus.DONE)) {
                    processMessage(messageExchange);
                }
            } catch (Exception e) {
                logger.log(Level.SEVERE,e.getMessage(),e);
            }
        }
    }

    /**
     * Stop the listener and the soap server
     */
    public void stop() {
        receiver.stop();
        running = false;
    }

    /**
     * Calls a remote web service using Xlinker API
     * 
     * @param service
     *            Name Name of the service to be called
     * @param opName
     *            Name of the operation to be called
     * @param portName
     *            Name of the port to be called
     * @param serviceDesc
     *            Service description dom document
     * @param inEnv
     *            Input SOAP envelop for the call
     * @return A string representation of the response SOAP envelop
     * @throws Exception
     */
    protected SOAPMessage callWS(QName serviceName, String opName,
            String portName, Document serviceDesc, String input,
            Set<DataHandler> attachments) throws Exception {
        SOAPMessage reqMessage = createRequestMessage();
        SOAPEnvelope reqEnv = reqMessage.getSOAPPart().getEnvelope();
        putContentInSOAPEnvelope(reqEnv, input);
        for (DataHandler dh : attachments) {
            AttachmentPart attachment = reqMessage.createAttachmentPart(dh);
            reqMessage.addAttachmentPart(attachment);
        }
        URL serviceURL = new URL(getServiceAddressFromWSDLDocument(serviceDesc,
                serviceName));
        SOAPMessage responseMessage = null;
        SOAPConnection connection = createSOAPConnection();
        responseMessage = connection.call(reqMessage, serviceURL);
        return responseMessage;
    }

    /**
     * Create a new empty soap message
     * @return a new soap message
     * @throws SOAPException If an error occurs during soap message creation
     */
    protected SOAPMessage createRequestMessage() throws SOAPException {
        MessageFactory msgFactory = MessageFactory.newInstance();
        SOAPMessage reqMessage = msgFactory.createMessage();
        reqMessage.getMimeHeaders().setHeader("SOAPAction", "");
        return reqMessage;
    }

    /**
     * Create a new soap connection
     * @return  a new soap connection
     * @throws SOAPException If a connection factory instance is not available, or if the factory could not create a new connection 
     */
    protected SOAPConnection createSOAPConnection()
        throws SOAPException {
        SOAPConnectionFactory factory = SOAPConnectionFactory.newInstance();
        return factory.createConnection();
    }

    /**
     * @see org.objectweb.petals.component.common.xmlmanipulation.SourceHelper#createSoapFault(Throwable, String)
     */
    protected String createSoapFault(Exception e, String reason) {
        return SourceHelper.createSoapFault(e, reason);
    }

    /**
     * 
     * @see org.objectweb.petals.component.common.xmlmanipulation.SourceHelper#createSource(java.lang.Source)
     */
    protected Source createSource(String input) {
        return SourceHelper.createSource(input);
    }

    /**
     * 
     * @see org.objectweb.petals.component.common.xmlmanipulation.SourceHelper#createString(javax.xml.transform.Source)
     */
    protected String createString(Source source) throws Exception {
        return SourceHelper.createString(source);
    }

    /**
     * Put all attachments from a normalized message in a Set
     * 
     * @param nm
     *            the normalized message containing the attachments
     * @return Returns a Set of DataHandlers
     */
    @SuppressWarnings("unchecked")
    protected Set<DataHandler> extractAttachments(NormalizedMessage nm) {
        Set<DataHandler> attachments = new HashSet<DataHandler>();
        for (String attachmentName : (Set<String>) nm.getAttachmentNames()) {
            attachments.add(nm.getAttachment(attachmentName));
        }
        return attachments;
    }

    /**
     * @param nm
     *            The normalized message in which will put the content of the
     *            SOAP Body
     * @param message
     *            The SOAP Message from which data will be extracted
     * @throws Exception
     *             When the SOAP Message is invalid
     */
    protected void fillNormalizedMessageFromSOAPMessage(NormalizedMessage nm,
            SOAPMessage message) throws Exception {
        String source = SOAPHelper.createStringFromSOAPEnvelope(message
                .getSOAPPart().getEnvelope());
        nm.setContent(SourceHelper.createSource(source));
    }

    /**
     * 
     * @throws WSDLException
     * @throws TransformerException 
     * @see org.objectweb.petals.component.common.xmlmanipulation.XMLHelper#getPortNameFromWSDLDocument(Document)(org.w3c.dom.Document )
     */
    protected String getPortNameFromWSDLDocument(Document serviceDesc,
            QName serviceName) throws WSDLException, TransformerException {
        return WSDLHelper.getPortNameFromWSDLDocument(serviceDesc, serviceName);
    }

    /**
     * 
     * @throws WSDLException
     * @throws TransformerException 
     * @see org.objectweb.petals.component.common.xmlmanipulation.XMLHelper#getServiceAddressFromWSDLDocument(org.w3c.dom.Document)
     */
    protected String getServiceAddressFromWSDLDocument(Document serviceDesc,
            QName serviceName) throws WSDLException, TransformerException {
        return WSDLHelper.getServiceAddressFromWSDLDocument(serviceDesc,
                serviceName);
    }

    /**
     * 
     * @throws TransformerException 
     * @see org.objectweb.petals.component.common.xmlmanipulation.XMLHelper#hasOperationNamed(org.w3c.dom.Document,
     *      java.lang.String)
     */
    protected boolean hasOperationNamed(Document serviceDesc, String operation,
            QName serviceName) throws WSDLException, TransformerException {
        return WSDLHelper
                .hasOperationNamed(serviceDesc, operation, serviceName);
    }

    /**
     * 
     * @throws TransformerException 
     * @see org.objectweb.petals.component.common.xmlmanipulation.XMLHelper#isInOutOperation(org.w3c.dom.Document,
     *      java.lang.String)
     */
    protected boolean isInOutOperation(Document serviceDesc, String operation,
            QName serviceName) throws WSDLException, TransformerException {
        return WSDLHelper.isInOutOperation(serviceDesc, operation, serviceName);
    }

    /**
     * Process calls for external services
     * 
     * @param msgExchange
     *            The current Message exchange
     * @param serviceDesc
     *            the dom representation of the service description
     * @param service
     *            The QName of the requested service
     * @param operation
     *            The operation requested
     * @throws MessagingException
     *             If an error occured during processing or while sending Error
     */
    protected void processExternalCall(MessageExchange msgExchange,
            Document serviceDesc, QName service, String operation)
        throws MessagingException {
        // Checks if this operation is described in the service
        // description
        try {
            if (hasOperationNamed(serviceDesc, operation, service)) {
                processOperation(msgExchange, serviceDesc, operation, service);
            } else {
                sendError(msgExchange, ERROR_UNKNOWN_OPERATION, false);
            }
        } catch (TransformerException e) {
            throw new MessagingException(ERROR_PARSE_WSDL, e);
        } catch (WSDLException e) {
            throw new MessagingException(ERROR_PARSE_WSDL, e);
        } catch (Exception e) {
            throw new MessagingException(ERROR_REMOTE_CALL, e);
        }
    }

    /**
     * Calls distant in only operation and set the return status on the
     * MessageExchange. An error is sent to the consumer if the MessageExchange
     * type and the operation type do not match
     * 
     * @param msgExchange
     *            the current msgExchange
     * @param operation
     *            The name of the operation to call
     * @param service
     *            The name of the service to call
     * @param serviceDesc
     *            The service description of this service
     * @param nm
     *            The input message, received from the JBI container
     * @throws Exception
     *             If the webservice call failed
     */
    protected void processInOnlyOperation(MessageExchange msgExchange,
            String operation, QName service, Document serviceDesc,
            NormalizedMessage nm, Set<DataHandler> attachments)
        throws Exception {
        if (msgExchange instanceof InOnly
                || msgExchange instanceof InOptionalOut
                || msgExchange instanceof RobustInOnly) {
            callWS(service, operation, getPortNameFromWSDLDocument(serviceDesc,
                    service), serviceDesc, createString(nm.getContent()),
                    attachments);
            msgExchange.setStatus(ExchangeStatus.DONE);
            channel.send(msgExchange);
        } else {
            sendError(
                    msgExchange,
                    "In-only operation but not in-only compatible message exchange",
                    true);
        }
    }

    /**
     * Calls distant in out operation and set the return message on the
     * MessageExchange. An error is sent to the consumer if the MessageExchange
     * type and the operation type do not match
     * 
     * @param msgExchange
     *            the current msgExchange
     * @param operation
     *            The name of the operation to call
     * @param service
     *            The name of the service to call
     * @param serviceDesc
     *            The service description of this service
     * @param nm
     *            The input message, received from the JBI container
     * @throws Exception
     *             If the webservice call failed
     */
    protected void processInOutOperation(MessageExchange msgExchange,
            String operation, QName service, Document serviceDesc,
            NormalizedMessage nm, Set<DataHandler> attachments)
        throws Exception {
        if (msgExchange instanceof InOut
                || msgExchange instanceof InOptionalOut) {
            SOAPMessage result = callWS(service, operation,
                    getPortNameFromWSDLDocument(serviceDesc, service),
                    serviceDesc, createString(nm.getContent()), attachments);
            fillNormalizedMessageFromSOAPMessage(nm, result);
            msgExchange.setMessage(nm, "out");
            channel.send(msgExchange);
        } else {
            sendError(
                    msgExchange,
                    "In-out operation but not in-out compatible message exchange",
                    true);
        }
    }

    /**
     * Handle incoming message and call the distant service
     * 
     * @param messageExchange
     *            The incoming MessageExchange
     * @throws MessagingException
     *             If an error occured in a method interacting with the JBI
     *             container
     */
    protected void processMessage(MessageExchange msgExchange)
        throws MessagingException {
        if (!msgExchange.getStatus().equals(ExchangeStatus.ERROR)) {
            String operation = msgExchange.getOperation().getLocalPart();
            QName service = msgExchange.getService();
            Document serviceDesc = suManager.getServiceDescription(msgExchange
                    .getEndpoint());

            // If we have a service description for the requested endpoint
            if (serviceDesc != null) {
                // Message for an external service
                processExternalCall(msgExchange, serviceDesc, service,
                        operation);
            } else {
                // No such endpoint registered
                sendError(msgExchange, ERROR_UNREGISTERED_ENDPOINT, false);
            }
        } else {
            sendDone(msgExchange);
        }

    }

    /**
     * Process an operation
     * 
     * @param msgExchange
     *            The current Message Exchange
     * @param serviceDesc
     *            The dom representation of the service description
     * @param operation
     *            The requested operation
     * @param service
     *            The QName of the requested service
     * @throws MessagingException
     */
    protected void processOperation(MessageExchange msgExchange,
            Document serviceDesc, String operation, QName service)
        throws MessagingException {
        logger.log(Level.INFO, "Handling for enpoint :"
                + msgExchange.getEndpoint());
        try {
            String targetURI = getServiceAddressFromWSDLDocument(serviceDesc,
                    service);
            logger.log(Level.INFO, "Destination address is :" + targetURI);

            // Send the request to the distant WS
            NormalizedMessage nm = msgExchange.getMessage("IN");
            Set<DataHandler> attachments = extractAttachments(nm);
            if (isInOutOperation(serviceDesc, operation, service)) {
                processInOutOperation(msgExchange, operation, service,
                        serviceDesc, nm, attachments);
            } else {
                processInOnlyOperation(msgExchange, operation, service,
                        serviceDesc, nm, attachments);
            }
        } catch (Exception e) {
            sendFault(msgExchange, ERROR_REMOTE_CALL, e);
        }
    }

    
    /**
     * @see org.objectweb.petals.binding.xlinker.utils.SOAPHelper#putContentInSOAPEnvelope(SOAPEnvelope, String)
     */
    protected void putContentInSOAPEnvelope(SOAPEnvelope reqEnv, String input)
        throws Exception {
        org.objectweb.petals.binding.xlinker.utils.SOAPHelper
                .putContentInSOAPEnvelope(reqEnv, input);
    }

    /**
     * Send a response message with a status set to "DONE".
     * 
     * @param msgExchange
     *            current MessageExchange used to send and receive messages
     * @throws MessagingException
     *             If an error occured during calls on the JBI container
     */
    protected void sendDone(MessageExchange msgExchange)
        throws MessagingException {
        msgExchange.setStatus(ExchangeStatus.DONE);
        channel.send(msgExchange);
    }

    /**
     * Send a response message with status set to "ERROR" to the service
     * consumer. An exception is created with the given reason as message. This
     * exception is set in the MessageExchange object. The modified
     * MessageExchange is then returned to the service consumer.
     * 
     * @param msgExchange
     *            current MessageExchange used to send and receive messages
     * @param reason
     *            The reason why the error was raised
     * @param severe
     *            Log the error at warning or severe level
     * @throws MessagingException
     *             If an error occured during calls on the JBI container
     */
    protected void sendError(MessageExchange msgExchange, String reason,
            boolean severe) throws MessagingException {
        Exception e = new Exception(reason);
        if ((msgExchange instanceof InOnly)
                && !(msgExchange instanceof RobustInOnly)) {
            sendDone(msgExchange);
        } else {
            msgExchange.setError(e);
            msgExchange.setStatus(ExchangeStatus.ERROR);
            channel.send(msgExchange);
        }
        if (!severe) {
            logger.log(Level.WARNING, e.getMessage(), e);
        } else {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
    }

    /**
     * Send a response message with status set to "ERROR" to the service
     * consumer. A fault is created with the given reason as content. This fault
     * is set in the MessageExchange object. The modified MessageExchange is
     * then returned to the service consumer.
     * 
     * @param msgExchange
     *            current MessageExchange used to send and receive messages
     * @param reason
     *            The reason why the error was raised
     * @param e
     *            The exception which created the fault
     * @throws MessagingException
     *             If an error occured during calls on the JBI container
     */
    protected void sendFault(MessageExchange msgExchange, String reason,
            Exception e) throws MessagingException {
        if ((msgExchange instanceof InOnly)
                && !(msgExchange instanceof RobustInOnly)) {
            sendDone(msgExchange);
        } else {
            Fault fault = msgExchange.createFault();
            String faultString = createSoapFault(e, reason);
            fault.setContent(createSource(faultString));
            msgExchange.setFault(fault);
            msgExchange.setStatus(ExchangeStatus.ERROR);
            channel.send(msgExchange);
        }
    }

    /**
     * Start the Xlinker SOAP server
     */
    protected void startSOAPReceiver() {
        String rootDir = context.getInstallRoot();
        String configFile = rootDir + File.separator + "META-INF"
                + File.separator + "SoapEngine.config";
        System.setProperty(Constants.PROPERTY_NAME_ROOT_PATH, rootDir);
        System
                .setProperty(Constants.PROPERTY_NAME_CONFIG_FILE_PATH,
                        configFile);
        receiver = new StandaloneWSServer("localhost", 8085);
        receiver.init();
    }
}
