/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XlinkerServiceUnitHandler.java 154 19 avr. 2006 wjoseph $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.xlinker.handler;

import java.io.File;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.management.DeploymentException;
import javax.xml.namespace.QName;

import org.objectweb.petals.binding.xlinker.XlinkerBC;
import org.objectweb.petals.binding.xlinker.servicemanager.XlinkerServiceUnitManager;
import org.objectweb.petals.component.common.serviceunitmanager.handler.ExternalServiceManager;
import org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager;
import org.objectweb.petals.component.common.util.WSDLHelper;
import org.objectweb.petals.component.common.util.XMLHelper;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

/**
 * @version $Rev: 250 $ $Date: 2006-04-21 14:20:57 +0200 (ven, 21 avr 2006) $
 * @since Petals 1.0
 * @author wjoseph - EBM Webousourcing
 * 
 */
public class XlinkerServiceUnitHandler extends ExternalServiceManager {

    protected static final String FAILED_DEPLOY_SERVICE = "Error during service deployment";

    protected static final String FAILED_UNDEPLOY_SERVICE = "Error during service undeployment";

    protected static final String ERROR_GET_SERVICE_NAME = "Could not retrieve the service QName";

    /**
     * Default contructor
     */
    public XlinkerServiceUnitHandler() {
        SERVICE_UNIT_TYPE = "xlinker_service";
    }

    /**
     * Create a new XlinkerServiceUnitHandler and set the context and logger.
     * 
     * @param context
     *            the context of the component creating the handler
     * @param logger
     *            the logger this class will use
     */
    public XlinkerServiceUnitHandler(ComponentContext context, Logger logger) {
        super(context, logger);
        SERVICE_UNIT_TYPE = "xlinker_service";
    }

    /*
     * If the service unit is of the right type, we check if the service unit
     * package is valid
     * 
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.serviceunitmanager.handler.ExternalServiceManager#deploy(java.lang.String,
     *      java.lang.String)
     */
    @Override
    public String deploy(String serviceUnitName, String serviceUnitType
            , String serviceUnitRootPath) {
        String result = super.deploy(serviceUnitName, serviceUnitType, serviceUnitRootPath);
        if (result != null && !validateServicePackage(serviceUnitRootPath)) {
            result = null;
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.objectweb.petals.component.common.serviceunitmanager.handler.ExternalServiceManager#start(java.lang.String,
     *      org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager)
     */
    @Override
    public void start(String serviceUnitName, PetalsServiceUnitManager epHandler)
        throws DeploymentException {
        start(serviceUnitName, (XlinkerServiceUnitManager) epHandler);
    }

    /**
     * Start a service unit. Before been started, the service package is checked
     * to see if it is complete and if the provided WSDL is valid
     * 
     * @param serviceUnitName
     *            The name of the service unit to be deployed
     * @param epHandler
     *            XLinkerServiceUnitManager
     * @throws DeploymentException
     *             if endpoint activation fails, if service description is not
     *             valid, if service unit package is not complete
     */
    public void start(String serviceUnitName,
            XlinkerServiceUnitManager epHandler) throws DeploymentException {

        logger.log(Level.FINE, "start  serviceUnitName  " + serviceUnitName);

        QName serviceName = null;
        String serviceUnitRootPath = serviceUnitInstallationRootPath.get(serviceUnitName);
        File wsarFile = getWsarFile(serviceUnitRootPath);
        XlinkerBC xlink = epHandler.getXlinkerBC();
        try {
            serviceName = getServiceName(serviceUnitRootPath);
            deployWsar(xlink, wsarFile, serviceName);
        } catch (Exception ex) {
            logger.log(Level.SEVERE, FAILED_DEPLOY_SERVICE + ex);
            try {
                undeployWsar(xlink, serviceName);
            } catch (Exception e) {
                throw new DeploymentException(FAILED_DEPLOY_SERVICE,
                        new DeploymentException(FAILED_UNDEPLOY_SERVICE, e));
            }
            throw new DeploymentException(FAILED_DEPLOY_SERVICE, ex);
        }
    }

    /**
     * Deploy the wsarFile in the StandalonWSServer of the binding component
     * 
     * @param xlink
     *            The component
     * @param wsarFile
     *            The wsar file
     * @param serviceName
     *            The service QName
     * @throws Exception
     *             if on of the parameter is null or if an error occured during
     *             the deployment process
     */
    protected void deployWsar(XlinkerBC xlink, File wsarFile, QName serviceName)
        throws Exception {
        xlink.getListener().getReceiver().getAutoDeployer().deploy(
                wsarFile.getAbsolutePath(), serviceName.getLocalPart());
    }

    /**
     * Returns the first .wsar File found at service unit root path
     * @param serviceUnitRootPath the service unit root path
     * @return a wsar File or null if no .wsar file was found
     */
    protected File getWsarFile(String serviceUnitRootPath) {
        File result = null;
        File[] files = new File(serviceUnitRootPath).listFiles();
        for (File file : files) {
            if (file.getName().endsWith(".wsar")) {
                result = file;
                break;
            }
        }
        return result;
    }

    /**
     * Undeploy the wsar from the StandaloneWSServer matching the service name
     * 
     * @param xlink
     *            the binding component
     * @param serviceName
     *            The service QName
     * @throws Exception
     *             if one of the parameter is null or if an exception was raised
     *             during undeployment process in the AutoDeployer
     */
    protected void undeployWsar(XlinkerBC xlink, QName serviceName)
        throws Exception {
        xlink.getListener().getReceiver().getAutoDeployer().undeploy(
                serviceName.getLocalPart());
    }

    /**
     * Check if the service unit is valid
     * @param serviceUnitRootPath the service unit root path
     * @return true if the service unit is valid, orelse false
     */
    protected boolean validateServicePackage(String serviceUnitRootPath) {
        boolean result = false;
        File[] files = new File(serviceUnitRootPath).listFiles();
        int i = 0;
        for (File file : files) {
            if (!file.isDirectory() && file.getName().endsWith(".wsar")) {
                i++;
            } else {
                if (!file.isDirectory()) {
                    i--;
                }
            }
        }
        if (i == 1) {
            result = true;
        }
        return result;
    }

    /**
     * Parse the deployment descriptor and return the service QName.
     * Only the service name of the first provides element is returned.
     * @param serviceUnitRootPath the service unit root path
     * @return the service QName 
     * @throws DeploymentException if no jbi.xml file or if no provides element is found 
     */
    protected QName getServiceName(String serviceUnitRootPath) throws DeploymentException {
        File jbiFile = getJbiDescriptor(serviceUnitRootPath);
        if (jbiFile != null) {
            Document jbiDescriptor = WSDLHelper.createDocumentFromWSDL(jbiFile);
            Node node = XMLHelper.findChild(jbiDescriptor, "consumes", true);
            if (node != null) {
                return QName.valueOf(XMLHelper.getAttributeValue(node,
                        "service-name"));
            }
        }
        throw new DeploymentException(ERROR_GET_SERVICE_NAME);
    }

    /**
     * Look into the META-INF directory of the service unit for the jbi.xml file
     * @param serviceUnitRootPath the service unit root path
     * @return A file pointing on the deployment descriptor or null if not found
     */
    protected File getJbiDescriptor(String serviceUnitRootPath) {
        File result = null;
        File jbiDescriptor = new File(serviceUnitRootPath + File.separator
                + "META-INF" + File.separator + "jbi.xml");
        if (jbiDescriptor.exists()) {
            result = jbiDescriptor;
        }
        return result;
    }

}
