/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.xquarebc;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.management.DeploymentException;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.common.serviceunitmanager.handler.ExternalServiceManager;
import org.objectweb.petals.component.common.serviceunitmanager.manager.PetalsServiceUnitManager;
import org.objectweb.petals.component.common.util.ManagementMessageUtil;
import org.objectweb.petals.component.common.util.WSDLHelper;
import org.objectweb.petals.component.common.util.XMLHelper;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.xquark.bridge.XQBridge;
import org.xquark.jdbc.datasource.JDBCDataSource;

/**
 * Service Unit handler for the XQuare component.
 * 
 * At startup, lookups for a single *.wsdl file and a single
 * *.properties file. The first provides the service name, and the
 * second provides properties that are given to the XQuare
 * component and from there to the XQuareBCJBIProcessor and
 * XQuareBCListener to use (look at their documentation about
 * which properties are available).
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.0
 * @author Marc Dutoo - Open Wide
 *
 */
public class XQuareSUHandler extends ExternalServiceManager {

    public static final String JNDI_DATASOURCE = "jndidatasource.name";
    public static final String DATABASE_URL = "database.url";
    public static final String DATABASE_USER = "database.user";
    public static final String DATABASE_PASSWORD = "database.password";
    public static final String DATABASE_MAPPING = "database.mapping";
    public static final String XML_CONNECTION_BASE_URI = "xmlconnection.baseuri";
    public static final String WRAP_QUERY_RESULTS_IN_ROOT_TAG = "queryResults.wrapInRootTag";
    public static final String XQUARE_BRIDGE = "xquare_bridge_object";
    public static final String STANDALONE_DATASOURCE = "standalone_datasource_object";
    
    public static final String LISTENER_PROP_PREFIX = "newdatalistener.";
    public static final String LISTENER_SERVICE_PROP = LISTENER_PROP_PREFIX + "service";
    public static final String LISTENER_OPERATION_PROP = LISTENER_PROP_PREFIX + "operation";
    
    protected HashMap<String, Properties> serviceToPropertiesMap;
    protected HashMap<String, String> serviceUnitToServiceMap;

    public XQuareSUHandler() {
        super();
        SERVICE_UNIT_TYPE = "xquare_service";
    }

    public XQuareSUHandler(ComponentContext context, Logger logger,
        HashMap<String, Properties> serviceToPropertiesMap) {
        super(context, logger);
        this.context = context;
        this.logger = logger;
        this.serviceToPropertiesMap = serviceToPropertiesMap;
        this.serviceUnitToServiceMap = new HashMap<String, String>();
        SERVICE_UNIT_TYPE = "xquare_service";
    }

    /* (non-Javadoc)
     * @see org.objectweb.petals.component.common.serviceunitmanager.handler.PetalsServiceUnitHandler#deploy(java.lang.String, java.lang.String, java.lang.String)
     */
    public String deploy(String serviceUnitName, String serviceUnitType,
        String serviceUnitRootPath) {
        return super.deploy(serviceUnitName, serviceUnitType,
            serviceUnitRootPath);
    }

    public void init(String serviceUnitName, String serviceUnitRootPath)
    throws DeploymentException {
        super.init(serviceUnitName, serviceUnitRootPath);
    }

    public String undeploy(String serviceUnitName, String serviceUnitRootPath)
    throws DeploymentException {
        // closing bridge
        String serviceName = this.serviceUnitToServiceMap.get(serviceUnitName);
        Properties serviceProps = serviceToPropertiesMap.get(serviceName);
        XQBridge bridge = (XQBridge) serviceProps.get(XQUARE_BRIDGE);
        if (bridge != null) {
            try {
                bridge.close();
            } catch (Throwable t) {
                logger.warning("Error closing bridge when undeploying service " + serviceName);
            }
        }
        // closing datasource (case of standalone datasource)
        JDBCDataSource xquareDataSource = (JDBCDataSource) serviceProps.get(STANDALONE_DATASOURCE);
        if (xquareDataSource != null) {
            try {
                xquareDataSource.close();
            } catch (Throwable t) {
                logger.warning("Error closing datasource when undeploying service " + serviceName);
            }
        }
        return ManagementMessageUtil.getComponentTaskResult(context
            .getComponentName(), "undeploy",
            ManagementMessageUtil.TASK_RESULT_SUCCESS);
    }

    public void shutDown(String serviceUnitName,
        PetalsServiceUnitManager epHandler) throws DeploymentException {
        super.shutDown(serviceUnitName, epHandler);
    }

    public void start(String serviceUnitName, PetalsServiceUnitManager epHandler)
    throws DeploymentException {
        String serviceUnitRootPath = serviceUnitInstallationRootPath
        .get(serviceUnitName);
        File[] files = new File(serviceUnitRootPath).listFiles();
        Document serviceDesc = null;
        String propFileName = null;
        QName service=  null;

        for (File file : files) {
            if (file.getName().endsWith(".wsdl")) {
                // found the service definition
                serviceDesc = WSDLHelper.createDocumentFromWSDL(file);
            } else if (file.getName().endsWith(".properties")){
                // found the service configuration
                propFileName = file.getAbsolutePath();
            }
        }

        // activating found endpoints
        if (serviceDesc != null) {
            try {
                activateEndpointsFromJBIDescription(epHandler, serviceDesc,
                    serviceUnitName, serviceUnitRootPath);
                service = WSDLHelper.getServiceNameFromWSDLDocument(serviceDesc)[0];
            } catch (Exception ex) {
                logger.log(Level.SEVERE, FAILED_ACTIVATE_ENDPOINT + ex);
                throw new DeploymentException(FAILED_ACTIVATE_ENDPOINT, ex);
            }
        } else {
            throw new DeploymentException(INCOMPLETE_SERVICE_UNIT_PACKAGE);
        }


        // Now configuring the component for the service & properties found
        if (service!=null && propFileName != null) {
            Properties xquareProps = new Properties();
            try {
                xquareProps.load(new FileInputStream(propFileName));
            } catch (IOException e) {
                logger.log(Level.SEVERE, FAILED_ACTIVATE_ENDPOINT, e);
                throw new DeploymentException(FAILED_ACTIVATE_ENDPOINT, e);
            }

            // some property polishing :
            // 1. Correcting database mapping path
            String databaseMappingFilepath = xquareProps.getProperty(DATABASE_MAPPING, "[none]");
            File databaseMappingFile = new File(databaseMappingFilepath);
            if (!databaseMappingFile.exists() && !databaseMappingFile.isAbsolute()) {
                // getting the absolute filepath
                databaseMappingFilepath = serviceUnitRootPath
                + File.separatorChar + databaseMappingFilepath;
                databaseMappingFile = new File(databaseMappingFilepath);
                xquareProps.setProperty(DATABASE_MAPPING, databaseMappingFilepath);
            }
            if (!databaseMappingFile.exists() || databaseMappingFile.isDirectory()
                    || !databaseMappingFile.canRead()) {
                throw new DeploymentException("Unable to find database mapping file at "
                    + databaseMappingFilepath);
            }
            // 1. Correcting base URI
            if (!xquareProps.contains(XML_CONNECTION_BASE_URI)) {
                xquareProps.setProperty(XML_CONNECTION_BASE_URI, "file://" + serviceUnitRootPath);
            }

            // adding the new service unit to ours 
            serviceUnitToServiceMap.put(serviceUnitName, service.getLocalPart());
            serviceToPropertiesMap.put(service.getLocalPart(), xquareProps);
        }

        // The following is not used for now, would be for in-jbi.xml configuration
        /*databaseMappingFilepath = "auction.map";
        databaseUrl = "jdbc:mysql://localhost/petals-samples-xquare_auction";
        databaseUser = "petals-samples-xquare";
        databasePassword = "petals-samples-xquare";

        String jbiFilePath = serviceUnitRootPath
              + File.separator + "META-INF" + File.separator + "jbi.xml";
        initConfig(new File(jbiFilePath));

        File databaseMappingFile = new File(databaseMappingFilepath);
        if (!databaseMappingFile.isAbsolute()) {
            databaseMappingFilepath = serviceUnitRootPath
                  + File.separatorChar + databaseMappingFilepath;
            databaseMappingFile = new File(databaseMappingFilepath);
        }
        if (!databaseMappingFile.exists() || databaseMappingFile.isDirectory()
                || !databaseMappingFile.canRead()) {
            throw new DeploymentException("Unable to find database mapping file at "
                + databaseMappingFilepath);
        }*/
    }

    public void stop(String serviceUnitName, PetalsServiceUnitManager epHandler)
    throws DeploymentException {
        super.stop(serviceUnitName, epHandler);
    }

    /**
     * NOT USED FOR NOW, property file used instead
     * jbiXML define attibuts to configure files to allow echange files.
     * @param jbiXml
     * @return properties
     * properties has all attributs to call an internal Petals service, 
     * the source directory, destination directory, service and method to be called
     */
    protected void initConfig(File jbiXml) {
        Properties configProperties = new Properties();
        Document jbiDoc = WSDLHelper.createDocumentFromWSDL(jbiXml);
        Node node = XMLHelper.findChild(jbiDoc, "services", true);
        if (node != null) {

            Node providesNode = XMLHelper.findChild(node, "provides", true);
            if (providesNode != null) {
                Node databaseUrlNode = XMLHelper.findChild(node,
                    "database_url", true);

                if (databaseUrlNode != null) {
                    String databaseUrl = databaseUrlNode.getTextContent();
                    configProperties.put(DATABASE_URL, databaseUrl);
                }
                
                // add other database & xquare properties too
            }

            Node consumesNode = XMLHelper.findChild(node, "consumes", true);
            if (consumesNode != null) {

                String service = XMLHelper.getAttributeValue(consumesNode,
                    "service-name");
                if (service != null) {
                    configProperties.put(XQuareSUHandler.LISTENER_SERVICE_PROP, service);
                }

                String operation = XMLHelper.getAttributeValue(consumesNode,
                    "operation-name");
                if (operation != null) {
                    configProperties.put(XQuareSUHandler.LISTENER_OPERATION_PROP, operation);
                }

                // add delimiter props too
            }

        }
    }

}
