/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.xquarebc.listeners;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.transform.Source;

import org.objectweb.petals.binding.xquarebc.XQuareBCException;
import org.xquark.bridge.XQBridge;

/**
 * Implements a strategy for detecting "new data",
 * getting it (as XML) and updating what data is known, by monitoring
 * a given field as its "new data delimiter".
 * 
 * What it actually does is first executing two SQL queries to
 * get the "new", current data delimiter value and the "old",
 * known data delimiter value. Then it executes an XQuery through
 * the XQuareBCJBIProcessor, after having the "##OLD_DATA_DELIMITER##"
 * and "##OLD_DATA_DELIMITER##" expressions replaced accordingly
 * within its XQuery code. Finally it executes an SQL insertion
 * to update the "old", known data delimiter to the current "new" one.
 * 
 * NB. All These SQL and XQuery queries, insertions and initial value
 * are configured in XQuareBC's listener's configuration. Look in the
 * javadoc of XQuareBCListener for more details.
 * 
 * NB. Must be instanciated for each new data detection.
 * NB. This is the only implementation of NewDataStrategy for now.
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.0
 * @author Marc Dutoo - Open Wide
 * 
 */
public class DelimiterNewDataStrategy implements NewDataStrategy {
    
    public static final String OLD_DATA_DELIMITER_SELECT_PROP = "oldDataDelimiterSelect";
    public static final String NEW_DATA_DELIMITER_SELECT_PROP = "newDataDelimiterSelect";
    public static final String INIT_DATA_DELIMITER_PROP = "initDataDelimiter";
    public static final String NEW_DATA_XQUERY_PROP = "newDataXquery";
    public static final String NEW_DATA_DELIMITER_UPDATES_PROP = "newDataDelimiterUpdates";

    protected static final String DELIMITER_PROP_PREFIX = "delimiter.";
    protected static final String CUSTOM_TABLE_SUFFIX = "_xqbc";
    protected static final String OLD_DATA_DELIMITER_TOKEN = "##OLD_DATA_DELIMITER##";
    protected static final String NEW_DATA_DELIMITER_TOKEN = "##NEW_DATA_DELIMITER##";

    private Logger logger;
    protected Connection conn;
    
    protected String oldDataDelimiterSelect;
    protected String newDataDelimiterSelect;
    protected String initDataDelimiter;
    protected ArrayList<String> newDataDelimiterUpdates = new ArrayList<String>();
    protected String newDataXquery;
    
    protected String oldDataDelimiter;
    protected String newDataDelimiter;

    protected XQuareBCJBIProcessor processor;
    public Properties serviceProps;
    protected XQBridge xqBridge;

    
    /**
     * Creates a new DelimiterNewDataStrategy
     * @param processor
     * @param conn
     * @param serviceProps
     * @param listenerPropPrefix
     * @param logger 
     * @throws XQuareBCException
     */
    public DelimiterNewDataStrategy(XQuareBCJBIProcessor processor,
        Connection conn, Properties serviceProps, String listenerPropPrefix,
        Logger logger)
        throws XQuareBCException {
        this.logger = logger;
        this.conn = conn;
        
        // for query :
        this.processor = processor;
        this.serviceProps = serviceProps;
        this.xqBridge= processor.getXQBridge(serviceProps);
        
        // init from props :
        listenerPropPrefix = listenerPropPrefix + DELIMITER_PROP_PREFIX;
        this.oldDataDelimiterSelect = (String) serviceProps.get(
            listenerPropPrefix + OLD_DATA_DELIMITER_SELECT_PROP);
        this.newDataDelimiterSelect = (String) serviceProps.get(
            listenerPropPrefix + NEW_DATA_DELIMITER_SELECT_PROP);
        this.initDataDelimiter = (String) serviceProps.get(
            listenerPropPrefix + INIT_DATA_DELIMITER_PROP);
        this.newDataXquery = (String) serviceProps.get(
            listenerPropPrefix + NEW_DATA_XQUERY_PROP);
        
        // initing new data delimiter update statements :
        String newDataDelimiterUpdateString = (String) serviceProps.get(
            listenerPropPrefix + NEW_DATA_DELIMITER_UPDATES_PROP);
        StringTokenizer stok = new StringTokenizer(newDataDelimiterUpdateString, ";");
        // splitting it to execute multiple updates if any
        // ex. for delimiter = max id : 1. cleaning the new data detection table, 2. filling it with the new max id
        while (stok.hasMoreTokens()) {
            String newDataDelimiterUpdate = stok.nextToken().trim();
            newDataDelimiterUpdates.add(newDataDelimiterUpdate);
        }
    }
    
    
    /* (non-Javadoc)
     * @see org.objectweb.petals.binding.xquarebc.listeners.NewDataStrategy#detectNewData()
     */
    public boolean detectNewData() throws Exception {
        // getting the old data delimiter :
        Statement newDataStmt = conn.createStatement();
        newDataStmt.execute(this.oldDataDelimiterSelect);
        ResultSet newDataRs = newDataStmt.getResultSet();
        if (newDataRs.next()) {
            oldDataDelimiter = newDataRs.getString(1);
        }
        if (oldDataDelimiter == null) {
            oldDataDelimiter = initDataDelimiter;
        }

        // getting the new data delimiter :
        Statement getNewMaxIdStmt = conn.createStatement();
        getNewMaxIdStmt.execute(this.newDataDelimiterSelect);
        ResultSet getNewMaxIdRs = getNewMaxIdStmt.getResultSet();
        if (getNewMaxIdRs.next()) {
            newDataDelimiter = getNewMaxIdRs.getString(1);
        }
        
        // there is new data if they are different
        return (newDataDelimiter != null
                && !newDataDelimiter.equals(oldDataDelimiter));
    }

    
    /* (non-Javadoc)
     * @see org.objectweb.petals.binding.xquarebc.listeners.NewDataStrategy#getNewData()
     */
    public Source getNewData() throws Exception {
        // formatting result message body :
        String formattedXquery = replaceDelimiterTokens(newDataXquery);
        logger.log(Level.FINE, "DelimiterNewDataStrategy formattedXquery : "
            + formattedXquery);
        Source bodySource = processor.doQuery(formattedXquery,
            xqBridge, serviceProps);
        return bodySource;
    }
    

    /* (non-Javadoc)
     * @see org.objectweb.petals.binding.xquarebc.listeners.NewDataStrategy#updateDataKnown()
     */
    public void updateDataKnown() throws Exception {
        // building SQL for updating the data delimiter to the new value
        // splitting it to execute multiple updates if any
        // ex. for delimiter = max id : 1. cleaning the new data detection table, 2. filling it with the new max id
        for (String newDataDelimiterUpdate : newDataDelimiterUpdates) {
            newDataDelimiterUpdate = replaceDelimiterTokens(newDataDelimiterUpdate);
            conn.createStatement().executeUpdate(newDataDelimiterUpdate);
            // conn.commit(); // would require autocommit, but there is no need
        }
    }


    protected String replaceDelimiterTokens(String sqlOrXquery) {
        sqlOrXquery = sqlOrXquery.replaceAll(OLD_DATA_DELIMITER_TOKEN, oldDataDelimiter);
        sqlOrXquery = sqlOrXquery.replaceAll(NEW_DATA_DELIMITER_TOKEN, newDataDelimiter);
        return sqlOrXquery;
    }
    
}
