/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id : $ 
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.xquarebc.listeners;

import java.sql.Connection;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.InOnly;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.sql.DataSource;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;
import org.objectweb.petals.binding.xquarebc.XQuareBCException;
import org.objectweb.petals.binding.xquarebc.XQuareSUHandler;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.objectweb.petals.component.common.util.WSDLHelper;
import org.w3c.dom.Document;

/**
 * This listener "listens" to new data in any XQuareBC service unit's target database
 * and builds upon the XQuery capabilities of the companion XQuareBCJBIProcessor in
 * order to get them, format them as XML and send it as a message to the configured
 * service. The "listening" part is done by regularly looking for "new data" as the
 * results of a configured SQL query.
 * <p/>
 * 
 * WARNING : polling is not a best practice to integrate databases. Databases are
 * usually application-driven, and those applications (or their users) should drive
 * the "new data" lookup process. This implementation is provided as fitting the most
 * generic needs.
 * 
 * <h4>How it works</h4>
 * <p/>
 * This default implementation identifies "new data" by a given field that must be
 * incremental - that is, be different and always greated in newly inserted records.
 * The usual case for it is being a record identifier or primary key, be it works
 * well with other kind of fields as well, like timestamps. Concretely, it executes
 * the configured "oldDataDelimiterSelect" to get the old "new data delimiting value",
 * then it executes the configured "newDataDelimiterSelect" to get the new "new data
 * delimiting value" as available in the actual data. If they are different,
 * the configured XQuery is then executed through the XQuareBCJBIProcessor and its
 * results formatted as an XML document and sent as a JBI message to the configured
 * JBI service. Finally the "old data delimiting value" is updated to the new one.
 * <br/>
 * Notes :
 * <br/>
 * This requires a new table to be added to the target database, that will contain
 * a single record with a single field being the last seen "maximum data delimiting
 * value". This table may start empty, in which case the "initDataDelimiter" will be
 * used the first time.
 * <br/>
 * Before executing the XQuery, in its code the ##OLD_DATA_DELIMITER## qnd
 * ##NEW_DATA_DELIMITER## expressions are replaced by the values found in the
 * preceding executions of "oldDataDelimiterSelect" and "newDataDelimiterSelect".
 * <br/>
 * Many such listeners may be deployed in different service units to listen
 * to different kinds of "new data" in the same database.
 * <br/>
 * Since it uses the XQuareBCJBIProcessor's "query" capabilities, look at its
 * documentation for more details on how it behaves.
 * <p/>
 * 
 * <h4>Configuration</h4>
 * The datasource to the target database and the XQuareBCJBIProcessor's XQuery
 * capabilities are configured through the XQuareBCJBIProcessor (look at its
 * javadoc for details).
 * The configuration specific to this listener is configured through the same
 * *.properties file provided by a given service unit, and uses the following
 * additional properties :
 * 
 * <ul>
 * <li>newdatalistener.service : The fully qualified name of the service to which the new data
 * messages will be sent. Ex. {http://petals.objectweb.org/}HelloworldService</li>
 * <li>newdatalistener.operation : This service's operation. Ex. sayHello</li>
 * <li>newdatalistener.delimiter.oldDataDelimiterSelect : An SQL query that must return the
 * single value of the saved, old data delimiter. Ex. select max(bid_date) from bids_xqbc</li>
 * <li>newdatalistener.delimiter.newDataDelimiterSelect : An SQL query that must return the
 * single value of the current "new data" delimiter. Ex. select max(bid_date) from bids</li>
 * <li>newdatalistener.delimiter.initDataDelimiter : The new data delimiter initial value if
 * none in its own table and field. Ex. 1970-01-01</li>
 * <li>newdatalistener.delimiter.newDataDelimiterUpdates : An SQL insert that must update the
 * value of the "new data" delimiter from "old" to "new". Note that the ##NEW_DATA_DELIMITER##
 * expression is replaced within the SQL code before its execution. Ex. delete from bids_xqbc;\
 * <li>insert into bids_xqbc (bid_date) values ("##NEW_DATA_DELIMITER##")
 * <li>newdatalistener.delimiter.newDataXquery : An XQuery that must return and format the
 * current new data. Note that the ##OLD_DATA_DELIMITER## and ##NEW_DATA_DELIMITER##
 * expressions are replaced within the SQL code before its execution. Ex. for $i in collection("bids")/bids \
 * <li>where $i/bid_date > xs:date("##OLD_DATA_DELIMITER##") and $i/bid_date <= xs:date("##NEW_DATA_DELIMITER##") return \
 * <bid>{ $i/bid_date }</bid></li>
 * </ul>
 * 
 * @version $Rev: 250 $Date: {date}
 * @since Petals 1.0
 * @author Marc Dutoo - Open Wide
 * 
 */
public class XQuareBCListener implements Runnable {
    
    private ComponentContext context;
    private DeliveryChannel channel;
    private Logger logger;
    private Map<String, Properties> serviceToPropertiesMap;

    protected XQuareBCJBIProcessor processor;
    
    private boolean running;
    

    /**
     * Creates a new XQuareBCListener.
     * @param context
     * @param channel
     * @param logger
     * @param serviceToPropertiesMap
     * @param processor
     */
    public XQuareBCListener(ComponentContext context, DeliveryChannel channel,
        Logger logger, HashMap<String, Properties> serviceToPropertiesMap,
        XQuareBCJBIProcessor processor) {
        super();
        this.context = context;
        this.channel = channel; // needed to write to
        this.logger = logger;
        this.serviceToPropertiesMap = serviceToPropertiesMap;
        this.processor = processor; // used to get access to datasource etc.
    }
    

    /**
     * Main listener "polling" method.
     */
    public void run() {

        running = true;

        while (running) {
            try {
                // do it for all properties i.e. for all service units deployed
                // NB. Using own instance of map so several xquare emitter SAs can coexist
                Map<String, Properties> myServiceToPropertiesMap = new HashMap<String, Properties>(serviceToPropertiesMap);
                for (Properties serviceProps : myServiceToPropertiesMap.values()) {
                    process(serviceProps);
                   
                   // NB. another way to do it would be :
                   //List<String> xqueries = getXqueries(serviceProps); // cache
                   //for (String xquery : xqueries) {
                   //    process((String) xquery);
                   //}
                }
                try {
                    Thread.sleep(2000); // NB. could be in a prop
                } catch (InterruptedException e) {
                    this.logger.log(Level.WARNING, e.getMessage());
                }
            } catch (RuntimeException e) {
                this.logger.log(Level.WARNING, "Error while processing "
                    + "configured emission service properties", e);
            }
        }

    }


    /**
     * Stops the polling loop
     *
     */
    public void stopProcessing() {
        running = false;
    }

    

    /**
     * Sends an inOnly message
     * @param bodySource
     * @param service
     * @param operation
     * @return
     * @throws XQuareBCException
     */
    public boolean handleInOnlyMessage(Source bodySource, QName service,
        String operation) throws XQuareBCException {

        try {
            InOnly msg = null;
            msg = channel.createExchangeFactory().createInOnlyExchange();
            
            // Create and send messageExchange
            NormalizedMessage inNm = msg.createMessage();
            // NB. no attach
            inNm.setContent(bodySource);
            msg.setMessage(inNm, "IN");
            msg.setService(service);
            msg.setOperation(new QName(operation));
                
            channel.sendSync(msg);
            if (ExchangeStatus.DONE.equals(msg.getStatus())) {
                return true;
            } else {
                return false;
            }

        } catch (Exception e) {
            String msg = "Error while emitting data as an InOnly message "
                + "to service " + service + " and operation " + operation;
            logger.log(Level.SEVERE, msg, e);
            throw new XQuareBCException(msg, e);
        }
    }

    /**
     * Sends an inOut message
     * @param bodySource
     * @param service
     * @param operation
     * @return
     * @throws XQuareBCException
     */
    public boolean handleInOutMessage(Source bodySource, QName service,
        String operation) throws XQuareBCException {
        InOut msg = null;
        String response = "";
        try {

            msg = channel.createExchangeFactory().createInOutExchange();

            NormalizedMessage inNm = msg.createMessage();
            // NB. no attach
            inNm.setContent(bodySource);
            msg.setMessage(inNm, "IN");
            msg.setService(service);
            msg.setOperation(new QName(operation));
            
            channel.sendSync(msg);

            // Verify if request was success,if not, handle the error message
            if (msg.getStatus().equals(ExchangeStatus.ERROR)) {
                // fault
                // TODO handle response !! output it as a file ??
                response = SourceHelper.createString(msg.getFault()
                    .getContent());
                logger.log(Level.SEVERE, "XQuareBC data emission "
                    + "to service " + service + " and operation " + operation
                    + " received error response :" + response);
            } else if (msg.getMessage("OUT") != null) {
                // success
                // if there wasn't errors , binding component handle out reponse
                NormalizedMessage outNm = msg.getMessage("OUT");
                response = SourceHelper.createString(outNm.getContent());
                logger.log(Level.FINE, "XQuareBC data emission "
                    + "to service " + service + " and operation " + operation
                    + " received response :" + response);
            } else {
                logger.log(Level.WARNING, "XQuareBC data emission "
                    + "to service " + service + " and operation " + operation
                    + " received no (null) response message");
            }
            // Send status DONE to end exchange
            msg.setStatus(ExchangeStatus.DONE);
            channel.send(msg);
            return true;

        } catch (Exception e) {
            String errMsg = "Error while emitting data as an InOut message "
                + "to service " + service + " and operation " + operation;
            logger.log(Level.SEVERE, errMsg, e);
            throw new XQuareBCException(errMsg, e);
        }
    }

    
    /**
     * Tries to find and process new data according to the configuration
     * of the given service unit. 
     * @param inOut
     * @param service
     * @param operation
     * @param serviceProps
     * @throws Exception
     */
    public void handleNewData(boolean inOut, QName service,
        String operation, Properties serviceProps) throws Exception {
        
        // getting the datasource
        DataSource ds = processor.getDataSource(serviceProps);
        Connection conn = ds.getConnection();
        // conn.setAutoCommit(false); // would be better but...
        
        // getting a new data detection and handling strategy
        NewDataStrategy newDataStrategy = new DelimiterNewDataStrategy(processor,
            conn, serviceProps, XQuareSUHandler.LISTENER_PROP_PREFIX,
            logger);
        
        // now detecting new data if any :
        if (!newDataStrategy.detectNewData()) {
            // no data or no new data (when looping without having new data in between) : do nothing
            return;
        }
        // now querying for new data and getting the formatted new data message :
        Source bodySource = newDataStrategy.getNewData();
        logger.log(Level.FINE, "XQuareBC data emission "
            + "to service " + service + " and operation " + operation
            + " sent data :" + SourceHelper.createString(bodySource));

        boolean written = false;
        try {
            if (!inOut) {
                written = handleInOnlyMessage(bodySource, service, operation);
            } else {
                // NB. in fact inOut and out very similar...
                written = handleInOutMessage(bodySource, service, operation);
            }
        } catch (XQuareBCException e) {
            this.logger.log(Level.SEVERE, e.getMessage());
        }
        if (written) {
            // updating known, "old" data
            newDataStrategy.updateDataKnown();
        }
    }
    

    /**
     * Logs that the given service is not available.
     * @param response
     * @param service
     */
    public void notFoundService(String response, String errorMessage) {
        this.logger.log(Level.SEVERE, "Service not found : " + errorMessage);
    }

    
    /**
     * Gets the targeted service configured in the given service unit
     * properties. If it does not exist (or its configured operation doesn't),
     * calls notFoundService() ; else calls handleNewData().
     * @param serviceProps
     */
    protected void process(Properties serviceProps) {
        boolean found = false;
        boolean isInOut = false;

        String serviceName = serviceProps.getProperty(
            XQuareSUHandler.LISTENER_SERVICE_PROP);
        String operation = serviceProps.getProperty(
            XQuareSUHandler.LISTENER_OPERATION_PROP);
        
        if (serviceName == null || serviceName.length() == 0
                || operation == null || operation.length() == 0) {
            // no newdatalistener for this xquare service
            return;
        }
        
        // looking for valid endpoint
        QName service = QName.valueOf(serviceName);
        String errorMessage = "Looking for endpoints for service " + service
            + " with operation " + operation + " : ";
        ServiceEndpoint[] serviceEndpoints = this.context.getEndpointsForService(service);
        ServiceEndpoint serviceEndpoint = null;
        if (serviceEndpoints.length > 0) {
            // using first endpoint found
            serviceEndpoint = serviceEndpoints[0];
            Document serviceDesc;
            try {
                serviceDesc = context.getEndpointDescriptor(serviceEndpoint);

                // Determines if the operation exists and if it is an inOnly or
                // inOut operation
                try {
                    found = WSDLHelper.hasOperationNamed(serviceDesc,
                        operation, service);
                    if (!found) {
                        errorMessage += "its WSDL definition (for endpoint "
                            + serviceEndpoint.getEndpointName() + ") has no such operation";
                    }
                    isInOut = WSDLHelper.isInOutOperation(serviceDesc, operation,
                        QName.valueOf(serviceName));
                } catch (Exception e) {
                    found = false;
                    isInOut = false;
                    errorMessage += "Error while looking in the wsdl definition "
                        + "for the required operation and its MEP for the endpoint "
                        + serviceEndpoint.getEndpointName();
                    this.logger.log(Level.FINE, errorMessage , e);
                }

            } catch (JBIException e) {
                errorMessage += "Error getting descriptor of endpoint "
                    + serviceEndpoint.getEndpointName();
                this.logger.log(Level.FINE, errorMessage, e);
            }
        } else {
            errorMessage += "No endpoint found";
        }
        if (!found) {
            String response = SourceHelper.createSoapFault(
                new XQuareBCException(errorMessage), "300");
            notFoundService(response, errorMessage);
            return;
        }
        
        // Handle the dataTable
        try {
            handleNewData(isInOut, service, operation, serviceProps);
        } catch (Exception e) {
            this.logger.log(Level.SEVERE, "Error while handling new data "
                + "for target service " + service + " with operation "
                + operation, e);
        }

    }

}
