package org.objectweb.rentacar.centraloffice.services.consultation;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;

import org.apache.log4j.Logger;
import org.objectweb.rentacar.centraloffice.services.CentralOfficeException;
import org.objectweb.rentacar.persistance.bo.Agency;
import org.objectweb.rentacar.persistance.bo.AgencyCriteria;
import org.objectweb.rentacar.persistance.bo.AgencyVO;
import org.objectweb.rentacar.persistance.bo.Car;
import org.objectweb.rentacar.persistance.bo.CarCriteria;
import org.objectweb.rentacar.persistance.bo.CarVO;
import org.objectweb.rentacar.persistance.dao.AgencyDAO;
import org.objectweb.rentacar.persistance.dao.CarDAO;
import org.objectweb.rentacar.persistance.dao.DAOException;

/**
 * This class expose services 
 * that allow to retrieve information in central office database.
 * @author ofabre
 *
 */
@WebService
public class ConsultationImpl {
	
	private Logger logger = Logger.getLogger(ConsultationImpl.class);
	
	/**
	 * Retrieve all agencies that match criteria given in parameter
	 * @param agencyCriteria criteria of the searched agencies
	 * @return The list of all agencies that match criteria
	 * @throws CentralOfficeException
	 */
	@WebMethod
	public AgencyVO[] retreiveAgencies(
							@WebParam(name="agencyCriteria")AgencyCriteria agencyCriteria) 
							throws CentralOfficeException{
		
		List<AgencyVO> agencies = new ArrayList<AgencyVO>();
		List<Agency> agenciesBO = null;
		try {
			logger.info("Central Office . retreiveAgencies : "+agencyCriteria.toString());
			agenciesBO = AgencyDAO.getInstance().retreiveAgencies(agencyCriteria);
		} catch (DAOException e) {
			logger.error("Error when retrieving agencies matching criteria :"+agencyCriteria.toString(), e);
			throw new CentralOfficeException("Error when retrieving agencies matching criteria :"+agencyCriteria.toString(), e);
		}
		for (Agency agency : agenciesBO) {
			agencies.add(new AgencyVO(agency));
		}
		
		return agencies.toArray(new AgencyVO[0]);
	}
	
	/**
	 * Retrieve the list of cars that match the given criterium 
	 * and that are available during the given period for all agencies
	 * @param criterium the car criteria for the research
	 * @param startDate the starting date
	 * @param endDate the ending date
	 * @return a CarVO table
	 * @throws CentralOfficeException 
	 */
	@WebMethod
	public CarVO[] consult(@WebParam(name="criterium")CarCriteria criterium, 
							@WebParam(name="startDate")Date startDate, 
							@WebParam(name="endDate")Date endDate) throws CentralOfficeException {
		Set<CarVO> availableCarInAgency;
		logger.info("Central Office . consult : "+criterium.toString()+", "+startDate.toString()+", "+endDate.toString());
		availableCarInAgency = availableCarsByCriterium(criterium, startDate, endDate);
		return availableCarInAgency.toArray(new CarVO[0]);
	}
	
	/**
	 * Retrieve the list of cars that match the given criterium 
	 * and that are available during the given period
	 * @param criterium the car criteria for the research
	 * @param start the starting date
	 * @param end the ending date
	 * @return
	 * @throws CentralOfficeException
	 */
	private Set<CarVO> availableCarsByCriterium(CarCriteria criterium, 
			Date start, Date end) throws CentralOfficeException {
			Set<CarVO> availableCars= new HashSet<CarVO>();
			boolean type=false, model= false, brand=false;
			
			if( criterium.getType()== null)
				type=true;
			
			if(criterium.getModel()== null )
				model=true;
			
			if(criterium.getBrand()==null)
				brand=true;

			try {
					Set<Car> cars= getAvailableCars(start, end);
					Set<CarVO> carsVO= new HashSet<CarVO>();
						
					for (Car car : cars) {
						carsVO.add(new CarVO(car));
					}
					
					for (CarVO carVO : carsVO) {
						if(type || carVO.getType().equals(criterium.getType()))
								if(model || carVO.getModel().equals(criterium.getModel()))
										if(brand || carVO.getBrand().equals(criterium.getBrand()))
													availableCars.add(carVO);
						
				}
				
			} catch (DAOException e) {
				logger.error("Error while retrieving available cars", e);
				throw new CentralOfficeException("Error while retrieving available cars", e);
			}
			
			return availableCars;
		}
	
	/**
	 * Returns all available cars in all agency for the given period
	 * @param start Starting date of the search
	 * @param end Ending date of the search
	 * @return A Set of avalaible cars in the agency for this period
	 * @throws DAOException
	 */
	private Set<Car> getAvailableCars(Date start, Date end) throws DAOException{
		Set<Car> result = getAllAgencyCars();
		result.removeAll(getReservedCars(start,end));
		return result;
	}
	
	/**
	 * Returns all cars of all agencies
	 * @return A Set of cars belonging to this agency
	 * @throws DAOException
	 */
	private Set<Car> getAllAgencyCars() throws DAOException{
		return CarDAO.getInstance().getAllAgencyCars();
	}
	
	/**
	 * Returns all reserved cars for the given period
	 * @param start Starting date of the search
	 * @param end Ending date of the search
	 * @return A Set of reserved cars for this period
	 * @throws DAOException
	 */
	private Set<Car> getReservedCars(Date start, Date end) throws DAOException{
		return (Set<Car>) CarDAO.getInstance().getReservedCarsByDates(start,end);
	}

}
