package org.objectweb.rentacar.persistance.database;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.sql.DriverManager;
import java.sql.SQLException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.dbunit.DatabaseUnitException;
import org.dbunit.database.DatabaseConnection;
import org.dbunit.database.IDatabaseConnection;
import org.dbunit.dataset.IDataSet;
import org.dbunit.dataset.xml.FlatXmlDataSet;
import org.dbunit.operation.DatabaseOperation;
import org.hibernate.HibernateException;
import org.hibernate.cfg.AnnotationConfiguration;
import org.hibernate.tool.hbm2ddl.SchemaExport;
import org.hsqldb.Server;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

/**
 * Database manager
 */
public class DatabaseManager {

	/**
	 * Instance of the server
	 */
	private Server server = new Server();

	/**
	 * Start the database server
	 * @param host host
	 * @param port port
	 * @param dbName name of the database
	 * @param dbPath path of the database
	 */
	public void runServer(String host, String port, String dbName, String dbPath) {
		server.setAddress(host);
		server.setDatabaseName(0, dbName);
		server.setDatabasePath(0, dbPath);
		server.setPort(Integer.parseInt(port));
		server.setSilent(true);
		server.setTrace(false);
		server.start();
	}

	/**
	 * Shutdown the hsqldb server
	 */
	public void shutdownServer() {
		server.stop();
	}

	/**
	 * Build database schema
	 * @param url
	 * @param driver
	 * @param dialect
	 * @param userName
	 * @param password
	 * @throws DatabaseException 
	 */
	public void buildDatabaseSchema(String url, String driver, String dialect, String userName, String password) throws DatabaseException {
		AnnotationConfiguration conf = new AnnotationConfiguration();
		try {
			DocumentBuilderFactory fact = DocumentBuilderFactory.newInstance();
			DocumentBuilder builder = fact.newDocumentBuilder();
			Document document = builder.parse(DatabaseManager.class.getClassLoader().getResourceAsStream("hibernate.cfg.xml"));
			conf.configure(document);
			conf.setProperty("hibernate.connection.username", userName);
			conf.setProperty("hibernate.connection.password", password);
			conf.setProperty("hibernate.connection.url", url);
			conf.setProperty("hibernate.dialect", dialect);
			conf.setProperty("hibernate.connection.driver_class", driver);
			SchemaExport export = new SchemaExport(conf);
			export.create(false, true);
		}
		catch (HibernateException e) {
			throw new DatabaseException("Error building the schema", e);
		}
		catch (SAXException e) {
			throw new DatabaseException("Error building the schema", e);
		}
		catch (IOException e) {
			throw new DatabaseException("Error building the schema", e);
		}
		catch (ParserConfigurationException e) {
			throw new DatabaseException("Error building the schema", e);
		}
	}

	/**
	 * @param host
	 * @param port
	 * @param dbName
	 * @param dsFileName
	 * @throws DatabaseException
	 */
	public void fillDatabase(String host, String port, String dbName, String dsFileName) throws DatabaseException {		
		IDatabaseConnection connection = null;
		try {
			Class.forName("org.hsqldb.jdbcDriver");
			connection = (IDatabaseConnection) new DatabaseConnection(DriverManager.getConnection("jdbc:hsqldb:hsql://" + host + " :" + port + "/" + dbName, "sa", ""));
			IDataSet dataSet = new FlatXmlDataSet(DatabaseManager.class.getClassLoader().getResourceAsStream(dsFileName));
			DatabaseOperation.DELETE.execute(connection, dataSet);
			DatabaseOperation.CLEAN_INSERT.execute(connection, dataSet);
		}
		catch (DatabaseUnitException e) {
			throw new DatabaseException("Error filling the database", e);
		}
		catch (SQLException e) {
			throw new DatabaseException("Error filling the database", e);
		}
		catch (FileNotFoundException e) {
			throw new DatabaseException("Error filling the database", e);
		}
		catch (IOException e) {
			throw new DatabaseException("Error filling the database", e);
		}
		catch (Exception e) {
			throw new DatabaseException("Error filling the database", e);
		}
		finally {
			try {
				if (connection != null) {
					connection.close();
				}
			}
			catch (SQLException e) {
				throw new DatabaseException("Error filling the database", e);
			}
		}
	}

}
