/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: ComponentContextImpl.java 1191 2006-11-02 11:16:29Z msauvage $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.component.context;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.Set;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.management.MBeanNames;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.MBeanServer;
import javax.naming.InitialContext;
import javax.xml.namespace.QName;

import org.objectweb.petals.jbi.component.lifecycle.ComponentLifeCycle;
import org.objectweb.petals.jbi.management.service.EndpointService;
import org.objectweb.petals.jbi.management.service.LifeCycleManagerService;
import org.objectweb.petals.jbi.messaging.DeliveryChannelImpl;
import org.objectweb.petals.jbi.registry.AbstractEndpoint;
import org.objectweb.petals.jbi.registry.ConsumerEndpoint;
import org.objectweb.petals.jbi.registry.InternalEndpoint;
import org.objectweb.petals.jbi.routing.Router;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;
import org.objectweb.petals.util.PetalsRuntimeException;
import org.objectweb.petals.util.StringHelper;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;

/**
 * @version $Revision: 1191 $ $Date: 2006-08-04 12:11:16 +0200 (ven., 04 août
 *          2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 * @author wjoseph - eBMWebsourcing
 */
public class ComponentContextImpl implements
        javax.jbi.component.ComponentContext {

    protected ConsumerEndpoint address;

    protected MBeanNames beanNames;

    protected MBeanServer beanServer;

    protected DeliveryChannelImpl deliveryChannel;

    protected JBIDescriptor descriptor;

    /**
     * The endpoint service is used to resolved the interface name in a
     * distributed context
     */
    protected EndpointService endpointService;

    /**
     * references of registered external serviceEndpoints
     */
    protected Set<ServiceEndpoint> externalEndpoints;

    protected String installationRoot;

    /**
     * references of registered internal serviceEndpoints
     */
    protected Set<AbstractEndpoint> internalEndpoints;

    protected Component jbiComponent;

    /**
     * Monolog Logger instance.
     */
    protected org.objectweb.util.monolog.api.Logger logger;

    protected LifeCycleManagerService manager;

    protected Logger rootLogger;

    protected String workspaceRoot;

    protected InitialContext namingContext;

    /**
     * The container router instance.
     */
    private Router router;

    public ComponentContextImpl() {
    }

    /**
     * Constructor of the ComponentContextImpl class
     * 
     * @param descriptor
     *            component descriptor
     * @param installationRoot
     *            Path to the component installation root
     * @param workspaceRoot
     *            path to the component workspace
     * @param beanNames
     * @param beanServer
     * @param manager
     *            current lifeCycleManagerService
     * @param address
     *            consumer endpoint
     * @param endpointService
     * @param logger
     *            InstallationService monolog logger
     * @param router
     *            Petals router
     */
    public ComponentContextImpl(JBIDescriptor descriptor,
            String installationRoot, String workspaceRoot,
            MBeanNames beanNames, MBeanServer beanServer,
            LifeCycleManagerService manager, ConsumerEndpoint address,
            EndpointService endpointService,
            org.objectweb.util.monolog.api.Logger logger, Router router,
            InitialContext namingContext) {
        super();
        if (installationRoot == null) {
            throw new IllegalArgumentException(
                    "installationRoot must not be null");
        } else {
            if (installationRoot.length() == 0) {
                throw new IllegalArgumentException(
                        "installationRoot must not be empty String");
            }
        }
        if (beanNames == null) {
            throw new IllegalArgumentException("beanNames must not be null");
        }
        if (beanServer == null) {
            throw new IllegalArgumentException("beanServer must not be null");
        }
        if (workspaceRoot == null) {
            throw new IllegalArgumentException("workspaceRoot must not be null");
        } else {
            if (workspaceRoot.length() == 0) {
                throw new IllegalArgumentException(
                        "workspaceRoot must not be empty String");
            }
        }
        this.descriptor = descriptor;
        this.installationRoot = installationRoot;
        this.beanNames = beanNames;
        this.beanServer = beanServer;
        this.manager = manager;
        this.address = address;
        this.internalEndpoints = new HashSet<AbstractEndpoint>();
        this.externalEndpoints = new HashSet<ServiceEndpoint>();
        this.endpointService = endpointService;
        this.logger = logger;
        this.router = router;
        this.namingContext = namingContext;
    }

    /**
     * @see javax.jbi.component.ComponentContext#activateEndpoint(javax.xml.namespace.QName,
     *      java.lang.String)
     */
    public ServiceEndpoint activateEndpoint(QName serviceName,
            String endpointName) throws JBIException {
        if (serviceName == null) {
            throw new IllegalArgumentException("serviceName must be non-null");
        }
        if (endpointName == null) {
            throw new IllegalArgumentException("endpointName must be non-null");
        }
        InternalEndpoint se = endpointService.activateEndpoint(serviceName,
                endpointName, address);
        internalEndpoints.add(se);
        return se;
    }

    /**
     * @see javax.jbi.component.ComponentContext#deactivateEndpoint(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public void deactivateEndpoint(ServiceEndpoint endpoint)
        throws JBIException {
        if (endpoint == null) {
            throw new IllegalArgumentException("endpoint must be non-null");
        }
        endpointService.deactivateEndpoint(endpoint);
        internalEndpoints.remove(endpoint);
    }

    /**
     * deregister all endpoints (internal and external)
     * 
     */
    public void deregisterAllEndpoints() throws JBIException {
        for (AbstractEndpoint se : internalEndpoints) {
            endpointService.deactivateEndpoint(se);
            internalEndpoints.remove(se);
        }
        for (ServiceEndpoint se : externalEndpoints) {
            endpointService.deregisterExternalEndpoint(se);
            externalEndpoints.remove(se);
        }
    }

    /**
     * @see javax.jbi.component.ComponentContext#deregisterExternalEndpoint(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public void deregisterExternalEndpoint(ServiceEndpoint externalEndpoint)
        throws JBIException {
        endpointService.deregisterExternalEndpoint(externalEndpoint);
        externalEndpoints.remove(externalEndpoint);
    }

    public ConsumerEndpoint getAddress() {
        return address;
    }

    public Component getComponent() {
        return jbiComponent;
    }

    /**
     * @see javax.jbi.component.ComponentContext#getComponentName()
     */
    public String getComponentName() {
        return descriptor.getComponent().getIdentification().getName();
    }

    /**
     * @see javax.jbi.component.ComponentContext#getDeliveryChannel()
     */
    public DeliveryChannel getDeliveryChannel() throws MessagingException {
        if (deliveryChannel == null) {
            deliveryChannel = createDeliveryChannel();
        } else {
            if (deliveryChannel.isOpened()) {
                throw new MessagingException(
                        "The previous DeliveryChannel is still opened");
            } else {
                deliveryChannel = createDeliveryChannel();
            }
        }
        return deliveryChannel;
    }

    public DeliveryChannelImpl getDeliveryChannelImpl() {
        return deliveryChannel;
    }

    /**
     * @see javax.jbi.component.ComponentContext#getEndpoint(javax.xml.namespace.QName,
     *      java.lang.String)
     */
    public ServiceEndpoint getEndpoint(QName service, String name) {
        if (service == null) {
            throw new IllegalArgumentException("service must not be null");
        }
        if (name == null) {
            throw new IllegalArgumentException("name must not be null");
        }
        return endpointService.getEndpoint(service, name, false);
    }

    /**
     * @see javax.jbi.component.ComponentContext#getEndpointDescriptor(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public Document getEndpointDescriptor(ServiceEndpoint endpoint)
        throws JBIException {
        Document result = null;
        if (endpoint == null) {
            throw new IllegalArgumentException("endpoint must not be null");
        }
        if (getEndpoint(endpoint.getServiceName(), endpoint.getEndpointName()) != null) {
            /*
             * Checks if the endpoint exist and if it is activated
             */
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoint;
            if (abstractEndpoint.getEndpointService() == null) {
                abstractEndpoint.setEndpointService(endpointService);
            }
            result = abstractEndpoint.getEndpointDescriptor();
        }
        return result;
    }

    /**
     * @see javax.jbi.component.ComponentContext#getEndpoints(javax.xml.namespace.QName)
     */
    public ServiceEndpoint[] getEndpoints(QName interfaceName) {
        return endpointService.getInternalEndpointsForInterface(interfaceName);
    }

    /**
     * @see javax.jbi.component.ComponentContext#getEndpointsForService(javax.xml.namespace.QName)
     */
    public ServiceEndpoint[] getEndpointsForService(QName serviceName) {
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        return endpointService.getInternalEndpointsForService(serviceName);
    }

    /**
     * @see javax.jbi.component.ComponentContext#getExternalEndpoints(javax.xml.namespace.QName)
     */
    public ServiceEndpoint[] getExternalEndpoints(QName interfaceName) {
        if (interfaceName == null) {
            throw new IllegalArgumentException(
                    "Interface name must be non null");
        }
        return endpointService.getExternalEndpointsForInterface(interfaceName);
    }

    /**
     * @see javax.jbi.component.ComponentContext#getExternalEndpointsForService(javax.xml.namespace.QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForService(QName serviceName) {
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        return endpointService.getExternalEndpointsForService(serviceName);
    }

    /**
     * @see javax.jbi.component.ComponentContext#getInstallRoot()
     */
    public String getInstallRoot() {
        if (StringHelper.isEmpty(installationRoot)) {
            throw new PetalsRuntimeException(
                    "Installation root must be non-null and non empty "
                            + "for a correctly installed component. Check component installation !");
        }
        return installationRoot;
    }

    /**
     * @see javax.jbi.component.ComponentContext#getLogger(java.lang.String,
     *      java.lang.String)
     */
    public Logger getLogger(String suffix, String resourceBundleName)
        throws MissingResourceException, JBIException {
        String loggerName = rootLogger.getName();
        if (suffix == null) {
            throw new IllegalArgumentException("suffix must not be null");
        } else {
            if (suffix.length() > 0) {
                loggerName += "." + suffix;
            }
        }
        Logger l = Logger.getLogger(loggerName, resourceBundleName);
        if (l == null) {
            l = Logger.getAnonymousLogger();
        }
        if (l == null) {
            throw new JBIException("No logger can be created");
        }
        return l;

    }

    /**
     * @see javax.jbi.component.ComponentContext#getMBeanNames()
     */
    public MBeanNames getMBeanNames() {
        return beanNames;
    }

    /**
     * @see javax.jbi.component.ComponentContext#getMBeanServer()
     */
    public MBeanServer getMBeanServer() {
        return beanServer;
    }

    /**
     * @see javax.jbi.component.ComponentContext#getNamingContext()
     */
    public InitialContext getNamingContext() {
        return namingContext;
    }

    /**
     * TODO not implemented
     * 
     * @see javax.jbi.component.ComponentContext#getTransactionManager()
     */
    public Object getTransactionManager() {
        throw new NoSuchMethodError("not yet implemented in Petals");
    }

    /**
     * @see javax.jbi.component.ComponentContext#getWorkspaceRoot()
     */
    public String getWorkspaceRoot() {
        return workspaceRoot;
    }

    /**
     * @see javax.jbi.component.ComponentContext#registerExternalEndpoint(javax.jbi.servicedesc.ServiceEndpoint)
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint)
        throws JBIException {
        if (externalEndpoint == null) {
            throw new IllegalArgumentException(
                    "externalEndpoint must not be null");
        }
        endpointService.registerExternalEndpoint(externalEndpoint);
        externalEndpoints.add(externalEndpoint);
    }

    // //////////////////////////////////////////////////////
    // Implementation specific
    // //////////////////////////////////////////////////////

    /**
     * TODO resolveEndpointReference : refactor
     * 
     * @see javax.jbi.component.ComponentContext#resolveEndpointReference(org.w3c.dom.DocumentFragment)
     */
    public ServiceEndpoint resolveEndpointReference(DocumentFragment epr) {
        if (epr == null) {
            throw new IllegalArgumentException("epr must not be null");
        }
        ServiceEndpoint result = null;
        // Search for all engines components if one can resolve the EPR.
        Map engines = manager.getEngineCompoLifeCycles();

        synchronized (engines) {
            for (Iterator iter = engines.values().iterator(); iter.hasNext()
                    && result == null;) {
                ComponentLifeCycle lifeCycle = (ComponentLifeCycle) iter.next();

                result = lifeCycle.getComponent().resolveEndpointReference(epr);
            }
        }

        if (result == null) {
            // Search for all binding components if one can resolve the EPR.
            Map bindings = manager.getBindingCompoLifeCycles();

            synchronized (bindings) {
                for (Iterator iter = bindings.values().iterator(); iter
                        .hasNext()
                        && result == null;) {
                    ComponentLifeCycle lifeCycle = (ComponentLifeCycle) iter
                            .next();

                    result = lifeCycle.getComponent().resolveEndpointReference(
                            epr);
                }
            }
        }
        return result;

    }

    public void setComponent(Component component) {
        this.jbiComponent = component;
    }

    public void setDeliveryChannel(DeliveryChannelImpl deliveryChannel) {
        this.deliveryChannel = deliveryChannel;
    }

    public void setRootLogger(Logger logger) {
        rootLogger = logger;
    }

    protected DeliveryChannelImpl createDeliveryChannel() {
        return new DeliveryChannelImpl(this, router, logger);
    }
    
    public JBIDescriptor getJBIDescriptor() {
        return this.descriptor;
    }

}
