/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: LifeCycle.java,v 1.2 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.component.lifecycle;

import javax.jbi.JBIException;
import javax.management.AttributeChangeNotification;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanNotificationInfo;
import javax.management.Notification;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationBroadcasterSupport;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectName;

import org.objectweb.petals.jbi.component.event.StateChangeFailedEvent;
import org.objectweb.petals.jbi.component.event.StateChangedEvent;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.PetalsRuntimeException;

/**
 * Implementation of the <code>LifeCycle</code> mbean. This lifecycle mbean
 * sends notifications about changes of state.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 */
public abstract class LifeCycleAbstract implements LifeCycleMBean,
        NotificationBroadcaster {

    protected Object activitySynchronizer;

    protected boolean initialized;

    protected LoggingUtil log;

    protected ObjectName mBeanName;

    protected long sequenceNumber;

    protected String state;

    NotificationBroadcasterSupport notifSupport = new NotificationBroadcasterSupport();

    public LifeCycleAbstract(ObjectName mBeanName, LoggingUtil log) {
        this.log = log;

        log.call();

        activitySynchronizer = new Object();

        state = SHUTDOWN;

        initialized = false;

        this.mBeanName = mBeanName;

    }

    public void addNotificationListener(NotificationListener arg0,
            NotificationFilter arg1, Object arg2)
        throws IllegalArgumentException {
        notifSupport.addNotificationListener(arg0, arg1, arg2);
    }

    public abstract void doInit() throws JBIException;

    public abstract void doShutdown() throws JBIException;

    // ////////////////////////////////////////////
    // LifeCycle implementation
    // ////////////////////////////////////////////

    public abstract void doStart() throws JBIException;

    public abstract void doStop() throws JBIException;

    /**
     * @see javax.jbi.management.LifeCycleMBean#getCurrentState()
     */
    public String getCurrentState() {
        if (!(LifeCycleMBean.SHUTDOWN.equals(state)
                || LifeCycleMBean.STARTED.equals(state)
                || LifeCycleMBean.STOPPED.equals(state) || LifeCycleMBean.UNKNOWN
                .equals(state))) {
            throw new PetalsRuntimeException(
                    "State '" + state + "' isn't defined by JBI specification !");
        }
        return state;
    }

    //
    // Notifications support
    //

    public ObjectName getMBeanName() {
        return mBeanName;
    }

    public MBeanNotificationInfo[] getNotificationInfo() {
        String[] types = new String[] {AttributeChangeNotification.ATTRIBUTE_CHANGE};

        String name = AttributeChangeNotification.class.getName();

        String description = "Notification about a change of the lifeCycle state or a state change failure.";

        MBeanNotificationInfo info = new MBeanNotificationInfo(types, name,
                description);

        return new MBeanNotificationInfo[] {info};
    }

    public void removeNotificationListener(NotificationListener arg0)
        throws ListenerNotFoundException {
        notifSupport.removeNotificationListener(arg0);
    }

    public void setMBeanName(ObjectName beanName) {
        mBeanName = beanName;
    }

    // ////////////////////////////////////////////
    // ElementListener implementation
    // ////////////////////////////////////////////

    /**
     * Set the <code>State</code> reference of the current state. An
     * <code>LifeCycleStateEventAbstract</code> is thrown.
     * 
     * @param state
     */
    public synchronized void setState(String state) throws JBIException {
        stateChanged(new StateChangedEvent(this, this.state, state));

        this.state = state;
    }

    /**
     * Shutdown the component. A notification will be thrown in case of
     * runtimeException or incoherent state.
     * 
     * @see javax.jbi.management.LifeCycleMBean#shutdown()
     * @throws JBIException
     *             in case of runtimeException or incoherent state.
     */
    public void shutDown() throws JBIException {
        log.start();

        if (STOPPED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    setState(SHUTDOWNING);

                    doShutdown();
                    
                    initialized = false;

                    setState(SHUTDOWN);
                } catch (Throwable re) {
                    setState(UNKNOWN);

                    String msg = "An exception occured during the shutdown.";

                    JBIException e = new JBIException(msg, re);

                    log.error(msg, e);

                    stateChangeFailed(new StateChangeFailedEvent(this, e));

                    throw e;
                }
            }
        } else {
            String msg = "The Object can not be shutdowned in this state.";

            JBIException e = new JBIException(msg);

            log.error(msg, e);

            stateChangeFailed(new StateChangeFailedEvent(this, e));

            throw e;
        }
        log.end();
    }

    /**
     * Start the component. A notification will be thrown in case of
     * runtimeException or incoherent state.
     * 
     * @see javax.jbi.management.LifeCycleMBean#start()
     * @throws JBIException
     *             in case of runtimeException or incoherent state.
     */
    public void start() throws JBIException {
        log.start();

        if (SHUTDOWN.equals(state) || STOPPED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {
                    if (!initialized) {
                        setState(INITIALIZING);

                        doInit();

                        initialized = true;
                    }
                    setState(STARTING);

                    doStart();

                    setState(STARTED);
                } catch (Throwable re) {
                    setState(UNKNOWN);

                    String msg = "An exception occured during the start.";

                    JBIException e = new JBIException(msg, re);

                    log.error(msg, e);

                    stateChangeFailed(new StateChangeFailedEvent(this, e));

                    throw e;
                }
            }
        } else {
            String msg = "The Object can not be started in this state.";

            JBIException e = new JBIException(msg);

            log.error(msg, e);

            stateChangeFailed(new StateChangeFailedEvent(this, e));

            throw e;
        }
        log.end();
    }

    /**
     * <i>stateChanged</i> is called when the state of a lifecycle element
     * changes. A JMX-notification is thrown.
     * 
     * @see petals.management.element.ElementStateListener#stateChanged(java.util.EventObject)
     */
    public synchronized void stateChanged(StateChangedEvent event) {
        log.call();
        if (mBeanName != null) {
            Notification n = new AttributeChangeNotification(mBeanName,
                    sequenceNumber++, System.currentTimeMillis(), "State of '"
                            + mBeanName.toString() + "' changed : "
                            + event.getOldState() + " -> "
                            + event.getNewState(), "state", "String", event
                            .getOldState(), event.getNewState());

            notifSupport.sendNotification(n);
        }
    }

    /**
     * <i>stateChangeFailed</i>> is called when the change of state of a
     * lifecycle element fails. A JMX-notification throws the relative
     * JBIException, .
     * 
     * @see petals.management.element.ElementStateListener#stateChanged(java.util.EventObject)
     */
    public synchronized void stateChangeFailed(StateChangeFailedEvent event) {
        log.call();
        if (mBeanName != null) {
            Notification n = new AttributeChangeNotification(mBeanName,
                    sequenceNumber++, System.currentTimeMillis(),
                    "State change of '" + mBeanName.toString() + "' failed : "
                            + event.getException().getMessage(), "exception",
                    "String", null, event.getException().toString());

            notifSupport.sendNotification(n);
        }
    }

    /**
     * Stop the component. A notification will be thrown in case of
     * runtimeException or incoherent state.
     * 
     * @see javax.jbi.management.LifeCycleMBean#stop()
     * @throws JBIException
     *             in case of runtimeException or incoherent state.
     */
    public void stop() throws JBIException {
        log.start();

        if (STARTED.equals(state)) {
            synchronized (activitySynchronizer) {
                try {

                    setState(STOPPING);

                    doStop();

                    setState(STOPPED);

                } catch (Throwable re) {
                    setState(UNKNOWN);

                    String msg = "An exception occured during the stop.";

                    JBIException e = new JBIException(msg, re);

                    log.error(msg, e);

                    stateChangeFailed(new StateChangeFailedEvent(this, e));

                    throw e;
                }
            }
        } else {
            String msg = "The Object can not be stoped in this state." + state;

            JBIException e = new JBIException(msg);

            log.error(msg, e);

            stateChangeFailed(new StateChangeFailedEvent(this, e));

            throw e;
        }
        log.end();
    }

}
