/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: AllSUDeploymentTask.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.deployment.deploy;

import java.io.InputStream;
import java.net.URI;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.ObjectName;

import org.objectweb.petals.jbi.component.lifecycle.ServiceAssemblyLifeCycle;
import org.objectweb.petals.jbi.management.deployment.DeploymentContextConstants;
import org.objectweb.petals.jbi.management.deployment.DeploymentServiceImpl;
import org.objectweb.petals.jbi.management.deployment.DeploymentTaskFactory;
import org.objectweb.petals.jbi.management.deployment.DeploymentUtils;
import org.objectweb.petals.jbi.management.service.LifeCycleManagerService;
import org.objectweb.petals.jbi.management.service.ManagementException;
import org.objectweb.petals.jbi.management.service.util.XMLResult;
import org.objectweb.petals.jbi.management.systemstate.SystemState;
import org.objectweb.petals.processor.Task;
import org.objectweb.petals.processor.TaskProcessor;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;
import org.objectweb.petals.tools.jbicommon.descriptor.ServiceUnit;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This task creates service assembly lifecycle element and for each Service
 * Unit included into the service assembly, executes the deployment process of a
 * service unit
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class AllSUDeploymentTask implements Task {

    /**
     * JBI Container :: Recovery Service
     */
    protected SystemState recoverySrv;

    /**
     * Platform Component :: Repository Service
     */
    protected RepositoryService repositorySrv;

    /**
     * JBI Container :: JMX Agent Admin
     */
    protected LifeCycleManagerService managerService;

    protected DeploymentServiceImpl deploymentService;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    public AllSUDeploymentTask(SystemState recoverySrv,
            RepositoryService repositorySrv,
            LifeCycleManagerService managerService,
            DeploymentServiceImpl deploymentService, LoggingUtil log) {
        super();
        this.recoverySrv = recoverySrv;
        this.repositorySrv = repositorySrv;
        this.managerService = managerService;
        this.deploymentService = deploymentService;
        this.log = log;
    }

    @SuppressWarnings("unchecked")
    public void execute(HashMap context) throws Exception {
        URI installationRoot = (URI) context
                .get(DeploymentContextConstants.SA_INSTALL_ROOT);

        JBIDescriptor descriptor = (JBIDescriptor) context
                .get(DeploymentContextConstants.SA_DESCRIPTOR);

        XMLResult xmlResult = (XMLResult) context
                .get(DeploymentContextConstants.XML_RESULT);

        Map<String, URI> suInstallRoots = (Map<String, URI>) context
                .get(DeploymentContextConstants.SU_INSTALL_ROOTS);

        List<TaskProcessor> suTaskProcessors = (List<TaskProcessor>) context
                .get(DeploymentContextConstants.SU_TASK_PROCESSORS);

        /*
         * Create service assembly lifecycle element
         */
        ServiceAssemblyLifeCycle saLifeCycle = createSALifeCycle(
                installationRoot, descriptor);

        /*
         * For each Service Unit included into the service assembly : execute
         * the deployment process of a service unit
         */
        deploySUS(installationRoot, saLifeCycle, descriptor, suInstallRoots,
                xmlResult, suTaskProcessors);

        /*
         * Fill Context
         */
        context.put(DeploymentContextConstants.SA_LIFECYCLE, saLifeCycle);
    }

    /**
     * * Deploy all Service Units to the JBI container
     * 
     * @param installationRoot
     *            the root {@link URI} of the service assembly
     * @param saLifeCycle
     *            the {@link ServiceAssemblyLifeCycle} of the service assembly
     * @param descriptor
     *            the service assembly jbi descriptor
     * @param suInstallRoots
     *            a map of all SU install roots
     * @param xmlResult
     *            xml task result
     * @throws ManagementException
     *             if an sus deployment fails caused by a severe error of the
     *             container
     * 
     */
    protected void deploySUS(URI installationRoot,
            ServiceAssemblyLifeCycle saLifeCycle, JBIDescriptor descriptor,
            Map<String, URI> suInstallRoots, XMLResult xmlResult,
            List<TaskProcessor> suTaskProcessors) throws Exception {
        List<ServiceUnit> serviceUnits = descriptor.getServiceAssembly()
                .getServiceUnits();
        if (serviceUnits != null) {
            StringBuffer componentOutput = new StringBuffer();
            for (ServiceUnit serviceUnit : serviceUnits) {

                /*
                 * Create Deployment context
                 */
                HashMap<String, Object> suDeploymentContext = new HashMap<String, Object>();
                suDeploymentContext.put(
                        DeploymentContextConstants.SA_LIFECYCLE, saLifeCycle);
                suDeploymentContext.put(
                        DeploymentContextConstants.SU_DESCRIPTOR, serviceUnit);
                suDeploymentContext.put(
                        DeploymentContextConstants.SU_INSTALL_ROOT,
                        suInstallRoots.get(DeploymentUtils
                                .getServiceUnitName(serviceUnit)));

                /*
                 * Create Deployment tasks
                 */
                DeploymentTaskFactory deploymentTaskFactory = deploymentService
                        .getDeploymentTaskFactory();
                Task suToComponentDeploymentTask = deploymentTaskFactory
                        .getSuToComponentDeploymentTask();
                Task createAndRegisterSULifeCycleTask = deploymentTaskFactory
                        .getCreateAndRegisterSULifeCycleTask();

                /*
                 * Create processor and add tasks
                 */
                TaskProcessor processor = new TaskProcessor(
                        suDeploymentContext, log);
                processor.addTask(suToComponentDeploymentTask);
                processor.addTask(createAndRegisterSULifeCycleTask);
                if (suTaskProcessors != null) {
                    suTaskProcessors.add(processor);
                }

                /*
                 * Launch processor
                 */
                if (!processor.process()) {
                    String msg = "Error occured during the deployment of a ServiceUnit";
                    throw new Exception(msg);
                }
                String xmlReturn = (String) suDeploymentContext
                        .get(DeploymentContextConstants.COMPONENT_XML_RESULT);
                if (!isValidComponentTask(xmlReturn, this.getClass()
                        .getClassLoader().getResourceAsStream(
                                "schema/management-message-component.xsd"))) {
                    xmlReturn = xmlResult.wrapComponentTaskResult(serviceUnit
                            .getTargetComponentName(), "deploy", xmlReturn);
                }
                componentOutput.append(xmlReturn);
            }
            xmlResult.addComponentTaskResult(componentOutput);
        }
    }

    /**
     * Creates a new {@link ServiceAssemblyLifeCycle}
     * 
     * @param installationRoot
     *            the root {@link URI} of the service assembly into the
     *            repository
     * @param descriptor
     *            the service assembly jbi descriptor
     * @return a new {@link ServiceAssemblyLifeCycle}
     * @throws ManagementException
     *             if deployment service MBean name retrieval fails
     */
    protected ServiceAssemblyLifeCycle createSALifeCycle(URI installationRoot,
            JBIDescriptor descriptor) throws ManagementException {
        // Retrieve the DeploymentServiceImpl ObjectName
        ObjectName objectName = managerService.getDeploymentServiceMBeanName();

        return new ServiceAssemblyLifeCycle(objectName, log, descriptor
                .getServiceAssembly(), repositorySrv, recoverySrv,
                installationRoot);

    }

    /**
     * Check if a component task result is valid against an xml schema
     * 
     * @param componentTask
     *            xml of the component task
     * @param xsd
     *            an {@link InputStream} of the xsd file
     * @return true if valid, false if not valid
     */
    protected boolean isValidComponentTask(String componentTask, InputStream xsd) {
        return XMLResult.isValidComponentTask(componentTask, xsd);
    }

    @SuppressWarnings("unchecked")
    public void undo(HashMap context) {
        List<TaskProcessor> suTaskProcessors = (List<TaskProcessor>) context
                .get(DeploymentContextConstants.SU_TASK_PROCESSORS);

        if (suTaskProcessors != null) {
            for (TaskProcessor processor : suTaskProcessors) {
                processor.unprocess();
            }
        }
    }

}
