/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SAAndSUExtractionTask.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.deployment.deploy;

import java.net.URI;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.jbi.management.deployment.DeploymentContextConstants;
import org.objectweb.petals.jbi.management.deployment.DeploymentUtils;
import org.objectweb.petals.jbi.management.service.ManagementException;
import org.objectweb.petals.jbi.management.service.PackageHandler;
import org.objectweb.petals.processor.Task;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;
import org.objectweb.petals.tools.jbicommon.descriptor.ServiceUnit;
import org.objectweb.petals.util.LoggingUtil;

/**
 * This task extracts service assembly package into the repository
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class SAAndSUExtractionTask implements Task {

    /**
     * Package Handler
     */
    protected PackageHandler packageHandler;

    /**
     * Platform Component :: Repository Service
     */
    protected RepositoryService repositorySrv;

    /**
     * logger wrapper
     */
    protected LoggingUtil log;

    public SAAndSUExtractionTask(PackageHandler packageHandler,
            RepositoryService repositorySrv, LoggingUtil log) {
        super();
        this.packageHandler = packageHandler;
        this.repositorySrv = repositorySrv;
        this.log = log;
    }

    @SuppressWarnings("unchecked")
    public void execute(HashMap context) throws Exception {
        JBIDescriptor descriptor = (JBIDescriptor) context
                .get(DeploymentContextConstants.SA_DESCRIPTOR);

        URI saArchiveURI = (URI) context
                .get(DeploymentContextConstants.ARCHIVE_URI);

        /*
         * Explode SA
         */
        URI saInstallRoot = explodeSA(saArchiveURI, descriptor);

        /*
         * Explodes SUs
         */
        Map<String, URI> suInstallRoots = expendSUS(saInstallRoot, descriptor);

        /*
         * Fill context
         */
        context.put(DeploymentContextConstants.SA_INSTALL_ROOT, saInstallRoot);
        context
                .put(DeploymentContextConstants.SU_INSTALL_ROOTS,
                        suInstallRoots);

    }

    /**
     * Expend service assembly package into the repository
     * 
     * @param saArchiveURI
     *            the service assembly archive {@link URI}
     * @param descriptor
     *            the service assembly descriptor
     * @return return installation {@link URI} of the expanded package, null if
     *         it fails
     * @throws ManagementException
     *             when it fails to expand service assembly package into
     *             repository
     */
    protected URI explodeSA(URI saArchiveURI, JBIDescriptor descriptor)
        throws ManagementException {

        String msg;
        URI installationRoot = null;

        try {
            installationRoot = repositorySrv.addServiceAssemblyPackage(
                    DeploymentUtils.getServiceAssemblyName(descriptor),
                    saArchiveURI);

        } catch (PetalsException pe) {
            msg = "Unexpected error adding JBI installation "
                    + "package into repository. ";
            log.error(msg, pe);
            throw new ManagementException(msg, pe);
        }

        if (installationRoot == null) {
            msg = "Unexpected Platform repository service error. The SA"
                    + "installation root must NOT be null:  "
                    + DeploymentUtils.getServiceAssemblyName(descriptor);
            log.error(msg);
            throw new ManagementException(msg);
        }

        return installationRoot;
    }

    /**
     * Explode service unit packages into the repository *
     * 
     * @param saInstallRoot
     *            the root {@link URI} of the service assembly
     * @param descriptor
     *            the service assembly descriptor
     * @return the map containing all su install roots
     * @throws ManagementException
     *             if the repository service doesn't manage to expend su archive
     */
    protected Map<String, URI> expendSUS(URI saInstallRoot,
            JBIDescriptor descriptor) throws ManagementException {
        Map<String, URI> suInstallRoots = new HashMap<String, URI>();
        List<ServiceUnit> sus = descriptor.getServiceAssembly()
                .getServiceUnits();
        for (ServiceUnit unit : sus) {
            URI suArchiveURI = getSUArchiveURI(unit, saInstallRoot);
            URI suInstallRoot = expandSUIntoSA(unit.getIdentification()
                    .getName(), suArchiveURI, DeploymentUtils
                    .getServiceAssemblyName(descriptor));
            suInstallRoots.put(unit.getIdentification().getName(),
                    suInstallRoot);
        }
        return suInstallRoots;
    }

    /**
     * Retrieve the service unit archive {@link URI}
     * 
     * @param su
     *            the {@link ServiceUnit} production element
     * @param saInstallRoot
     *            the sa installation root {@link URI}
     * @return the su archive {@link URI}
     */
    protected URI getSUArchiveURI(ServiceUnit su, URI saInstallRoot) {
        String suZipNAme = su.getTargetArtifactsZip();
        String suZipURL = "file://" + saInstallRoot.getPath() + suZipNAme;
        return packageHandler.processAndGetPackageURI(suZipURL.replaceAll(" ",
                "%20"), true);
    }

    /**
     * Expand a service unit package into its service assembly installRoot.
     * 
     * @param suId
     *            the unique identifier of the su
     * @param suZipLocation
     *            URI representing the su zip location
     * @param saId
     *            the unique identifier of the sa
     * @return SU installationRoot, null if it fails
     * @throws ManagementException
     *             if it fails to expend the su package to the repository
     */
    protected URI expandSUIntoSA(String suId, URI suZipLocation, String saId)
        throws ManagementException {
        log.start();
        String msg;
        URI result = null;

        try {
            result = repositorySrv.explodeSUIntoSAInstallFolder(suId,
                    suZipLocation, saId);
        } catch (PetalsException pe) {
            msg = "Unexpected error expending service JBI installation "
                    + "package into repository.";
            log.error(msg, pe);
            throw new ManagementException(msg, pe);
        }
        if (result == null) {
            msg = "Unexpected platform repository service error. The "
                    + "su root must NOT be null: " + suId;
            log.error(msg);
            throw new ManagementException(msg);
        }
        log.end();
        return result;
    }

    public void undo(HashMap context) {
        JBIDescriptor descriptor = (JBIDescriptor) context
                .get(DeploymentContextConstants.SA_DESCRIPTOR);

        try {
            repositorySrv.removeServiceAssemblyPackage(DeploymentUtils
                    .getServiceAssemblyName(descriptor));
        } catch (PetalsException e) {
            String msg = "Failed to revert a SAAndSUExtractionTask";
            log.error(msg, e);
        }

    }

}
