/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SAPackageCheckingTask.java 154 6 oct. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.deployment.deploy;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.ZipFile;

import javax.xml.namespace.QName;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.jbi.management.deployment.DeploymentContextConstants;
import org.objectweb.petals.jbi.management.deployment.DeploymentServiceImpl;
import org.objectweb.petals.jbi.management.deployment.ServiceAssemblyDataHandler;
import org.objectweb.petals.processor.Task;
import org.objectweb.petals.tools.jbicommon.descriptor.Connection;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;
import org.objectweb.petals.tools.jbicommon.descriptor.ServiceAssembly;
import org.objectweb.petals.tools.jbicommon.descriptor.ServiceUnit;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.StringHelper;
import org.objectweb.petals.util.ZipUtil;

/**
 * This task checks the validity of the service assembly package. Some checking
 * steps are described in the specification (section 6.6 of the JSR 208).
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class SAPackageCheckingTask implements Task {

	/**
	 * logger wrapper
	 */
	protected LoggingUtil log;

	protected DeploymentServiceImpl deploymentService;

	public SAPackageCheckingTask(LoggingUtil log,
			DeploymentServiceImpl deploymentService) {
		super();
		this.log = log;
		this.deploymentService = deploymentService;
	}

	public void execute(HashMap context) throws Exception {
		String msg = null;

		JBIDescriptor descriptor = (JBIDescriptor) context
				.get(DeploymentContextConstants.SA_DESCRIPTOR);

		URI saArchiveURI = (URI) context
				.get(DeploymentContextConstants.ARCHIVE_URI);

		/*
		 * Retrieve assembly production element into JBI descriptor
		 */
		ServiceAssembly assembly = descriptor.getServiceAssembly();
		if (assembly == null) {
			msg = "Given archive isn't a service assembly archive : "
					+ "missing service assembly production element into jbi descriptor";
			log.error(msg);
			throw new Exception(msg);
		}

		String saName = assembly.getIdentification().getName();

		/*
		 * Check if sa name is already used
		 */
		if (isSANameAlreadyUsed(saName)) {
			msg = "You are trying to deploy a sa with an identifier "
					+ "that is used by a previously successfuly deployed sa ("
					+ saName + ").";
			log.error(msg);
			throw new Exception(msg);
		}

		/*
		 * Check service assembly identifier
		 */
		if (StringHelper.isEmpty(saName)) {
			msg = "Unique identifier of a service assembly must be non null and non empty ("
					+ saArchiveURI.getPath() + ").";
			log.error(msg);
			throw new Exception(msg);
		}

		/*
		 * Check potential connection elements
		 */
		List<Connection> connections = assembly.getConnections();
		if (connections != null) {
			checkConnections(connections);
		}

		/*
		 * Check potential service unit elements
		 */
		List<ServiceUnit> sus = assembly.getServiceUnits();
		if (sus != null) {
			checkServiceUnits(sus, saName, saArchiveURI);
		}

	}

	/**
	 * Check the validity of connection elements
	 * 
	 * @param connections
	 *            a {@link List} of {@link Connection} element
	 * @throws Exception
	 *             if consumer interface and consumer endpoint and consumer
	 *             service are null or empty at the same time or if (provider
	 *             endpoint and provider service) are null or empty
	 */
	protected void checkConnections(List<Connection> connections)
			throws Exception {
		String msg = "";

		for (Connection connection : connections) {
			QName consInterface = connection.getConsumerInterfaceName();
			QName consService = connection.getConsumerServiceName();
			String consEndpoint = connection.getConsumerEndpointName();
			QName provService = connection.getProviderServiceName();
			String provEndpoint = connection.getProviderEndpointName();

			if (consInterface == null
					|| StringHelper.isEmpty(consInterface.getLocalPart())) {
				if (StringHelper.isEmpty(consEndpoint) || consService == null
						|| StringHelper.isEmpty(consService.getLocalPart())) {
					msg = "Consumer interface and (consumer service or consumer endpoint) "
							+ "must not be null or empty at the same time.";
					log.error(msg);
					throw new Exception(msg);
				}
			}
			if (StringHelper.isEmpty(provEndpoint)
					|| (provService == null || StringHelper.isEmpty(provService
							.getLocalPart()))) {
				msg = "Provider service and provider endpoint) "
						+ "must not be null or empty.";
				log.error(msg);
				throw new Exception(msg);
			}
		}
	}

	/**
	 * Check the validity of service unit elements
	 * 
	 * @param sus
	 *            a {@link List} of {@link ServiceUnit} element
	 * @param saName
	 *            the service assembly name
	 * @param saArchiveURI
	 *            the service assembly archive URI
	 * @throws Exception
	 *             if su identifier is null or empty or non unique within the
	 *             sa, if target artifact zip and target component are null or
	 *             empty, if not all declared su zip are presents in the sa
	 *             package or if su has already been deployed to the target
	 *             component
	 */
	protected void checkServiceUnits(List<ServiceUnit> sus, String saName,
			URI saArchiveURI) throws Exception {
		String msg = "";

		List<String> suIds = new ArrayList<String>();
		for (ServiceUnit su : sus) {
			String suIdentifier = su.getIdentification().getName();

			// Check su identifier : must be non null, non empty and unique
			// within the sa
			checkSUIdentifier(suIds, suIdentifier, saName);

			suIds.add(suIdentifier);

			// check target artifact zip and target component entries : must
			// be non null and non empty
			String artifactZip = su.getTargetArtifactsZip();
			String targetComponent = su.getTargetComponentName();
			checkTargetComponentAndZip(artifactZip, suIdentifier,
					targetComponent);

			// check that all declared su zip are presents in the sa package
			checkSUZipInSAZip(artifactZip, saArchiveURI);

			// Check if su hasn't already been deployed to the target
			// component
			if (isSUAlreadyDeployedOnTargetedComponent(su)) {
				msg = "You are trying to deploy a su ("
						+ su.getIdentification().getName()
						+ ")that have already been deployed on the targeted component ("
						+ su.getTargetComponentName() + ").";
				log.error(msg);
				throw new Exception(msg);
			}

		}
	}

	/**
	 * Check su identifier : must be non null, non empty and unique within the
	 * sa
	 * 
	 * @param suIds
	 *            the su identifier list within the sa
	 * @param suIdentifier
	 *            the service unit identifier
	 * @param saName
	 *            the service assembly name
	 * @throws Exception
	 *             if su identifier is null, empty and non unique within the sa
	 */
	protected void checkSUIdentifier(List<String> suIds, String suIdentifier,
			String saName) throws Exception {
		String msg = "";
		if (StringHelper.isEmpty(suIdentifier)) {
			msg = "Identifier of a service unit must be non null and non empty (SA name : "
					+ saName + ").";
			log.error(msg);
			throw new Exception(msg);
		}
		if (suIds.contains(suIdentifier)) {
			msg = "Identifier of a service unit must be unique within a single SA (SA name : "
					+ saName + ").";
			log.error(msg);
			throw new Exception(msg);
		}
	}

	/**
	 * Check target artifact zip and target component entries : must be non null
	 * and non empty
	 * 
	 * @param artifactZip
	 *            the artefact zip path into the sa archive
	 * @param suIdentifier
	 *            the service unit identifier
	 * @param targetComponent
	 *            the targeted component name
	 * @throws Exception
	 *             if target artifact zip or target component entries are null
	 *             or empty
	 */
	protected void checkTargetComponentAndZip(String artifactZip,
			String suIdentifier, String targetComponent) throws Exception {
		String msg = "";
		if (StringHelper.isEmpty(artifactZip)) {
			msg = "Target artifact zip of a service unit must be non null "
					+ "and non empty (SU name : " + suIdentifier + ").";
			log.error(msg);
			throw new Exception(msg);
		}
		if (StringHelper.isEmpty(targetComponent)) {
			msg = "Target component name of a service unit must be non null "
					+ "and non empty (SU name : " + suIdentifier + ").";
			log.error(msg);
			throw new Exception(msg);
		}
	}

	/**
	 * Check that all declared su zip are presents in the sa package
	 * 
	 * @param artifactZip
	 *            the artefact zip path into the sa archive
	 * @param saArchiveURI
	 *            the service assembly archive URI
	 * @throws Exception
	 *             if at least one su is missing into the sa archive or if zip
	 *             archive can't be correctly close or opened
	 */
	protected void checkSUZipInSAZip(String artifactZip, URI saArchiveURI)
			throws Exception {
		String msg = "";
		ZipFile zipArchive = openZipFile(saArchiveURI);
		try {
			if (!hasEntry(zipArchive, artifactZip)) {
				msg = "The specified target artifact zip in JBI "
						+ "Descriptor's service unit declaration was not "
						+ "found within deployment package Zip archive file : "
						+ saArchiveURI.getPath() + ".";
				log.error(msg);
				throw new Exception(msg);
			}
		} finally {
			try {
				zipArchive.close();
			} catch (IOException e) {
				throw new Exception(
						"Error while closing the Zip archive file.", e);
			}
		}
	}

	/**
	 * Load a {@link ZipFile} from an {@link URI}
	 * 
	 * @param archiveURI
	 *            {@link URI} of the {@link ZipFile}
	 * @return the targeted {@link ZipFile}
	 * @throws PetalsException
	 *             when {@link ZipFile} can't be opened
	 */
	protected ZipFile openZipFile(URI archiveURI) throws PetalsException {
		return ZipUtil.openZipFile(archiveURI);
	}

	/**
	 * Test if a zip file has a specific entry
	 * 
	 * @param archive
	 *            the sa {@link ZipFile}
	 * @param entryName
	 *            the entry name
	 * @return true if the entry is in the zip file
	 */
	protected boolean hasEntry(ZipFile zipArchive, String artifactZip) {
		return ZipUtil.hasEntry(zipArchive, artifactZip);
	}

	/**
	 * Check if the su is already deployed on the targeted component.
	 * 
	 * @param su
	 *            the service unit production element
	 * 
	 * @return true if the su is already deployed on the targeted component,
	 *         false otherwise
	 * @throws Exception
	 *             if deploymentService.isDeployedServiceUnit fails
	 */
	protected boolean isSUAlreadyDeployedOnTargetedComponent(ServiceUnit su)
			throws Exception {
		return deploymentService.isDeployedServiceUnit(su
				.getTargetComponentName(), su.getIdentification().getName());
	}

	/**
	 * Check if the sa name is already used by a previously successfuly deployed
	 * sa.
	 * 
	 * @param saName
	 *            the service assembly name
	 * @return true if the sa name is already used, false otherwise
	 */
	protected boolean isSANameAlreadyUsed(String saName) {

		Map<String, ServiceAssemblyDataHandler> deployedSACache = deploymentService
				.getDeployedSACache();

		return deployedSACache.containsKey(saName);
	}

	public void undo(HashMap context) {
		// Nothing to do

	}

}
