/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: EndpointService.java 11:27:27 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.service;

import java.util.ArrayList;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.fractal.fraclet.annotation.Requires;
import org.objectweb.petals.jbi.registry.AbstractEndpoint;
import org.objectweb.petals.jbi.registry.ConsumerEndpoint;
import org.objectweb.petals.jbi.registry.InternalEndpoint;
import org.objectweb.petals.jbi.registry.LinkedEndpoint;
import org.objectweb.petals.jbi.registry.Registry;
import org.objectweb.petals.jbi.registry.RegistryException;
import org.objectweb.petals.kernel.admin.DistributedJMXServer;
import org.objectweb.petals.kernel.admin.DistributedJMXServerFactory;
import org.objectweb.petals.tools.jbicommon.util.XMLUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.ParameterCheckHelper;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.petals.util.WSDLUtil;
import org.objectweb.util.monolog.api.Logger;
import org.w3c.dom.Document;

/**
 * This class is used to manage everything that deals with endpoint in the
 * distributed environment
 * 
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces=@Interface(name="service",signature=org.objectweb.petals.jbi.management.service.EndpointService.class))
public class EndpointServiceImpl implements EndpointService {

    /**
     * Name of the current container
     */
    protected String containerName;

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name="logger")
    protected Logger logger;

    /**
     * Registry
     */
    @Requires(name="registry",signature=org.objectweb.petals.jbi.registry.Registry.class)
    protected Registry registry;

    /**
     * Manager of the distributed JMX server
     */
    @Requires(name="dist-jmx-server",signature=org.objectweb.petals.kernel.admin.DistributedJMXServerFactory.class)
    protected DistributedJMXServerFactory serverManager;

    /**
     * @see EndpointService#activateEndpoint(QName, String, ConsumerEndpoint)
     */
    public InternalEndpoint activateEndpoint(QName serviceName,
            String endpointName, ConsumerEndpoint address) throws JBIException {
        log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (endpointName == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        if (address == null) {
            throw new IllegalArgumentException(
                    "Consumer endpoint must be non null");
        }
        InternalEndpoint se = new InternalEndpoint(serviceName, endpointName,
                address.getComponentName(), address.getContainerName(), this);
        se.setContainerUID(SystemUtil.getContainerUID());
        registry.registerInternalEndpoint(se);
        log.end();
        return se;
    }

    /**
     * @see EndpointService#createConnection(QName, QName, String)
     */
    public void createConnection(QName consInterface, QName provService,
            String provEndpoint) throws JBIException {
        log.start();
        if (consInterface == null) {
            throw new IllegalArgumentException(
                    "Interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException(
                    "Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException(
                    "Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface,
                provService, provEndpoint, this);
        registry.registerConnection(linkedEndpoint);
        log.end();
    }

    /**
     * @see EndpointService#createConnection(QName, String, QName, String)
     */
    public void createConnection(QName consService, String consEndpoint,
            QName provService, String provEndpoint) throws JBIException {
        log.start();
        if (consService == null) {
            throw new IllegalArgumentException(
                    "Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException(
                    "Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException(
                    "Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException(
                    "Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService,
                consEndpoint, provService, provEndpoint, this);
        registry.registerConnection(linkedEndpoint);
        log.end();
    }

    /**
     * @see EndpointService#deactivateEndpoint(ServiceEndpoint)
     */
    public void deactivateEndpoint(ServiceEndpoint endpoint)
        throws JBIException {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        registry.deregisterInternalEndpoint((AbstractEndpoint) endpoint);
        log.end();
    }

    /**
     * @see EndpointService#deleteConnection(QName, QName, String)
     */
    public void deleteConnection(QName consInterface, QName provService,
            String provEndpoint) throws JBIException {
        log.start();
        if (consInterface == null) {
            throw new IllegalArgumentException(
                    "Consumer interface name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException(
                    "Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException(
                    "Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consInterface,
                provService, provEndpoint, this);
        registry.deregisterConnection(linkedEndpoint);
        log.end();
    }

    /**
     * @see EndpointService#deleteConnection(QName, String, QName, String)
     */
    public void deleteConnection(QName consService, String consEndpoint,
            QName provService, String provEndpoint) throws JBIException {
        log.start();
        if (consService == null) {
            throw new IllegalArgumentException(
                    "Consumer service name must be non null");
        }
        if (consEndpoint == null) {
            throw new IllegalArgumentException(
                    "Consumer endpoint name must be non null");
        }
        if (provService == null) {
            throw new IllegalArgumentException(
                    "Provider service name must be non null");
        }
        if (provEndpoint == null) {
            throw new IllegalArgumentException(
                    "Provider endpoint name must be non null");
        }
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(consService,
                consEndpoint, provService, provEndpoint, this);
        registry.deregisterConnection(linkedEndpoint);
        log.end();
    }

    /**
     * @see EndpointService#deregisterExternalEndpoint(ServiceEndpoint)
     */
    public void deregisterExternalEndpoint(ServiceEndpoint endpoint)
        throws JBIException {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        registry.deregisterExternalEndpoint((AbstractEndpoint) endpoint);
        log.end();
    }

    /**
     * @see EndpointService#getEndpoint(QName, String)
     * 
     * <ol>
     * <li>Retreive the named endpoint</li>
     * <li>If found, update the endpoint service of the found endpoint with
     * this</li>
     * </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name) {
        log.call();
        return getEndpoint(service, name, true);
    }

    /**
     * @see EndpointService#getEndpoint(QName, String,boolean)
     * 
     * <ol>
     * <li>Retreive the named endpoint</li>
     * <li>If found, update the endpoint service of the found endpoint with
     * this</li>
     * </ol>
     */
    public ServiceEndpoint getEndpoint(QName service, String name,
            boolean resolveLink) {
        log.start();
        if (service == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        if (name == null) {
            throw new IllegalArgumentException("Endpoint name must be non null");
        }
        AbstractEndpoint abstractEndpoint = null;
        try {
            abstractEndpoint = registry.getInternalEndpoint(service, name,
                    resolveLink);
        } catch (RegistryException e) {
            // no endpoint found, return null;
        }
        if (abstractEndpoint != null
                && abstractEndpoint.getEndpointService() == null) {
            abstractEndpoint.setEndpointService(this);
        }
        log.end();
        return abstractEndpoint;
    }

    /**
     * @see EndpointService#getEndpointDescriptorForEndpoint(ServiceEndpoint)
     */
    public Document getEndpointDescriptorForEndpoint(ServiceEndpoint endpoint)
        throws JBIException {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        Document document = null;
        try {
            // Update the new endpoints
            updateNewEndpoints();

            // Recover the endpoint
            AbstractEndpoint abstractEndpoint = registry.getInternalEndpoint(
                    endpoint.getServiceName(), endpoint.getEndpointName());
            document = abstractEndpoint.getEndpointDescriptor();
        } catch (RegistryException e) {
            log.error("Error while validating endpoint "
                    + endpoint.getEndpointName(), e);
        }
        log.end();
        return document;
    }

    /**
     * @see EndpointService#getExternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForInterface(
            QName interfaceName) {
        log.start();
        if (interfaceName == null) {
            throw new IllegalArgumentException(
                    "Interface name must be non null");
        }
        ServiceEndpoint[] endpoints = null;
        try {
            endpoints = registry
                    .getExternalEndpointsForInterface(interfaceName);
            // Update the EndpointService field of the found endpoints
            for (int i = 0; i < endpoints.length; i++) {
                AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
                if (abstractEndpoint.getEndpointService() == null) {
                    abstractEndpoint.setEndpointService(this);
                    endpoints[i] = abstractEndpoint;
                }
            }
        } catch (RegistryException e) {
            log.error("Error while getting external endpoints for interface "
                    + interfaceName, e);
        }
        log.end();
        return endpoints;
    }

    /**
     * @see EndpointService#getExternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getExternalEndpointsForService(QName serviceName) {
        log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = null;
        endpoints = registry.getExternalEndpointsForService(serviceName);
        // Update the EndpointService field of the found endpoints
        for (int i = 0; i < endpoints.length; i++) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
            if (abstractEndpoint.getEndpointService() == null) {
                abstractEndpoint.setEndpointService(this);
                endpoints[i] = abstractEndpoint;
            }
        }
        log.end();
        return endpoints;
    }

    /**
     * Return the interface for the specified endpoint
     * 
     * @param endpoint
     *            endpoint on which we want the interface names; must be non
     *            null
     * @return array of interfaces QName
     */
    public QName[] getInterfacesForEndpoint(AbstractEndpoint endpoint) {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        QName[] interfs = new QName[0];
        AbstractEndpoint abstractEndpoint = null;
        try {
            updateNewEndpoints();
            abstractEndpoint = registry.getInternalEndpoint(endpoint
                    .getServiceName(), endpoint.getEndpointName());
            interfs = abstractEndpoint.getInterfaces();
        } catch (Exception e) {
            log.error("Error while getting interfaces for endpoint "
                    + endpoint.getEndpointName(), e);
        }
        log.end();
        return interfs;
    }

    /**
     * Update all the new endpoints and remove them
     * 
     * @throws JBIException
     *             if a problem occurs
     */
    protected void updateNewEndpoints() throws JBIException {
        for (AbstractEndpoint abstractEndpoint : registry
                .retrieveNewEndpoints()) {
            AbstractEndpoint endpoint = null;
            try {
                endpoint = updateEndpointDescriptionAndInterfaces(abstractEndpoint);
                registry.validateEndpoint(endpoint);
                // We clean the new endpoint context if no problem occured
                registry.cleanNewEndpoints();
            } catch (JBIException e) {
                log.error(
                        "Problem while getting description and interfaces for the endpoint "
                                + abstractEndpoint.getEndpointName(), e);
            }
        }
    }

    /**
     * @see EndpointService#getInternalEndpointsForInterface(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForInterface(
            QName interfaceName) {
        log.start();
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        try {
            // Update the new endpoints
            updateNewEndpoints();

            endpoints = registry
                    .getInternalEndpointsForInterface(interfaceName);
            for (int i = 0; i < endpoints.length; i++) {
                AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
                if (abstractEndpoint.getEndpointService() == null) {
                    abstractEndpoint.setEndpointService(this);
                    endpoints[i] = abstractEndpoint;
                }
            }
        } catch (RegistryException e) {
            log.error("Error while getting internal endpoints for interface "
                    + interfaceName, e);
        } catch (JBIException e) {
            log.error("Error while getting internal endpoints for interface "
                    + interfaceName, e);
        }
        log.end();
        return endpoints;
    }

    /**
     * @see EndpointService#getInternalEndpointsForService(QName)
     */
    public ServiceEndpoint[] getInternalEndpointsForService(QName serviceName) {
        log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException("Service name must be non null");
        }
        ServiceEndpoint[] endpoints = new ServiceEndpoint[0];
        endpoints = registry.getInternalEndpointsForService(serviceName);
        for (int i = 0; i < endpoints.length; i++) {
            AbstractEndpoint abstractEndpoint = (AbstractEndpoint) endpoints[i];
            if (abstractEndpoint.getEndpointService() == null) {
                abstractEndpoint.setEndpointService(this);
                endpoints[i] = abstractEndpoint;
            }
        }
        log.end();
        return endpoints;
    }

    /**
     * @see EndpointService#isExchangeWithConsumerOkayForComponent(InternalEndpoint,
     *      MessageExchange)
     */
    public boolean isExchangeWithConsumerOkayForComponent(
            InternalEndpoint internalEndpoint, MessageExchange exchange) {
        log.start();
        ParameterCheckHelper.isNullParameterWithLog(internalEndpoint,
                "Endpoint", log);
        boolean test = true;
        try {
            // Retreive a connection to the JMX server of the container were the
            // component is deployed
            DistributedJMXServer server = serverManager
                    .createDistributedJMXServer(internalEndpoint
                            .getContainerName());
            // Ask the component if it is ok to accept the exchange
            Boolean bool = (Boolean) server.invoke(server
                    .getAdminServiceMBeanName(),
                    DistributedJMXServer.IS_OK_WITH_CONS, new Object[] {
                        internalEndpoint.getComponentName(),
                        (ServiceEndpoint) internalEndpoint, exchange},
                    new String[] {String.class.getName(),
                        ServiceEndpoint.class.getName(),
                        MessageExchange.class.getName()});
            test = bool.booleanValue();
        } catch (Exception e) {
            log.error("Problem in isExchangeWithConsumerOkayForComponent", e);
        }
        log.end();
        return test;
    }

    /**
     * @see EndpointService#isExchangeWithProviderOkayForComponent(InternalEndpoint,
     *      MessageExchange)
     */
    public boolean isExchangeWithProviderOkayForComponent(
            InternalEndpoint internalEndpoint, MessageExchange exchange) {
        log.start();
        if (internalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        boolean test = true;
        try {
            // Retreive a connection to the JMX server of the container were the
            // component is deployed
            DistributedJMXServer server = serverManager
                    .createDistributedJMXServer(internalEndpoint
                            .getContainerName());
            // Retreive the endpoint description
            Boolean bool = (Boolean) server.invoke(server
                    .getAdminServiceMBeanName(),
                    DistributedJMXServer.IS_OK_WITH_PROV, new Object[] {
                        internalEndpoint.getComponentName(),
                        (ServiceEndpoint) internalEndpoint, exchange},
                    new String[] {String.class.getName(),
                        ServiceEndpoint.class.getName(),
                        MessageExchange.class.getName()});
            test = bool.booleanValue();
        } catch (Exception e) {
            log.error("Problem in isExchangeWithProviderOkayForComponent", e);
        }
        log.end();
        return test;
    }

    /**
     * @see EndpointService#registerExternalEndpoint(ServiceEndpoint)
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint)
        throws JBIException {
        log.start();
        if (externalEndpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        registry.registerExternalEndpoint(externalEndpoint);
        log.end();
    }

    @LifeCycle(on=LifeCycleType.START)
    public void start() throws IllegalLifeCycleException {
        log = new LoggingUtil(logger);
        log.start();
        containerName = SystemUtil.getContainerName();
        log.end();
    }

    @LifeCycle(on=LifeCycleType.STOP)
    public void stop() throws IllegalLifeCycleException {
        log.start();
        log.end();
    }

    /**
     * This method will update the endpoint description and its interfaces then
     * return it
     * 
     * @param endpoint
     *            endpoint to update
     * @return updated endpoint
     * @throws JBIException
     */
    private AbstractEndpoint updateEndpointDescriptionAndInterfaces(
            AbstractEndpoint endpoint) throws JBIException {
        log.start();
        try {
            // Retreive a connection to the JMX server of the container were the
            // endpoint is deployed
            DistributedJMXServer server = serverManager
                    .createDistributedJMXServer(endpoint.getContainerName());
            // Retreive the endpoint description
            Document document = (Document) server.invoke(server
                    .getAdminServiceMBeanName(),
                    DistributedJMXServer.GET_SERVICE_DESCRITION,
                    new Object[] {endpoint.getComponentName(),
                        (ServiceEndpoint) endpoint}, new String[] {
                        "java.lang.String",
                        "javax.jbi.servicedesc.ServiceEndpoint"});
            server.closeConnector();
            List<QName> interfaces = new ArrayList<QName>();
            if (document != null) {
                // Get the whole description
                String description = XMLUtil
                        .createStringFromDOMDocument(document);
                endpoint.setDescription(description);
                // Get the interfaces
                interfaces = WSDLUtil.getInterfacesForService(document,
                        endpoint.getServiceName());
            }

            if (interfaces.size() == 0) {
                QName unresolved = new QName(
                        "http://www.petals.objectweb.org/",
                        "UNRESOLVED_INTERFACE");
                log
                        .warning("Problem while extracting interfaces for the endpoint "
                                + endpoint.getEndpointName()
                                + "\n Set its interface to " + unresolved);
                interfaces.add(unresolved);
            }
            endpoint.setInterfaces(interfaces.toArray(new QName[0]));
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new JBIException(e);
        }
        log.end();
        return endpoint;
    }

    /**
     * @see EndpointService#isContainerStarted(AbstractEndpoint)
     */
    public boolean isContainerStarted(AbstractEndpoint endpoint) {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must be non null");
        }
        boolean isStarted = false;
        isStarted = serverManager.isContainerStarted(endpoint
                .getContainerName());
        log.end();
        return isStarted;
    }
}
