/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SystemRecovery.java 154 2 mai 2006 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.systemstate;

import java.io.File;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.management.LifeCycleMBean;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.ReflectionException;

import org.objectweb.petals.PetalsException;
import org.objectweb.petals.jbi.component.lifecycle.Installer;
import org.objectweb.petals.jbi.component.lifecycle.ServiceAssemblyLifeCycle;
import org.objectweb.petals.jbi.management.deployment.DeploymentContextConstants;
import org.objectweb.petals.jbi.management.deployment.DeploymentService;
import org.objectweb.petals.jbi.management.deployment.DeploymentTaskFactory;
import org.objectweb.petals.jbi.management.service.InstallationService;
import org.objectweb.petals.jbi.management.service.LifeCycleManagerService;
import org.objectweb.petals.jbi.management.service.ManagementException;
import org.objectweb.petals.jbi.management.service.PackageHandler;
import org.objectweb.petals.processor.TaskProcessor;
import org.objectweb.petals.repository.RepositoryService;
import org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Allows to restore system state after shutdown or crash.
 * 
 * @author Olivier Fabre - EBM Websourcing
 * 
 */
public class SystemRecovery {

    protected DeploymentService deploymentService;

    protected InstallationService installationService;

    protected LoggingUtil log;

    protected LifeCycleManagerService managerService;

    protected PackageHandler packageHandler;

    protected RepositoryService repositorySrv;

    protected StateStorage stateStorage;

    public SystemRecovery(InstallationService installationService,
            Logger logger, DeploymentService deploymentService,
            LifeCycleManagerService managerService, StateStorage stateStorage,
            RepositoryService repositorySrv) {
        super();
        this.deploymentService = deploymentService;
        this.installationService = installationService;
        this.managerService = managerService;
        this.log = new LoggingUtil(logger);
        this.stateStorage = stateStorage;
        this.repositorySrv = repositorySrv;
        this.packageHandler = new PackageHandler(logger, repositorySrv);
        createLostPlusFoundDir();
        createRepositoryDirectories();
    }

    protected void createRepositoryDirectories() {
        File componentsRep = new File(repositorySrv.getComponentsDirectory());
        File serviceAssRep = new File(repositorySrv
                .getServiceAssembliesDirectory());
        File sharedLibRep = new File(repositorySrv.getSharedLibsDirectory());

        // Create components directory in Petals repository if it doesn't exist
        if (!componentsRep.exists()) {
            if (!componentsRep.mkdirs()) {
                String msg = "Failed to create components dir in Petals repository";
                log.error(msg);
                throw new RuntimeException(msg);
            }
        }

        // Create sharedlibs directory in Petals repository if it doesn't exist
        if (!sharedLibRep.exists()) {
            if (!sharedLibRep.mkdirs()) {
                String msg = "Failed to create sharedlibs dir in Petals repository";
                log.error(msg);
                throw new RuntimeException(msg);
            }
        }

        // Create serviceassemblies directory in Petals repository if it doesn't
        // exist
        if (!serviceAssRep.exists()) {
            if (!serviceAssRep.mkdirs()) {
                String msg = "Failed to create serviceassemblies dir in Petals repository";
                log.error(msg);
                throw new RuntimeException(msg);
            }
        }
    }

    public boolean recoverAllComponent() throws ManagementException {
        log.start();
        String msg = null;

        /*
         * Retrieve all component directories in the components petals
         * repository. This list will be used to delete components that have
         * been unzipped in the Petals repository but no more persisted in the
         * system recory file (system-state.xml). These unknown component
         * directories will be put in the Petals "lost+found" directory.
         */

        List<File> unknownDirs = new ArrayList<File>(Arrays.asList(new File(
                repositorySrv.getComponentsDirectory()).listFiles()));

        /*
         * Retrieve all components to recover
         */

        List<ComponentState> cps = stateStorage.retrieveAllComponentStates();

        /*
         * Recover state for each component
         */
        for (ComponentState state : cps) {

            URI installationRoot = null;
            try {

                installationRoot = packageHandler.processAndGetPackageURI(state
                        .getInstallURL(), true);

                URI archiveURI = packageHandler.processAndGetPackageURI(state
                        .getArchiveURL(), true);

                JBIDescriptor descriptor = packageHandler
                        .loadDescriptor(archiveURI);

                recoverComponent(installationRoot, descriptor, state
                        .getInstallState(), state.getLifecycleState());
            } catch (PetalsException e) {
                msg = "A component can't be recovered : " + state.getName();
                log.error(msg, e);
                throw new ManagementException(msg, e);
            }

            // Delete component from the unknownDirs list because this
            // component is known now...
            if (unknownDirs != null) {
                File knownComp = new File(installationRoot).getParentFile();
                if (knownComp.exists()) {
                    unknownDirs.remove(knownComp);
                }
            }
        }

        // Now, move all unknown components from Petals repository to
        // "lost+found" dir
        moveUnknownDirsToLostPlusFoundDir(unknownDirs);

        log.end();
        return false;
    }

    public boolean recoverAllServiceAssembly() throws ManagementException {
        log.start();
        String msg = null;

        /*
         * Retrieve all sa directories in the sa petals repository. This list
         * will be used to delete sa that have been unzipped in the Petals
         * repository but no more persisted in the system recory file
         * (system-state.xml). These unknown sa directories will be put in the
         * Petals "lost+found" directory.
         */

        List<File> unknownDirs = new ArrayList<File>(Arrays.asList(new File(
                repositorySrv.getServiceAssembliesDirectory()).listFiles()));

        /*
         * Retrieve all service assemblies to recover
         */

        List<ServiceAssemblyState> sas = stateStorage
                .retrieveAllServiceAssemblyStates();

        /*
         * Recover state for each service assembly
         */

        for (ServiceAssemblyState state : sas) {
            URI installationRoot = null;
            try {

                installationRoot = packageHandler.processAndGetPackageURI(state
                        .getInstallURL(), true);

                URI archiveURI = packageHandler.processAndGetPackageURI(state
                        .getArchiveURL(), true);

                JBIDescriptor descriptor = packageHandler
                        .loadDescriptor(archiveURI);

                recoverServiceAssembly(archiveURI, descriptor, state
                        .getLifecycleState());
            } catch (PetalsException e) {
                msg = "A service assembly can't be recovered : "
                        + state.getName();
                log.error(msg, e);
                throw new ManagementException(msg, e);
            } catch (Exception e) {
                msg = "A service assembly can't be recovered : "
                        + state.getName();
                log.error(msg, e);
                throw new ManagementException(msg, e);
            }

            // Delete sa from the unknownDirs list because this
            // sa is known now...
            if (unknownDirs != null) {
                File knownComp = new File(installationRoot).getParentFile();
                if (knownComp.exists()) {
                    unknownDirs.remove(knownComp);
                }
            }
        }

        // Now, move all unknown sl from Petals repository to
        // "lost+found" dir
        moveUnknownDirsToLostPlusFoundDir(unknownDirs);

        log.end();
        return false;
    }

    public boolean recoverAllSharedLibrary() throws ManagementException {
        log.start();
        String msg = null;

        /*
         * Retrieve all sl directories in the sl petals repository. This list
         * will be used to delete sl that have been unzipped in the Petals
         * repository but no more persisted in the system recory file
         * (system-state.xml). These unknown sl directories will be put in the
         * Petals "lost+found" directory.
         */

        List<File> unknownDirs = new ArrayList<File>(Arrays.asList(new File(
                repositorySrv.getSharedLibsDirectory()).listFiles()));

        /*
         * Retrieve all shared library to recover
         */

        List<SharedLibraryState> sls = stateStorage
                .retrieveAllSharedLibraryStates();

        /*
         * Recover state for each shared library
         */
        for (SharedLibraryState state : sls) {
            URI installationRoot = null;
            try {

                installationRoot = packageHandler.processAndGetPackageURI(state
                        .getInstallURL(), true);

                URI archiveURI = packageHandler.processAndGetPackageURI(state
                        .getArchiveURL(), true);

                JBIDescriptor descriptor = packageHandler
                        .loadDescriptor(archiveURI);

                recoverSharedLib(installationRoot, descriptor);
            } catch (PetalsException e) {
                msg = "A Shared lib can't be recovered : " + state.getName();
                log.error(msg, e);
                throw new ManagementException(msg, e);
            }

            // Delete sl from the unknownDirs list because this
            // sl is known now...
            if (unknownDirs != null) {
                File knownComp = new File(installationRoot).getParentFile();
                if (knownComp.exists()) {
                    unknownDirs.remove(knownComp);
                }
            }
        }

        // Now, move all unknown sl from Petals repository to
        // "lost+found" dir
        moveUnknownDirsToLostPlusFoundDir(unknownDirs);

        log.end();
        return false;
    }

    /**
     * Create the lost+found dir in the petals file system if it doesn't exist,
     * and cleanup it if it exists
     * 
     */
    protected void createLostPlusFoundDir() {
        File lostPlusFoundFile = new File(repositorySrv
                .getLostPlusFoundDirectory());
        // Cleanup lost+found directory at Petals startup
        if (lostPlusFoundFile.exists()) {
            if (!lostPlusFoundFile.delete()) {
                log
                        .warning("\"lost+found\" can't be deleted, "
                                + "following Petals repository cleanup can have some troubles!");
            } else {
                if (!lostPlusFoundFile.mkdirs()) {
                    String msg = "Failed to create lost+found directory, "
                            + "following Petals repository cleanup will fail!";
                    log.warning(msg);
                }
            }
        } else {
            if (!lostPlusFoundFile.mkdirs()) {
                String msg = "Failed to create lost+found directory, "
                        + "following Petals repository cleanup will fail!";
                log.warning(msg);
            }
        }
    }

    /**
     * Move all unknown components from Petals repository to "lost+found" dir
     * 
     * @param unknownDirs
     *            the {@link List} of unknown dirs that have to be moved.
     */
    protected void moveUnknownDirsToLostPlusFoundDir(List<File> unknownDirs) {
        if (unknownDirs != null) {
            for (File unknownDir : unknownDirs) {

                if (!unknownDir.exists()) {
                    log
                            .warning("The unknown dir ("
                                    + unknownDir.getAbsolutePath()
                                    + ") doesn't exist.");
                }
                boolean moveDir = unknownDir.renameTo(new File(new File(
                        repositorySrv.getLostPlusFoundDirectory()), unknownDir
                        .getName()));
                if (!moveDir) {
                    log.warning("Failed to move unknown directory ("
                            + unknownDir.getAbsolutePath()
                            + ") to lost+found directory");
                }
            }
        }
    }

    /**
     * Recover the given component. Recovering a component consist in redoing
     * installation process of a component (see <code>InstallationService</code>),
     * calling the "install" method on the created Installer and putting the
     * component in its previous state by calling "start" or/and "stop" on the
     * created component lifecycle.
     * 
     * @param installationRoot
     * @param archiveURI
     * @param installState
     * @param lifecycleState
     * @throws ManagementException
     */
    protected void recoverComponent(URI installationRoot,
            JBIDescriptor descriptor, String installState, String lifecycleState)
        throws ManagementException {
        log.start();
        String msg = null;

        String compName = descriptor.getComponent().getIdentification()
                .getName();
        try {

            if (Installer.INSTALLED.equals(installState)) {
                /*
                 * Redo install
                 */
                ObjectName installerName = installationService
                        .loadNewInstaller(installationRoot, descriptor);
                MBeanServer beanServer = managerService.getMBeanServer();

                Object[] nullObjects = new Object[0];
                String[] nullStrings = new String[0];
                ObjectName clcName = (ObjectName) beanServer.invoke(
                        installerName, "install", nullObjects, nullStrings);

                if (LifeCycleMBean.STOPPED.equals(lifecycleState)) {
                    /*
                     * Redo start followed by stop
                     */
                    beanServer.invoke(clcName, "start", nullObjects,
                            nullStrings);
                    beanServer
                            .invoke(clcName, "stop", nullObjects, nullStrings);
                } else if (LifeCycleMBean.STARTED.equals(lifecycleState)) {
                    /*
                     * Redo start
                     */
                    beanServer.invoke(clcName, "start", nullObjects,
                            nullStrings);
                }
            }
        } catch (InstanceNotFoundException e) {
            msg = "Component can't be recovered : " + compName;
            log.error(msg, e);
            throw new ManagementException(msg, e);
        } catch (MBeanException e) {
            msg = "Component can't be recovered : " + compName;
            log.error(msg, e);
            throw new ManagementException(msg, e);
        } catch (ReflectionException e) {
            msg = "Component can't be recovered : " + compName;
            log.error(msg, e);
            throw new ManagementException(msg, e);
        } catch (PetalsException e) {
            msg = "Component can't be recovered : " + compName;
            log.error(msg, e);
            throw new ManagementException(msg, e);
        }
        log.end();
    }

    /**
     * Recover the given Service Assembly. Recovering a Service Assembly consist
     * in redoing deployment process of a Service Assembly (see
     * <code>DeploymentService</code>) and putting the Service Assembly in
     * its previous state by calling "start" or/and "stop" on the created
     * component lifecycle.
     * 
     * @param installationRoot
     * @param archiveURI
     * @param lifecycleState
     * @throws ManagementException
     */
    protected void recoverServiceAssembly(URI archiveURI,
            JBIDescriptor descriptor, String lifecycleState) throws Exception {
        log.start();
        String msg = null;

        String saName = descriptor.getServiceAssembly().getIdentification()
                .getName();
        try {

            /*
             * Redo deployment
             */

            // Create deployment execution context
            HashMap<String, Object> deploymentContext = createDeploymentContext(
                    archiveURI, descriptor);

            // Create deployment processor and fill with tasks
            TaskProcessor processor = createDeploymentProcessor(deploymentContext);

            // Launch the process
            if (!processor.process()) {
                msg = "An error occured during the deployment "
                        + "of the following service assembly: "
                        + saName;
                throw new Exception(msg);
            }

            if (LifeCycleMBean.STOPPED.equals(lifecycleState)) {
                /*
                 * Redo start followed by stop
                 */
                ServiceAssemblyLifeCycle salc = managerService
                        .getServiceAssemblyByName(saName);
                salc.start();
                salc.stop();
            } else if (LifeCycleMBean.STARTED.equals(lifecycleState)) {
                /*
                 * Redo start
                 */
                ServiceAssemblyLifeCycle salc = managerService
                        .getServiceAssemblyByName(saName);
                salc.start();
            }

        } catch (PetalsException e) {
            msg = "ServiceAssembly can't be recovered : " + saName;
            log.error(msg, e);
            throw new ManagementException(msg, e);
        } catch (JBIException e) {
            msg = "ServiceAssembly can't be recovered : " + saName;
            log.error(msg, e);
            throw new ManagementException(msg, e);
        }
        log.end();
    }

    /**
     * Create the deployment processor and add all required task for the
     * complete deployment of a service assembly
     * 
     * @param deploymentContext
     *            the deployment context
     * @return the filled deployment processor
     */
    protected TaskProcessor createDeploymentProcessor(
            HashMap<String, Object> deploymentContext) {
        TaskProcessor processor = new TaskProcessor(deploymentContext, log);

        DeploymentTaskFactory deploymentTaskFactory = deploymentService
                .getDeploymentTaskFactory();

        processor.addTask(deploymentTaskFactory.getPackageCheckingTask());
        processor.addTask(deploymentTaskFactory.getResultCreationTask());
        processor.addTask(deploymentTaskFactory
                .getSaAndSUInstallRootRetrievalTask());
        processor.addTask(deploymentTaskFactory.getAllSUDeploymentTask());
        processor.addTask(deploymentTaskFactory
                .getSaLifeCycleRegistrationTask());
        processor.addTask(deploymentTaskFactory
                .getAllConnectionRegistrationTask());
        processor.addTask(deploymentTaskFactory.getDeploymentSuccessTask());

        return processor;
    }

    /**
     * Create the deployment context
     * 
     * @param archiveURI
     *            the sa archive URI
     * @param descriptor
     *            the sa jbi descriptor
     * @return the Deployment Context HashMap
     */
    protected HashMap<String, Object> createDeploymentContext(URI archiveURI,
            JBIDescriptor descriptor) {
        HashMap<String, Object> deploymentContext = new HashMap<String, Object>();
        deploymentContext.put(DeploymentContextConstants.ARCHIVE_URI,
                archiveURI);
        deploymentContext.put(DeploymentContextConstants.SA_DESCRIPTOR,
                descriptor);
        return deploymentContext;
    }

    /**
     * Recover the given shared library. Recovering a shared lib consist in
     * redoing installation process of a shared lib (see
     * <code>InstallationService</code>).
     * 
     * @param installationRoot
     * @param archiveURI
     * @throws ManagementException
     */
    protected void recoverSharedLib(URI installationRoot,
            JBIDescriptor descriptor) throws ManagementException {
        log.start();

        /*
         * Redo install
         */
        try {
            installationService.installSharedLibrary(installationRoot,
                    descriptor);
        } catch (PetalsException e) {
            String msg = "ServiceAssembly can't be recovered : "
                    + installationRoot.toString();
            log.error(msg, e);
            throw new ManagementException(msg, e);
        }

        log.end();
    }

}
