/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramRegistry.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.registry;

import java.util.ArrayList;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.LinkRef;
import javax.naming.Name;
import javax.naming.NameClassPair;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.NotContextException;
import javax.xml.namespace.QName;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.petals.util.JNDIUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.NamingHelper;
import org.objectweb.petals.util.PropertyUtil;
import org.objectweb.petals.util.StringHelper;
import org.objectweb.util.monolog.api.Logger;

/**
 * This registry is the registry used in a distrubed environment
 * 
 * It used a JNDI distributed directory to manage the endpoint storage
 * 
 * @author Adrien LOUIS
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces=@Interface(name="service",signature=org.objectweb.petals.jbi.registry.Registry.class))
public class JNDIRegistry implements Registry {

    /**
     * Name of the endpoints context
     */
    public static final String ENDPOINTS_REF = "endpoints";

    /**
     * Name of the interfaces context
     */
    public static final String INTERFACES_REF = "interfaces";

    /**
     * Name of the new endpoints context
     */
    public static final String NEW_ENDPOINTS_REF = "new-endpoints";

    /**
     * Name of the service context
     */
    public static final String SERVICES_REF = "services";

    /**
     * Name of the user context
     */
    public static final String USERS_REF = "users";

    /**
     * Endpoint context
     */
    protected Context endpointsContext;

    /**
     * Interfaces context
     */
    protected Context interfacesContext;

    /**
     * Logger wrapper
     */
    protected LoggingUtil log;

    /**
     * Monolog Logger instance.
     */
    @Monolog(name="logger")
    protected Logger logger;

    /**
     * New endpoint context
     */
    protected Context newEndpointContext;

    /**
     * Root context
     */
    protected InitialContext rootContext;

    /**
     * Service context
     */
    protected Context servicesContext;

    /**
     * Fractal Component Current State.
     */
    protected String state;

    /**
     * Users context
     */
    protected Context usersContext;

    /**
     * @see Registry#cleanNewEndpoints()
     */
    public void cleanNewEndpoints() throws RegistryException {
        try {
            NamingEnumeration<NameClassPair> names = rootContext
                    .list(NEW_ENDPOINTS_REF);
            while (names.hasMore()) {
                NameClassPair nameClassPair = (NameClassPair) names.next();
                newEndpointContext.unbind(nameClassPair.getName());
            }
        } catch (NamingException e) {
            log.error("Problem while cleaning new endpoints context", e);
            throw new RegistryException(
                    "Problem while cleaning new endpoints context", e);
        }
    }

    /**
     * @see Registry#deregisterConnection(LinkedEndpoint)
     */
    public void deregisterConnection(LinkedEndpoint linkedEndpoint)
        throws RegistryException {
        log.start();
        if (linkedEndpoint == null) {
            throw new IllegalArgumentException(
                    "The connection must be non null");
        }
        if (linkedEndpoint.getInterfaces() != null) {
            Name interf = NamingHelper.qNameToName(linkedEndpoint
                    .getInterfaces()[0]);
            if (!JNDIUtil.isBound(rootContext, INTERFACES_REF, interf)) {
                log.error("The connection interface name "
                        + linkedEndpoint.getInterfaces()[0]
                        + " is not registered");
                throw new RegistryException("The connection interface name "
                        + linkedEndpoint.getInterfaces()[0]
                        + " is not registered");
            } else {

                try {
                    // check if the provider endpoint is still bound in the
                    // new endpoint context if so its interfaces have not
                    // been resolved
                    if (JNDIUtil.isBound(rootContext, NEW_ENDPOINTS_REF,
                            linkedEndpoint.getToEndpointName())) {
                        Context interfaceCtx = (Context) interfacesContext
                                .lookup(interf);

                        interfaceCtx.unbind(linkedEndpoint.getToEndpointName());
                    } else {
                        // we need to check that we do not remove it from
                        // one of its interfaces context
                        InternalEndpoint internalEndpoint = (InternalEndpoint) getInternalEndpoint(
                                linkedEndpoint.getToServiceName(),
                                linkedEndpoint.getToEndpointName());
                        boolean isNotInInt = true;
                        for (QName i : internalEndpoint.getInterfaces()) {
                            if (linkedEndpoint.getInterfaces()[0].equals(i)) {
                                isNotInInt = false;
                            }
                        }
                        if (isNotInInt) {
                            Context interfaceCtx = (Context) interfacesContext
                                    .lookup(interf);
                            interfaceCtx.unbind(linkedEndpoint
                                    .getToEndpointName());
                            if (!interfacesContext.list(interf)
                                    .hasMoreElements()) {
                                interfacesContext.destroySubcontext(interf);
                            }
                        }
                    }

                } catch (NamingException e) {
                    log.error(e.getMessage() ,e);
                    throw new RegistryException(
                            "problem while unregistering connection with interfacename "
                                    + linkedEndpoint.getInterfaces()[0]
                                    + " to "
                                    + linkedEndpoint.getToEndpointName() + " ");
                }
            }
        } else {
            try {
                if (!JNDIUtil.isBound(rootContext, ENDPOINTS_REF,
                        linkedEndpoint.getEndpointName())) {
                    log.error("The provider endpoint "
                            + linkedEndpoint.getEndpointName()
                            + " is not registered in the endpoints");
                    throw new RegistryException("The provider endpoint "
                            + linkedEndpoint.getEndpointName()
                            + " is not registered in the endpoints");
                }

                // If the prov service name is different from the cons service
                // name we removed it
                if (!linkedEndpoint.getToServiceName().equals(
                        linkedEndpoint.getServiceName())) {
                    retrieveServiceContext(
                            NamingHelper.qNameToName(linkedEndpoint
                                    .getServiceName())).unbind(
                            linkedEndpoint.getEndpointName());
                    // If the service context is empty, we need to remove it
                    Name serviceName = NamingHelper.qNameToName(linkedEndpoint
                            .getServiceName());
                    if (!servicesContext.list(serviceName).hasMore()) {
                        servicesContext.destroySubcontext(serviceName);
                    }
                }

                // If the prov endpoint name is different from the cons endpoint
                // name we removed it
                if (!linkedEndpoint.getEndpointName().equals(
                        linkedEndpoint.getToEndpointName())) {
                    endpointsContext.unbind(linkedEndpoint.getEndpointName());
                }

            } catch (NamingException e) {
                log.error(e.getMessage(), e);
                throw new RegistryException(
                        "problem while registering connection with service name "
                                + linkedEndpoint.getServiceName()
                                + " and endpoint name "
                                + linkedEndpoint.getEndpointName() + " to "
                                + linkedEndpoint.getToEndpointName() + " ");
            }
        }
        log.end();
    }

    /**
     * @see org.objectweb.petals.jbi.registry.Registry#deregisterExternalEndpoint(ServiceEndpoint)
     *      Uses the method deregisterInternalEndpoint
     */
    public void deregisterExternalEndpoint(ServiceEndpoint externalEndpoint)
        throws RegistryException {
        log.start();
        if (externalEndpoint == null) {
            throw new IllegalArgumentException(
                    "The external endpoint must be non null");
        }
        deregisterInternalEndpoint((AbstractEndpoint) externalEndpoint);
        log.end();
    }

    /**
     * @see Registry#deregisterInternalEndpoint(AbstractEndpoint)
     *      <ol>
     *      <li> Unbind the endpoint from the endpoint context</li>
     *      <li> If the endpoint is still in the new-endpoints context, unbind
     *      the endpoint from the new-endpoints context</li>
     *      <li> Unbind the endpoint from the service context with contains it</li>
     *      <li> If the service subcontext is empty, remove it from the services
     *      context</li>
     *      <li> Unbind the endpoint from the interface context with contains it</li>
     *      <li> If the interface subcontext is empty, remove it from the
     *      interfaces context</li>
     *      </ol>
     */
    public void deregisterInternalEndpoint(AbstractEndpoint address)
        throws RegistryException {
        log.start();
        if (address == null) {
            throw new IllegalArgumentException("The endpoint must be non null");
        }
        try {
            // Check if the endpoint is already registered before deregistering
            // it
            if (!JNDIUtil.isBound(rootContext, ENDPOINTS_REF, address
                    .getEndpointName())) {
                throw new RegistryException("The endpoint "
                        + address.getEndpointName()
                        + " must be previously registered");
            }

            //
            // 1.Unbind from interfaces context
            if (address.getInterfaces() != null) {
                for (QName interfaceName : address.getInterfaces()) {
                    Name name = NamingHelper.qNameToName(interfaceName);
                    if (JNDIUtil.isBound(rootContext, INTERFACES_REF, name)) {
                        Context interfaceCtx = (Context) interfacesContext
                                .lookup(name);

                        interfaceCtx.unbind(address.getEndpointName());

                        if (!interfacesContext.list(name).hasMoreElements()) {
                            interfacesContext.destroySubcontext(name);
                        }
                    } else {
                        log.error("Could not deregister internal endpoint "
                                + address.getEndpointName() + " its interface "
                                + interfaceName + " is not registered");
                        throw new RegistryException(
                                "Could not deregister internal endpoint "
                                        + address.getEndpointName()
                                        + " its interface " + interfaceName
                                        + " is not registered");
                    }
                }
            }

            //
            // 2.Unbind from services context
            if (!StringHelper.isEmpty(NamingHelper.qNameToString(address
                    .getServiceName()))) {
                Name serviceName = NamingHelper.qNameToName(address
                        .getServiceName());

                if (JNDIUtil.isBound(rootContext, SERVICES_REF, serviceName)) {
                    Context serviceCtx = (Context) servicesContext
                            .lookup(serviceName);

                    serviceCtx.unbind(address.getEndpointName());

                    // destroy service subcontext if no more endpoints serve
                    // this
                    // services
                    if (!servicesContext.list(serviceName).hasMore()) {
                        servicesContext.destroySubcontext(serviceName);
                    }
                } else {
                    log.error("Could not deregister internal endpoint "
                            + address.getEndpointName() + " its service "
                            + address.getServiceName() + " is not registered");
                    throw new RegistryException(
                            "Could not deregister internal endpoint "
                                    + address.getEndpointName()
                                    + " its service "
                                    + address.getServiceName()
                                    + " is not registered");
                }
            }

            //
            // 3.Remove the element from endpoints map
            endpointsContext.unbind(address.getEndpointName());

            // 3.1.Unbind from new-endpoints context
            if (JNDIUtil.isBound(rootContext, NEW_ENDPOINTS_REF, address
                    .getEndpointName())) {
                newEndpointContext.unbind(address.getEndpointName());
            }

        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new RegistryException(e);
        }
        log.end();
    }

    /**
     * @see Registry#getExternalEndpoint(QName, String)
     */
    public AbstractEndpoint getExternalEndpoint(QName service, String name)
        throws RegistryException {
        log.call();
        return retrieveEndpoint(service, name, ExternalEndpoint.class.getName());
    }

    /**
     * @see Registry#getExternalEndpointsForInterface(QName)
     */
    public AbstractEndpoint[] getExternalEndpointsForInterface(
            QName interfaceName) throws RegistryException {
        log.call();
        return retrieveEndpointsForInterface(interfaceName,
                ExternalEndpoint.class.getName());
    }

    /**
     * @see Registry#getExternalEndpointsForService(QName)
     */
    public AbstractEndpoint[] getExternalEndpointsForService(QName serviceName) {
        log.call();
        return retrieveEndpointsForService(serviceName, ExternalEndpoint.class
                .getName());
    }

    /**
     * @see Registry#getInternalEndpoint(QName, String)
     */
    public AbstractEndpoint getInternalEndpoint(QName service, String name)
        throws RegistryException {
        log.call();
        return retrieveEndpoint(service, name, InternalEndpoint.class.getName());
    }

    /**
     * @see Registry#getInternalEndpoint(QName, String,boolean)
     */
    public AbstractEndpoint getInternalEndpoint(QName service, String name,
            boolean resolveLink) throws RegistryException {
        log.call();
        return retrieveEndpoint(service, name,
                InternalEndpoint.class.getName(), resolveLink);
    }

    /**
     * Return an array of all endpoints for an interface
     * 
     * @param interfaceName
     * @return a non null array
     */
    public AbstractEndpoint[] getInternalEndpointsForInterface(
            QName interfaceName) throws RegistryException {
        log.call();
        return retrieveEndpointsForInterface(interfaceName,
                InternalEndpoint.class.getName());
    }

    /**
     * @see Registry#getInternalEndpointsForService(QName)
     */
    public AbstractEndpoint[] getInternalEndpointsForService(QName serviceName) {
        log.call();
        return retrieveEndpointsForService(serviceName, InternalEndpoint.class
                .getName());
    }

    /**
     * @see org.objectweb.petals.jbi.registry.Registry#getUsersContext()
     */
    public Context getUsersContext() {
        return usersContext;
    }

    /**
     * @see Registry#registerConnection(LinkedEndpoint)
     */
    public void registerConnection(LinkedEndpoint linkedEndpoint)
        throws RegistryException {
        log.start();
        if (linkedEndpoint == null) {
            throw new IllegalArgumentException(
                    "The connection must be non null");
        }
        // if (!JNDIUtil.isBound(rootContext, ENDPOINTS_REF, linkedEndpoint
        // .getToEndpointName())) {
        // log.error("The provider endpoint "
        // + linkedEndpoint.getToEndpointName() + " is not registered");
        // throw new RegistryException("The provider endpoint "
        // + linkedEndpoint.getToEndpointName() + " is not registered");
        // } else {
        if (linkedEndpoint.getInterfaces() != null) {
            try {
                LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/"
                        + linkedEndpoint.getToEndpointName());
                if (!JNDIUtil.isBound(rootContext, INTERFACES_REF, NamingHelper
                        .qNameToName(linkedEndpoint.getInterfaces()[0]))) {
                    retrieveInterfaceContext(
                            NamingHelper.qNameToName(linkedEndpoint
                                    .getInterfaces()[0])).bind(
                            linkedEndpoint.getToEndpointName(), linkRef);
                } else {
                    retrieveInterfaceContext(
                            NamingHelper.qNameToName(linkedEndpoint
                                    .getInterfaces()[0])).rebind(
                            linkedEndpoint.getToEndpointName(), linkRef);
                }
            } catch (NamingException e) {
                log.error(e.getMessage(), e);
                throw new RegistryException(
                        "problem while registering connection with interfacename "
                                + linkedEndpoint.getInterfaces()[0] + " to "
                                + linkedEndpoint.getToEndpointName() + " ");
            }
        } else {
            try {
                LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/"
                        + linkedEndpoint.getToEndpointName());
                if (!JNDIUtil.isBound(rootContext, SERVICES_REF, NamingHelper
                        .qNameToName(linkedEndpoint.getServiceName()))) {
                    retrieveServiceContext(
                            NamingHelper.qNameToName(linkedEndpoint
                                    .getServiceName())).bind(
                            linkedEndpoint.getEndpointName(), linkRef);
                } else {
                    retrieveServiceContext(
                            NamingHelper.qNameToName(linkedEndpoint
                                    .getServiceName())).rebind(
                            linkedEndpoint.getEndpointName(), linkRef);
                }
                if (!JNDIUtil.isBound(rootContext, ENDPOINTS_REF,
                        linkedEndpoint.getEndpointName())) {
                    endpointsContext.bind(linkedEndpoint.getEndpointName(),
                            linkRef);
                } else {
                    endpointsContext.rebind(linkedEndpoint.getEndpointName(),
                            linkRef);
                }

            } catch (NamingException e) {
                log.error(e.getMessage(), e);
                throw new RegistryException(
                        "problem while registering connection with service name  "
                                + linkedEndpoint.getServiceName()
                                + " and endpoint name "
                                + linkedEndpoint.getEndpointName() + " to "
                                + linkedEndpoint.getToEndpointName() + " ");
            }
        }
        // }
        log.end();
    }

    /**
     * @see Registry#registerExternalEndpoint(ServiceEndpoint)
     */
    public void registerExternalEndpoint(ServiceEndpoint externalEndpoint)
        throws RegistryException {
        log.start();
        registerEndpoint(externalEndpoint);
        log.end();
    }

    /**
     * @see Registry#registerInternalEndpoint(AbstractEndpoint)
     */
    public void registerInternalEndpoint(AbstractEndpoint endpoint)
        throws RegistryException {
        log.start();
        registerEndpoint(endpoint);
        log.end();
    }

    /**
     * @see Registry#retrieveNewEndpoints()
     */
    public List<AbstractEndpoint> retrieveNewEndpoints()
        throws RegistryException {
        log.start();
        List<AbstractEndpoint> endpoints = new ArrayList<AbstractEndpoint>();
        try {
            NamingEnumeration<Binding> bindings = rootContext
                    .listBindings(NEW_ENDPOINTS_REF);

            while (bindings.hasMoreElements()) {
                Binding binding = bindings.next();
                Object object = binding.getObject();
                if (object instanceof LinkRef) {
                    AbstractEndpoint abstractEndpoint = (AbstractEndpoint) rootContext
                            .lookup(((LinkRef) object).getLinkName());
                    if (abstractEndpoint.getDescription() == null) {
                        endpoints.add(abstractEndpoint);
                    }
                }
            }
        } catch (Exception e) {
            String msg = "Problem while retrieving all the new endpoints";
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
        log.end();
        return endpoints;
    }

    /**
     * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
     */
    @LifeCycle(on=LifeCycleType.START)
    public void start() throws IllegalLifeCycleException {
        log = new LoggingUtil(logger);
        try {
            initContexts();
        } catch (JBIException e) {
            throw new IllegalLifeCycleException(e.getMessage());
        }
    }

    /**
     * @see Registry#validateEndpoint(AbstractEndpoint)
     */
    public void validateEndpoint(AbstractEndpoint endpoint)
        throws RegistryException {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("The endpoint must be non null");
        }
        try {
            // Add the interfaces to the interfaces context
            for (QName name : endpoint.getInterfaces()) {
                Name nameInterface = NamingHelper.qNameToName(name);

                Context interfaceCtx = retrieveInterfaceContext(nameInterface);
                // Check if the endpoint is already bound
                if (JNDIUtil.isBound(interfacesContext, nameInterface, endpoint
                        .getEndpointName())) {
                    // It is already bound, so we update it
                    LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/"
                            + endpoint.getEndpointName());
                    interfaceCtx.rebind(endpoint.getEndpointName(), linkRef);
                } else {
                    // It is not bound so we bind it
                    LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/"
                            + endpoint.getEndpointName());
                    interfaceCtx.bind(endpoint.getEndpointName(), linkRef);
                }
            }
            // Update the endpoint in the endpoint context
            endpointsContext.rebind(endpoint.getEndpointName(), endpoint);

            // Remove the endpoint from the new endpoints context
            // newEndpointContext.unbind(endpoint.getEndpointName());
        } catch (Exception e) {
            log.error("Error while validating endpoint "
                    + endpoint.getEndpointName(), e);
            throw new RegistryException(e);
        }
        log.end();
    }

    /**
     * Bind the endpoint in the endpointcontext
     * 
     * @param endpoint
     *            endpoint to bind
     * @throws NamingException
     */
    @SuppressWarnings("unchecked")
    protected void bindInEndpointContext(AbstractEndpoint endpoint)
        throws NamingException {
        log.call();
        if (JNDIUtil.isBound(rootContext, ENDPOINTS_REF, endpoint
                .getEndpointName())) {
            throw new NamingException("Endpoint " + endpoint.getEndpointName()
                    + " is already bound");
        } else {
            endpointsContext.bind(endpoint.getEndpointName(), endpoint);
        }
    }

    /**
     * Bind the endpoint in the newendpointcontext
     * 
     * @param endpoint
     *            endpoint to bind
     * @throws NamingException
     */
    protected void bindInNewEndpointContext(AbstractEndpoint endpoint)
        throws NamingException {
        log.call();
        LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/"
                + endpoint.getEndpointName());
        if (!JNDIUtil.isBound(rootContext, NEW_ENDPOINTS_REF, endpoint
                .getEndpointName())) {
            newEndpointContext.bind(endpoint.getEndpointName(), linkRef);
        }
    }

    /**
     * Bind the endpoint in the servicecontext. If no serviceName is specified,
     * do nothing.
     * 
     * @param endpoint
     *            endpoint to bind
     * @throws NamingException
     */
    protected void bindInServiceContext(AbstractEndpoint endpoint)
        throws NamingException {
        log.call();
        Name serviceName = NamingHelper.qNameToName(endpoint.getServiceName());
        if (!StringHelper.isEmpty(NamingHelper.qNameToString(endpoint
                .getServiceName()))) {
            LinkRef linkRef = new LinkRef("/" + ENDPOINTS_REF + "/"
                    + endpoint.getEndpointName());
            Context servContext = retrieveServiceContext(serviceName);
            if (!JNDIUtil.isBound(servicesContext, serviceName, endpoint
                    .getEndpointName())) {
                servContext.bind(endpoint.getEndpointName(), linkRef);
            }
        }
    }

    /**
     * Create and / or reference contexts "containers", "endpoints", "services"
     * and "interfaces" in the global directory.
     * 
     * @throws JBIException
     *             problems while creating or lookup contexts
     */
    protected void initContexts() throws RegistryException {
        log.start();
        try {
            rootContext = new InitialContext(PropertyUtil
                    .retrieveJNDIProperties());
            if (JNDIUtil.isBound(rootContext, "/", ENDPOINTS_REF)) {
                endpointsContext = (Context) rootContext.lookup(ENDPOINTS_REF);
            } else {
                endpointsContext = rootContext.createSubcontext(ENDPOINTS_REF);
            }
            if (JNDIUtil.isBound(rootContext, "/", SERVICES_REF)) {
                servicesContext = (Context) rootContext.lookup(SERVICES_REF);
            } else {
                servicesContext = rootContext.createSubcontext(SERVICES_REF);
            }
            if (JNDIUtil.isBound(rootContext, "/", NEW_ENDPOINTS_REF)) {
                newEndpointContext = (Context) rootContext
                        .lookup(NEW_ENDPOINTS_REF);
            } else {
                newEndpointContext = rootContext
                        .createSubcontext(NEW_ENDPOINTS_REF);
            }
            if (JNDIUtil.isBound(rootContext, "/", INTERFACES_REF)) {
                interfacesContext = (Context) rootContext
                        .lookup(INTERFACES_REF);
            } else {
                interfacesContext = rootContext
                        .createSubcontext(INTERFACES_REF);
            }
            if (JNDIUtil.isBound(rootContext, "/", USERS_REF)) {
                usersContext = (Context) rootContext.lookup(USERS_REF);
            } else {
                usersContext = rootContext.createSubcontext(USERS_REF);
            }
            if (endpointsContext == null || servicesContext == null
                    || interfacesContext == null || usersContext == null) {
                throw new NotContextException();
            }
        } catch (Exception e) {
            log.error("Problem initating the contexts", e);
            throw new RegistryException(e);
        }
        log.end();
    }

    /**
     * @see Registry#registerExternalEndpoint(ServiceEndpoint)
     */
    protected void registerEndpoint(ServiceEndpoint endpoint)
        throws RegistryException {
        log.start();
        if (endpoint == null) {
            throw new IllegalArgumentException("Endpoint must not be null");
        }
        if (JNDIUtil.isBound(rootContext, ENDPOINTS_REF, endpoint
                .getEndpointName())) {
            // Retrieve a potential matching endpoint
            AbstractEndpoint registeredEndpoint = retrieveEndpoint(endpoint
                    .getEndpointName());

            // registeredEndpoint == null ==> the bound object isn't a real
            // endpoint (link or anything else)
            if (registeredEndpoint == null) {
                String msg = "You are trying to register an endpoint with a name that "
                        + "is already bound to an other object "
                        + "(link or anything else) in the JNDI Registry.";
                log.error(msg);
                throw new RegistryException(msg);
            }

            // An endpoint can't be registered with the same endpoint name as a
            // previously registered endpoint that have different container name
            // or component name
            if (!(((AbstractEndpoint) endpoint).getContainerName().equals(
                    registeredEndpoint.getContainerName()) && ((AbstractEndpoint) endpoint)
                    .getComponentName().equals(
                            registeredEndpoint.getComponentName()))) {
                String msg = "The endpoint "
                        + endpoint.getEndpointName()
                        + " is already registered for an other component/container";
                log.error(msg);
                throw new RegistryException(msg);
            }

        } else {
            try {
                bindInEndpointContext((AbstractEndpoint) endpoint);

                bindInNewEndpointContext((AbstractEndpoint) endpoint);

                bindInServiceContext((AbstractEndpoint) endpoint);
            } catch (Exception e) {
                log.error("Problem while registering endpoint "
                        + endpoint.getEndpointName(), e);
                throw new RegistryException(
                        "Problem while registering endpoint "
                                + endpoint.getEndpointName());
            }
        }
        log.end();
    }

    /**
     * Get the internal/external service endpoint for the named activated
     * endpoint, if any. If the Endpoint is a link, resolve it and return the
     * real endpoint
     * 
     * @param service
     *            the {@link QName} of the service
     * @param name
     *            the endpoint name
     * @param endpointClassName
     *            class name of the endpoint
     * @return the matching endpoint
     * @throws RegistryException
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint retrieveEndpoint(QName service, String name,
            String endpointClassName) throws RegistryException {
        log.call();
        AbstractEndpoint endpoint = retrieveEndpoint(service, name,
                endpointClassName, true);
        return endpoint;
    }

    /**
     * Get the internal/external service endpoint for the named activated
     * endpoint, if any
     * 
     * @param service
     *            qualified service name
     * @param name
     *            endpoint name
     * @param endpointClassName
     *            class name of the endpoint
     * @param resolveLink
     *            if the Endpoint is a link, resolve it (true) or return the
     *            link (false)
     * @return
     * @throws RegistryException
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint retrieveEndpoint(QName service, String name,
            String endpointClassName, boolean resolveLink)
        throws RegistryException {
        log.start();
        if (service == null) {
            throw new IllegalArgumentException(
                    "The service name must be non null");
        }
        if (name == null) {
            throw new IllegalArgumentException(
                    "The endpoint name must be non null");
        }
        AbstractEndpoint result = null;
        try {
            Object outObj = endpointsContext.lookup(name);

            // real endpoint
            if (outObj instanceof AbstractEndpoint) {
                AbstractEndpoint endpoint = (AbstractEndpoint) outObj;
                if (endpoint.getClass().getName().equals(endpointClassName)) {
                    result = endpoint;
                }
            }
            // link endpoint
            else if (outObj instanceof LinkRef) {
                Object obj = rootContext.lookup(((LinkRef) outObj)
                        .getLinkName());

                if (obj instanceof AbstractEndpoint) {
                    AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                    if (endpoint.getClass().getName().equals(endpointClassName)) {
                        // return the real endpoint
                        if (resolveLink) {
                            result = endpoint;
                        }
                        // return a link
                        else {
                            result = new LinkedEndpoint(service, name, endpoint
                                    .getServiceName(), endpoint
                                    .getEndpointName(), null);
                        }
                    }
                }
            }
        } catch (NamingException e1) {
            // Do nothing
        }
        log.end();
        return result;
    }

    /**
     * Retrieve an endpoint in the JNDI registry without making distinction
     * between Internal and External one.
     * 
     * @param name
     *            endpoint name
     * @return an {@link AbstractEndpoint} if the matching endpoint is a real
     *         endpoint, null otherwise (for exemple, the matching endpoint is a
     *         link)
     * @throws RegistryException
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint retrieveEndpoint(String name)
        throws RegistryException {
        log.start();
        if (name == null) {
            throw new IllegalArgumentException(
                    "The endpoint name must be non null");
        }
        AbstractEndpoint result = null;
        try {
            Object outObj = endpointsContext.lookup(name);

            // real endpoint
            if (outObj instanceof AbstractEndpoint) {
                result = (AbstractEndpoint) outObj;
            }
        } catch (NamingException e1) {
            String msg = "Error while looking up an Endpoint in the JNDI Registry";
            log.error(msg, e1);
            throw new RegistryException(msg, e1);
        }
        log.end();
        return result;
    }

    /**
     * Retreive the list of endpoint that implements the interface name
     * 
     * @param interfaceName
     *            the {@link QName} of the interface
     * @param endpointType
     *            the type (class name) of the endpoint (InternalEndpoint or
     *            ExternalEndpoint)
     * @return an Array of matching endpoints
     * @throws RegistryException
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint[] retrieveEndpointsForInterface(
            QName interfaceName, String endpointType) throws RegistryException {
        log.start();
        List<AbstractEndpoint> result = new ArrayList<AbstractEndpoint>();
        try {
            if (interfaceName == null) {
                // We must return all the endpoints of the specified type
                NamingEnumeration<Binding> bindings = rootContext
                        .listBindings(ENDPOINTS_REF);

                while (bindings.hasMoreElements()) {
                    Binding binding = (Binding) bindings.next();
                    if (binding.getObject() instanceof AbstractEndpoint) {
                        AbstractEndpoint endpoint = (AbstractEndpoint) binding
                                .getObject();
                        if (endpoint.getClass().getName().equals(endpointType)) {
                            result.add(endpoint);
                        }
                    }
                }
            } else {
                // We must return only the endpoints for the specified
                // interface
                Name name = NamingHelper.qNameToName(interfaceName);
                if (JNDIUtil.isBound(rootContext, INTERFACES_REF, name)) {
                    NamingEnumeration<Binding> bindings = interfacesContext
                            .listBindings(name);

                    while (bindings.hasMoreElements()) {
                        Binding binding = bindings.next();
                        Object object = binding.getObject();
                        if (object instanceof LinkRef) {
                            LinkRef linkRef = (LinkRef) object;
                            Object obj = rootContext.lookup(linkRef
                                    .getLinkName());
                            if (obj instanceof AbstractEndpoint) {
                                AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                                if (endpoint.getClass().getName().equals(
                                        endpointType)) {
                                    result.add(endpoint);
                                }
                            }
                        }
                    }
                }
            }
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new RegistryException(
                    "Problem while getting endpoints for the specified interface "
                            + interfaceName);
        }
        log.end();
        return result.toArray(new AbstractEndpoint[0]);
    }

    /**
     * @see Registry#getExternalEndpointsForService(QName)
     */
    @SuppressWarnings("unchecked")
    protected AbstractEndpoint[] retrieveEndpointsForService(QName serviceName,
            String endpointType) {
        log.start();
        if (serviceName == null) {
            throw new IllegalArgumentException(
                    "The service name must be non null");
        }
        AbstractEndpoint[] result = new AbstractEndpoint[0];
        Name name = NamingHelper.qNameToName(serviceName);
        if (JNDIUtil.isBound(rootContext, SERVICES_REF, name)) {
            try {
                NamingEnumeration<Binding> bindings = servicesContext
                        .listBindings(NamingHelper.qNameToName(serviceName));
                List<AbstractEndpoint> endpoints = new ArrayList<AbstractEndpoint>();
                while (bindings.hasMoreElements()) {
                    Binding binding = bindings.next();
                    Object object = binding.getObject();
                    if (object instanceof LinkRef) {
                        LinkRef linkRef = (LinkRef) object;
                        Object obj = rootContext.lookup(linkRef.getLinkName());
                        if (obj instanceof AbstractEndpoint) {
                            AbstractEndpoint endpoint = (AbstractEndpoint) obj;
                            if (endpoint.getClass().getName().equals(
                                    endpointType)) {
                                endpoints.add(endpoint);
                            }
                        }
                    }
                }
                result = endpoints.toArray(new AbstractEndpoint[0]);
            } catch (NamingException e) {
                log.error(e.getMessage(), e);
            }
        }
        log.end();
        return result;
    }

    /**
     * Return the interfaceContext associated with this interfaceName. If the
     * subcontext does not exists, it is created.
     * 
     * @param serviceName
     * @throws NamingException
     */
    protected Context retrieveInterfaceContext(Name interfaceName)
        throws NamingException {
        log.call();
        Context interfContext = null;
        if (JNDIUtil.isBound(rootContext, INTERFACES_REF, interfaceName)) {
            interfContext = (Context) interfacesContext.lookup(interfaceName);
        } else {
            interfContext = interfacesContext.createSubcontext(interfaceName);
        }
        return interfContext;
    }

    /**
     * Return the serviceContext associated with this serviceName. If the
     * subcontext does not exists, it is created.
     * 
     * @param serviceName
     * @throws NamingException
     */
    protected Context retrieveServiceContext(Name serviceName)
        throws NamingException {
        log.call();
        Context serviceContext = null;
        if (JNDIUtil.isBound(rootContext, SERVICES_REF, serviceName)) {
            serviceContext = (Context) servicesContext.lookup(serviceName);
        } else {
            serviceContext = servicesContext.createSubcontext(serviceName);
        }
        return serviceContext;
    }

}
