/*
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: AddressResolver.java,v 1.1 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.routing;

import javax.jbi.component.Component;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.management.service.EndpointService;
import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.jbi.registry.AbstractEndpoint;
import org.objectweb.petals.jbi.registry.InternalEndpoint;
import org.objectweb.petals.util.SystemUtil;

/**
 * The <code>AddressResolver</code> resolves implicit addresses such as
 * <code>interface</code> or <code>service</code> into an
 * <code>endpoint</code>. Rules are used to help the
 * <code>AddressResolver</code> in the choice of the <code>endpoint</code>,
 * if more than one implement the implicit address.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * 
 */
public class AddressResolver {

    protected EndpointService endpointService;

    public AddressResolver(EndpointService endpointService) {
        this.endpointService = endpointService;
    }

    /**
     * Resolve the address if necessary. Resolution consist in finding an
     * endpoint satisfying <code>interface</code> or <code>service</code>.
     * the source of the exchange and the exchange help the
     * <code>AddressResolver</code> to find the endpoint. The
     * <code>containerName</code> and <code>ComponentName</code> properties
     * are set. If <code>containerName</code> and <code>ComponentName</code>
     * already exist, do nothing. The check of acceptance from the two part is
     * done here (the Component method isExchangeWithXXXOkay()).
     * 
     * @param source
     * @param exchange
     * @throws RoutingException
     *             no destination can be found, or problem accessing the
     *             registry
     */
    public void resolveAddress(ComponentContextImpl source,
        MessageExchangeImpl exchange) throws RoutingException {
        ServiceEndpoint givenEndpoint = exchange.getEndpoint();
        QName givenServiceName = exchange.getService();
        QName givenInterfaceName = exchange.getInterfaceName();

        // Construct an array of potential matching endpoints.
        // If the endpoint has been set in the messageExchange,
        // this array contains only one element, which is this endpoint.
        ServiceEndpoint[] endpoints = null;
        ServiceEndpoint targetedEndpoint = null;

        if (givenEndpoint != null) {
            // try to resolve address with a connection
            // TODO minimize registry access, refactor
            if (givenEndpoint instanceof AbstractEndpoint
                && ((AbstractEndpoint) givenEndpoint).getType() == AbstractEndpoint.EndpointType.INTERNAL) {
                // Assume it is a real endpoint
                //FIXME don't check if it is still registered
                targetedEndpoint = givenEndpoint;
            } else {
                targetedEndpoint = resolveLinkedAddressForEndpoint(
                    givenEndpoint.getServiceName(), givenEndpoint
                        .getEndpointName());

                // if no connection, try to retrieve the real endpoint object
                // from the registry
                if (targetedEndpoint == null) {
                    targetedEndpoint = endpointService.getEndpoint(
                        givenEndpoint.getServiceName(), givenEndpoint
                            .getEndpointName());
                    // no endpoint found, we can not find a destination
                    if (targetedEndpoint == null) {
                        throw new RoutingException("The specified endpoint ("
                            + givenEndpoint
                            + ") does not match a registered endpoint.");
                    }
                }
            }
        } else if (givenServiceName != null) {
            // find an Endpoint with the specified ServiceName
            endpoints = endpointService
                .getInternalEndpointsForService(givenServiceName);

            // if there is several endpoints that matching
            // the given information, we have to choose one.
            targetedEndpoint = electEndpoint(endpoints, source, exchange);

            if (targetedEndpoint == null) {
                throw new RoutingException(
                    "No endpoint found for the specified service : "
                        + givenServiceName + ".");
            }
        } else if (givenInterfaceName != null) {
            // try to resolve address with a connection
            targetedEndpoint = resolveLinkedAddressForInterface(givenInterfaceName);

            if (targetedEndpoint == null) {
                // find an Endpoint with the specified InterfaceName
                endpoints = endpointService
                    .getInternalEndpointsForInterface(givenInterfaceName);

                // if there is several endpoints that matching
                // the given information, we have to choose one.
                targetedEndpoint = electEndpoint(endpoints, source, exchange);

                if (targetedEndpoint == null) {
                    throw new RoutingException(
                        "No endpoint found for the specified interface : "
                            + givenInterfaceName + ".");
                }
            }
        }

        // Now, we can set the real endpoint that will
        // serves the MessageExchange
        exchange.setEndpoint(targetedEndpoint);
    }

    /**
     * Retrieve a potential mapping between an interface name and an endpoint.
     * 
     * @param interfaceName
     * @return the linked endpoint, null if no connection retrieved
     */
    protected ServiceEndpoint resolveLinkedAddressForInterface(
        QName interfaceName) {
        ServiceEndpoint[] internalEndpoints = endpointService
            .getInternalEndpointsForInterface(interfaceName);
        if (internalEndpoints.length > 0) {
            return internalEndpoints[0];
        }
        return null;
    }

    /**
     * Retrieve a potential mapping between a fully qualified endpoint name
     * (service name + endpoint name)and a real endpoint.
     * 
     * @param serviceName
     * @param endpointName
     * @return the linked endpoint, null if no connection retrieved
     */
    protected ServiceEndpoint resolveLinkedAddressForEndpoint(
        QName serviceName, String endpointName) throws RoutingException {
        return endpointService.getEndpoint(serviceName, endpointName);
    }

    /**
     * Elects the address in the list of endpoints as the message recipient,
     * with the defined strategy. Ask the consumer and the provider if they are
     * Okay for the exchange.
     * 
     * <br>
     * Strategy : <br>
     * Each endpoint is tested, with the following rules.
     * <li> if the endpoint is hosted by a shutdown-container, keep it, but try
     * to find another 'active' endpoint.</li>
     * <li> if isExchangeValidation server.properties is set to true, test the
     * "isExchangeWithXXXOkay" (for an endpoint on an active container). If
     * okay, return the endpoint.</li>
     * <br>
     * If no 'active' endpoint is found, but one 'inactive' is found, return it;
     * so the message will be send, but will be delivered when the corresponding
     * container will start again.
     * 
     * @param addresses
     * @return the elected recipient
     */
    protected ServiceEndpoint electEndpoint(ServiceEndpoint[] endpoints,
        ComponentContextImpl ctxSource, MessageExchange exchange) {

        ServiceEndpoint result = null;
        ServiceEndpoint inactiveEndpoint = null;

        if (endpoints != null && endpoints.length > 0) {
            Component consumer = ctxSource.getComponent();

            for (int i = 0; i < endpoints.length && result == null; i++) {
                InternalEndpoint providerEP = (InternalEndpoint) endpoints[i];

                // Check if the container of the endpoint is started
                if (endpointService.isContainerStarted(providerEP)) {
                    // test if isExchangeWithXXXOkay has to be used.
                    if (SystemUtil.isExchangeValidation()) {
                        // apply isExchangeWithXXXOkay
                        if (consumer.isExchangeWithProviderOkay(providerEP,
                            exchange)
                            && providerEP.getEndpointService()
                                .isExchangeWithConsumerOkayForComponent(
                                    providerEP, exchange)) {
                            result = providerEP;
                        }
                    } else {
                        result = providerEP;
                    }
                }
                // the container of the endpoint is inactive,
                // keep the endpoint to use it if no active is found
                else {
                    inactiveEndpoint = providerEP;
                }
            }
            // if no active endpoint was found,
            // maybe an inactive endpoint was found.
            if (result == null) {
                result = inactiveEndpoint;
            }
        }
        return result;
    }
}
