/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 * --------------------------------------------------------------------------
 * $Id: JoramConnection.java,v 0.3 2005/07/22 10:24:27 alouis Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.jbi.transport;

import java.net.ConnectException;
import java.net.UnknownHostException;
import java.util.concurrent.ArrayBlockingQueue;

import javax.jbi.JBIException;
import javax.jms.Connection;
import javax.jms.DeliveryMode;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageConsumer;
import javax.jms.MessageListener;
import javax.jms.MessageProducer;
import javax.jms.Session;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.objectweb.joram.client.jms.ConnectionFactory;
import org.objectweb.joram.client.jms.Topic;
import org.objectweb.joram.client.jms.admin.AdminException;
import org.objectweb.joram.client.jms.admin.AdminModule;
import org.objectweb.joram.client.jms.local.LocalConnectionFactory;

import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.util.JNDIUtil;

/**
 * @author Adrien LOUIS - EBM WebSourcing
 * 
 */
public class JoramConnection {

    private int POOLSIZE = 8;

    protected static final String TOPIC_NAME = "JoramConnection-";

    protected String host;

    protected InitialContext ictx;

    protected int id;

    protected String pwd;

    protected Connection connection;

    protected ArrayBlockingQueue<Session> sessionsPool;

    protected Session session;

    protected Topic topic;

    protected int tcp;

    protected String user;

    protected boolean connectionStarted = false;

    protected Serializer serializer = new ObjectSerializer();

    public JoramConnection(int id, int tcp, String user, String pwd,
        String host, InitialContext ictx) {
        this.id = id;
        this.tcp = tcp;
        this.user = user;
        this.pwd = pwd;
        this.host = host;
        this.ictx = ictx;
    }

    public static String containerNameToTopicName(String containerName) {
        return TOPIC_NAME + containerName;
    }

    /**
     * Create and bind the Topic, if necessary, and Subscribeon the topic of
     * this server. The specified MessageListener is called when a message
     * incomes
     * 
     * @param aML
     * @throws JMSException
     * @throws JBIException
     */
    public void start(MessageListener aML) throws JMSException,
        ConnectException, JBIException {

        ConnectionFactory connectionFactory = null;
        try {
            connectAdminModule();
            topic = bindTopic();
            connectionFactory = (ConnectionFactory) LocalConnectionFactory
                .create();
            disconnectAdminModule();
        } catch (Exception e) {
            throw new JBIException(e);
        }

        // FIXME optim multithread
        //connectionFactory.getParameters().multiThreadSync = true;
        //connectionFactory.getParameters().multiThreadSyncDelay = 1;

        connection = connectionFactory.createConnection(user, pwd);
        session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
        MessageConsumer consumer = session.createDurableSubscriber(topic, ""
            + id);
        consumer.setMessageListener(aML);

        // create the sessions pool to send messages
        sessionsPool = new ArrayBlockingQueue<Session>(POOLSIZE);
        for (int i = 0; i < POOLSIZE; i++) {
            sessionsPool.add(connection.createSession(false,
                Session.CLIENT_ACKNOWLEDGE));
        }

        connection.start();
        connectionStarted = true;
    }

    /**
     * Close jms and jndi connections. Clean sessions pool The Topic
     * subscribtion is not removed.
     * 
     * @throws JMSException
     *             can not close the jms connection or session
     * @throws NamingException
     *             can not close the jndi context access
     */
    public void stop() throws JMSException, NamingException {

        for (Session session : sessionsPool) {
            session.close();
        }
        if (connectionStarted) {
            connection.stop();
            connectionStarted = false;
        }
    }

    /**
     * Create and bind the topic for this server in the global JNDI (create or
     * retrieve it). Connection to the Joram AdminModule must be done.
     * 
     * @return
     * @throws NamingException
     * @throws ConnectException
     * @throws AdminException
     */
    protected Topic bindTopic() throws NamingException, ConnectException,
        AdminException {
        Topic t = null;

        String topicName = TOPIC_NAME + id;

        boolean alreadyBound = JNDIUtil.isBound(ictx, "/", topicName);

        if (alreadyBound) {
            // retreive the existing topic
            t = (Topic) ictx.lookup(topicName);
        } else {
            // create and bind a new topic
            t = Topic.create(topicName);
            t.setFreeReading();
            t.setFreeWriting();
            ictx.bind(topicName, t);
        }

        return t;
    }

    protected void unbindTopic() throws NamingException {

        String topicName = TOPIC_NAME + id;

        boolean alreadyBound = JNDIUtil.isBound(ictx, "/", topicName);

        if (alreadyBound) {
            // retreive the existing topic
            ictx.unbind(topicName);
        }
    }

    /**
     * send the MessageExchange to the remoteTopic. the exchange is tranformed
     * into a jms Message
     * 
     * @param remoteTopic
     * @param messageExchange
     * @param timeToLive
     * @param persistence
     * @throws TransportException
     */
    public void sendTo(String containerName,
        MessageExchangeImpl messageExchange, long timeToLive, boolean persistence)
        throws TransportException {

        Session distantSession = null;
        MessageProducer producer;

        try {
            // find the topic destination in jndi
            Topic remoteTopic = findTopicForContainer(containerName);

            // retrieve a session - block until a session is available
            distantSession = sessionsPool.take();

            // create a producer
            producer = distantSession.createProducer(remoteTopic);

            // set the TextMessage object with MessageExchange object
            Message msg = serializer.jbi2jms(messageExchange, distantSession);

            // set persistence to true or false
            int persistenceMode = (persistence) ? DeliveryMode.PERSISTENT : DeliveryMode.NON_PERSISTENT;
            
            producer.send(msg, persistenceMode, 0, timeToLive);

            producer.close();

            sessionsPool.offer(distantSession);
            
            distantSession =null;

        } catch (InterruptedException e) {
            throw new TransportException(
                "Problem accessing the JMS sessions pool.", e);
        } catch (JMSException e) {
            throw new TransportException(
                "Can not send the message to the JMS destination.", e);
        } finally {
            // release the session
            if (distantSession != null) {
                sessionsPool.offer(distantSession);
            }
        }
    }

    /**
     * Unsubscribe from the Topic. The topic is removed from the Joram
     * configuration and from the Jndi server.
     * 
     * @throws JMSException
     * @throws NamingException
     * @throws AdminException
     * @throws UnknownHostException
     * @throws ConnectException
     */
    public void unsubscribe() throws JMSException, NamingException,
        ConnectException, UnknownHostException, AdminException {

        if (connectionStarted)
            connection.stop();

        session.close();
        connection.start();

        session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
        session.unsubscribe("" + id);
        session.close();

        connection.close();
        connectionStarted = false;

        unbindTopic();
    }

    /**
     * Find in the JNDI context the registered Topic associated to the specified
     * containerName
     * 
     * @param containerName
     * @return the topic related to the containerName
     * @throws TransportException
     *             topic not found, or jndi access exception
     */
    protected Topic findTopicForContainer(String containerName)
        throws TransportException {
        Topic remoteTopic = null;
        try {
            remoteTopic = (Topic) ictx.lookup(JoramConnection
                .containerNameToTopicName(containerName));
        } catch (NamingException e) {
            throw new TransportException(
                "Can not find the Topic associate to the destination container '"
                    + containerName + "'", e);
        }
        return remoteTopic;
    }

    /**
     * @see AdminModule#connect(String, int, String, String, int, String)
     */
    protected void connectAdminModule() throws ConnectException,
        UnknownHostException, AdminException {
        AdminModule.connect(host, tcp, user, pwd, 3);
    }

    /**
     * @see AdminModule#disconnect()
     */
    protected void disconnectAdminModule() {
        AdminModule.disconnect();
    }

}
