/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PetalsAdminService.java 16:40:52 alouis $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.kernel.admin.service;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.naming.Binding;
import javax.naming.InitialContext;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.fractal.fraclet.annotation.Requires;
import org.objectweb.petals.jbi.management.deployment.DeploymentService;
import org.objectweb.petals.jbi.management.service.InstallationService;
import org.objectweb.petals.jbi.transport.Transporter;
import org.objectweb.petals.kernel.admin.ContainerInformation;
import org.objectweb.petals.util.JNDIUtil;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.PropertyUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * This class is used for Petals global Administration
 * 
 * @author alouis - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.kernel.admin.service.PetalsAdminServiceMBean.class))
public class PetalsAdminService implements PetalsAdminServiceMBean {

	protected static final String CONTAINERS_REF = "containers";

	/**
	 * Logger wrapper
	 */
	protected LoggingUtil log;

	/**
	 * Monolog
	 */
	@Monolog(name = "logger")
	protected Logger logger;

	/**
	 * JBI :: JMX Deployment
	 */
	@Requires(name = "deployment", signature = org.objectweb.petals.jbi.management.deployment.DeploymentService.class)
	protected DeploymentService deployment;

	/**
	 * JBI :: JMX Installation
	 */
	@Requires(name = "installation", signature = org.objectweb.petals.jbi.management.service.InstallationService.class)
	protected InstallationService installation;

	@Requires(name = "transporter", signature = org.objectweb.petals.jbi.transport.Transporter.class)
	protected Transporter transporter;

	protected Component rootComponent;

	/**
	 * uninstalled serviceassemblies, components and sharelibraries endpoints
	 * are removed the container is removed from the global configuration
	 * 
	 * @throws Exception
	 */
	public void shutdownContainer() throws Exception {
		log.start();
		log
				.info("Shutdown the server. It will be removed from the Petals network.");
		try {
			log.info("  1/4 Undeploy Service Assemblies.");
			deployment.shutdown();
		} catch (Exception e) {
			log.error("Can not shutdown the DeploymentService.", e);
		}
		try {
			log.info("  2/4 Uninstall Shared Libraries and Components.");
			installation.shutdown();
		} catch (Exception e) {
			log.error("Can not shutdown the InstallationService.", e);
		}
		try {
			log
					.info("  3/4 Shutdown the Transporter (remove the server from the global configuration).");
			transporter.shutdown();
		} catch (Exception e) {
			log.error("Can not shutdown the Transporter.", e);
		}

		log.info("  4/4 Stopping the server...");
		System.exit(0);
		log.end();
	}

	public void stopContainer() throws Exception {
		log.start();
		log.info("Stopping the server...");
		System.exit(0);
		log.end();
	}

	/**
	 * Retrieve the collection of all the started servers configuration from
	 * this domain.
	 * 
	 * @return Collection of server information ContainerInformation
	 */
	public Collection<Map<String, String>> getAllServersConfiguration() {
		return retrieveAllServersConfiguration(false);
	}

	/**
	 * Retrieve the collection of all the started servers configuration from
	 * this domain.
	 * 
	 * @return Collection of server information ContainerInformation
	 */
	public Collection<Map<String, String>> getAllStartedServersConfiguration() {
		return retrieveAllServersConfiguration(true);
	}

	/**
	 * Retrieve the collection of all the started servers configuration from
	 * this domain.
	 * 
	 * @param startedOnly
	 *            return only started servers information
	 * @return Collection of server information ContainerInformation
	 */
	protected Collection<Map<String, String>> retrieveAllServersConfiguration(
			boolean startedOnly) {
		log.start();
		Set<Map<String, String>> containers = new HashSet<Map<String, String>>();
		try {
			InitialContext rootContext = new InitialContext(PropertyUtil
					.retrieveJNDIProperties());
			if (JNDIUtil.isBound(rootContext, "/", CONTAINERS_REF)) {
				NamingEnumeration<Binding> enums = rootContext
						.listBindings(CONTAINERS_REF);
				while (enums.hasMoreElements()) {
					Binding binding = (Binding) enums.next();
					ContainerInformation containerInformation = (ContainerInformation) binding
							.getObject();
					if (!startedOnly
							|| (startedOnly && containerInformation.isStarted())) {
						containers
								.add(createConfigurationMap(containerInformation));
					}
				}
			}
		} catch (NamingException e) {
			log
					.debug("Problem while trying to retrieve all the containers information : context may not be initialized");
		}
		log.end();
		return containers;
	}

	/**
	 * Create a Map from a ContainerInformation object
	 * 
	 * @param info
	 * @return a Map of the server properties
	 */
	protected Map<String, String> createConfigurationMap(
			ContainerInformation info) {
		Map<String, String> map = new HashMap<String, String>();
		if (info != null) {
			map.put(CONF_HOST, info.getHost());
			map.put(CONF_HTML_PORT, info.getHtmlPort());
			map.put(CONF_JMX_LOGIN, info.getJmxLogin());
			map.put(CONF_JMX_PASSWORD, info.getJmxPassword());
			map.put(CONF_JMX_PORT, info.getJmxPort());
			map.put(CONF_JNDI_FACTORY, info.getJndiFactory());
			map.put(CONF_JNDI_PORT, info.getJndiPort());
			map.put(CONF_JORAM_DOMAIN, info.getJoramDomain());
			map.put(CONF_JORAM_DOMAIN_PORT, info.getJoramDomainPort());
			map.put(CONF_JORAM_ID, info.getJoramId());
			map.put(CONF_JORAM_LOGIN, info.getJoramLogin());
			map.put(CONF_JORAM_PASSWORD, info.getJoramPassword());
			map.put(CONF_JORAM_TCP_PORT, info.getJoramTCPPort());
			map.put(CONF_NAME, info.getName());
			map
					.put(
							CONF_STATE,
							(info.getState() == ContainerInformation.ContainerState.STARTED) ? "started"
									: "stopped");
			map.put(CONF_UID, "" + info.getUid());
		}
		return map;
	}

	// ----------------------------------------------------------------------
	// Fractal LifecycleController implementation
	// ----------------------------------------------------------------------

	/**
	 * @see org.objectweb.fractal.api.control.LifeCycleController#startFc()
	 */
	@LifeCycle(on=LifeCycleType.START)
	public void start() throws IllegalLifeCycleException {
		log = new LoggingUtil(logger);
	}

}
