/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: SocketServerThread.java 11:55:50 AM ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.kernel.registry.thread;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;

import javax.naming.Binding;
import javax.naming.NameClassPair;
import javax.naming.NamingException;

import org.objectweb.petals.kernel.registry.RegistryServer;
import org.objectweb.petals.kernel.registry.msg.request.RegistryRequest;
import org.objectweb.petals.kernel.registry.msg.response.AckResponse;
import org.objectweb.petals.kernel.registry.msg.response.CreateSubcontextResponse;
import org.objectweb.petals.kernel.registry.msg.response.ExceptionResponse;
import org.objectweb.petals.kernel.registry.msg.response.ListBindingsResponse;
import org.objectweb.petals.kernel.registry.msg.response.ListResponse;
import org.objectweb.petals.kernel.registry.msg.response.LookupLinkResponse;
import org.objectweb.petals.kernel.registry.msg.response.LookupResponse;
import org.objectweb.petals.kernel.registry.msg.response.RegistryResponse;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;

/**
 * This class is the socket server for the distributed JNDI directory
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public class SocketServerThread extends Thread {

    /**
     * Logger
     */
    protected LoggingUtil log;

    /**
     * Boolean used to control the lifecycle of the socket server
     */
    protected boolean run = true;

    /**
     * Start time of this server
     */
    protected long sender;

    /**
     * JNDI server
     */
    protected RegistryServer server;

    /**
     * Socket server
     */
    protected ServerSocket serverSocket;

    public SocketServerThread(RegistryServer server) {
        super("SocketServerThread" + server.getStartTime());
        this.server = server;
        this.log = server.getLog();
        this.sender = server.getStartTime();
    }

    @SuppressWarnings("unchecked")
    public void run() {
        log.call();
        try {
            serverSocket = new ServerSocket();
            serverSocket.bind(new InetSocketAddress(SystemUtil.getHost(),
                Integer.parseInt(SystemUtil.getJndiPort())));

            log.info("Start JNDI server on " + SystemUtil.getHost() + ":"
                + SystemUtil.getJndiPort());

            RegistryRequest request = null;
            RegistryResponse response = null;
            ObjectInputStream objectInputStream = null;
            ObjectOutputStream objectOutputStream = null;

            Socket socket = null;
            while (run) {
                socket = serverSocket.accept();
                // log.debug(server.getStartTime() + " Accept a socket");

                try {
                    objectInputStream = new ObjectInputStream(socket
                        .getInputStream());
                    request = (RegistryRequest) objectInputStream.readObject();
                } catch (Exception e) {
                    log.error("Unable to read JNDIRequest", e);
                }

                if (request != null) {
                    try {
                        // log.debug(server.getStartTime() + " Request type : "
                        // + request.getType());
                        Object object = server.processRequest(request);
                        switch (request.getType()) {
                            /**
                             * BIND
                             */
                            case bind:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            /**
                             * CREATE SUBCONTEXT
                             */
                            case createSubcontext:
                                response = new CreateSubcontextResponse(
                                    (String) object, sender, 0);
                                request.setSender(sender);
                                break;
                            /**
                             * DESTROY SUBCONTEXT
                             */
                            case destroySubcontext:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            /**
                             * LIST
                             */
                            case list:
                                response = new ListResponse(
                                    (NameClassPair[]) object, sender, 0);
                                break;
                            /**
                             * LIST BINDINGS
                             */
                            case listBindings:
                                response = new ListBindingsResponse(
                                    (Binding[]) object, sender, 0);
                                break;
                            /**
                             * LOOKUP
                             */
                            case lookup:
                                response = new LookupResponse(object, sender,
                                    request.getSender());
                                break;
                            /**
                             * LOOKUP LINK
                             */
                            case lookupLink:
                                response = new LookupLinkResponse(object,
                                    sender, request.getSender());
                                break;
                            /**
                             * PING
                             */
                            case ping:
                                response = new AckResponse(sender, request
                                    .getSender());
                                break;
                            /**
                             * REBIND
                             */
                            case rebind:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            /**
                             * RENAME
                             */
                            case rename:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            /**
                             * UNBIND
                             */
                            case unbind:
                                response = new AckResponse(sender, 0);
                                request.setSender(sender);
                                break;
                            default:
                                break;
                        }
                        if (request.getSender() != 0) {
                            // Send the request to the other registries if the
                            // sender is not null
                            try {
                                server.getChannel().send(request);
                            } catch (Exception e) {
                                log
                                    .error(
                                        "Error while send the request to the other registries",
                                        e);
                            }
                        }
                    } catch (NamingException e) {
                        log.error(e.getMessage(), e);
                        response = new ExceptionResponse(e, sender, request
                            .getSender());
                    }
                }
                // log.debug("Send response " + response.getType());
                objectOutputStream = new ObjectOutputStream(socket
                    .getOutputStream());
                objectOutputStream.writeObject(response);
                objectOutputStream.flush();
                objectOutputStream.close();
            }
        } catch (SocketException e) {
            if (!"Socket closed".equals(e.getMessage())) {
                log.error(e.getMessage(), e);
            }
        } catch (IOException e) {
            log.error("Error while starting socket server", e);
            throw new RuntimeException("Error while starting socket server", e);
        }
    }

    public void terminate() {
        log.start();
        run = false;
        // Close the server socket
        try {
            serverSocket.close();
            this.interrupt();
        } catch (Exception e) {
            log.error("Problem while closing server socket " + e);
        }
        log.end();
    }

}
