/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JoramRegistryTest.java 10:05:48 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.registry;

import java.lang.reflect.Method;
import java.util.List;

import javax.jbi.JBIException;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.LinkRef;
import javax.naming.Name;
import javax.naming.NameClassPair;
import javax.naming.NamingException;
import javax.xml.namespace.QName;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.petals.jbi.registry.mock.MockContext;
import org.objectweb.petals.jbi.registry.mock.MockInitialContext;
import org.objectweb.petals.jbi.registry.mock.MockLinkRef;
import org.objectweb.petals.kernel.registry.jndi.NamingEnumerationImpl;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.NamingHelper;
import org.objectweb.util.monolog.api.Logger;

/**
 * Test of the JoramRegistry
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public class JNDIRegistryTest extends TestCase {

    private MockContext endpointsContext = new MockContext();

    private MockContext interfacesContext = new MockContext();

    private MockContext newEndpointContext = new MockContext();

    private JNDIRegistry registry;

    private MockInitialContext rootContext;

    private MockContext servicesContext = new MockContext();

    public void setUp() throws NamingException {
        registry = new JNDIRegistry();
        rootContext = new MockInitialContext();

        rootContext.bind(JNDIRegistry.ENDPOINTS_REF, endpointsContext);
        rootContext.bind(JNDIRegistry.INTERFACES_REF, interfacesContext);
        rootContext.bind(JNDIRegistry.NEW_ENDPOINTS_REF, newEndpointContext);
        rootContext.bind(JNDIRegistry.SERVICES_REF, servicesContext);

        registry.rootContext = rootContext;

        Logger logger = EasyMock.createMock(Logger.class);
        registry.logger = logger;
        registry.log = EasyMock.createMock(LoggingUtil.class);

        registry.endpointsContext = endpointsContext;
        registry.interfacesContext = interfacesContext;
        registry.newEndpointContext = newEndpointContext;
        registry.servicesContext = servicesContext;
    }

    public void testDeregisterConnection() throws RegistryException {
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(new QName("serv"),
            "endpoint", new QName("toserv"), "endpoint", null);
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "toserv"), "endpoint", "compo", "0", null);
        endpointsContext.getBindings().put("endpoint", internalEndpoint);
        registry.registerConnection(linkedEndpoint);
        registry.deregisterConnection(linkedEndpoint);
        assertEquals(interfacesContext.getBindings().size(), 0);
    }

    public void testDeregisterConnectionWithInterfaces()
        throws RegistryException {
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(new QName("inter"),
            new QName("toserv"), "endpoint", null);
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "toserv"), "endpoint", "compo", "0", null);
        internalEndpoint.setInterfaces(new QName[] {new QName("test")});
        endpointsContext.getBindings().put("endpoint", internalEndpoint);
        registry.registerConnection(linkedEndpoint);
        registry.deregisterConnection(linkedEndpoint);
        assertEquals(interfacesContext.getBindings().size(), 0);
    }

    public void testDeregisterConnectionExceptionNull() {
        try {
            registry.deregisterConnection(null);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    public void testDeregisterConnectionWithInterfacesStillNewEndpoints()
        throws RegistryException {
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(new QName("inter"),
            new QName("toserv"), "endpoint", null);
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "toserv"), "endpoint", "compo", "0", null);
        endpointsContext.getBindings().put("endpoint", internalEndpoint);
        newEndpointContext.getBindings().put("endpoint", internalEndpoint);
        registry.registerConnection(linkedEndpoint);
        registry.deregisterConnection(linkedEndpoint);
        assertEquals(interfacesContext.getBindings().size(), 1);
        assertEquals(((MockContext) interfacesContext.getBindings()
            .get("inter")).getBindings().size(), 0);
    }

    public void testDeregisterConnectionWithInterfacesNotRegistered()
        throws RegistryException {
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(new QName("inter"),
            new QName("toserv"), "endpoint", null);
        try {
            registry.deregisterConnection(linkedEndpoint);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    public void testDeregisterExternalEndpoint() throws JBIException,
        NamingException {
        AbstractEndpoint endpoint = new InternalEndpoint(new QName("test"),
            "endpoint", "axis2bc", "0", null);
        testRegisterInternalEndpoint();
        registry.deregisterExternalEndpoint(endpoint);
        assertEquals(endpointsContext.getBindings().size(), 0);
    }

    public void testDeregisterExternalEndpointException() {
        try {
            registry.deregisterExternalEndpoint(null);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    public void testDeregisterInternalEndpoint() throws JBIException,
        NamingException {
        AbstractEndpoint endpoint = new InternalEndpoint(new QName("test"),
            "endpoint", "axis2bc", "0", null);
        testRegisterInternalEndpoint();
        registry.deregisterInternalEndpoint(endpoint);
        assertEquals(endpointsContext.getBindings().size(), 0);
    }

    public void testDeregisterInternalEndpointException() {
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "serv"), "endpoint", "compo", "0", null);
        try {
            registry.deregisterInternalEndpoint(internalEndpoint);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    public void testDeregisterInternalEndpointExceptionInterface()
        throws RegistryException, NamingException {
        AbstractEndpoint endpoint = new InternalEndpoint(new QName("test"),
            "endpoint", "axis2bc", "0", null);
        endpoint.setInterfaces(new QName[] {new QName("foo")});
        registry.registerInternalEndpoint(endpoint);
        MockContext context = new MockContext();
        interfacesContext.bind("inter", context);
        MockLinkRef linkRef = new MockLinkRef("endpoint");
        linkRef.setClassName(InternalEndpoint.class.getName());
        context.bind("endpoint", linkRef);
        try {
            registry.deregisterInternalEndpoint(endpoint);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    public void testDeregisterInternalEndpointExceptionNull() {
        try {
            registry.deregisterInternalEndpoint(null);
            fail();
        } catch (Exception e) {
            // do nothing
        }
    }

    public void testDeregisterInternalEndpointExceptionService()
        throws RegistryException {
        AbstractEndpoint endpoint = new InternalEndpoint(new QName("test"),
            "endpoint", "axis2bc", "0", null);
        registry.registerInternalEndpoint(endpoint);
        servicesContext.getBindings().clear();
        try {
            registry.deregisterInternalEndpoint(endpoint);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    public void testDeregisterInternalEndpointWithInterfaces()
        throws JBIException, NamingException {
        AbstractEndpoint endpoint = new InternalEndpoint(new QName("test"),
            "endpoint", "axis2bc", "0", null);
        endpoint.setInterfaces(new QName[] {new QName("inter")});
        registry.registerInternalEndpoint(endpoint);
        MockContext context = new MockContext();
        interfacesContext.bind("inter", context);
        MockLinkRef linkRef = new MockLinkRef("endpoint");
        linkRef.setClassName(InternalEndpoint.class.getName());
        context.bind("endpoint", linkRef);
        registry.deregisterInternalEndpoint(endpoint);
        assertEquals(endpointsContext.getBindings().size(), 0);
    }

    public void testGetExternalEndpoint() throws JBIException, NamingException {
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "service"), "endpoint", "compo", "0", null);
        ExternalEndpoint externalEndpoint = new ExternalEndpoint(
            internalEndpoint, "compo", "0");
        endpointsContext.getBindings().put("endpoint", externalEndpoint);
        AbstractEndpoint endpoint = registry.getExternalEndpoint(new QName(
            "test"), "endpoint");
        assertEquals(endpoint.containerName, "0");
    }

    public void testGetExternalEndpointsForInterface() throws JBIException,
        NamingException {
        ExternalEndpoint externalEndpoint = new ExternalEndpoint(
            new ConsumerEndpoint("compo", "0"), "compo", "0");
        rootContext.bind("test-link", externalEndpoint);
        MockContext context = (MockContext) interfacesContext
            .createSubcontext("test");
        MockLinkRef linkRef = new MockLinkRef("test-link");
        linkRef.setClassName(ExternalEndpoint.class.getName());
        context.bind("inter1", linkRef);
        AbstractEndpoint[] endpoints = registry
            .getExternalEndpointsForInterface(new QName("test"));
        assertEquals(endpoints[0], externalEndpoint);
    }

    public void testGetExternalEndpointsForInterfaceNotRegisteredInterface()
        throws RegistryException {
        assertEquals(0, registry.getExternalEndpointsForInterface(new QName(
            "test")).length);
    }

    public void testGetExternalEndpointsForService() throws JBIException,
        NamingException {
        ExternalEndpoint externalEndpoint = new ExternalEndpoint(
            new ConsumerEndpoint("compo", "0"), "compo", "0");
        rootContext.bind("test-link", externalEndpoint);
        MockContext context = (MockContext) servicesContext
            .createSubcontext("test");
        MockLinkRef linkRef = new MockLinkRef("test-link");
        linkRef.setClassName(ExternalEndpoint.class.getName());
        context.bind("inter1", linkRef);
        AbstractEndpoint[] endpoints = registry
            .getExternalEndpointsForService(new QName("test"));
        assertEquals(endpoints[0], externalEndpoint);
    }

    public void testGetInternalEndpoint() throws JBIException, NamingException {
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "service"), "endpoint", "compo", "0", null);
        endpointsContext.getBindings().put("endpoint", internalEndpoint);
        AbstractEndpoint endpoint = registry.getInternalEndpoint(new QName(
            "test"), "endpoint");
        assertEquals(endpoint.containerName, "0");
    }

    public void testGetInternalEndpointsForInterface() throws JBIException,
        NamingException {
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "service"), "endpoint", "compo", "0", null);
        rootContext.bind("test-link", internalEndpoint);
        MockContext context = (MockContext) interfacesContext
            .createSubcontext("test");
        MockLinkRef linkRef = new MockLinkRef("test-link");
        linkRef.setClassName(InternalEndpoint.class.getName());
        context.bind("inter1", linkRef);
        AbstractEndpoint[] endpoints = registry
            .getInternalEndpointsForInterface(new QName("test"));
        assertEquals(endpoints[0], internalEndpoint);
    }

    public void testGetInternalEndpointsForInterface1() throws JBIException,
        NamingException {
        testRegisterInternalEndpoint();
        AbstractEndpoint[] endpoints = registry
            .getInternalEndpointsForInterface(null);
        assertEquals(endpoints[0].containerName, "0");
    }

    public void testGetInternalEndpointsForService() throws JBIException,
        NamingException {
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "serv"), "endpoint", "compo", "0", null);
        rootContext.bind("test-link", internalEndpoint);
        MockContext context = (MockContext) servicesContext
            .createSubcontext("test");
        MockLinkRef linkRef = new MockLinkRef("test-link");
        linkRef.setClassName(InternalEndpoint.class.getName());
        context.bind("inter1", linkRef);
        AbstractEndpoint[] endpoints = registry
            .getInternalEndpointsForService(new QName("test"));
        assertEquals(endpoints[0], internalEndpoint);
    }

    public void testRegisterConnectionExceptionNull() {
        try {
            registry.registerConnection(null);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    public void testRegisterConnectionWithInterfaces() throws RegistryException {
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(new QName("inter"),
            new QName("toserv"), "endpoint", null);
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "toserv"), "endpoint", "compo", "0", null);
        endpointsContext.getBindings().put("endpoint", internalEndpoint);
        registry.registerConnection(linkedEndpoint);
        assertEquals(interfacesContext.getBindings().size(), 1);
    }

    public void testRegisterConnectionWithoutInterface()
        throws RegistryException {
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(
            new QName("service"), "endpoint", new QName("toservice"),
            "toendpoint", null);
        registry.registerInternalEndpoint(new InternalEndpoint(new QName(
            "toservice"), "toendpoint", "compo", "0", null));
        registry.registerConnection(linkedEndpoint);
        assertNotNull(endpointsContext.getBindings().get("endpoint"));
    }

    @SuppressWarnings("unchecked")
    public void testRegisterExternalEndpoint() throws JBIException {
        ExternalEndpoint externalEndpoint = new ExternalEndpoint(
            new ConsumerEndpoint("compo", "0"), "compo", "0");
        externalEndpoint.setServiceName(new QName("test"));
        registry.registerExternalEndpoint(externalEndpoint);
        // Check if the endpointcontext
        AbstractEndpoint endpoint = (AbstractEndpoint) endpointsContext
            .getBindings().values().iterator().next();
        assertEquals(externalEndpoint, endpoint);
        // Check if the newendpointcontext
        LinkRef linkRef = (LinkRef) newEndpointContext.getBindings().values()
            .iterator().next();
        assertNotNull(linkRef);
    }

    /**
     * test registration failure of an endpoint already bound
     * (by another component or container)
     * @throws JBIException
     */
    public void testRegisterExternalEndpointException() {
        ExternalEndpoint externalEndpoint = new ExternalEndpoint(
            new InternalEndpoint(new QName("test"), "endpoint", "compo", "0",
                null), "compo", "0");
        endpointsContext.getBindings().put("endpoint", externalEndpoint);
        
        ExternalEndpoint externalEndpoint2 = new ExternalEndpoint(
            new InternalEndpoint(new QName("test"), "endpoint", "compo", "0",
                null), "compo", "1");
        
        try {
            registry.registerExternalEndpoint(externalEndpoint2);
            fail();
        } catch (RegistryException e) {
        }
    }

    public void testRegisterExternalEndpointExceptionNull() {
        try {
            registry.registerExternalEndpoint(null);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    /**
     * Check if the endpoint is registered in all the contexts
     * 
     * @throws JBIException
     * @throws NamingException
     */
    @SuppressWarnings("unchecked")
    public void testRegisterInternalEndpoint() throws JBIException,
        NamingException {
        AbstractEndpoint endpoint = new InternalEndpoint(new QName("test"),
            "endpoint", "axis2bc", "0", null);
        registry.registerInternalEndpoint(endpoint);
        // Check if the endpointcontext
        AbstractEndpoint ep = (AbstractEndpoint) endpointsContext.getBindings()
            .values().iterator().next();
        assertEquals(endpoint, ep);
        // Check if the newendpointcontext
        LinkRef linkRef = (LinkRef) newEndpointContext.getBindings().values()
            .iterator().next();
        assertNotNull(linkRef);
    }

    /**
     * test registration failure of an endpoint already bound
     * (by another component or container)
     * @throws JBIException
     */
    public void testRegisterInternalEndpointAlreadyRegistered() {
        InternalEndpoint internalEndpoint = new InternalEndpoint(new QName(
            "serv"), "endpoint", "compo", "0", null);
        endpointsContext.getBindings().put("endpoint", internalEndpoint);

        InternalEndpoint internalEndpoint2 = new InternalEndpoint(new QName(
        "serv"), "endpoint", "compo", "1", null);

        try {
            registry.registerInternalEndpoint(internalEndpoint2);
            fail();
        } catch (Exception e) {

        }
    }

    public void testRegisterInternalEndpointExceptionNull() {
        try {
            registry.registerInternalEndpoint(null);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    @SuppressWarnings("unchecked")
    public void testRegisterInternalEndpointWithLinkedEndpoint()
        throws JBIException, NamingException {
        AbstractEndpoint endpoint = new LinkedEndpoint(new QName("service1"),
            "endpoint1", new QName("service2"), "endpoint2", null);
        registry.registerInternalEndpoint(endpoint);
        // Check if the endpointcontext
        AbstractEndpoint ep = (AbstractEndpoint) endpointsContext.getBindings()
            .values().iterator().next();
        assertEquals(endpoint, ep);
        // Check if the newendpointcontext
        LinkRef linkRef = (LinkRef) newEndpointContext.getBindings().values()
            .iterator().next();
        assertNotNull(linkRef);
    }

    public void testRetreiveNewEndpoints() throws JBIException, NamingException {
        AbstractEndpoint endpoint = new InternalEndpoint(new QName("test"),
            "endpoint", "axis2bc", "0", null);
        testRegisterInternalEndpoint();
        rootContext.bind("/endpoints/endpoint", endpoint);
        List<AbstractEndpoint> endpoints = registry.retrieveNewEndpoints();
        assertEquals(((AbstractEndpoint) endpoints.get(0)).containerName, "0");
    }

    public void testUnRegisterConnectionWithoutInterface()
        throws RegistryException {
        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(
            new QName("service"), "endpoint", new QName("toservice"),
            "toendpoint", null);
        registry.registerInternalEndpoint(new InternalEndpoint(new QName(
            "toservice"), "toendpoint", "compo", "0", null));
        registry.registerConnection(linkedEndpoint);
        assertNotNull(endpointsContext.getBindings().get("endpoint"));
        registry.deregisterConnection(linkedEndpoint);
        assertNull(endpointsContext.getBindings().get("endpoint"));
    }

    public void testValidateEndpoint() throws JBIException {
        InternalEndpoint endpoint = new InternalEndpoint(new QName("service"),
            "endpoint", "compo", "0", null);
        endpoint.setInterfaces(new QName[] {new QName("inter")});
        endpoint.setDescription("<xml><test />");
        registry.validateEndpoint(endpoint);
        assertEquals(interfacesContext.getBindings().size(), 1);
    }

    public void testValidateEndpointExceptionNull() {
        try {
            registry.validateEndpoint(null);
            fail();
        } catch (Exception e) {
            // Do nothing
        }
    }

    public void testValidateEndpointWithInterfaces() throws JBIException {
        InternalEndpoint endpoint = new InternalEndpoint(new QName("service"),
            "endpoint", "compo", "0", null);
        endpoint.setInterfaces(new QName[] {new QName("inter")});
        endpoint.setDescription("<xml><test />");
        interfacesContext.getBindings().put("inter", new MockContext());
        registry.validateEndpoint(endpoint);
        assertEquals(interfacesContext.getBindings().size(), 1);
    }

    public void testRetrieveNewEndpointsException() throws NamingException {
        LoggingUtil loggingUtil = EasyMock.createNiceMock(LoggingUtil.class);
        InitialContext rootContext = EasyMock.createMock(InitialContext.class);
        NamingException namingException = new NamingException();

        loggingUtil.start();
        EasyMock.expect(rootContext.listBindings("new-endpoints")).andThrow(
            namingException);

        EasyMock.replay(loggingUtil);
        EasyMock.replay(rootContext);

        registry.log = loggingUtil;
        registry.rootContext = rootContext;

        try {
            registry.retrieveNewEndpoints();
            fail();
        } catch (RegistryException e) {
            // do nothing
        }
    }

    public void testStart() throws SecurityException, NoSuchMethodException,
        IllegalLifeCycleException {
        JNDIRegistry registry = EasyMock.createMock(JNDIRegistry.class,
            new Method[] {JNDIRegistry.class.getDeclaredMethod("initContexts",
                new Class[0])});
        Logger logger = EasyMock.createMock(Logger.class);

        registry.logger = logger;

        registry.start();

        assertNotNull(registry.log);
    }

    public void testDeregisterConnectionWithInterfacesException()
        throws NamingException {
        LoggingUtil loggingUtil = EasyMock.createNiceMock(LoggingUtil.class);
        InitialContext rootContext = EasyMock.createMock(InitialContext.class);
        NamingException namingException = new NamingException();
        Context interContext = EasyMock.createMock(Context.class);

        EasyMock.expect(rootContext.list("interfaces")).andReturn(
            new NamingEnumerationImpl<NameClassPair>(
                new NameClassPair[] {new NameClassPair("inter", "test")}));
        EasyMock.expect(rootContext.list("new-endpoints")).andReturn(
            new NamingEnumerationImpl<NameClassPair>(
                new NameClassPair[] {new NameClassPair("toep", "ep")}));
        EasyMock.expect(
            interContext.lookup(NamingHelper.qNameToName(new QName("inter"))))
            .andThrow(namingException);

        EasyMock.replay(loggingUtil);
        EasyMock.replay(rootContext);
        EasyMock.replay(interContext);

        registry.log = loggingUtil;
        registry.rootContext = rootContext;
        registry.interfacesContext = interContext;

        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(new QName("inter"),
            new QName("toserv"), "toep", null);

        try {
            registry.deregisterConnection(linkedEndpoint);
            fail();
        } catch (RegistryException e) {
            // do nothing
        }
    }

    public void testDeregisterConnectionWithServiceEndpointNotBound()
        throws NamingException {
        LoggingUtil loggingUtil = EasyMock.createMock(LoggingUtil.class);
        InitialContext rootContext = EasyMock.createMock(InitialContext.class);

        loggingUtil.start();
        EasyMock.expect(rootContext.list("endpoints")).andReturn(
            new NamingEnumerationImpl<NameClassPair>(new NameClassPair[0]));
        loggingUtil
            .error("The provider endpoint ep is not registered in the endpoints");

        EasyMock.replay(loggingUtil);
        EasyMock.replay(rootContext);

        registry.log = loggingUtil;
        registry.rootContext = rootContext;

        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(new QName("serv"),
            "ep", new QName("toserv"), "toep", null);

        try {
            registry.deregisterConnection(linkedEndpoint);
            fail();
        } catch (RegistryException e) {
            // do nothing
        }
    }

    public void testDeregisterConnectionWithServiceException()
        throws NamingException, SecurityException, NoSuchMethodException {
        JNDIRegistry registry = EasyMock.createMock(JNDIRegistry.class,
            new Method[] {JNDIRegistry.class.getDeclaredMethod(
                "retrieveServiceContext", new Class[] {Name.class})});
        LoggingUtil loggingUtil = EasyMock.createNiceMock(LoggingUtil.class);
        InitialContext rootContext = EasyMock.createMock(InitialContext.class);
        NamingException namingException = new NamingException();

        EasyMock.expect(rootContext.list("endpoints")).andReturn(
            new NamingEnumerationImpl<NameClassPair>(
                new NameClassPair[] {new NameClassPair("ep", "foo")}));
        EasyMock.expect(
            registry.retrieveServiceContext(NamingHelper.qNameToName(new QName(
                "serv")))).andThrow(namingException);

        EasyMock.replay(registry);
        EasyMock.replay(loggingUtil);
        EasyMock.replay(rootContext);

        registry.log = loggingUtil;
        registry.rootContext = rootContext;

        LinkedEndpoint linkedEndpoint = new LinkedEndpoint(new QName("serv"),
            "ep", new QName("toserv"), "toep", null);

        try {
            registry.deregisterConnection(linkedEndpoint);
            fail();
        } catch (RegistryException e) {
            // do nothing
        }
    }
}
