/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JNDIServerTest.java 9:42:06 AM ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.kernel.registry;

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.util.HashMap;
import java.util.Map;

import javax.naming.Binding;
import javax.naming.LinkRef;
import javax.naming.NameClassPair;

import junit.framework.TestCase;

import org.easymock.classextension.EasyMock;
import org.objectweb.petals.kernel.registry.mock.MockConfigurationService;
import org.objectweb.petals.kernel.registry.msg.request.BindRequest;
import org.objectweb.petals.kernel.registry.msg.request.CreateSubcontextRequest;
import org.objectweb.petals.kernel.registry.msg.request.DestroySubcontextRequest;
import org.objectweb.petals.kernel.registry.msg.request.ListBindingsRequest;
import org.objectweb.petals.kernel.registry.msg.request.ListRequest;
import org.objectweb.petals.kernel.registry.msg.request.LookupLinkRequest;
import org.objectweb.petals.kernel.registry.msg.request.LookupRequest;
import org.objectweb.petals.kernel.registry.msg.request.RebindRequest;
import org.objectweb.petals.kernel.registry.msg.request.RegistryRequest;
import org.objectweb.petals.kernel.registry.msg.request.RenameRequest;
import org.objectweb.petals.kernel.registry.msg.request.UnbindRequest;
import org.objectweb.petals.kernel.registry.msg.response.RegistryResponse;
import org.objectweb.petals.kernel.registry.msg.response.RegistryResponse.ResponseType;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Test of the JNDIServer
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public class ServerAloneTest extends TestCase {

    private RegistryServer server;

    public void setUp() throws InterruptedException {
        SystemUtil.setJndiPort("16400");
        String groupName = "petals"
            + System.nanoTime();
        server = new RegistryServer();
        Logger logger = EasyMock.createMock(Logger.class);
        server.logger = logger;
        server.configurationService = new MockConfigurationService();
        server.log = EasyMock.createMock(LoggingUtil.class);
        server.registryUtil.setLog(EasyMock.createMock(LoggingUtil.class));
        server.setGroupName(groupName);

        server.startServer();
        // Wait for the server to start
        Thread.sleep(1000);
    }

    public void tearDown() {
        server.stopServer();
    }

    public void testBind() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        RegistryRequest request = new BindRequest("/", "key", "value", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.ack)) {
            fail();
        } else {
            assertEquals(server.data.get("/").get("key"), "value");
        }
    }

    @SuppressWarnings("unchecked")
    public void testBindInSubcontext() throws Exception {

        testCreateSubcontext();

        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        RegistryRequest request = new BindRequest("/sub/", "key", "value", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.ack)) {
            fail();
        } else {
            assertEquals(server.data.get("/sub/").get("key"), "value");
        }
    }

    public void testCreateSubcontext() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        RegistryRequest request = new CreateSubcontextRequest("/", "sub", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.createSubcontext)) {
            fail();
        } else {
            assertEquals(response.getArg1(), "/sub/");
            assertNotNull(server.data.get("/").get("sub"));
            assertNotNull(server.data.get("/sub/"));
        }
    }

    public void testDestroySubcontext() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        Map<String, Object> map = new HashMap<String, Object>();
        map.put("sub", new HashMap<String, Object>());

        server.data.put("/", map);

        Map<String, Object> subMap = new HashMap<String, Object>();
        subMap.put("key", "value");
        server.data.put("/sub/", subMap);

        RegistryRequest request = new DestroySubcontextRequest("/", "sub", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.ack)) {
            fail();
        } else {
            assertEquals(server.data.get("/").size(), 0);
            assertNull(server.getData().get("/sub/"));
        }
    }

    public void testList() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        Map<String, Object> sub = new HashMap<String, Object>();
        sub.put("key", "value");
        server.data.put("/sub/", sub);

        RegistryRequest request = new ListRequest("/", "sub", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.list)) {
            fail();
        } else {
            assertEquals("key", ((NameClassPair) ((NameClassPair[]) response
                .getArg1())[0]).getName());
        }
    }

    public void testListBindings() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        Map<String, Object> sub = new HashMap<String, Object>();
        sub.put("key", "value");
        server.data.put("/sub/", sub);

        RegistryRequest request = new ListBindingsRequest("/", "sub", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.listBindings)) {
            fail();
        } else {
            assertEquals("key", ((Binding) ((Binding[]) response.getArg1())[0])
                .getName());
            assertEquals("value",
                ((Binding) ((Binding[]) response.getArg1())[0]).getObject());
        }
    }

    public void testLookup() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);
        Map<String, Object> map = new HashMap<String, Object>();
        map.put("key", "value");
        server.data.put("/", map);

        RegistryRequest request = new LookupRequest("/", "key", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.lookup)) {
            fail();
        } else {
            assertEquals(response.getArg1(), "value");
        }
    }

    public void testLookupLink() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        Map<String, Object> map = new HashMap<String, Object>();
        map.put("key", "value");
        server.data.put("/", map);

        RegistryRequest request = new LookupLinkRequest("/", "key", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.lookupLink)) {
            fail();
        } else {
            assertEquals(response.getArg1(), "value");
        }
    }

    public void testLookupLinkWithLink() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        Map<String, Object> subMap = new HashMap<String, Object>();
        LinkRef linkRef = new LinkRef("/key");
        subMap.put("link", linkRef);

        server.data.put("/sub/", subMap);
        Map<String, Object> map = new HashMap<String, Object>();
        map.put("key", "value");
        server.data.put("/", map);

        RegistryRequest request = new LookupLinkRequest("/sub/", "link", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.lookupLink)) {
            fail();
        } else {
            assertEquals(response.getArg1(), "value");
        }
    }

    public void testRebind() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        Map<String, Object> map = new HashMap<String, Object>();
        map.put("key", "value");
        server.data.put("/", map);

        RegistryRequest request = new RebindRequest("/", "key", "value1", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.ack)) {
            fail();
        } else {
            assertEquals(server.data.get("/").get("key"), "value1");
        }
    }

    public void testRename() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);
        Map<String, Object> map = new HashMap<String, Object>();
        map.put("key", "value");
        server.data.put("/", map);

        RegistryRequest request = new RenameRequest("/", "key", "key1", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.ack)) {
            fail();
        } else {
            assertEquals(server.data.get("/").get("key1"), "value");
        }
    }

    public void testUnbind() throws Exception {
        Socket socket = new Socket();
        socket
            .connect(new InetSocketAddress(SystemUtil.getHost(), 16400), 1000);

        Map<String, Object> map = new HashMap<String, Object>();
        map.put("key", "value");
        server.data.put("/", map);

        RegistryRequest request = new UnbindRequest("/", "key", 0, 0);

        ObjectOutputStream objectOutputStream = new ObjectOutputStream(socket
            .getOutputStream());
        objectOutputStream.writeObject(request);
        objectOutputStream.flush();

        ObjectInputStream objectInputStream = new ObjectInputStream(socket
            .getInputStream());
        RegistryResponse response = (RegistryResponse) objectInputStream
            .readObject();

        if (!response.getType().equals(ResponseType.ack)) {
            fail();
        } else {
            assertEquals(server.data.get("/").size(), 0);
        }
    }

}
