/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: RepositoryImpl.java 1428 2006-12-22 09:31:10Z ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.repository;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.io.filefilter.PrefixFileFilter;
import org.objectweb.fractal.fraclet.annotation.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.Interface;
import org.objectweb.fractal.fraclet.annotation.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.LifeCycleType;
import org.objectweb.fractal.fraclet.annotation.Monolog;
import org.objectweb.fractal.fraclet.annotation.Provides;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.petals.util.ZipUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Platform Component Repository service implementation.
 * 
 * @version $Rev: 1428 $ $Date: 2006-12-22 10:31:10 +0100 (ven, 22 déc 2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 * @author alouis - EBM Websourcing
 * @author ofabre - EBM Websourcing
 * @author ddesjardins - eBMWebsourcing
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = org.objectweb.petals.repository.RepositoryService.class))
public class RepositoryImpl implements RepositoryService {

	/**
	 * Path to component install directory
	 */
	private static final String INSTALL_PATH = "install";

	/**
	 * Path to repository location within Petals installation directory.
	 */
	private static final String REPOSITORY_PATH = "repository";

	/**
	 * Path to components directory into petals repository
	 */
	private static final String COMPONENTS_PATH = "components";

	/**
	 * Path to components directory into petals repository
	 */
	private static final String SHAREDLIBS_PATH = "shared-libraries";

	/**
	 * Path to lost+found directory into petals repository
	 */
	private static final String LOSTPLUSFOUND_PATH = "lost+found";

	/**
	 * Path to components directory into petals repository
	 */
	private static final String SERVICEASSEMBLIES_PATH = "service-assemblies";

	/**
	 * The complete repository directory path.
	 */
	private static String repositoryDirectory;

	/**
	 * Path to component working directory
	 */
	private static final String WORK_PATH = "work";

	/**
	 * Entity types (components, shared libs or service assemblies)
	 */
	public static enum EntityType {
		COMPONENT_TYPE, SL_TYPE, SA_TYPE
	};

	/**
	 * logger wrapper
	 */
	protected LoggingUtil log;

	/**
	 * Monolog Logger instance.
	 */
	@Monolog(name = "logger")
	protected Logger logger;

	/**
	 * DateFormat used for repository folders
	 */
	private SimpleDateFormat folderDateFormat;

	/**
	 * Date Pattern length
	 */
	private int datePatternLength;

	/**
	 * @see org.objectweb.petals.repository.RepositoryService#addPackage(String,
	 *      java.net.URI)
	 */
	protected URI addPackage(String entityId, URI zipArchiveLocation,
			EntityType entityType) throws PetalsException {

		File rootFile = null;

		// don't care about replacing existing data
		switch (entityType) {
		case COMPONENT_TYPE:
			rootFile = createNewComponentRoot(entityId);
			break;
		case SA_TYPE:
			rootFile = createNewServiceAssemblyRoot(entityId);
			break;
		case SL_TYPE:
			rootFile = createNewSharedLibRoot(entityId);
			break;
		}
		File installationRoot = new File(rootFile, INSTALL_PATH);
		File workRoot = new File(rootFile, WORK_PATH);

		// recreate dirs necessary for adding the package into the repository :
		// install
		if (!installationRoot.mkdirs()) {
			throw new PetalsException(
					"Failed to create installation root dir: "
							+ installationRoot.getAbsolutePath());
		}
		// recreate dirs necessary for adding the package into the repository :
		// work
		if (!workRoot.mkdirs()) {
			throw new PetalsException("Failed to create work dir: "
					+ workRoot.getAbsolutePath());
		}

		// explode package Zip archive file into repository
		ZipFile zipArchive = null;

		try {
			File archiveFile = new File(zipArchiveLocation);
			zipArchive = new ZipFile(archiveFile);
		} catch (ZipException ze) {
			throw new PetalsException("Unable to open Zip archive file: "
					+ zipArchiveLocation.toString(), ze);
		} catch (IOException ioe) {
			throw new PetalsException("Unexpected exception.", ioe);
		}

		ZipUtil.explodeIntoDirectory(zipArchive, installationRoot);
		return installationRoot.toURI();

	}

	public URI addComponentPackage(String componentID, URI zipArchiveLocation)
			throws PetalsException {
		return addPackage(componentID, zipArchiveLocation,
				EntityType.COMPONENT_TYPE);
	}

	public URI addSharedLibPackage(String sharedLibID, URI zipArchiveLocation)
			throws PetalsException {
		return addPackage(sharedLibID, zipArchiveLocation, EntityType.SL_TYPE);
	}

	public URI addServiceAssemblyPackage(String serviceAssemblyID,
			URI zipArchiveLocation) throws PetalsException {
		return addPackage(serviceAssemblyID, zipArchiveLocation,
				EntityType.SA_TYPE);
	}

	/**
	 * @see org.objectweb.petals.repository.RepositoryService#explodeSUIntoSAInstallFolder(String,
	 *      URI, String)
	 */
	public URI explodeSUIntoSAInstallFolder(String suId, URI suZipLocation,
			String saId) throws PetalsException {
		File installationRoot = new File(getServiceAssemblyInstallRoot(saId),
				suId);

		if (!installationRoot.exists() && !installationRoot.mkdirs()) {
			throw new PetalsException(
					"Failed to create installation root dir: "
							+ installationRoot.getAbsolutePath());
		}
		// explode package Zip archive file into repository
		ZipFile zipArchive = null;
		try {
			File archiveFile = new File(suZipLocation);
			zipArchive = new ZipFile(archiveFile);
		} catch (ZipException ze) {
			throw new PetalsException("Unable to open Zip archive file: "
					+ suZipLocation.toString(), ze);
		} catch (IOException ioe) {
			throw new PetalsException("Unexpected exception.", ioe);
		}
		ZipUtil.explodeIntoDirectory(zipArchive, installationRoot);
		return installationRoot.toURI();
	}

	public String getComponentsDirectory() {
		return getRepositoryDirectory() + File.separator + COMPONENTS_PATH;
	}

	public String getServiceAssembliesDirectory() {
		return getRepositoryDirectory() + File.separator
				+ SERVICEASSEMBLIES_PATH;
	}

	public String getSharedLibsDirectory() {
		return getRepositoryDirectory() + File.separator + SHAREDLIBS_PATH;
	}

	public String getLostPlusFoundDirectory() {
		return SystemUtil.getPetalsInstallDirectory().getAbsolutePath()
				+ File.separator + LOSTPLUSFOUND_PATH;
	}

	public File getComponentInstallRoot(String componentId)
			throws PetalsException {
		String path = getComponentRoot(componentId) + File.separator
				+ INSTALL_PATH;
		return new File(path);
	}

	public File getComponentRoot(String componentId) throws PetalsException {
		File result = null;
		File componentsDirectory = new File(getComponentsDirectory());
		String[] componentsFilePaths = componentsDirectory
				.list(new PrefixFileFilter(componentId));

		String[] componentsFilePathsWithGoodLength = filterPathsOnlength(
				componentId, componentsFilePaths);

		String selectedPath = null;
		try {
			selectedPath = searchForMostRecentFilePath(
					componentsFilePathsWithGoodLength, componentId);
		} catch (ParseException e) {
			throw new PetalsException(
					"Can't retrieve component root folder for the following component: "
							+ componentId, e);
		}
		if (selectedPath != null) {
			result = new File(componentsDirectory, selectedPath);
		}
		return result;
	}

	private String[] filterPathsOnlength(String componentId,
			String[] componentsFilePaths) {
		List<String> goodFilePaths = new ArrayList<String>();

		for (String path : componentsFilePaths) {
			if (path.length() == componentId.length() + datePatternLength + 1) {
				goodFilePaths.add(path);
			}
		}

		return goodFilePaths.toArray(new String[0]);
	}

	/**
	 * Retrieve the most recent path in an array of folder path, according to
	 * the included date information.
	 * 
	 * @param filePaths
	 *            the file paths to compare
	 * @param entityId
	 *            an entity (component, shared lib or service assembly) id
	 * @return the most recent file path, null if given filePaths array is null
	 * @throws ParseException
	 *             if an error occurs during Date extraction
	 */
	protected String searchForMostRecentFilePath(String[] filePaths,
			String entityId) throws ParseException {
		String mostRecentPath = null;
		Date mostRecentDate = null;
		if (filePaths != null) {
			for (String path : filePaths) {
				String baseName = FilenameUtils.getBaseName(path);
				Date currentDate = extractDateFromName(baseName, entityId);
				if (mostRecentDate != null) {
					if (currentDate.after(mostRecentDate)) {
						mostRecentDate = currentDate;
						mostRecentPath = path;
					}
				} else {
					mostRecentDate = currentDate;
					mostRecentPath = path;
				}
			}
		}
		return mostRecentPath;
	}

	/**
	 * Extract a Date from a file name.
	 * 
	 * @param baseName
	 *            a file name
	 * @param entityId
	 *            an entity (component, shared lib or service assembly) id
	 * @return the file Date, non null
	 * @throws ParseException
	 *             if an error occurs during Date extraction
	 */
	protected Date extractDateFromName(String baseName, String entityId)
			throws ParseException {
		String dateString = baseName.substring(baseName.indexOf(entityId)
				+ entityId.length() + 1);

		return folderDateFormat.parse(dateString);
	}

	public File createNewComponentRoot(String componentId) {
		String path = getComponentsDirectory() + File.separator + componentId
				+ "-"
				+ folderDateFormat.format(new Date(System.currentTimeMillis()));
		return new File(path);
	}

	public File getComponentWorkRoot(String componentId) throws PetalsException {
		String path = getComponentRoot(componentId) + File.separator
				+ WORK_PATH;
		return new File(path);
	}

	public File getSharedLibRoot(String slId) throws PetalsException {
		File result = null;
		File slsDirectory = new File(getSharedLibsDirectory());
		String[] slFilePaths = slsDirectory.list(new PrefixFileFilter(slId));

		String[] slFilePathsWithGoodLength = filterPathsOnlength(slId,
				slFilePaths);

		String selectedPath = null;
		try {
			selectedPath = searchForMostRecentFilePath(
					slFilePathsWithGoodLength, slId);
		} catch (ParseException e) {
			throw new PetalsException(
					"Can't retrieve shared lib root folder for the following shared lib: "
							+ slId, e);
		}
		if (selectedPath != null) {
			result = new File(slsDirectory, selectedPath);
		}
		return result;
	}

	public File createNewSharedLibRoot(String slId) {
		String path = getSharedLibsDirectory() + File.separator + slId + "-"
				+ folderDateFormat.format(new Date(System.currentTimeMillis()));
		return new File(path);
	}

	public File getSharedLibWorkRoot(String slId) throws PetalsException {
		String path = getSharedLibRoot(slId) + File.separator + WORK_PATH;
		return new File(path);
	}

	public File getSharedLibInstallRoot(String slId) throws PetalsException {
		String path = getSharedLibRoot(slId) + File.separator + INSTALL_PATH;
		return new File(path);
	}

	public File getServiceAssemblyRoot(String saId) throws PetalsException {
		File result = null;
		File sasDirectory = new File(getServiceAssembliesDirectory());
		String[] saFilePaths = sasDirectory.list(new PrefixFileFilter(saId));

		String[] saFilePathsWithGoodLength = filterPathsOnlength(saId,
				saFilePaths);

		String selectedPath = null;
		try {
			selectedPath = searchForMostRecentFilePath(
					saFilePathsWithGoodLength, saId);
		} catch (ParseException e) {
			throw new PetalsException(
					"Can't retrieve service assembly root folder for the following service assembly: "
							+ saId, e);
		}
		if (selectedPath != null) {
			result = new File(sasDirectory, selectedPath);
		}
		return result;
	}

	public File createNewServiceAssemblyRoot(String saId) {
		String path = getServiceAssembliesDirectory() + File.separator + saId
				+ "-"
				+ folderDateFormat.format(new Date(System.currentTimeMillis()));
		return new File(path);
	}

	public File getServiceAssemblyWorkRoot(String saId) throws PetalsException {
		String path = getServiceAssemblyRoot(saId) + File.separator + WORK_PATH;
		return new File(path);
	}

	public File getServiceAssemblyInstallRoot(String saId)
			throws PetalsException {
		String path = getServiceAssemblyRoot(saId) + File.separator
				+ INSTALL_PATH;
		return new File(path);
	}

	/**
	 * @throws PetalsException
	 * @see org.objectweb.petals.repository.RepositoryService#removePackage(java.lang.String)
	 */
	protected boolean removePackage(String entityID, EntityType entityType)
			throws PetalsException {
		boolean removed = true;

		File entityRoot = null;

		switch (entityType) {
		case COMPONENT_TYPE:
			entityRoot = getComponentRoot(entityID);
			break;
		case SA_TYPE:
			entityRoot = getServiceAssemblyRoot(entityID);
			break;
		case SL_TYPE:
			entityRoot = getSharedLibRoot(entityID);
			break;
		}

		try {
			FileUtils.deleteDirectory(entityRoot);
		} catch (IOException e) {
			log
					.warning("Failed to delete old Petals entity folder. "
							+ "This folder will be deleted on a future Petals restarting.");
		}

		return removed;
	}

	public boolean removeComponentPackage(String componentId)
			throws PetalsException {
		return removePackage(componentId, EntityType.COMPONENT_TYPE);
	}

	public boolean removeSharedLibPackage(String slId) throws PetalsException {
		return removePackage(slId, EntityType.SL_TYPE);
	}

	public boolean removeServiceAssemblyPackage(String saId)
			throws PetalsException {
		return removePackage(saId, EntityType.SA_TYPE);
	}

	protected String getRepositoryDirectory() {
		if (repositoryDirectory == null) {
			repositoryDirectory = SystemUtil.getPetalsInstallDirectory()
					.getAbsolutePath()
					+ File.separator + REPOSITORY_PATH;
		}
		return repositoryDirectory;
	}

	@LifeCycle(on = LifeCycleType.START)
	protected void start() {
		log = new LoggingUtil(logger);
		log.start();
		String datePattern = "MM-dd-yyyy-hhmmss";
		folderDateFormat = new SimpleDateFormat(datePattern);
		datePatternLength = datePattern.length();
		log.end();
	}

	@LifeCycle(on = LifeCycleType.STOP)
	protected void stop() {
		log.start();

		logger = null;

		log.end();
	}

}