/*
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): OUAZZANI CHAHDI Anass
 * --------------------------------------------------------------------------
 * $Id: VolumeProbe.java,v 1.1 2006/11/17 14:25:42 mdutoo Exp $
 * --------------------------------------------------------------------------  
 */

package org.objectweb.petals.monitoring.router.volume;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.stream.StreamSource;

import org.objectweb.petals.jbi.messaging.MessageExchangeImpl;
import org.objectweb.petals.monitoring.router.RouterMonitor;

/**
 * @version $Rev: 617 $ $Date: 2006/11/17 14:25:42 $
 * @since Petals 1.0
 * @author alouis, oanass, dutoo
 * 
 * Calculate the messages size : Summ of all the normalizedMessages of the
 * Exchange
 */
public class VolumeProbe implements RouterMonitor, Serializable {

    private static final long serialVersionUID = -1466370911446522100L;

    private List<Marker> markers = new ArrayList<Marker>();

    public List<Marker> getMarkers() {
        return markers;
    }

    /**
     * Clear the probe
     * 
     */
    public synchronized void clear() {
        this.markers.clear();
    }

    /**
     * Calculate the size of message exchange and add the result to the list of
     * marker whose each element contains an id of a message and the size
     * corresponding to this last.
     * 
     * @param exchange
     */
    public synchronized void beforeMessageSent(MessageExchangeImpl _exchange) {

        long date = System.currentTimeMillis();

        int size = 0;
        try {
            ByteArrayOutputStream bo = new ByteArrayOutputStream();
            ObjectOutputStream o = new ObjectOutputStream(bo);
            
            // TODO more efficient way of estimating the size than serialization
            o.writeObject(_exchange.getConsumerEndpoint());
            o.writeObject(_exchange.getEndpoint());
            o.writeObject(_exchange.getError());
            o.writeObject(_exchange.getExchangeId());
            o.writeObject(_exchange.getInterfaceName());
            o.writeObject(_exchange.getOperation());
            o.writeObject(_exchange.getPattern());
            //o.writeObject(_exchange.getRole()); // NB. Role not serializable and not part of what is sent
            o.writeObject(_exchange.getService());
            //o.writeObject(_exchange.getStatus()); // NB. Role not serializable and not part of what is sent
            
            // content
            NormalizedMessage aboutToBeSentMsg = null;
            if (MessageExchange.Role.CONSUMER.equals(_exchange.getRole())) {
                aboutToBeSentMsg = _exchange.getMessage("IN");
            } else if (javax.jbi.messaging.MessageExchange.Role.PROVIDER.equals(_exchange.getRole())) {
                aboutToBeSentMsg = _exchange.getMessage("OUT");
                
            }
            if (aboutToBeSentMsg != null) {
                javax.xml.transform.Source source = aboutToBeSentMsg.getContent();
                if (source instanceof StreamSource) {
                    // using clean way for streams (avoids them to be left in a read state)
                    // (bug #306435)
                    InputStream in = ((StreamSource) source).getInputStream();
                    size += getSizeAndReset(in);
                }
                
                // attachments
                for (Iterator attNameIt = aboutToBeSentMsg.getAttachmentNames().iterator(); attNameIt.hasNext();) {
                    String attName = (String) attNameIt.next();
                    javax.activation.DataHandler attDh = aboutToBeSentMsg.getAttachment(attName);
                    java.io.InputStream in = attDh.getInputStream();
                    // using clean way for streams (avoids them to be left in a read state)
                    size += getSizeAndReset(in);
                }
            }
            
            // properties
            for (java.util.Iterator propNameIt = _exchange.getPropertyNames().iterator(); propNameIt.hasNext();) {
                String propName = (String) propNameIt.next();
                Object prop = _exchange.getProperty(propName);
                if (prop instanceof Serializable) {
                   o.writeObject(prop);
                }
                // TODO if not serializable ??
            }
            
            o.flush();
            //o.close();
            size = bo.size();
            this.markers.add(new Marker(_exchange.getExchangeId(), size, date));
        } catch (Exception e) {
            e.printStackTrace();
            //TODO log
        }
    }
    
    /**
     * Helps in cleanly estimating the size of the serialized InputStream.
     * The given InputStream will be readable again afterwards, thanks to
     * InputStream.reset() .
     * @param in
     * @return
     * @throws IOException 
     */
    private int getSizeAndReset(InputStream in) throws IOException {
        // NB. an alternative to this method would be to use SourceHelper.createSource()
        in.reset(); // resetting in case it was already read, but should not happen
        int size = 0;
        //in.mark(readLimit);
        byte[] byteBuf = new byte[1024];
        int byteReadNb;
        //java.io.ByteArrayOutputStream bos = new java.io.ByteArrayOutputStream();
        while ((byteReadNb = in.read(byteBuf)) != -1) {
            //bos.write(byteBuf, 0, byteReadNb);
            size += byteReadNb;
        }
        //String s = bos.toString();
        //o.writeObject(s);
        // IMPORTANT the following is required, else the it won't be readable again !!!
        // (bug #306435)
        in.reset();
        return size;
    }

    /**
     * Do nothing
     * 
     * @param exchange
     */
    public synchronized void afterMessageSent(MessageExchangeImpl exchange) {
        // Do nothing
    }

    /**
     * Do nothing
     */
    public synchronized void beforeMessageReceived(MessageExchangeImpl message) {
        // Do nothing

    }

    /**
     * Do nothing
     */
    public synchronized void afterMessageReceived(MessageExchangeImpl message) {
        // Do nothing

    }

    /**
     * This class contains information about a message (message id and size)
     * 
     * @author oanass
     * 
     */
    public class Marker implements Serializable {
        /**
         * 
         */
        private static final long serialVersionUID = 3130161625771314655L;

        String id;

        int size;

        long date;

        public String getId() {
            return id;
        }

        public int getSize() {
            return size;
        }

        public void setSize(int size) {
            this.size = size;
        }

        /**
         * 
         * @param id :
         *            id of message.
         * @param size :
         *            size of message.
         * 
         * The object Marker Contain the id of message that was sent or received
         * and her size.
         */
        public Marker(String id, int size, long date) {
            this.id = id;
            this.size = size;
            this.date = date;
        }
    }
}
