/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: JMXUtil.java 17:54:02 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.tools.webadmin.util;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;

import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;

import org.objectweb.petals.tools.ant.util.JBIJMXConnectorUtil;
import org.objectweb.petals.tools.webadmin.bean.JBIEntity;
import org.objectweb.petals.tools.webadmin.exception.WebAdminException;

/**
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public class JMXUtil {

    
    /**
     * Host for the JMX connection
     */
    private String host = "localhost";

    /**
     * Password the JMX connection
     */
    private String password = "";

    /**
     * Port for the JMX connection
     */
    private String port = "8081";

    /**
     * Username for the JMX connection
     */
    private String username = "";

    public JMXUtil(InputStream propStream) {
        init(propStream);
    }

    /**
     * Change the state of a component
     * 
     * @param component
     *            to change state
     * @param action
     *            change state
     * @throws WebAdminException
     */
    public void changeCompoState(JBIEntity component, String action)
        throws WebAdminException {
        JMXConnector connector = null;
        try {
            connector = retrieveConnector();
            MBeanServerConnection connection = connector
                .getMBeanServerConnection();
            Object[] objects = new Object[0];
            String[] strings = new String[0];
            connection.invoke(component.getObjectName(), action, objects,
                strings);
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
    }

    /**
     * Change the state of a service assembly
     * 
     * @param sa
     *            service assembly to change state
     * @param action
     *            change state
     * @throws WebAdminException
     */
    public void changeServiceAssemblyState(JBIEntity sa, String action)
        throws WebAdminException {
        JMXConnector connector = null;
        try {
            connector = retrieveConnector();
            MBeanServerConnection connection;
            connection = connector.getMBeanServerConnection();
            Object[] objects = new Object[1];
            objects[0] = sa.getName();
            String[] strings = new String[1];
            strings[0] = "java.lang.String";
            connection.invoke(JBIJMXConnectorUtil
                .getDeploymentServiceMBeanName(connection), action, objects,
                strings);
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
    }

    /**
     * Install a service assembly
     * 
     * @param file
     *            archive file
     * @throws WebAdminException
     */
    public void deployServiceAssembly(String file) throws WebAdminException {
        JMXConnector connector = null;
        try {
            String url = "file:"
                + file;
            connector = retrieveConnector();
            MBeanServerConnection connection = connector
                .getMBeanServerConnection();
            Object[] objects = new Object[1];
            objects[0] = url;
            String[] strings = new String[1];
            strings[0] = "java.lang.String";
            connection.invoke(JBIJMXConnectorUtil
                .getDeploymentServiceMBeanName(connection), "deploy", objects,
                strings);
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
    }

    public String getHost() {
        return host;
    }

    public String getPassword() {
        return password;
    }

    public String getPort() {
        return port;
    }

    /**
     * Get the information of the service assembly
     * 
     * @param saName
     *            service assembly name
     * @throws WebAdminException
     */
    public Map<String, String> getServiceAssemblyInformation(String saName)
        throws WebAdminException {
        JMXConnector connector = null;
        Map<String, String> serviceUnits = new HashMap<String, String>();
        try {
            connector = retrieveConnector();
            MBeanServerConnection connection = connector
                .getMBeanServerConnection();
            Object[] objects = new Object[1];
            objects[0] = saName;
            String[] sig = new String[1];
            sig[0] = "java.lang.String";
            Set<String> composForSA = new TreeSet<String>(Arrays
                .asList((String[]) connection.invoke(JBIJMXConnectorUtil
                    .getDeploymentServiceMBeanName(connection),
                    "getComponentsForDeployedServiceAssembly", objects, sig)));
            for (String compoName : composForSA) {
                objects = new Object[1];
                objects[0] = compoName;
                sig = new String[1];
                sig[0] = "java.lang.String";
                Set<String> units = new TreeSet<String>(Arrays
                    .asList((String[]) connection.invoke(JBIJMXConnectorUtil
                        .getDeploymentServiceMBeanName(connection),
                        "getDeployedServiceUnitList", objects, sig)));
                for (String string : units) {
                    serviceUnits.put(string, compoName);
                }
            }
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
        return serviceUnits;
    }

    public String getUsername() {
        return username;
    }

    /**
     * Install a component
     * 
     * @param file
     *            archive file
     * @throws WebAdminException
     */
    public void installComponent(String file) throws WebAdminException {
        JMXConnector connector = null;
        try {
            String url = "file:"
                + file;
            connector = retrieveConnector();
            MBeanServerConnection connection;
            connection = connector.getMBeanServerConnection();
            Object[] objects = new Object[1];
            objects[0] = url;
            String[] strings = new String[1];
            strings[0] = "java.lang.String";
            Object result = connection.invoke(JBIJMXConnectorUtil
                .getInstallationServiceMBeanName(connection),
                "loadNewInstaller", objects, strings);
            if (result instanceof ObjectName) {
                ObjectName installerName = (ObjectName) result;
                // Install the component
                objects = new Object[0];
                strings = new String[0];
                connection.invoke(installerName, "install", objects, strings);
            } else {
                throw new WebAdminException("file is not a component");
            }
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
    }

    /**
     * Install a shared library
     * 
     * @param file
     *            archive file
     * @throws WebAdminException
     */
    public void installSharedLibrary(String file) throws WebAdminException {
        JMXConnector connector = null;
        try {
            String url = "file:"
                + file;
            connector = retrieveConnector();
            MBeanServerConnection connection = connector
                .getMBeanServerConnection();
            Object[] objects = new Object[1];
            objects[0] = url;
            String[] strings = new String[1];
            strings[0] = "java.lang.String";
            Object result = connection.invoke(JBIJMXConnectorUtil
                .getInstallationServiceMBeanName(connection),
                "installSharedLibrary", objects, strings);
            if (result == null) {
                throw new WebAdminException("file is not a shared library");
            }
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
    }

    /**
     * List the deployed JBI components
     * 
     * @return map of components
     * @throws WebAdminException
     */
    public Map<String, JBIEntity> listComponents() throws WebAdminException {
        Map<String, JBIEntity> components = new HashMap<String, JBIEntity>();
        JMXConnector connector = null;
        try {
            connector = retrieveConnector();
            MBeanServerConnection connection = connector
                .getMBeanServerConnection();
            Object result = connection.getAttribute(JBIJMXConnectorUtil
                .getAdminServiceMBeanName(connection), "BindingComponents");
            if (result instanceof ObjectName[]) {
                ObjectName[] engines = (ObjectName[]) result;
                for (ObjectName objectName : engines) {
                    JBIEntity component = new JBIEntity();
                    String objState = (String) connection.getAttribute(
                        objectName, "CurrentState");
                    String name = objectName.getKeyProperty("name");
                    component.setName(name);
                    component.setState(objState);
                    component.setObjectName(objectName);
                    // TODO find a way to get the description of a component
                    component.setDescription("");
                    component.setType(JBIEntity.BINDING_COMPONENT);
                    components.put(component.getName(), component);
                }
            }
            result = connection.getAttribute(JBIJMXConnectorUtil
                .getAdminServiceMBeanName(connection), "EngineComponents");
            if (result instanceof ObjectName[]) {
                ObjectName[] engines = (ObjectName[]) result;
                for (ObjectName objectName : engines) {
                    JBIEntity component = new JBIEntity();
                    String objState = (String) connection.getAttribute(
                        objectName, "CurrentState");
                    String name = objectName.getKeyProperty("name");
                    component.setName(name);
                    component.setState(objState);
                    component.setObjectName(objectName);
                    // TODO find a way to get the description of a component
                    component.setDescription("");
                    component.setType(JBIEntity.SERVICE_ENGINE);
                    components.put(component.getName(), component);
                }
            }
            connector.close();
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
        return components;
    }

    /**
     * List the deployed JBI service assemblies
     * 
     * @return map of service assemblies
     * @throws WebAdminException
     */
    public Map<String, JBIEntity> listServiceAssemblies()
        throws WebAdminException {
        Map<String, JBIEntity> serviceAssemblies = new HashMap<String, JBIEntity>();
        JMXConnector connector = null;
        try {
            connector = retrieveConnector();
            MBeanServerConnection connection = connector
                .getMBeanServerConnection();
            Object result = connection.getAttribute(JBIJMXConnectorUtil
                .getDeploymentServiceMBeanName(connection),
                "DeployedServiceAssemblies");
            if (result instanceof String[]) {
                String[] assemblies = (String[]) result;
                for (String objectName : assemblies) {
                    Object[] objects = new Object[1];
                    objects[0] = objectName;
                    String[] sig = new String[1];
                    sig[0] = "java.lang.String";
                    String objState = (String) connection.invoke(
                        JBIJMXConnectorUtil
                            .getDeploymentServiceMBeanName(connection),
                        "getState", objects, sig);
                    JBIEntity sa = new JBIEntity();
                    sa.setName(objectName);
                    sa.setState(objState);
                    // TODO find a way to get the description of a component
                    sa.setDescription("");
                    sa.setType(JBIEntity.SERVICE_ASSEMBLY);
                    serviceAssemblies.put(sa.getName(), sa);
                }
            }
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
        return serviceAssemblies;
    }

    /**
     * Retrieve a JMX connector
     * 
     * @return a JMXConnector
     * @throws IOException
     */
    public JMXConnector retrieveConnector() throws IOException {
        return JBIJMXConnectorUtil
            .getConnection(host, port, username, password);
    }

    public void setHost(String host) {
        this.host = host;
    }

    public void setPassword(String password) {
        this.password = password;
    }

    public void setPort(String port) {
        this.port = port;
    }

    public void setUsername(String username) {
        this.username = username;
    }

    /**
     * Undeploy a service assembly
     * 
     * @param saName
     *            service assembly name
     * @throws WebAdminException
     */
    public void undeployServiceAssembly(String saName) throws WebAdminException {
        JMXConnector connector = null;
        try {
            connector = retrieveConnector();
            MBeanServerConnection connection = connector
                .getMBeanServerConnection();
            Object[] objects = new Object[1];
            objects[0] = saName;
            String[] strings = new String[1];
            strings[0] = "java.lang.String";
            String result = (String) connection.invoke(JBIJMXConnectorUtil
                .getDeploymentServiceMBeanName(connection), "undeploy",
                objects, strings);
            if (result.indexOf("FAILED") > -1) {
                throw new WebAdminException(result);
            }
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
    }

    /**
     * Undeploy a component
     * 
     * @param component
     *            to undeploy
     * @throws WebAdminException
     */
    public void uninstallCompo(JBIEntity component) throws WebAdminException {
        JMXConnector connector = null;
        try {
            connector = retrieveConnector();
            MBeanServerConnection connection = connector
                .getMBeanServerConnection();
            Object[] objects = new Object[2];
            objects[0] = component.getName();
            objects[1] = true;
            String[] strings = new String[2];
            strings[0] = "java.lang.String";
            strings[1] = "boolean";
            connection.invoke(JBIJMXConnectorUtil
                .getInstallationServiceMBeanName(connection),
                "unloadInstaller", objects, strings);
        } catch (Exception e) {
            try {
                if (connector != null) {
                    connector.close();
                }
            } catch (IOException e1) {
                // Do nothing
            }
            throw new WebAdminException(e);
        }
    }

    /**
     * Init the JMXUtil
     * 
     * @param propStream
     *            properties stream
     */
    protected void init(InputStream propStream) {
        if (propStream != null) {
            Properties properties = new Properties();
            try {
                properties.load(propStream);
                if (properties.get("host") != null) {
                    host = (String) properties.get("host");
                }
                if (properties.get("port") != null) {
                    port = (String) properties.get("port");
                }
                if (properties.get("username") != null) {
                    username = (String) properties.get("username");
                }
                if (properties.get("password") != null) {
                    password = (String) properties.get("password");
                }
            } catch (Exception e) {
                // Do nothing
            }
        }
    }

}
