/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): EBM WebSourcing
 */
package org.objectweb.petals.tools.ant;

import java.util.Arrays;
import java.util.Set;
import java.util.TreeSet;

import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;

import org.apache.tools.ant.BuildException;
import org.objectweb.petals.tools.ant.managers.JBIAntTaskAbstract;
import org.objectweb.petals.tools.ant.util.JBIJMXConnectorUtil;

/**
 * Task used to list the service engines
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public class ListServiceEnginesTask extends JBIAntTaskAbstract {

    /**
     * XML namespace
     */
    private static final String XMLNS = "http://java.sun.com/xml/ns/jbi/component-info-list";

    /**
     * List the Engines that have Service Units deployed to them as part of the
     * Service Assembly deployed in the JBI Environment. Without this parameter,
     * no Service Assembly dependency is verified.
     */
    private String serviceAssemblyName;

    /**
     * If supplied, only the named service engine is reported. If the engine
     * doesn’t exist, an empty component-list report is given.
     */
    private String serviceEngineName;

    /**
     * List the Engines that are dependent on the specified shared library
     * Without this parameter, no shared library dependency is verified.
     */
    // private String sharedLibraryName;
    /**
     * Service engines state Valid state are : shutdown, stated, stopped
     */
    private String state;

    /**
     * If supplied, set the given Ant property name value to the XML version of
     * this listing task.
     */
    private String xmlOutput;

    public ListServiceEnginesTask() {
        super();
    }

    /**
     * Execute the ant task
     * 
     * @throws IOException
     */
    public void execute() throws BuildException {
        try {
            if (state != null
                    && (!state.equalsIgnoreCase("shutdown")
                            || !state.equalsIgnoreCase("started") || !state
                            .equalsIgnoreCase("stopped"))) {
                throw new BuildException(
                        "Valid states are : 'shutdown', 'started', or 'stopped'");
            }
            JMXConnector connector = JBIJMXConnectorUtil.getConnection(host,
                    port, username, password);
            MBeanServerConnection connection = connector
                    .getMBeanServerConnection();
            // Output
            StringBuffer output = new StringBuffer();
            output
                    .append("-------------------------------------------------------\n");
            output.append("---------------- Service Engines --------------\n");
            output
                    .append("-------------------------------------------------------\n");
            StringBuffer outputXML = new StringBuffer();
            outputXML.append("<?xml version='1.0' encoding='utf-8'?>\n");
            outputXML.append("<component-info-list ");
            outputXML.append("xmlns='" + ListServiceEnginesTask.XMLNS
                    + "' version='1.0'>\n");
            // Get the engines component names
            Object result = connection.getAttribute(JBIJMXConnectorUtil
                    .getAdminServiceMBeanName(connection), "EngineComponents");
            Set<String> composForSA = new TreeSet<String>();
            if (serviceAssemblyName != null) {
                // Get getComponentsForDeployedServiceAssembly
                Object[] objects = new Object[1];
                objects[0] = serviceAssemblyName;
                String[] sig = new String[1];
                sig[0] = "java.lang.String";
                Object resultSA = connection
                        .invoke(JBIJMXConnectorUtil
                                .getDeploymentServiceMBeanName(connection),
                                "getComponentsForDeployedServiceAssembly",
                                objects, sig);
                if (resultSA instanceof String[]) {
                    composForSA.addAll(Arrays.asList((String[]) resultSA));
                }
            }
            String engineName = serviceEngineName;
            String engineState = state;
            if (result instanceof ObjectName[]) {
                ObjectName[] engines = (ObjectName[]) result;
                for (ObjectName objectName : engines) {
                    String objState = (String) connection.getAttribute(
                            objectName, "CurrentState");
                    String name = objectName.getKeyProperty("name");
                    // TODO extract the description
                    String description = "";
                    // Set the variable to have a positive test
                    if (state == null) {
                        engineState = objState;
                    }
                    if (serviceEngineName == null) {
                        engineName = name;
                    }
                    if (serviceAssemblyName == null) {
                        composForSA.add(name);
                    }
                    if (name.equals(engineName) && objState.equals(engineState)
                            && composForSA.contains(name)) {
                        // Add the component to the output
                        output.append("Name : " + name + "\n");
                        output.append("State : " + objState + "\n");
                        output.append("Description : " + description + "\n\n");
                        outputXML
                                .append("\t<component-info type='service-engine' name='"
                                        + name
                                        + "' state='"
                                        + objState
                                        + "'>\n");
                        outputXML.append("\t\t<description>");
                        outputXML.append(description);
                        outputXML.append("</description>\n");
                        outputXML.append("\t</component-info>\n");
                    }
                }
                outputXML.append("</component-info-list>");
                if (xmlOutput == null) {
                    // Display the list on the console
                    try {
                        log(output.toString());
                    } catch (NullPointerException e) {
                        // Exception thrown when execute outisde of the ant
                        // context
                    }
                } else {
                    try {
                        // Fill the property xmlOutput
                        getProject().setNewProperty(xmlOutput,
                                outputXML.toString());
                    } catch (Exception e) {
                        // Exception thrown if executed outsite of a Ant context
                    }
                }
            }
            connector.close();
        } catch (Exception e) {
            if (Boolean.parseBoolean(failOnError)) {
                throw new BuildException(e.getMessage(), e.getCause());
            }
        }
    }

    public void setServiceAssemblyName(String serviceAssemblyName) {
        this.serviceAssemblyName = serviceAssemblyName;
    }

    public void setServiceEngineName(String serviceEngineName) {
        this.serviceEngineName = serviceEngineName;
    }

    // public void setSharedLibraryName(String sharedLibraryName) {
    // this.sharedLibraryName = sharedLibraryName;
    // }

    public void setState(String state) {
        this.state = state;
    }

    public void setXmlOutput(String xmlOutput) {
        this.xmlOutput = xmlOutput;
    }
}
