/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: ComponentDescription.java 12:14:20 rmarins $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.tools.jbicommon.descriptor;

import java.util.List;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * Component production is used to describe a component installation package.
 * 
 * @version $Rev: 76 $ $Date: 2006-01-17 18:15:44 +0100 (mar., 17 janv. 2006) $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class ComponentDescription extends ExtensibleJbiElement {

    /**
     * Component type: <code>"binding-component"</code>
     */
    private static final String BC_TYPE = "binding-component";

    /**
     * Compoment type: <code>"service-engine"</code>
     */
    private static final String SE_TYPE = "service-engine";

    /**
     * Component's bootstrap class loading delegation model:
     * <code>"self-first"</code> or <code>"parent-first"</code>.
     */
    private String bootstrapClassLoaderDelegation;

    /**
     * Fully qualified name for bootstrap's <code>Bootstrap</code>
     * implementation.
     */
    private String bootstrapClassName;

    /**
     * Bootstrap Class Path.
     */
    private List<String> bootstrapClassPath;

    /**
     * Component's class loading delegation model: <code>"self-first"</code>
     * or <code>"parent-first"</code>.
     */
    private String componentClassLoaderDelegation;

    /**
     * Fully qualified name for component's <code>Component</code>
     * implementation.
     */
    private String componentClassName;

    /**
     * Component Class Path.
     */
    private List<String> componentClassPath;

    /**
     * Component identification data.
     */
    private Identification identification;

    /**
     * List of Shared Libraries used by this component.
     */
    private List<SharedLibraryList> sharedLibraryList;

    /**
     * Compoment type: SERVICE_ENGINE_TYPE or BINDING_COMPOMENT_TYPE.
     */
    private String type;

    /**
     * Default constructor.
     */
    public ComponentDescription() { // NOPMD by gblondelle
        super();
    }

    @Override
    public boolean equals(final Object other) {
        if (!(other instanceof ComponentDescription)) {
            return false; // NOPMD by gblondelle
        }
        ComponentDescription castOther = (ComponentDescription) other;
        return new EqualsBuilder().append(bootstrapClassLoaderDelegation,
                castOther.bootstrapClassLoaderDelegation).append(
                bootstrapClassName, castOther.bootstrapClassName).append(
                bootstrapClassPath, castOther.bootstrapClassPath).append(
                componentClassLoaderDelegation,
                castOther.componentClassLoaderDelegation).append(
                componentClassName, castOther.componentClassName).append(
                componentClassPath, castOther.componentClassPath).append(
                identification, castOther.identification).append(
                sharedLibraryList, castOther.sharedLibraryList).append(type,
                castOther.type).isEquals();
    }

    /**
     * Returns the class loading delegation model of component's bootstrap used
     * during component installation. If <code>null</code> the "parent-first"
     * default delegation model must be used.
     * 
     * @return <code>String</code> the delegation model:
     *         <code>"parent-first"</code> or <code>"self-first"</code>.
     */
    public String getBootstrapClassLoaderDelegation() {
        return bootstrapClassLoaderDelegation;
    }

    /**
     * Returns the fully-qualified class name for the component's
     * <code>Bootstrap</code> implementation. This is used during component
     * installation.
     * 
     * @return <code>String</code> the fully-qualified class name.
     */
    public String getBootstrapClassName() {
        return bootstrapClassName;
    }

    //
    // Getters and Setters
    //

    /**
     * Returns the class path used during installation processing
     * (bootstrapping) of the component.
     * 
     * @return <code>List<String></code> the bootstrap class path.
     */
    public List<String> getBootstrapClassPath() {
        return bootstrapClassPath;
    }

    /**
     * Returns the class loading delegation model of this component used during
     * its execution phase. If <code>null</code> the "parent-first" default
     * delegation model must be used.
     * 
     * @return <code>String</code> the delegation model:
     *         <code>"parent-first"</code> or <code>"self-first"</code>.
     */
    public String getComponentClassLoaderDelegation() {
        return componentClassLoaderDelegation;
    }

    /**
     * Returns the fully-qualified class name for the component's
     * <code>Component</code> implementation. This is used during component's
     * execution phase of its lifecycle.
     * 
     * @return <code>String</code> the fully-qualified class name.
     */
    public String getComponentClassName() {
        return componentClassName;
    }

    /**
     * Returns the class path used after the component is installed. This class
     * path, along with shared-libraries, constitutes the full class path
     * avaiable to the component through its default class loader context as
     * provided by JBI.
     * 
     * @return <code>List<String></code> the component class path.
     */
    public List<String> getComponentClassPath() {
        return componentClassPath;
    }

    /**
     * Returns component's identification data.
     * 
     * @return <code>Identification</code> the component's identification.
     */
    public Identification getIdentification() {
        return identification;
    }

    /**
     * Returns a list of shared libraries (by name and optional version) used by
     * the component. The implementation may use the version information to
     * verify that the shared library version declared by the component matches
     * the version of the shared library installed in the JBI environment.
     * 
     * @return <code>List<SharedLibraryList></code> used by this component.
     */
    public List<SharedLibraryList> getSharedLibraryList() {
        return sharedLibraryList;
    }

    /**
     * The component type string representation specifying whether this is a
     * service engine or a binding component.
     * 
     * @return <code>String</code> the component type.
     */
    public String getType() {
        return type;
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().append(bootstrapClassLoaderDelegation)
                .append(bootstrapClassName).append(bootstrapClassPath).append(
                        componentClassLoaderDelegation).append(
                        componentClassName).append(componentClassPath).append(
                        identification).append(sharedLibraryList).append(type)
                .toHashCode();
    }

    /**
     * Returns <code>true</code> if this component type is a Binding
     * Component. Otherwise returns <code>false</code>.
     */
    public boolean isBindingComponent() {
        return BC_TYPE.equals(type);
    }

    /**
     * Returns <code>true</code> if this component type is a Service Engine.
     * Otherwise returns <code>false</code>.
     */
    public boolean isServiceEngine() {
        return SE_TYPE.equals(type);
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this).append(
                "bootstrapClassLoaderDelegation",
                bootstrapClassLoaderDelegation).append("bootstrapClassName",
                bootstrapClassName).append("bootstrapClassPath",
                bootstrapClassPath).append("componentClassLoaderDelegation",
                componentClassLoaderDelegation).append("componentClassName",
                componentClassName).append("componentClassPath",
                componentClassPath).append("identification", identification)
                .append("sharedLibraryList", sharedLibraryList).append("type",
                        type).toString();
    }

    /**
     * Set the Bootstrap class loading delegation model.
     * 
     * @param delegationModel
     *            The delegation model to set: <code>"parent-first"</code> or
     *            <code>"self-first"</code>.
     */
    protected void setBootstrapClassLoaderDelegation(
            final String delegationModel) {
        this.bootstrapClassLoaderDelegation = delegationModel;
    }

    /**
     * Set the class name of component's <code>Bootstrap</code>
     * implementation.
     * 
     * @param className
     *            <code>String</code> fully-qualified class name.
     */
    protected void setBootstrapClassName(final String className) {
        this.bootstrapClassName = className;
    }

    /**
     * Set the class path used by this component's <code>Bootstrap</code>
     * during installation process.
     * 
     * @param classPath
     *            <code>List<String></code> the bootstrap class path.
     */
    protected void setBootstrapClassPath(final List<String> classPath) {
        this.bootstrapClassPath = classPath;
    }

    /**
     * Set the Component class loading delegation model.
     * 
     * @param delegationModel
     *            the delegation model to set: <code>"parent-first"</code> or
     *            <code>"self-first"</code>.
     */
    protected void setComponentClassLoaderDelegation(
            final String delegationModel) {
        this.componentClassLoaderDelegation = delegationModel;
    }

    /**
     * Set the class name of this component implementation.
     * 
     * @param className
     *            <code>String</code> fully-qualified class name.
     */
    protected void setComponentClassName(final String className) {
        this.componentClassName = className;
    }

    /**
     * Set the class path used by this component during its execution
     * processing.
     * 
     * @param classPath
     *            <code>List<String></code> the component class path.
     */
    protected void setComponentClassPath(final List<String> classPath) {
        this.componentClassPath = classPath;
    }

    /**
     * Set the component identification data.
     * 
     * @param identification
     *            The identification to set.
     */
    protected void setIdentification(final Identification identification) {
        this.identification = identification;
    }

    //
    // Implementation methods
    //

    /**
     * Set shared library list used by this component.
     * 
     * @param sharedLibraryList
     *            <code>List<SharedLibraryList></code> shared library list to
     *            set.
     */
    protected void setSharedLibraryList(
            final List<SharedLibraryList> sharedLibraryList) {

        this.sharedLibraryList = sharedLibraryList;
    }

    /**
     * Set the component type: <code>SERVICE_ENGINE_TYPE</code> or
     * <code>BINDING_COMPONENT_TYPE</code>.
     * 
     * @param type
     *            <code>String</code> the component type to set.
     */
    protected void setType(final String type) {
        this.type = type;
    }
}
