/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: Connection.java 12:14:20 rmarins $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.tools.jbicommon.descriptor;

import javax.xml.namespace.QName;

import org.apache.commons.lang.builder.EqualsBuilder;
import org.apache.commons.lang.builder.HashCodeBuilder;
import org.apache.commons.lang.builder.ToStringBuilder;

/**
 * The connection element declares a mapping from consumer-provided service
 * address to a provider's service endpoint address. This mapping applies to
 * consumption of services by all components. The consumer subelement is used to
 * declare the type of consumer address that is to be mapped (or <i>connected</i>).
 * This declaration can take two forms:
 * <ul>
 * <li><b>service type.</b> This specifies a service type, by supplying an
 * interface-name attribute. All applicable exchanges that are addressed to the
 * given service type are to be sent to the given service provider endpoint
 * instead.</li>
 * <li><b>service endpoint.</b> This specifies a service endpoint, by
 * supplying a service-name attribute and an endpoint-name attribute. All
 * applicable exchanges that are addressed to the given service endpoint are to
 * be sent to the given service provider endpoint instead.</li>
 * </ul>
 * The provider attributes is used to declare the service provider endpoint to
 * which the consumer is to be connected.
 * 
 * @version $Rev: 438 $ $Date: 2006-05-19 11:12:26Z $
 * @since Petals 1.0
 * @author <a href="mailto:rmarins@fossilec.com">Rafael Marins</a>
 */
public class Connection {

    /**
     * Consumer Endpoint Name for a service endpoint connection.
     */
    private String consumerEndpointName; // NOPMD by gblondelle

    /**
     * Consumer Interface Name for a service type connection.
     */
    private QName consumerInterfaceName; // NOPMD by gblondelle

    /**
     * Consumer Service Name for a service endpoint connection.
     */
    private QName consumerServiceName; // NOPMD by gblondelle

    /**
     * Provider Endpoint Name to be mapped the connection.
     */
    private String providerEndpointName; // NOPMD by gblondelle

    /**
     * Provider Service Name to be mapped the connection.
     */
    private QName providerServiceName; // NOPMD by gblondelle

    /**
     * Default constructor.
     */
    public Connection() { // NOPMD by gblondelle
        super();
    }

    @Override
    public boolean equals(final Object other) {
        if (!(other instanceof Connection)) {
            return false; // NOPMD by gblondelle
        }
        Connection castOther = (Connection) other;
        return new EqualsBuilder().append(consumerInterfaceName,
                castOther.consumerInterfaceName).append(consumerServiceName,
                castOther.consumerServiceName).append(consumerEndpointName,
                castOther.consumerEndpointName).append(providerServiceName,
                castOther.providerServiceName).append(providerEndpointName,
                castOther.providerEndpointName).isEquals();
    }

    /**
     * Returns the consumer Endpoint Name for a <b>service endpoint</b>
     * connection.
     * 
     * @return <code>String</code> the endpoint name.
     */
    public String getConsumerEndpointName() {
        return consumerEndpointName;
    }

    /**
     * Returns the consumer Interface Name for a <b>service type</b>
     * connection.
     * 
     * @return <code>QName</code> the interface name.
     */
    public QName getConsumerInterfaceName() {
        return consumerInterfaceName;
    }

    //
    // Getters and Setters
    //

    /**
     * Returns the consumer Service Name for a <b>service endpoint</b>
     * connection.
     * 
     * @return <code>QName</code> the service name.
     */
    public QName getConsumerServiceName() {
        return consumerServiceName;
    }

    /**
     * Returns the provider Endpoint Name which the connection is to be mapped.
     * 
     * @return <code>String</code> the endpoint name.
     */
    public String getProviderEndpointName() {
        return providerEndpointName;
    }

    /**
     * Returns the provider Service Name which the connection is to be mapped.
     * 
     * @return <code>QName</code> the service name.
     */
    public QName getProviderServiceName() {
        return providerServiceName;
    }

    @Override
    public int hashCode() {
        return new HashCodeBuilder().append(consumerInterfaceName).append(
                consumerServiceName).append(consumerEndpointName).append(
                providerServiceName).append(providerEndpointName).toHashCode();
    }

    @Override
    public String toString() {
        return new ToStringBuilder(this).append("consumerInterfaceName",
                consumerInterfaceName).append("consumerServiceName",
                consumerServiceName).append("consumerEndpointName",
                consumerEndpointName).append("providerServiceName",
                providerServiceName).append("providerEndpointName",
                providerEndpointName).toString();
    }

    /**
     * Set the consumer Endpoint Name for a <b>service endpoint</b> connection.
     * 
     * @param endpointName
     *            the endpoint name.
     */
    protected void setConsumerEndpointName(final String endpointName) {
        this.consumerEndpointName = endpointName;
    }

    /**
     * Set the consumer Interface Name for a <b>service type</b> connection.
     * 
     * @param interfaceName
     *            the interface name.
     */
    protected void setConsumerInterfaceName(final QName interfaceName) {
        this.consumerInterfaceName = interfaceName;
    }

    /**
     * Set the consumer Service Name for a <b>service endpoint</b> connection.
     * 
     * @param serviceName
     *            the service name.
     */
    protected void setConsumerServiceName(final QName serviceName) {
        this.consumerServiceName = serviceName;
    }

    /**
     * Set the provider Endpoint Name to be mapped.
     * 
     * @param endpointName
     *            the endpoint name.
     */
    protected void setProviderEndpointName(final String endpointName) {
        this.providerEndpointName = endpointName;
    }

    /**
     * Set the provder Service Name to be mapped.
     * 
     * @param serviceName
     *            the service name.
     */
    protected void setProviderServiceName(final QName serviceName) {
        this.providerServiceName = serviceName;
    }

}
