/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PackageAssemblyMojo.java 16:56:43 ddesjardins $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.tools.jbiplugin;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.StringWriter;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;
import java.util.HashSet;
import java.util.Set;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipOutputStream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.apache.commons.io.FileUtils;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.embedder.MavenEmbedder;
import org.apache.maven.model.Dependency;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProject;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Make a JBI Archive of a project
 * 
 * @goal jbi
 * @description Make a JBI from a JBI component
 * @requiresDependencyResolution runtime
 * 
 * @author ddesjardins - eBMWebsourcing
 */
public class JBIPluginMojo extends AbstractMojo {

    /**
     * @component
     */
    protected ArtifactResolver artifactResolver;

    /**
     * Component Jar
     * 
     * @parameter expression="${project.build.directory}/${project.artifactId}-${project.version}.jar"
     * @editable true
     * @required
     * @description Name of the generated component Jar
     */
    protected File componentJar;

    /**
     * JBI conf directory
     * 
     * @parameter expression="${basedir}/src/conf/"
     * @required
     * @description Path of the JBI conf directory
     */
    protected File jbiDirectory;

    /**
     * JBI Name
     * 
     * @parameter expression="${project.artifactId}-${project.version}"
     * @editable true
     * @required
     * @description Name of the generated jbi archive
     */
    protected String jbiName;

    /**
     * SubDependencies inclusion
     * 
     * @parameter expression="false"
     * @editable true
     * @required
     * @description Boolean to include the dependencies of the dependencies
     */
    protected boolean includeSubDependencies;

    /**
     * Verbose
     * 
     * @parameter expression="false"
     * @editable true
     * @required
     * @description Boolean to switch to verbose mode
     */
    protected boolean verbose;

    /**
     * Update JBI XML file
     * 
     * @parameter expression="false"
     * @editable true
     * @required
     * @description Boolean to force the update of the JBI XML file to add all
     *              the dependencies of the archive
     */
    protected boolean updateJBIXml;

    /**
     * Local maven repository.
     * 
     * @parameter expression="${localRepository}"
     * @required
     * @readonly
     */
    protected ArtifactRepository localRepository;

    /**
     * Output directory
     * 
     * @parameter expression="${project.build.directory}"
     * @required
     */
    protected File outputDirectory;

    /**
     * List of jar to exclude (comma separated values)
     * 
     * @parameter expression="foo"
     * @required
     * @editable true
     * @description List of jar to exclude (comma separated values)
     */
    protected String toExcludes;

    /**
     * The Maven Project.
     * 
     * @parameter expression="${project}"
     * @required
     * @readonly
     */
    protected MavenProject project;

    /**
     * Set of entries in the zip
     */
    private Set<String> entries = new HashSet<String>();

    /**
     * List of entries to exclude
     */
    private Set<String> entriesToExclude = new HashSet<String>();

    /**
     * Recurse a directory to add its content to a zip file if updateJBIXml file
     * is true do not add the jbi.xml file to the archive we will added it after
     * 
     * @param zipOutputStream
     *            zip file
     * @param directory
     *            directory to recurse
     * @param entryDirectoryName
     *            directory to put the content in the zip
     * @throws Exception
     */
    private void recurseDirectory(ZipOutputStream zipOutputStream,
        File directory, String entryDirectoryName) throws Exception {
        for (File file : directory.listFiles()) {
            if (file.isFile()) {
                if (!updateJBIXml
                    || (updateJBIXml && !file.getName().equals("jbi.xml"))) {
                    ZipEntry zipEntry = new ZipEntry(entryDirectoryName
                        + File.separator + file.getName());
                    DataInputStream dis = new DataInputStream(
                        new FileInputStream(file));
                    byte[] content = new byte[dis.available()];
                    dis.readFully(content);
                    zipOutputStream.putNextEntry(zipEntry);
                    zipOutputStream.write(content);
                    zipOutputStream.closeEntry();
                }
            } else {
                if (!file.getName().startsWith(".")) {
                    // Avoid to add the .svn directory
                    recurseDirectory(zipOutputStream, file, entryDirectoryName
                        + File.separator + file.getName());
                }
            }
        }
    }

    /**
     * Extract the list of jar to exclude from the archive
     * 
     */
    private void extractExcludeValues() {
        // We remove the space
        if (!"foo".equals(toExcludes)) {
            toExcludes = toExcludes.replace(" ", "");
            for (String jar : toExcludes.split(",")) {
                entriesToExclude.add(jar);
            }
        }
    }

    /**
     * Execute the plugin
     */
    public void execute() throws MojoExecutionException {
        try {
            project.getArtifact().setFile(new File(outputDirectory
                + File.separator + jbiName + ".zip"));
            if (jbiDirectory.exists()) {
                if (verbose) {
                    System.out.println("Start building JBI archive "
                        + new File(outputDirectory
                            + File.separator + jbiName + ".zip")
                            .getAbsolutePath());
                }
                new File(outputDirectory
                    + File.separator + jbiName + ".zip").delete();

                // Extract the exclude values
                extractExcludeValues();

                // Add the jbi files
                ZipOutputStream zipOutputStream = new ZipOutputStream(
                    new FileOutputStream(new File(outputDirectory
                        + File.separator + jbiName + ".zip")));
                recurseDirectory(zipOutputStream, jbiDirectory, "META-INF");
                zipOutputStream.flush();

                // Add the component jar
                ZipEntry zipEntry = new ZipEntry(componentJar.getName());
                DataInputStream dis = new DataInputStream(new FileInputStream(
                    componentJar));
                byte[] content = new byte[dis.available()];
                dis.readFully(content);
                zipOutputStream.putNextEntry(zipEntry);
                zipOutputStream.write(content);
                zipOutputStream.closeEntry();

                // Include a project
                includeProject(zipOutputStream, project);

                if (updateJBIXml) {
                    // Update the JBI XML file
                    updateJBIXmlFile(zipOutputStream);
                }

                zipOutputStream.flush();
                if (project.getDependencies().size() > 0) {
                    zipOutputStream.close();
                }
                if (verbose) {
                    System.out.println("JBI archive building done.\n");
                }
            }
        } catch (Exception e) {
            e.printStackTrace();
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }
    }

    /**
     * Clone a set of String
     * 
     * @param setToClone
     * @return
     */
    private Set<String> cloneSet(Set<String> setToClone) {
        Set<String> outSet = new HashSet<String>();
        for (String string : setToClone) {
            outSet.add(string);
        }
        return outSet;
    }

    /**
     * Update the JBI xml file to added all the dependencies of the JBI archive
     * 
     * @param zipOutputStream
     * @throws FileNotFoundException
     */
    private void updateJBIXmlFile(ZipOutputStream zipOutputStream)
        throws FileNotFoundException {
        InputStream inputStream = null;
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory
                .newInstance();
            DocumentBuilder builder = factory.newDocumentBuilder();
            Document document = builder.parse(new File(jbiDirectory
                + File.separator + "jbi.xml"));
            Element element = document.getDocumentElement();

            // Check the component-class-path

            Element compoClassPath = (Element) element.getElementsByTagName(
                "component-class-path").item(0);
            NodeList compoElems = compoClassPath
                .getElementsByTagName("path-element");
            Set<String> elemsToAdd = cloneSet(entries);
            for (int i = 0; i < compoElems.getLength(); i++) {
                Element elem = (Element) compoElems.item(i);
                if (entries.contains(elem.getTextContent())) {
                    elemsToAdd.remove(elem.getTextContent());
                }
            }
            if (elemsToAdd.size() > 0) {
                // We need to add the missing elements
                for (String jar : elemsToAdd) {
                    Element elem = document.createElement("path-element");
                    elem.setTextContent(jar);
                    compoClassPath.appendChild(elem);
                }
            }
            // add the main component-jar in the component-class-path
            Element mainElem = document.createElement("path-element");
            mainElem.setTextContent(componentJar.getName());
            compoClassPath.appendChild(mainElem);
            
            // Check the bootstrap-class-path

            Element bootClassPath = (Element) element.getElementsByTagName(
                "bootstrap-class-path").item(0);
            NodeList bootElems = bootClassPath
                .getElementsByTagName("path-element");
            elemsToAdd = cloneSet(entries);
            for (int i = 0; i < bootElems.getLength(); i++) {
                Element elem = (Element) bootElems.item(i);
                if (elemsToAdd.contains(elem.getTextContent())) {
                    elemsToAdd.remove(elem.getTextContent());
                }
            }
            if (elemsToAdd.size() > 0) {
                // We need to add the missing elements
                for (String jar : elemsToAdd) {
                    Element elem = document.createElement("path-element");
                    elem.setTextContent(jar);
                    bootClassPath.appendChild(elem);
                }
            }
            // add the main component-jar in the bootstrap-class-path
            mainElem = document.createElement("path-element");
            mainElem.setTextContent(componentJar.getName());
            bootClassPath.appendChild(mainElem);
            

            document.normalize();
            Source source = new DOMSource(document);
            StringWriter out = new StringWriter();
            Result resultStream = new StreamResult(out);
            TransformerFactory tFactory = TransformerFactory.newInstance();
            Transformer transformer;
            transformer = tFactory.newTransformer();
            transformer.transform(source, resultStream);

            inputStream = new ByteArrayInputStream(out.toString().getBytes());

        } catch (Exception e) {
            if (verbose) {
                System.out
                    .println("Problem while updating jbi.xml file. We add it like it is");
            }
            inputStream = new FileInputStream(new File(jbiDirectory
                + File.separator + "jbi.xml"));
        }

        // Add the jbi.xml entry
        try {
            ZipEntry zipEntry = new ZipEntry("META-INF/jbi.xml");
            DataInputStream dis = new DataInputStream(inputStream);
            byte[] content = new byte[dis.available()];
            dis.readFully(content);
            zipOutputStream.putNextEntry(zipEntry);
            zipOutputStream.write(content);
            zipOutputStream.closeEntry();
        } catch (Exception e) {
            if (verbose) {
                System.out.println("Fail to add jbi.xml file");
            }
        }
    }

    /**
     * Recurse the dependencies of a dependencies to add them to the component
     * archive
     * 
     * @param dependency
     */
    private void recurseDepencies(ZipOutputStream zipOutputStream,
        Dependency dependency) throws Exception {

        MavenEmbedder maven = new MavenEmbedder();
        ClassLoader classLoader = Thread.currentThread()
            .getContextClassLoader();
        maven.setClassLoader(classLoader);
        maven.setLogger(null);
        maven.start();

        String pomName = dependency.getArtifactId()
            + "-" + dependency.getVersion() + ".pom";

        File pomFile = new File(localRepository.getBasedir()
            + File.separator
            + dependency.getGroupId().replace(".", File.separator)
            + File.separator + dependency.getArtifactId() + File.separator
            + dependency.getVersion() + File.separator + pomName);
        if (pomFile.exists()) {
            MavenProject project = null;
            try {
                project = maven.readProject(pomFile);
            } catch (Exception e) {
                String jarName=null;
                try {
                    jarName = localRepository.getBasedir()
                    + File.separator
                    + dependency.getGroupId().replace(".", File.separator)
                    + File.separator + dependency.getArtifactId()
                    + File.separator + dependency.getVersion() + File.separator
                    + dependency.getArtifactId() + "-"
                    + dependency.getVersion() + ".jar";
                    
                    JarFile jarFile = new JarFile(jarName);
                    String pomEntryName = "META-INF/maven/"
                        + dependency.getGroupId() + "/"
                        + dependency.getArtifactId() + "/pom.xml";
                    ZipEntry jarEntry = jarFile.getEntry(pomEntryName);
                    if (jarEntry != null) {
                        // Copy pom.xml file
                        File tmpFile = new File(outputDirectory
                            + File.separator + "temp.xml");
                        copyStreamToFile(jarFile.getInputStream(jarEntry), tmpFile);
                        project = maven.readProject(tmpFile);
                        FileUtils.forceDelete(tmpFile);
                    }
                } catch (ZipException e1) {
                    if (verbose) {
                        System.out.println("Problem while adding "
                            + jarName + " to JBI archive");
                    }
                }
            }
            if (project != null) {
                includeProject(zipOutputStream, project);
            }
        }
    }

    /**
     * Include a project dependencies to the archive
     * 
     * @param zipOutputStream
     * @param project
     * @throws Exception
     */
    private void includeProject(ZipOutputStream zipOutputStream,
        MavenProject project) throws Exception {

        for (Object object : project.getDependencies()) {
            if (object instanceof Dependency) {
                Dependency dep = (Dependency) object;
                String jarName = dep.getArtifactId()
                    + "-" + dep.getVersion() + ".jar";

                String scope = "test";
                if (dep.getScope() == null
                    || (dep.getScope() != null && !dep.getScope()
                        .equals("test"))) {
                    scope = "other";
                }
                if (!"test".equals(scope)
                    && !entriesToExclude.contains(jarName)) {

                    // Add the dep to the zip
                    zipDependency(zipOutputStream, dep);
                    // Recurse the dep
                    recurseDepencies(zipOutputStream, dep);
                }
            }
        }
    }

    /**
     * Copy a stream to a file
     * 
     * @param inputStream
     * @param tmpFile
     * @throws Exception
     */
    private void copyStreamToFile(InputStream inputStream, File tmpFile)
        throws Exception {
        ReadableByteChannel inputChannel = Channels.newChannel(inputStream);
        FileOutputStream fos = new FileOutputStream(tmpFile);
        FileChannel outputChannel = fos.getChannel();
        ByteBuffer buffer = ByteBuffer.allocate(4096);
        int readBytes;
        while ((readBytes = inputChannel.read(buffer)) != -1) {
            if (readBytes > 0) {
                buffer.flip();
                outputChannel.write(buffer);
                buffer.clear();
            }
        }
        inputChannel.close();
        outputChannel.close();
        fos.close();
    }

    /**
     * Add a dependency to a zip stream
     * 
     * @param zipOutputStream
     * @param dependency
     */
    private void zipDependency(ZipOutputStream zipOutputStream,
        Dependency dependency) {
        String jarName = dependency.getArtifactId()
            + "-" + dependency.getVersion() + ".jar";
        try {
            if (!entries.contains(jarName)) {
                ZipEntry zipEntry = new ZipEntry(jarName);
                DataInputStream dis = new DataInputStream(new FileInputStream(
                    new File(localRepository.getBasedir()
                        + File.separator
                        + dependency.getGroupId().replace(".", File.separator)
                        + File.separator + dependency.getArtifactId()
                        + File.separator + dependency.getVersion()
                        + File.separator + jarName)));
                byte[] content = new byte[dis.available()];
                dis.readFully(content);
                zipOutputStream.putNextEntry(zipEntry);
                zipOutputStream.write(content);
                zipOutputStream.closeEntry();
                entries.add(jarName);
                if (verbose) {
                    System.out.println("     "
                        + jarName + " added to JBI archive");
                }
            }
        } catch (Exception e) {
            entriesToExclude.add(jarName);
            if (verbose) {
                System.out.println("Problem while adding "
                    + jarName + " to JBI archive");
            }
        }
    }
}
