package petalsclipse.wizards;

import static petalsclipse.wizards.NewPetalsProjectWizardPage.JBI_PROJECT;
import static petalsclipse.wizards.NewPetalsProjectWizardPage.MODULE_PROJECT;
import static petalsclipse.wizards.NewPetalsProjectWizardPage.WAR_PROJECT;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.StringWriter;
import java.lang.reflect.InvocationTargetException;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.ui.INewWizard;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.IWorkbenchWizard;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * This is a sample new wizard. Its role is to create a new file resource in the
 * provided container. If the container resource (a folder or a project) is
 * selected in the workspace when the wizard is opened, it will accept it as the
 * target container. The wizard creates one file with the extension "mpe". If a
 * sample multi-page editor (also available as a template) is registered for the
 * same extension, it will be able to open it.
 */

public class NewPetalsProjectWizard extends Wizard implements INewWizard {
	private NewPetalsProjectWizardPage page;

	private ISelection selection;

	/**
	 * Constructor for SampleNewWizard.
	 */
	public NewPetalsProjectWizard() {
		super();
		setNeedsProgressMonitor(true);
	}

	/**
	 * Adding the page to the wizard.
	 */

	public void addPages() {
		page = new NewPetalsProjectWizardPage(selection);
		addPage(page);
	}

	/**
	 * This method is called when 'Finish' button is pressed in the wizard. We
	 * will create an operation and run it using wizard as execution context.
	 */
	public boolean performFinish() {
		final String parentName = page.getParentProjectName();
		final String projectName = page.getProjectName();
		final String projectType = page.getProjectType();
		final String projectCompleteName = page.getProjectCompleteName();
		final String projectFolderName = page.getProjectFolderName();
		IRunnableWithProgress op = new IRunnableWithProgress() {
			public void run(IProgressMonitor monitor)
					throws InvocationTargetException {
				try {
					doFinish(parentName, projectName, projectFolderName,
							projectCompleteName, projectType, monitor);
				} catch (CoreException e) {
					throw new InvocationTargetException(e);
				} finally {
					monitor.done();
				}
			}
		};
		try {
			getContainer().run(true, false, op);
		} catch (InterruptedException e) {
			return false;
		} catch (InvocationTargetException e) {
			Throwable realException = e.getTargetException();
			MessageDialog.openError(getShell(), "Error", realException
					.getMessage());
			return false;
		}
		return true;
	}

	/**
	 * The worker method. It will find the container, create the file if missing
	 * or just replace its contents, and open the editor on the newly created
	 * file.
	 */

	private void doFinish(String parentName, String projectName,
			String projectFolderName, String projectCompleteName,
			String projectType, IProgressMonitor monitor) throws CoreException {
		// Create project directory
		File parentDir = new File(new File(parentName).getAbsolutePath()
				.replace("pom.xml", ""));

		File projectDir = new File(parentDir + File.separator
				+ projectFolderName);
		projectDir.mkdir();

		monitor.beginTask("Creating " + projectDir, 2);

		try {
			// Load pom type file
			DocumentBuilderFactory factory = DocumentBuilderFactory
					.newInstance();
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document document = builder.parse(this.getClass().getClassLoader()
					.getResourceAsStream("pom-type.xml"));
			Element docElement = document.getDocumentElement();

			// Load the parent pom file
			Document parentDocument = builder.parse(new File(parentName));
			Element parentDocElement = parentDocument.getDocumentElement();

			// Set the packaging
			String packaging = "jar";
			if (projectType.equals(WAR_PROJECT)) {
				packaging = "war";
			} else if (projectType.equals(MODULE_PROJECT)) {
				packaging = "pom";
			}
			NodeList packs = docElement.getElementsByTagName("packaging");
			Element packElement = (Element) packs.item(0);
			packElement.setTextContent(packaging);

			// Set the parent pom name
			Element curParentElem = (Element) docElement.getElementsByTagName(
					"parent").item(0);
			Element curParentArtifactID = (Element) curParentElem
					.getElementsByTagName("artifactId").item(0);

			// Set the parent artifactId
			int pos = 0;
			if (parentDocElement.getElementsByTagName("parent").getLength() > 0) {
				pos = 1;
			}
			Element parentArtifactID = (Element) parentDocElement
					.getElementsByTagName("artifactId").item(pos);
			curParentArtifactID.setTextContent(parentArtifactID
					.getTextContent());

			Element nameElem = (Element) docElement
					.getElementsByTagName("name").item(0);
			nameElem.setTextContent(projectCompleteName);

			Element artifactElem = (Element) docElement.getElementsByTagName(
					"artifactId").item(1);
			artifactElem.setTextContent(projectName);

			StringBuffer xmlPomBuffer = new StringBuffer(
					createStringFromDOMDocument(document));

			if (projectType.equals(JBI_PROJECT)) {
				// Insert the JBI dependency
				xmlPomBuffer.insert(xmlPomBuffer.indexOf("</dependencies>"),
						jbiDependencyXML());
				// Insert the JBI plugin part
				xmlPomBuffer.insert(xmlPomBuffer.indexOf("</plugins>"),
						jbiPluginXML());
			} else if (projectType.equals(MODULE_PROJECT)) {
				xmlPomBuffer.insert(
						xmlPomBuffer.indexOf("</dependencies>") + 17,
						modulesXML());
			}

			// Update parent project module list
			Element modulesElement = (Element) parentDocElement
					.getElementsByTagName("modules").item(0);
			Element module = parentDocument.createElement("module");
			module.setTextContent(projectFolderName);
			modulesElement.appendChild(module);

			// Write parent project pom
			copyStreamToFile(new ByteArrayInputStream(
					createStringFromDOMDocument(parentDocument).getBytes()),
					new File(parentName));

			// Write project pom
			copyStreamToFile(new ByteArrayInputStream(xmlPomBuffer.toString()
					.getBytes()), new File(projectDir + File.separator
					+ "pom.xml"));

			// Write Eclipse settings
			new File(projectDir + File.separator + ".settings").mkdir();
			copyStreamToFile(this.getClass().getClassLoader()
					.getResourceAsStream("org.eclipse.jdt.core.prefs"),
					new File(projectDir + File.separator + ".settings"
							+ File.separator + "org.eclipse.jdt.core.prefs"));

			// Update Eclipse .project file
			Document projectDocument = builder.parse(this.getClass()
					.getClassLoader().getResourceAsStream("project"));
			Element projNameElement = (Element) projectDocument
					.getDocumentElement().getElementsByTagName("name").item(0);
			projNameElement.setTextContent(projectName);
			// Write Eclipse .project file
			copyStreamToFile(new ByteArrayInputStream(
					createStringFromDOMDocument(projectDocument).getBytes()),
					new File(projectDir + File.separator + ".project"));

			// Update the Eclispe .classpath file
			if (projectType.equals(JBI_PROJECT)) {
				// We need to update the classpath
				DataInputStream dis = new DataInputStream(this.getClass()
						.getClassLoader().getResourceAsStream("classpath"));
				byte[] content = new byte[dis.available()];
				dis.readFully(content);
				StringBuffer classpathBuf = new StringBuffer(
						new String(content));
				classpathBuf.insert(classpathBuf.indexOf("</classpath>"),
						jbiClasspathXML());
				copyStreamToFile(new ByteArrayInputStream(classpathBuf
						.toString().getBytes()), new File(projectDir
						+ File.separator + ".classpath"));
			} else {
				copyStreamToFile(this.getClass().getClassLoader()
						.getResourceAsStream("classpath"), new File(projectDir
						+ File.separator + ".classpath"));
			}
			
			//ResourcesPlugin.getWorkspace().loadProjectDescription(arg0);
			//ResourcesPlugin.getWorkspace().newProjectDescription(arg0);
			
		} catch (Exception e1) {
			throwCoreException("Error while creating sub-project");
		}
		monitor.worked(1);

		// IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
		// IResource resource = root.findMember(new Path(parentName));
		// if (!resource.exists() || !(resource instanceof IContainer)) {
		// throwCoreException("Container \"" + parentName
		// + "\" does not exist.");
		// }
		// IContainer container = (IContainer) resource;
		// final IFile file = container.getFile(new Path(projectName));
		// try {
		// InputStream stream = openContentStream();
		// if (file.exists()) {
		// file.setContents(stream, true, true, monitor);
		// } else {
		// file.create(stream, true, monitor);
		// }
		// stream.close();
		// } catch (IOException e) {
		// }
		//
		// monitor.setTaskName("Opening file for editing...");
		// getShell().getDisplay().asyncExec(new Runnable() {
		// public void run() {
		// IWorkbenchPage page = PlatformUI.getWorkbench()
		// .getActiveWorkbenchWindow().getActivePage();
		// try {
		// IDE.openEditor(page, file, true);
		// } catch (PartInitException e) {
		// }
		// }
		// });
		// monitor.worked(1);
	}

	/**
	 * We will initialize file contents with a sample text.
	 */

	private void throwCoreException(String message) throws CoreException {
		IStatus status = new Status(IStatus.ERROR, "TestWizard", IStatus.OK,
				message, null);
		throw new CoreException(status);
	}

	/**
	 * We will accept the selection in the workbench to see if we can initialize
	 * from it.
	 * 
	 * @see IWorkbenchWizard#init(IWorkbench, IStructuredSelection)
	 */
	public void init(IWorkbench workbench, IStructuredSelection selection) {
		this.selection = selection;
	}

	private String jbiPluginXML() {
		StringBuffer out = new StringBuffer();
		out.append("\t\t<plugin>\n");
		out.append("\t\t\t<groupId>org.objectweb.petals</groupId>\n");
		out.append("\t\t\t<artifactId>petals-tools-jbi-plugin</artifactId>\n");
		out.append("\t\t\t<version>${project.version}</version>\n");
		out.append("\t\t\t<extensions>true</extensions>\n");
		out.append("\t\t\t<executions>\n");
		out.append("\t\t\t\t<execution>\n");
		out.append("\t\t\t\t\t<phase>install</phase>\n");
		out.append("\t\t\t\t\t<configuration>\n");
		out
				.append("\t\t\t\t\t\t\t<includeSubDependencies>false</includeSubDependencies>\n");
		out.append("\t\t\t\t\t\t\t<updateJBIXml>false</updateJBIXml>\n");
		out.append("\t\t\t\t\t</configuration>\n");
		out.append("\t\t\t\t\t<goals>\n");
		out.append("\t\t\t\t\t\t<goal>jbi</goal>\n");
		out.append("\t\t\t\t\t</goals>\n");
		out.append("\t\t\t\t</execution>\n");
		out.append("\t\t\t</executions>\n");
		out.append("\t\t</plugin>\n");
		return out.toString();
	}

	private String jbiDependencyXML() {
		StringBuffer out = new StringBuffer();
		out.append("\t\t<dependency>\n");
		out.append("\t\t\t<groupId>javax.jbi</groupId>\n");
		out.append("\t\t\t<artifactId>jbi</artifactId>\n");
		out.append("\t\t</dependency>\n");
		return out.toString();
	}

	private String modulesXML() {
		StringBuffer out = new StringBuffer();
		out.append("\t<!-- ======= -->\n");
		out.append("\t<!-- MODULES -->\n");
		out.append("\t<!-- ======= -->\n");
		out.append("\n");
		out.append("\t<modules>\n");
		out.append("\t</modules>\n");
		return out.toString();
	}

	private String jbiClasspathXML() {
		return "<classpathentry kind=\"var\" path=\"M2_REPO/javax/jbi/jbi/1.0/jbi-1.0.jar\"/>";
	}

	/**
	 * Copy a stream to a file
	 * 
	 * @param inputStream
	 * @param tmpFile
	 * @throws Exception
	 */
	private void copyStreamToFile(InputStream inputStream, File tmpFile)
			throws Exception {
		ReadableByteChannel inputChannel = Channels.newChannel(inputStream);
		FileOutputStream fos = new FileOutputStream(tmpFile);
		FileChannel outputChannel = fos.getChannel();
		ByteBuffer buffer = ByteBuffer.allocate(4096);
		int readBytes;
		while ((readBytes = inputChannel.read(buffer)) != -1) {
			if (readBytes > 0) {
				buffer.flip();
				outputChannel.write(buffer);
				buffer.clear();
			}
		}
		inputChannel.close();
		outputChannel.close();
		fos.close();
	}

	/**
	 * Create a String result from a DOM document
	 * 
	 * @param document
	 *            the DOM Document
	 * @return a String representation of the DOM Document
	 * @throws TransformerException
	 */
	private String createStringFromDOMDocument(Document document)
			throws TransformerException {
		document.normalize();
		Source source = new DOMSource(document);
		StringWriter out = new StringWriter();
		Result resultStream = new StreamResult(out);
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer;
		transformer = tFactory.newTransformer();
		transformer.transform(source, resultStream);
		String result = out.toString();
		return result;
	}
}