/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XlinkerJava2WsdlMojo.java 154 4 juil. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.tools.xlinkerplugin;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.ByteBuffer;
import java.nio.channels.Channels;
import java.nio.channels.FileChannel;
import java.nio.channels.ReadableByteChannel;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.wsdl.Definition;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLWriter;
import javax.xml.namespace.QName;

import org.act.soap.util.TMConstants;
import org.act.soap.util.XMLUtils;
import org.act.soap.wsdd.Deployment;
import org.act.soap.wsdd.Provider;
import org.act.soap.wsdd.Transport;
import org.act.soap.wsdd.Transports;
import org.act.soap.wsdd.TypeMapping;
import org.act.soap.wsdd.TypeMappings;
import org.act.soap.wsdl.java2wsdl.Emitter;
import org.act.soap.wsdl.java2wsdl.Types;
import org.apache.commons.io.FileUtils;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.embedder.MavenEmbedder;
import org.apache.maven.model.Dependency;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Build a wsar archive from java interface and/or implementation. This archive
 * can be targeted to Xlinker binding component or to classic Xlinker web app.
 * 
 * @goal java2wsar
 * @description Build a wsar archive from java interface and/or implementation.
 *              This archive can be targeted to Xlinker binding component or to
 *              classic Xlinker web app.
 * @aggregator
 * @requiresDependencyResolution runtime
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class XlinkerJava2WsarMojo extends AbstractMojo {

    /**
     * @component
     */
    protected ArtifactResolver artifactResolver;

    /**
     * Local maven repository.
     * 
     * @parameter expression="${localRepository}"
     * @required
     * @readonly
     */
    protected ArtifactRepository localRepository;

    /**
     * The Maven Project.
     * 
     * @parameter expression="${project}"
     * @required
     * @readonly
     */
    protected MavenProject project;

    /**
     * The project basedir.
     * 
     * @parameter expression="${basedir}"
     * @required
     * @readonly
     */
    protected String basedir;

    /**
     * Java class qualified name of the interface from which the wsdl file will
     * be generated. For exemple : "org.objectweb.Echo"
     * 
     * @parameter expression="org.objectweb.Echo"
     * @required
     */
    protected String interfaceClassName;

    /**
     * Java class qualified name of the implementation of the ws interface. For
     * exemple : "org.objectweb.EchoImpl"
     * 
     * @parameter expression="org.objectweb.EchoImpl"
     * @required
     */
    protected String implementationClassName;

    /**
     * The default targetNamespace for the generated wsdl. For exemple :
     * "http://objectweb.org"
     * 
     * @parameter expression="http://objectweb.org"
     * @required
     */
    protected String defaultNamespace;

    /**
     * URI of the targeted web service (mapped to the "location" attribute of
     * the <code>wsdlsoap:address<code> tag of the generated wsdl). 
     * For exemple : "http://localhost:8080/xlinker/EchoService"
     * 
     * @parameter expression="http://localhost:8080/xlinker/EchoService"
     * @required
     */
    protected String uri;

    /**
     * Type of service provider ("java" ==> java class provider || "petals" ==>
     * petals component provider)
     * 
     * @parameter expression="java"
     * @required
     */
    protected String providerType;

    /**
     * Invocation style of the web service ("rpc" or "document")
     * 
     * @parameter expression="rpc"
     * 
     */
    protected String serviceInvokeStyle;

    /**
     * Output directory (classicaly ${basedir}/target)
     * 
     * @parameter expression="${project.build.directory}"
     * @required
     * @readonly
     */
    protected File baseOutputDirectory;

    /**
     * List of jar to exclude (comma separated values). "foo" value skip this
     * parameter
     * 
     * @parameter expression="foo"
     * @required
     * @editable true
     * @description List of jar to exclude (comma separated values)
     */
    protected String toExcludes;

    /**
     * Verbose
     * 
     * @parameter expression="false"
     * @editable true
     * @required
     * @description Boolean to switch to verbose mode
     */
    protected boolean verbose;

    /**
     * Project classpath.
     * 
     * @parameter expression="${project.compileClasspathElements}"
     * @required
     * @readonly
     */
    private List classpathElements;

    /**
     * The directory for compiled classes.
     * 
     * @parameter expression="${project.build.outputDirectory}"
     * @required
     * @readonly
     */
    private File outputDirectory;

    /**
     * Temporary directory where wsar file are created
     */
    private String tempLocation;

    /**
     * Temporary META-INF directory location
     */
    private String tempMETAINFLocation;

    /**
     * The name of the service, extracted from the interface class name. Pattern :
     * interfaceClassName+"Service"
     */
    private String serviceName;

    /**
     * Set of entries in the zip
     */
    private Set<String> entries = new HashSet<String>();

    /**
     * List of entries to exclude
     */
    private Set<String> entriesToExclude = new HashSet<String>();

    private Types types = null;

    Emitter emitter;

    // Provider types
    private final static String JAVA_PROVIDER = "java";

    private final static String PETALS_PROVIDER = "petals";

    /**
     * Execute the plugin
     */
    public void execute() throws MojoExecutionException, MojoFailureException {

        try {
            // init plugin
            init();

            // Create the wsdl file (META-INF/InterfaceName.wsdl)
            createWsdlFile();

            // Create the descriptor file (META-INF/service.xml)
            createWSDDFile();

            // Create the archive file (ServiceName.wsar)
            createWsarFile();

        } catch (Exception e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }

    }

    /**
     * Create a new classloader linked to all project dependencies, with the
     * current thread classloader as parent
     * 
     * @return the new classloader
     * @throws MalformedURLException
     *             if project dependencies location are malformed
     */
    protected ClassLoader buildClassloader() throws MalformedURLException {

        // Get the current thread classloader
        ClassLoader parentClassLoader = Thread.currentThread()
                .getContextClassLoader();

        // Build all classpath url with project dependencies
        URL[] urls = null;
        if (null != this.classpathElements) {
            urls = new URL[this.classpathElements.size() + 1];
            int i = 0;
            for (Iterator iter = this.classpathElements.iterator(); iter
                    .hasNext();) {
                String element = (String) iter.next();
                urls[i] = new URL("file:/" + element);
                i++;
            }
            urls[i] = this.outputDirectory.toURL();
        } else {
            urls = new URL[1];
            urls[0] = this.outputDirectory.toURL();
        }

        // Create the new classloader that has the current thread classloader
        // for parent
        URLClassLoader classLoader = new URLClassLoader(urls, parentClassLoader);

        return classLoader;
    }

    /**
     * Create the wsdl file of the wsar.
     * 
     * @throws MalformedURLException
     */
    protected void createWsdlFile() throws Exception {
        // Build a new ClassLoader containing classpath entries
        ClassLoader classLoader = buildClassloader();

        // Set the current thread classloader to the new classloader
        Thread.currentThread().setContextClassLoader(classLoader);

        // Generate the wsdl file
        generateWSDL();

        // Reset the current thread classloader to the previous classloader
        Thread.currentThread().setContextClassLoader(classLoader.getParent());

    }

    /**
     * Generate the wsdl file and put it in the specified directory
     * 
     * @throws Exception
     */
    private void generateWSDL() throws Exception {
        emitter = new Emitter();
        emitter.setStyle(serviceInvokeStyle);
        emitter.setServiceClassName(interfaceClassName);
        emitter.setTargetNamespace(defaultNamespace);
        emitter.setLocationURI(uri);
        Definition definition = emitter.emit();
        types = emitter.getTypes();

        File outFile = new File(tempMETAINFLocation + File.separator
                + retrieveInterfaceShortName() + ".wsdl");
        File outDir = outFile.getParentFile();
        if (null != outDir && !outDir.exists()) {
            outDir.mkdirs();
        }
        FileOutputStream fileOutputStream = new FileOutputStream(outFile);
        WSDLFactory factory = WSDLFactory.newInstance();
        WSDLWriter writer = factory.newWSDLWriter();
        writer.writeWSDL(definition, fileOutputStream);
    }

    /**
     * Init the mojo : create tempLocation, tempMETAINFLocation ands serviceName
     * 
     */
    protected void init() {
        this.tempLocation = this.basedir + File.separator + "target"
                + File.separator + "generated" + File.separator + "wsarTemp";
        this.tempMETAINFLocation = this.tempLocation + File.separator
                + "META-INF";
        File tempMETAINFFile = new File(tempMETAINFLocation);
        if (null != tempMETAINFLocation && !tempMETAINFFile.exists()) {
            tempMETAINFFile.mkdirs();
        }
        this.serviceName = retrieveInterfaceShortName() + "Service";
    }

    /**
     * Copy a stream to a file
     * 
     * @param inputStream
     *            the input stream
     * @param tmpFile
     *            the output file
     * @throws Exception
     */
    private void copyStreamToFile(InputStream inputStream, File tmpFile)
        throws Exception {
        ReadableByteChannel inputChannel = Channels.newChannel(inputStream);
        FileOutputStream fos = new FileOutputStream(tmpFile);
        FileChannel outputChannel = fos.getChannel();
        ByteBuffer buffer = ByteBuffer.allocate(4096);
        int readBytes;
        while ((readBytes = inputChannel.read(buffer)) != -1) {
            if (readBytes > 0) {
                buffer.flip();
                outputChannel.write(buffer);
                buffer.clear();
            }
        }
        inputChannel.close();
        outputChannel.close();
        fos.close();
    }

    /**
     * Create the wsar archive file.
     * 
     * @throws Exception
     */
    private void createWsarFile() throws Exception {
        project.getArtifact().setFile(
                new File(baseOutputDirectory + File.separator + serviceName
                        + ".wsar"));
        if (verbose) {
            System.out.println("Start building WSAR archive "
                    + new File(baseOutputDirectory + File.separator
                            + serviceName + ".wsar").getAbsolutePath());
        }
        new File(baseOutputDirectory + File.separator + serviceName + ".wsar")
                .delete();

        // Extract the exclude values
        extractExcludeValues();

        // Add created files (service.xml and wsdl file) to zip
        ZipOutputStream zipOutputStream = new ZipOutputStream(
                new FileOutputStream(new File(baseOutputDirectory
                        + File.separator + serviceName + ".wsar")));
        recurseDirectory(zipOutputStream, new File(tempMETAINFLocation),
                "META-INF");
        zipOutputStream.flush();

        // Add classes and dependencies only if providerType is "java"
        if (XlinkerJava2WsarMojo.JAVA_PROVIDER.equals(providerType)) {
            // Add classes to zip
            recurseDirectory(zipOutputStream, outputDirectory, "classes");
            zipOutputStream.flush();

            // Include project dependencies
            includeProject(zipOutputStream, project);
            zipOutputStream.flush();
        }

        zipOutputStream.close();
        if (verbose) {
            System.out.println("WSAR archive building done.\n");
        }

    }

    /**
     * Create the wsdd (service descriptor) file of the wsar.
     * 
     * @throws Exception
     */
    private void createWSDDFile() throws Exception {
        // Create a new Deployment element
        Deployment deployment = generateWSDD();

        // Create the xml file mapping this element
        Document wsddDoc = XMLUtils.createDocument();
        Element elmWSDD = deployment.toElement(wsddDoc);
        wsddDoc.appendChild(elmWSDD);
        String wsddStr = XMLUtils.nodeToPrettyString(wsddDoc);

        // Write String to file
        File outputXMLFile = new File(this.tempMETAINFLocation + File.separator
                + "service.xml");
        ByteArrayInputStream in = new ByteArrayInputStream(wsddStr.getBytes());
        copyStreamToFile(in, outputXMLFile);

    }

    /**
     * Extract the list of jar to exclude from the archive
     * 
     */
    private void extractExcludeValues() {
        // We remove the space
        if (!"foo".equals(toExcludes)) {
            toExcludes = toExcludes.replace(" ", "");
            for (String jar : toExcludes.split(",")) {
                entriesToExclude.add(jar);
            }
        }
    }

    /**
     * Generate the wsdd bean(<code>Deployment</code> bean) that represents
     * the service descriptor xml file.
     * 
     * @return the <code>Deployment</code> bean
     * @throws Exception
     */
    protected Deployment generateWSDD() throws Exception {
        // Create a new deployment descriptor bean
        Deployment wsddDeployment = new Deployment();
        wsddDeployment.setServiceName(this.serviceName);
        wsddDeployment.setWSDLLocation(retrieveInterfaceShortName() + ".wsdl");

        // Create transport bean
        Transport transport = new Transport();
        transport.setType("http");
        transport.addParameter("URLPattern", "/" + this.serviceName);
        Transports transports = new Transports();
        transports.addTransport(transport);
        wsddDeployment.setTransports(transports);

        // Create provider bean. Two cases : "petals" or "java"
        // Java case :
        if (XlinkerJava2WsarMojo.JAVA_PROVIDER.equals(providerType)) {
            Provider provider = new Provider();
            provider.setType("java");
            provider.addParameter("ClassName", this.implementationClassName);
            wsddDeployment.setProvider(provider);
        }
        // Petals case :
        else if (XlinkerJava2WsarMojo.PETALS_PROVIDER.equals(providerType)) {
            Provider provider = new Provider();
            provider.setType("petals");
            wsddDeployment.setProvider(provider);
        } else {
            throw new Exception("Given provider type not supported : "
                    + providerType + ". Only java and petals type are allowed.");
        }

        // Generate types
        if (types != null) {
            Map schemaElements = types.getAllSchemaElements();
            org.act.soap.encoding.TypeMappingImpl typemapping = (org.act.soap.encoding.TypeMappingImpl) types
                    .getTypeMapping();
            // Generate all types mappings
            TypeMappings tms = new TypeMappings();
            // Iterate over types to add to type mappings
            for (Iterator i = schemaElements.keySet().iterator(); i.hasNext();) {
                QName xmlType = (QName) i.next();
                if (!xmlType.getNamespaceURI().equals(TMConstants.URI_ACT_SOAP)) {
                    Class javaType = typemapping.getJavaType(xmlType);
                    if (javaType != null) {
                        Class serClass = typemapping.getSerializer(javaType,
                                xmlType).getClass();
                        Class dserClass = typemapping.getDeserializer(javaType,
                                xmlType).getClass();
                        TypeMapping tm = new TypeMapping();
                        tm.setSerializer(serClass.getName());
                        tm.setDeserializer(dserClass.getName());
                        tm.setName(javaType.getName());
                        tm.setQName(xmlType);
                        tm.setEncodingStyle(TMConstants.DEFAULT_ENCODINGSTYLE);
                        tm.setLanguageSpecificType(javaType.getName());
                        tms.addTypeMapping(tm);
                    }
                }
            }
            if (tms.getAllTypeMappings().size() > 0) {
                wsddDeployment.setTypeMappings(tms);
            }
        }

        return wsddDeployment;
    }

    /**
     * Include project dependencies to the archive
     * 
     * @param zipOutputStream
     *            the archive zip stream
     * @param project
     *            the maven project
     * @throws Exception
     */
    private void includeProject(ZipOutputStream zipOutputStream,
            MavenProject project) throws Exception {

        for (Object object : project.getDependencies()) {
            if (object instanceof Dependency) {
                Dependency dep = (Dependency) object;
                String jarName = dep.getArtifactId() + "-" + dep.getVersion()
                        + ".jar";

                String scope = "test";
                if (dep.getScope() == null
                        || (dep.getScope() != null && !dep.getScope().equals(
                                "test"))) {
                    scope = "other";
                }
                if (!"test".equals(scope)
                        && !entriesToExclude.contains(jarName)) {

                    // Add the dep to the zip
                    zipDependency(zipOutputStream, dep);
                    // Recurse the dep
                    recurseDepencies(zipOutputStream, dep);
                }
            }
        }
    }

    /**
     * Recurse the dependencies of a dependency to add them to the archive
     * 
     * @param zipOutputStream
     *            the archive stream
     * @param dependency
     *            the dependency
     * @throws Exception
     */
    private void recurseDepencies(ZipOutputStream zipOutputStream,
            Dependency dependency) throws Exception {

        MavenEmbedder maven = new MavenEmbedder();
        ClassLoader classLoader = Thread.currentThread()
                .getContextClassLoader();
        maven.setClassLoader(classLoader);
        maven.setLogger(null);
        maven.start();

        String pomName = dependency.getArtifactId() + "-"
                + dependency.getVersion() + ".pom";

        File pomFile = new File(localRepository.getBasedir() + File.separator
                + dependency.getGroupId().replace(".", File.separator)
                + File.separator + dependency.getArtifactId() + File.separator
                + dependency.getVersion() + File.separator + pomName);
        if (pomFile.exists()) {
            MavenProject project = null;
            try {
                project = maven.readProject(pomFile);
            } catch (Exception e) {
                JarFile jarFile = new JarFile(localRepository.getBasedir()
                        + File.separator
                        + dependency.getGroupId().replace(".", File.separator)
                        + File.separator + dependency.getArtifactId()
                        + File.separator + dependency.getVersion()
                        + File.separator + dependency.getArtifactId() + "-"
                        + dependency.getVersion() + ".jar");
                String pomEntryName = "META-INF/maven/"
                        + dependency.getGroupId() + "/"
                        + dependency.getArtifactId() + "/pom.xml";
                ZipEntry jarEntry = jarFile.getEntry(pomEntryName);
                if (jarEntry != null) {
                    // Copy pom.xml file
                    File tmpFile = new File(baseOutputDirectory
                            + File.separator + "temp.xml");
                    copyStreamToFile(jarFile.getInputStream(jarEntry), tmpFile);
                    project = maven.readProject(tmpFile);
                    FileUtils.forceDelete(tmpFile);
                }
            }
            if (project != null) {
                includeProject(zipOutputStream, project);
            }
        }
    }

    /**
     * Recurse a directory to add its content to a zip file
     * 
     * @param zipOutputStream
     *            zip file
     * @param directory
     *            directory to recurse
     * @param entryDirectoryName
     *            directory in the zip where the content of the recursed
     *            directory have to be put
     * @throws Exception
     */
    private void recurseDirectory(ZipOutputStream zipOutputStream,
            File directory, String entryDirectoryName) throws Exception {
        for (File file : directory.listFiles()) {
            if (file.isFile()) {
                ZipEntry zipEntry = new ZipEntry(entryDirectoryName
                        + File.separator + file.getName());
                DataInputStream dis = new DataInputStream(new FileInputStream(
                        file));
                byte[] content = new byte[dis.available()];
                dis.readFully(content);
                zipOutputStream.putNextEntry(zipEntry);
                zipOutputStream.write(content);
                zipOutputStream.closeEntry();
            } else {
                if (!file.getName().startsWith(".")) {
                    // Avoid to add the .svn directory
                    recurseDirectory(zipOutputStream, file, entryDirectoryName
                            + File.separator + file.getName());
                }
            }
        }
    }

    /**
     * retrieve the interface short name (ex: "Bar") from the interface
     * qualified name (ex: "foo.Bar")
     * 
     * @return the interface short name
     */
    private String retrieveInterfaceShortName() {

        String interfaceShortName = "NOTWELLDEFINED";

        if (this.interfaceClassName.lastIndexOf(".") > -1) {
            interfaceShortName = this.interfaceClassName
                    .substring(this.interfaceClassName.lastIndexOf(".") + 1);
        }

        return interfaceShortName;
    }

    /**
     * Add a dependency to a zip stream
     * 
     * @param zipOutputStream
     *            the zip input stream
     * @param dependency
     *            the dependency to add
     */
    private void zipDependency(ZipOutputStream zipOutputStream,
            Dependency dependency) {
        String jarName = dependency.getArtifactId() + "-"
                + dependency.getVersion() + ".jar";
        try {
            if (!entries.contains(jarName)) {
                ZipEntry zipEntry = new ZipEntry("lib" + File.separator
                        + jarName);
                DataInputStream dis = new DataInputStream(new FileInputStream(
                        new File(localRepository.getBasedir()
                                + File.separator
                                + dependency.getGroupId().replace(".",
                                        File.separator) + File.separator
                                + dependency.getArtifactId() + File.separator
                                + dependency.getVersion() + File.separator
                                + jarName)));
                byte[] content = new byte[dis.available()];
                dis.readFully(content);
                zipOutputStream.putNextEntry(zipEntry);
                zipOutputStream.write(content);
                zipOutputStream.closeEntry();
                entries.add(jarName);
                if (verbose) {
                    System.out.println("     " + jarName
                            + " added to WSAR archive");
                }
            }
        } catch (Exception e) {
            entriesToExclude.add(jarName);
            if (verbose) {
                System.out.println("Problem while added " + jarName
                        + " to JBI archive");
            }
        }
    }

}
