/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: XlinkerJava2WsdlMojo.java 154 4 juil. 06 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.tools.xlinkerplugin;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Iterator;
import java.util.List;

import org.act.soap.wsdl.java2wsdl.Java2WSDL;
import org.apache.maven.artifact.repository.ArtifactRepository;
import org.apache.maven.artifact.resolver.ArtifactResolver;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.project.MavenProject;

/**
 * Build a wsdl file from a java class
 * 
 * @goal java2wsdl
 * @description Build a wsdl file from a java class
 * @aggregator
 * @requiresDependencyResolution runtime
 * 
 * @author ofabre - EBM Websourcing
 * 
 */
public class XlinkerJava2WsdlMojo extends AbstractMojo {

    /**
     * @component
     */
    protected ArtifactResolver artifactResolver;

    /**
     * Local maven repository.
     * 
     * @parameter expression="${localRepository}"
     * @required
     * @readonly
     */
    protected ArtifactRepository localRepository;

    /**
     * The Maven Project.
     * 
     * @parameter expression="${project}"
     * @required
     * @readonly
     */
    protected MavenProject project;

    /**
     * The location where generated wsdl file will be put. For exemple :
     * "${basedir}/src/conf/wsdl/Echo.wsdl"
     * 
     * @parameter expression="${basedir}/src/conf/wsdl/Echo.wsdl"
     * @required
     */
    protected String wsdlLocation;

    /**
     * Java class qualified name from which the wsdl file will be generated. For
     * exemple : "org.objectweb.Echo"
     * 
     * @parameter expression="org.objectweb.Echo"
     * @required
     */
    protected String className;

    /**
     * The default targetNamespace for the generated wsdl. For exemple :
     * "http://objectweb.org"
     * 
     * @parameter expression="http://objectweb.org"
     * @required
     */
    protected String defaultNamespace;

    /**
     * URI of the targeted web service (mapped to the "location" attribute of
     * the <code>wsdlsoap:address</code> tag of the generated wsdl). 
     * For exemple : "http://localhost:8080/xlinker/EchoService"
     * 
     * @parameter expression="http://localhost:8080/xlinker/EchoService"
     * @required
     */
    protected String uri;

    /**
     * Invocation style of the web service ("rpc" or "document")
     * 
     * @parameter expression="rpc"
     * 
     */
    protected String serviceInvokeStyle;

    /**
     * Project classpath.
     * 
     * @parameter expression="${project.compileClasspathElements}"
     * @required
     * @readonly
     */
    private List classpathElements;

    /**
     * The directory for compiled classes.
     * 
     * @parameter expression="${project.build.outputDirectory}"
     * @required
     * @readonly
     */
    private File outputDirectory;

    /**
     * Execute the plugin
     */
    public void execute() throws MojoExecutionException, MojoFailureException {

        try {

            // Build a new ClassLoader containing classpath entries
            ClassLoader classLoader = buildClassloader();

            // Set the current thread classloader to the new classloader
            Thread.currentThread().setContextClassLoader(classLoader);

            // Build a new array of String parameters
            String[] parameters = {wsdlLocation, className, defaultNamespace,
                uri, serviceInvokeStyle};

            // Call the WSDL to JAVA generator
            Java2WSDL.main(parameters);

            // Reset the current thread classloader to the previous classloader
            Thread.currentThread().setContextClassLoader(
                    classLoader.getParent());

        } catch (Exception e) {
            throw new MojoExecutionException(e.getLocalizedMessage(), e);
        }

    }

    /**
     * Create a new classloader linked to all project dependencies, with the
     * current thread classloader as parent
     * 
     * @return the new classloader
     * @throws MalformedURLException
     *             if project dependencies location are malformed
     */
    protected ClassLoader buildClassloader() throws MalformedURLException {

        // Get the current thread classloader
        ClassLoader parentClassLoader = Thread.currentThread()
                .getContextClassLoader();

        // Build all classpath url with project dependencies
        URL[] urls = null;
        if (null != classpathElements) {
            urls = new URL[classpathElements.size() + 1];
            int i = 0;
            for (Iterator iter = classpathElements.iterator(); iter.hasNext();) {
                String element = (String) iter.next();
                urls[i] = new URL("file:/" + element);
                i++;
            }
            urls[i] = outputDirectory.toURL();
        } else {
            urls = new URL[1];
            urls[0] = outputDirectory.toURL();
        }

        // Create the new classloader that has the current thread classloader
        // for parent
        URLClassLoader classLoader = new URLClassLoader(urls, parentClassLoader);

        return classLoader;
    }

}
